program Conmap    
  !! invers calculates the coefficients for mapping a given 
  !!     Jordan curve to the disk 
  !! invert calculates the mapping
  !! eval evaluates the points for a mapping
  use sep
  use CMap
  implicit none
  integer                              :: nmap,nt,mm,n,j,k,ng,m,cnt
  integer, dimension(:),allocatable    :: index1,xtmp
  real(KIND=8)                         :: angler,dgam,dtau,anglers
  real(KIND=8)                         :: xlen,perim,glen
  real(KIND=8),dimension(:),allocatable:: a,b,c,as,bs,cs,yline,gdist,gline
  real                                 :: dx
  complex,dimension(:),allocatable     :: points,endpoints
  complex,dimension(:,:),allocatable   :: zout
  complex(KIND=8)                      :: z1 ,z2 ,z3 ,zrot1 ,zto0 ,zto1 ,zrot2
  complex(KIND=8)                      :: z1s,z2s,z3s,zrot1s,zto0s,zto1s,zrot2s
  complex(KIND=8)                      :: zint,zi,za,zwa,zc1,zc2,zc3,zc4
  complex(KIND=8),dimension(:),allocatable:: zend,zr,zn,z,zzz,zin,ztopo,znt,zsave

  call sep_init()
  call from_history("n1",nmap); ng=nmap/4
  call from_history("d1",dx  )
  call from_param("nt",nt,50);  n=nmap;  mm=ng*nt
  allocate( zend(4),zr(4),zn(7),index1(4),zzz(4),znt(7) )
  allocate( z(nmap+1),a(nmap+1),b(nmap+1),c(nmap+1))
  allocate( points(nmap),endpoints(5), zin( ng*nt ), zout(nt,ng) )
  allocate( as(nmap+1),bs(nmap+1),cs(nmap+1),xtmp(1) )
  allocate( yline(nt),gdist(ng),zsave(nmap+1),gline(ng) )

  zend=0.;zr=0.;zn=0.;zzz=0.;znt=0.;z=0.;z=0.;a=0.;b=0.;c=0.
  points=0.;endpoints=0.;zin=0.;zout=0.;as=0.;bs=0.;cs=0.
  yline=0.;gdist=0.;zsave=0.;gline=0.;index1=0;xtmp=0

  !! . . INPUT
  call sep_read(endpoints,"endpoints"); zzz=dcmplx(endpoints(1:4))
  call sep_read ( points ); z(1:nmap)=dcmplx(points)
  z(nmap+1)=dcmplx( endpoints(5) ) 
  write(0,*) 'Read in files'

  !! . . TESTS - Winding and index order
  call wind_test(z,nmap); 
  do j=1,4
     xtmp=minloc(cdabs(zzz(j)-z)); index1(j)=xtmp(1)
  end do
  call index_test(index1)
  write(0,*) 'Did tests'
write(0,*) index1
  !! . . JORDAN REGION --> DISK
  call invers(z,a,b,c,z1,z2,z3,zrot1,zto0,zto1,angler,zrot2,nmap,n)
write(0,*) z(1:5)
  zend=z(index1);z1s=z1;z2s=z2;z3s=z3;zrot1s=zrot1;zto0s=zto0;
  zto1s=zto1;anglers=angler;zrot2s=zrot2;as=a;bs=b;cs=c 
  zsave=z;
  write(0,*) 'Computed mapping Jordan region to disk'

  !! . . DISK --> SQUARE
  call get_length(zend,xlen) !! . . Find length of conformal rect
  call get_coeffs(z,zn,xlen) !! . . Mapping coefficients (new z)
  write(0,*) 'Computed length of rectangle',xlen
  do j=1,4
     xtmp=minloc(cdabs(zn(j)-z)); index1(j)=xtmp(1);
  end do

  !! . . INVERSE MAPPING POINTS
  call invers(z,a,b,c,z1,z2,z3,zrot1,zto0,zto1,angler,zrot2,nmap,n)
  zr=z(index1); za=(zr(3)-zr(2))/(zr(3)-zr(1))
  zwa=(zend(3)-zend(2))/(zend(3)-zend(1))
  zc1=zwa*zend(1)-za*zend(2);  zc2=za*zend(2)*zr(1)-zwa*zend(1)*zr(2)
  zc3=zwa-za;                  zc4=za*zr(1)-zwa*zr(2)
  zsave=(zsave*zc4-zc2)/(-zc3*zsave+zc1)
  write(0,*) 'Computed mapping from square to disk'
  !! . . Image of bndry points
  call eval(zsave,a,b,c,z1,z2,z3,zrot1,zto0,zto1,angler,zrot2,nmap,n)
  gline=dreal(zsave(1:ng))
  write(0,*) 'Computed image of topography on rectangle'

  !! . . SETUP OUTPUT GRID
  dgam=xlen/dble(ng+1);
  dtau=1./dble(nt); 
  yline=(/(k,k=nt,1,-1)/)*dtau
!  dtau=1./dble(nt-1); 
!  yline=(/(k,k=nt-1,0,-1)/)*dtau
 do j=1,ng
     zin(1+(j-1)*nt:j*nt)=dcmplx(gline(j),yline)
  end do
 ! do j=1,ng
 !    zin(1+(j-1)*nt:j*nt)=dcmplx(gline(j),yline/2.+.5)
 ! end do
  write(0,*) 'Set up output grid'
  write(0,*) 'nt',nt,'ng',ng
  write(0,*) 'dt',dtau,'dg',dgam

  !! . . SQUARE --> DISK
  call invert(zin,a,b,c,z1,z2,z3,zrot1,zto0,zto1,angler,zrot2,mm,n) 
  zin(1:mm)=(zin(1:mm)*zc1+zc2)/(zc3*zin(1:mm)+zc4)
  write(0,*) 'Mapped grid from rectangle to disk'

  !! . . DISK --> JORDAN REGION
  call eval(zin,as,bs,cs,z1s,z2s,z3s,zrot1s,zto0s,zto1s,anglers,zrot2s,mm,n)
  zout=reshape(zin,(/nt,ng/))
  zout(1,1:ng)=points(1:ng)
  write(0,*) 'Mapped Grid from disk to Jordan region'  

  !! . . Output to file
  call write_out(zout,nt,ng,dble(ng-1)*dx/(xlen)/dble(nt-1),dble(dx),dtau*dble(nt-1),xlen)
!  call write_out(zout,nt,ng,dble(ng-1)*dx/(2.*xlen)/dble(nt-1),dble(dx),dtau*dble(nt-1)/2.,xlen)
  call exit()
end program Conmap
