!< 
!-------------------------------------------------------------
! . . Phaseray.x
! . . Program that calculates exact rays based on the theory 
! . . of Terry L. Foreman in 
! . . An exact ray theoretical formulation of the Helmholtz
! . . equation, J. Acoust. Soc. Am. 86(1), July 1989
!
! USAGE:
! . . < in.H Phaseray.x par=parfile.p > out.H
! . .
! . . where parfile.p contains
! . . 
! . . 
!& end of self-documentation
!-------------------------------------------------------------
program ERT
  use SEP
  use ert_mod 
  implicit none
  integer             :: ng,ntau,nx,nz
  real                :: dz,og,dg,dx,ox,oz,eps,dtau
  complex,allocatable :: start(:),rays(:,:)
  real   ,allocatable :: x(:),z(:),data(:,:)
  logical             :: verbose
  call sep_init()
  call from_history("n1",ng           )  
  call from_history("o1",og           )
  call from_history("d1",dg           )
  call from_aux("tt","n1",nz          ) 
  call from_aux("tt","d1",dz          )
  call from_aux("tt","o1",oz          )
  call from_aux("tt","n2",nx          ) 
  call from_aux("tt","d2",dx          ) 
  call from_aux("tt","o2",ox          ) 

  call from_param("eps ",eps, 0.000001)
  call from_param("ntau",ntau,100     )
  call from_param("dtau",dtau,0.004   )
  call from_param("verb",verbose,.false.)

  call to_history("n1",ntau)
  call to_history("n2",ng  )
  call to_history("d1",dtau)
  call to_history("d2",dg  )
  call to_history("o1",0.  )
  call to_history("o2",og  )
  call to_history("esize",8)
  call auxputch("n1","i",ntau*2+ng*2,"boundary.H")
  call auxputch("d1","r",1.         ,"boundary.H")
  call auxputch("o1","r",1.         ,"boundary.H")
  call auxputch("esize","i",8       ,"boundary.H")

  call auxputch("n1","i",ntau,"mesh.H")
  call auxputch("d1","r",dtau,"mesh.H")
  call auxputch("o1","r",0.  ,"mesh.H")
  call auxputch("n2","i",ng  ,"mesh.H")
  call auxputch("d2","r",dg  ,"mesh.H")
  call auxputch("o2","r",og  ,"mesh.H")
  call auxputch("esize","i",8,"mesh.H")

  call hclose()

  allocate( data(nz,nx),start(ng),x(ng),z(ng)   )
  allocate( rays(ntau,ng) )

  call sep_read( data,"tt")
  call sep_read( start    )

  x  = real (start)  
  z  = aimag(start)

  call phaseray_init(ng,nx,nz,dx,dz,ox,oz,ntau,dtau,eps,verbose)
  call phaserays(data,rays,x,z)

  call sep_write(rays) 

  deallocate(rays,data,x,z)
end program ERT
