!< 
!-------------------------------------------------------------
! . . Exact_rest.x
! . . Program that calculates exact rays based on the theory 
! . . of Terry L. Foreman in 
! . . An exact ray theoretical formulation of the Helmholtz
! . . equation, J. Acoust. Soc. Am. 86(1), July 1989
!
! USAGE:
! . . < in.H Exact_r1.x par=parfile.p > out.H
! . .
! . . where parfile.p contains
! . . 
! . . xsp = x location of shotpoint
! . . R = Radius of circle to start about xsp
! . . begang = starting angle (in degrees) at Radius R about shot point xsp 
! . . endang = finishing angle (in degrees) at Radius R about shot point xsp 
! . . maxnum = maximum number of steps with numerical integrator
! . . eps = (0) a small number to prevent a division by zero
! . . 
!& end of self-documentation
!-------------------------------------------------------------
program Exact_rest
  use SEP
  use exact_rest_mod 
  implicit none
  integer                                :: nz,nx,nw,maxnum,i2,nray,iray,it,wantP
  integer,  dimension(:)                 :: ixx,izz
  real                                   :: dtau,begang,endang,dz,dx,dw,oz,ox,ow,eps,xsp,R,mxint,mnint
  real(kind=8),dimension(:),allocatable  :: x,z,xang,zang
  real        ,dimension(:,:),allocatable:: vel_in
  real(kind=8),dimension(:,:),allocatable:: vel
  complex,  dimension(:)  , allocatable  :: Initpos
  complex,  dimension(:,:),allocatable   :: Rayout,data
  logical                                :: adj,add,fromSP
  call sep_init()
  call from_history("n1",nz             ) 
  call from_history("n2",nx             )  
  call from_history("n3",nw             )  
  call from_history("d1",dz             ) 
  call from_history("d2",dx             )  
  call from_history("d3",dw             )  
  call from_history("o1",oz             ) 
  call from_history("o2",ox             )  
  call from_history("o3",ow             )
  call from_param("adj", adj, .false.   )
  call from_param("add", add, .false.   ) 
  call from_param("eps", eps, 0.000001  )
  call from_param("maxnum",maxnum,700   )
  call from_param("xsp", xsp            )
  call from_param("R",   R              )
  call from_param("begang",begang,-45.  ) 
  call from_param("endang",endang,45.   )
  call from_param("nray",nray,100       )
  call from_param("dtau",dtau           )
  call from_param("fromSP",fromSP,.true.)
  call from_param("mxint",mxint,5.      )
  call from_param("mnint",mnint,0.05    )
  call from_param("wantP",wantP,0       )
  allocate( x(nx),z(nz)                 )
  allocate( Rayout( maxnum,8192)        )
  allocate( data(nz,nx),vel(nz,nx)      )
  allocate( vel_in(nz,nx),Initpos(nray) )
  allocate( ixx(nx),izz(nz)             )
  allocate( xang(nray),zang(nray)       )
  ! . . Read in Data
  call sep_read( data                   )
  call sep_read( vel_in,"vel"           )
  call sep_read( Initpos,"initpoints"    )

  ! . . Initialize Modules
  call exact_mod_init(nz,nx,nw,dz,dx,dw,oz,ox,ow,dtau,mxint,fromSP,xsp,R,eps,mnint)
  vel=dble(vel_in)

  ! . . Starting points
  xang = dble( real(Initpos))
  zang = dble(aimag(Initpos))

  ! . . Create model dimensions
  ixx(1:nx) = (/ (i2,i2=1,nx) /);  izz(1:nz) = (/ (i2,i2=1,nz) /) 
  x=(ixx-1)*dx+ox               ;  z=(izz-1)*dz+oz

  ! . . Run interpolation program
  call Run_exact_interp(vel,data,Rayout,adj,add,maxnum,x,z,xang,zang,it)

  !! . . Write(0,*) out Shit
  call write_on_baby(Rayout,maxnum,it,endang,begang)

  deallocate(Rayout,xang,zang,data,vel,x,z)
end program Exact_rest
