module RCabcd2_NL_
  use sep
  use utils_
  implicit none
  complex, pointer :: rays(:,:)
  real,    pointer :: x(:,:),z(:,:)
  real,    pointer :: h1(:,:),h2(:,:)
  real,    pointer :: m1(:,:),m2(:,:)
  real,    pointer :: a(:,:),b(:,:),c(:,:),d(:,:)
  real,    pointer :: dang(:),azv(:),axv(:),jxv(:),jzv(:)
  type(myaxis) ::at,ag
  integer      ::it,ig,j
  real         ::gx,gz
  real         ::eps,peps
  call sep_init()
  call hinaxis(at,1)
  call hinaxis(ag,2)
  call to_history("esize",4)
  call to_history("n3",4)
  call from_param("peps",peps,1.)
  call sep_close()
  allocate(rays(at%n,ag%n))
  allocate(  h1(at%n,ag%n),h2(at%n,ag%n))
  allocate(  m1(at%n,ag%n),m2(at%n,ag%n))
  allocate(   x(at%n,ag%n))
  allocate(   z(at%n,ag%n))
  allocate(   a(at%n,ag%n),b(at%n,ag%n),c(at%n,ag%n),d(at%n,ag%n))
  allocate(   dang(ag%n),azv(at%n),axv(at%n),jxv(ag%n),jzv(ag%n) )
  call sep_read(rays)
  call sep_read(dang,"dang")
  z=aimag(rays)
  x= real(rays)

  !! h1 Calculate devivative along ray direction (tau)
  do ig=1,ag%n
     axv(1:at%n-1)=( x(2:at%n,ig)-x(1:at%n-2,ig) )/(at%d)
     azv(1:at%n-1)=( z(2:at%n,ig)-z(1:at%n-2,ig) )/(at%d)
     axv(at%n)    =( x(at%n,ig)  -x(at%n-1,ig)   )/(at%d)
     azv(at%n)    =( z(at%n,ig)  -z(at%n-1,ig)   )/(at%d)
     h1(:,ig)=sqrt(axv*axv+azv*azv)
  end do
  h1(1,:)=h1(2,:)  

  !! h2
  do it=1,at%n
     jxv(1)=( x(it,2)-x(it,1) )/dang(1)
     jzv(1)=( z(it,2)-z(it,1) )/dang(1)
     jxv(2:ag%n-1)=&
     ( x(it,3:ag%n)  -x(it,2:ag%n-1) )/(2.*dang(2:ag%n  ))&
     +( x(it,2:ag%n-1)-x(it,1:ag%n-2) )/(2.*dang(1:ag%n-1))
     jzv(2:ag%n-1)=&
     ( z(it,3:ag%n)  -z(it,2:ag%n-1) )/(2.*dang(2:ag%n  ))&
     +( z(it,2:ag%n-1)-z(it,1:ag%n-2) )/(2.*dang(1:ag%n-1))
     jxv(ag%n)= ( x(it,ag%n)  -x(it,ag%n-1) )/dang(ag%n  )
     jzv(ag%n)= ( z(it,ag%n)  -z(it,ag%n-1) )/dang(ag%n  )
     h2(it,:) = sqrt(jxv*jxv+jzv*jzv)
  end do
  h2(:,1)=h2(:,2)
  !! m1 = d(h2/h1)/dt = ( d(h2)/dt * h1 - h2 * d(h1)/dt )/ h1^2 
  do it=2,at%n-1
     do ig=1,ag%n
        m1(it,ig)=&
        ( h1(it,ig)*(h2(it+1,ig)-h2(it-1,ig)) - h2(it,ig)*&
        ( h1(it+1,ig)-h1(it-1,ig) ) )/(2.*at%d*h1(it,ig)*h1(it,ig))
     end do
  end do
  m1(1,:)    = m1(2,:) 
  m1(at%n,:) = m1(at%n-1,:)

  !! m2 = d(h1/h2)/dg = ( d(h1)/dg * h2 - h1 * d(h2)/dg )/ h2^2 
  do it=1,at%n
     do ig=2,ag%n-1
        m2(it,ig)= &
        0.5*(h2(it,ig)*((h1(it,ig+1)-h1(it,ig))/dang(ig)+&
        (h1(it,ig)-h1(it,ig-1))/dang(ig-1)))-   &
        0.5*(h1(it,ig)*((h2(it,ig+1)-h2(it,ig))/dang(ig)+&
        (h2(it,ig)-h2(it,ig-1))/dang(ig-1)))
     end do
  end do
  m2(:,1)    = m2(:,2)
  m2(:,ag%n) = m2(:,ag%n-1)
  m2(1,:)    = m2(2,:)
  m2(at%n,:) = m2(at%n-1,:)

  a=1./(h1*h1)
  where (h2 .eq. 0.)
     b=0.
     c=0.
     d=0.
     write(0,*) 'ZEROS'
  end where
  where (h2 .ne. 0.)
     b=m1/(h1*h2)
     c=m2/(h1*h2*h2*h2)
     d=1./(h2*h2)
  end where

  call sep_write(a)
  call sep_write(b)
  call sep_write(c)
  call sep_write(d)

  call exit(0)
end module RCabcd2_NL_









