program Laplace2d
  !     ===========================================================
  !     2-dimensionnal Laplace's equation
  !     ==========================================================
  use SEP
  implicit none
  integer                    :: nx,nz
  real                       :: oz,ox,dx,dz


  !! . . Solution paramters
  integer       :: xmin , xmax , zmin , zmax
  integer       :: Lxmin, Lxmax, Lzmin, Lzmax
  integer       :: Rxmin, Rxmax, Rzmin, Rzmax
  integer       :: Uxmin, Uxmax, Uzmin, Uzmax
  integer       :: Dxmin, Dxmax, Dzmin, Dzmax
  real                       :: maxdelta
  real, parameter            :: CONVERGENCE = 0.001
  integer                    :: ITERATIONS = 0
  integer                    :: I,J,ii
  real, dimension(:,:)       :: phiold,phinew
  real, dimension(:,:)       :: data_input

  !! . . Output file setup
  call sep_init()
  call from_history("n1",nz             ) 
  call from_history("n2",nx             )  
  call from_history("d1",dz             ) 
  call from_history("d2",dx             )  
  call from_history("o1",oz             ) 
  call from_history("o2",ox             )  
  call to_history("n1",nz)
  call to_history("n2",nx)
  call to_history("o1",oz)
  call to_history("o2",ox)
  call to_history("d1",dz)
  call to_history("d2",dx)
  call hclose()
  allocate(data_input(nz,nx) )
  call sep_read(data_input)

  !! . . Parameters setup
  xmin=2
  xmax=nx-1
  zmin=2
  zmax=nz-1
  Lxmin = xmin-1; Lxmax = xmax-1
  Lzmin = zmin  ; Lzmax = zmax
  Rxmin = xmin+1; Rxmax = xmax+1
  Rzmin = zmin  ; Rzmax = zmax
  Uxmin = xmin  ; Uxmax = xmax
  Uzmin = zmin-1; Uzmax = zmax-1
  Dxmin = xmin  ; Dxmax = xmax
  Dzmin = zmin+1; Dzmax = zmax+1
  allocate( phiold(nz,nx), phinew(nz,nx) )

!! . . Output file setup
  phinew(1:nz,1   ) = data_input(1:nz,1   )
  phinew(1:nz,nx  ) = data_input(1:nz,nx  )
  phinew(1   ,1:nx) = data_input(1   ,1:nx)
  phinew(  nz,1: 2) = data_input(  nz,1: 2)

  do ii=3,nx
     if ((data_input(nz,ii)-data_input(nz,ii-1))*(data_input(nz,ii-1)-data_input(nz,ii-2)) .lt. 0) then
        phinew(nz,ii) = data_input(nz,ii-1)
     else
        phinew(nz,ii) = data_input(nz,ii)
     end if
  end do

!! . . Set values to zero
  phinew(zmin:zmax,xmin:xmax) = 0
write(0,*) phinew(nz,:)
  JACOBI_ITERATION: DO

     ITERATIONS = ITERATIONS + 1 
     phiold = phinew        

     phinew(zmin:zmax,xmin:xmax)=0.25*(phiold(Lzmin:Lzmax,Lxmin:Lxmax)+ \
     phiold(Rzmin:Rzmax,Rxmin:Rxmax)+phiold(Uzmin:Uzmax,Uxmin:Uxmax)+ \
     phiold(Dzmin:Dzmax,Dxmin:Dxmax))

     maxdelta=maxVAL(ABS(phinew(zmin:zmax,xmin:xmax)-phiold(zmin:zmax,xmin:xmax)))

     IF (maxdelta < CONVERGENCE) exit JACOBI_ITERATION

  END DO JACOBI_ITERATION

  call sep_write(phinew)

  STOP
end program Laplace2d



