program MakeSWF
  use SEP
  use LMO_mod
  implicit none

  integer                                :: n1,n2,n3,n4,pad2,m1,m2,m3
  integer                                :: mx,qmod,mt,tpfact,xpfact
  real                                   :: p1,p2,d1,d2,d3,dt,dx,o1,o2,o3,ot,ox
  real, dimension(:,:)   ,allocatable    :: modl
  real, dimension(:),allocatable         :: avel
  real, dimension(:),allocatable         :: dvel

  call sep_init()
  call from_param("n1",n1,400           ) ! Num depth samples
  call from_param("d1",d1,0.12          ) ! Depth spacing
  call from_param("o1",o1,0.            ) ! First time loc.
  call from_param("n2",n2,120           ) ! Num x-dir samples
  call from_param("d2",d2,3.            ) ! X-dir spacing
  call from_param("o2",o2,0.            ) ! X-offset
  call from_param("n3",n3,1             ) ! Num y-dir samples
  call from_param("d3",d3,1.            ) ! Y-dir spacing
  call from_param("o3",o3,0.            ) ! Y offset
  call from_param("n4",n3,1             ) ! Num Plane Waves
  call from_param("p1",p1               ) ! Ray param - x-dir
  call from_param("p2",p2               ) ! Ray param - y-dir
  call to_history("n1",n1               )
  call to_history("n2",n2               )
  call to_history("n3",n3               )
  call to_history("d1",d1               )
  call to_history("d2",d2               )
  call to_history("d3",d3               )
  call to_history("o1",o1               )
  call to_history("o2",o2               )
  call to_history("o3",o3               )

  !! Find Padded Numbers for Complex model
  m1=pad2(n1)
  m2=pad2(n2)
  m3=pad2(n3)
  dt=d1;dx=d2;
  mt=m1;mx=m2;
  tpfact=2; xpfact=3;
  ox=-mx*dx
  ot=-mt*dt 

  !! Allocate arrays
  allocate( avel(m1), dvel(m1)     )
  allocate(   modl(tpfact*m1,xpfact*m2) )
  modl=0.;  

  !! Get Correct Linear Moveout Velocity
  call sep_read(avel,"avel")

  !! Call Linear Moveout Subroutine 
  call LMO_2D(mt,mx,dt,dx,ot,ox,p1,avel,modl,tpfact,xpfact)
  !! Take real component of model and write out
  call to_history("n1",tpfact*mt             )
  call to_history("n2",xpfact*mx             )
  call to_history("o1",-mt*dt               )
  call to_history("o2",-mx*dx           )
  call hclose()
  call sep_write(modl)

  deallocate(modl,avel,dvel)

end program MakeSWF
