program RWEab_
  use sep
  use util_

  implicit none

  complex, pointer :: rays(:,:)
  real,    pointer :: x(:,:),z(:,:)
  real,    pointer :: h1(:,:),h2(:,:)
  real,    pointer :: ss(:,:),aa(:,:),bb(:,:)
  real,    pointer :: ma(:,:),mb(:,:)

  type(myaxis) ::at,ag,ar
  integer      ::it,ig
  real         ::gx,gz,Jclip
  real         :: eps,peps
  character(len=128)   :: name

  integer          :: naref, nbref
  real             :: mina,maxa,dela
  real             :: minb,maxb,delb
  real             :: a0,b0
  integer          :: ia,ib,ii
  complex, pointer :: ab(:,:)
  real,    pointer :: mm(:,:)

  real             :: tiny

  call sep_init()
  call from_param("naref",naref,1)
  call from_param("nbref",nbref,1)
  call from_param("Jclip",Jclip,0.75)
  write(0,*) "naref=",naref
  write(0,*) "nbref=",nbref

  call hinaxis(at,1)
  call hinaxis(ag,2)

  call to_history("esize",4)
  call to_history("n3",5)

  call from_param("peps",peps,0.01)
  ar%n=naref*nbref
  ar%o=0.
  ar%d=1.
  call aouaxis(at,1,'abref')
  call aouaxis(ar,2,'abref')
  call auxputch("esize",'i',8,'abref')
  call sep_close()

  allocate(rays(at%n,ag%n))
  allocate(  h1(at%n,ag%n))
  allocate(  h2(at%n,ag%n))
  allocate(   x(at%n,ag%n))
  allocate(   z(at%n,ag%n))

  allocate(  ss(at%n,ag%n))
  allocate(  aa(at%n,ag%n))
  allocate(  bb(at%n,ag%n))

  allocate(  ab(at%n,naref*nbref))
  allocate(  mm(at%n,ag%n))
  allocate(  ma(at%n,ag%n))
  allocate(  mb(at%n,ag%n))

  !----------------------------------------------------------------
  call sep_read(rays)
  z=aimag(rays)
  x= real(rays)
  call sreed("slo",ss,size(ss)*4)
  !----------------------------------------------------------------

  !! h1 == alpha
  do it=1,at%n-1
     do ig=1,ag%n
        gx = (x(it+1,ig)-x(it,ig))/at%d
        gz = (z(it+1,ig)-z(it,ig))/at%d
        h1(it,ig) = sqrt(gx*gx+gz*gz)
     end do
  end do
  h1(at%n,:) = h1(at%n-1,:)
write(0,*) 'FREQ STRETCH VALUES: ',minval(h1),maxval(h1)

  !! h2 == J
  do it=1,at%n
     do ig=1+1,ag%n-1
        gx = (x(it,ig+1)-x(it,ig-1))/2./ag%d
        gz = (z(it,ig+1)-z(it,ig-1))/2./ag%d
        h2(it,ig) = sqrt(gx*gx+gz*gz)
     end do
  end do
  h2(:,1)    = h2(:,2)
  h2(:,ag%n) = h2(:,ag%n-1)
  h2(1,:)    = h2(2,:)
write(0,*) 'JACOBIAN VALUES: ',minval(h2),maxval(h2)

  eps=peps*(maxval(abs(h2))+minval(abs(h2)))/2.
  where(abs(h2)<eps)
     h2=eps
  end where

  aa=ss*h1
  bb=h1/h2
  mm=1.
  ma=1.
  mb=1.

  tiny=0.1

  !! compute reference velocities
  do it=1,at%n

     mina=minval(aa(it,:))/Jclip
     maxa=maxval(aa(it,:))*Jclip
     dela=(maxa-mina)/naref

     minb=max(Jclip*(minval(bb(it,:))),.4)
     maxb=min(Jclip*(maxval(bb(it,:))),1.5)
     delb=(maxb-minb)/nbref

     !! reference medium (a0 and b0)
     ii=1
     do ia=1,naref
        a0 = mina + (ia-1)*dela
        do ib=1,nbref
 !          if (ib .eq. 1) then
 !             ab(it,ii) = cmplx(a0,1.)
 !          else
              b0 = minb + (ib-1)*delb
              ab(it,ii) = cmplx(a0,b0)
 !          end if
           ii=ii+1
        end do
     end do
   
     !! mask for a
     do ia=1,naref
        a0    = mina + 0.5*dela + (ia-1)*dela
        where( a0-(0.5+tiny)*dela <= aa(it,:) .and. aa(it,:)<=a0+(0.5+tiny)*dela) 
           ma(it,:)=ia
        end where
     end do

     !! mask for b
     do ib=1,nbref
        b0    = minb + 0.5*delb + (ib-1)*delb
        where( b0-(0.5+tiny)*delb <= bb(it,:) .and. bb(it,:)<=b0+(0.5+tiny)*delb) 
           mb(it,:)=ib
        end where
     end do

  end do

  !! reference medium mask
  ii=1
  do ia=1,naref
     do ib=1,nbref
        where(ma==ia .and. mb==ib)
           mm=ii
        end where
        ii=ii+1
     end do
  end do

write(0,*) 'A factor Values: ', minval(aa),maxval(aa)
write(0,*) 'B factor Values: ', minval(bb),maxval(bb)

  !! write output
  call sep_write(aa)
  call sep_write(bb)
  call sep_write(mm)
  call sep_write(ma)
  call sep_write(mb)
  call sep_write(ab,'abref')

  call exit(0)
end program RWEab_









