program NewRWEsp
  use sep
  use util_
  use cc2rc
  use rwe_wemig
  use rwe_ab
  implicit none
  type(myaxis) :: az,ax,aw,as,at,ag,ar
  character(len=128)   :: name

  integer :: method
  integer  :: iz,ii,nrefa,nrefs,nrefj,niter
  logical              :: adj
  real,    allocatable :: slow_in(:,:)
  real,    allocatable :: ss(:,:),aa(:,:),jj(:,:),mm(:,:)
  real,    allocatable :: s0(:,:),a0(:,:),j0(:,:)
  complex, pointer     :: rwfall(:,:,:) !! receiver wavefield
  complex, pointer     :: swfall(:,:,:) !! source   wavefield
  real,    pointer     :: imgall(:,:,:) !! image
  complex, pointer     :: rwf(:,:) !! receiver wavefield
  complex, pointer     :: swf(:,:) !! source   wavefield
  real,    pointer     :: img(:,:) !! image
  complex, pointer     :: rays(:,:) !! rays input

  call sep_init()
  call initpar()
  call from_param("method",method,2)
  call from_param("nrefs",nrefs,1)
  call from_param("nrefa",nrefa,1)
  call from_param("nrefj",nrefj,1)
  call from_param("niter",niter,30)

  !! . . Read in axes information
  call hinaxis(aw,2) !! Frequency from rwf
  call hinaxis(as,3) !! Shots from rwf
  name="rays"; call ainaxis(at,1,name)
  name="rays"; call ainaxis(ag,2,name)
  name="slow"; call ainaxis(az,1,name)
  name="slow"; call ainaxis(ax,2,name)
  ar%n=nrefs*nrefa*nrefj; ar%o=1.; ar%d=1.

  !! . . Output parameters
write(0,*) 'MIGRATION PARAMETERS'
write(0,*) "Gamma      ",ag%n,ag%o,ag%d !! . . Receivers
write(0,*) "Tau        ",at%n,at%o,at%d !! . . Depths
write(0,*) "Frequency  ",aw%n,aw%o,aw%d !! . . Frequencies
write(0,*) "References ",ar%n,ar%o,ar%d !! . . References
write(0,*) "Shots      ",as%n,as%o,as%d !! . . Shots
write(0,*) "Slow Depths",az%n,az%o,az%d !! . . Slowness model Depths
write(0,*) "Slow Offset",ax%n,ax%o,ax%d !! . . Slowness model offsets
write(0,*) " "

  !! . . Number of references
write(0,*) "Number of s refs",nrefs
write(0,*) "Number of a refs",nrefa
write(0,*) "Number of j refs",nrefj
write(0,*) " "

  !! . . Allocate array variables
  allocate(swf(ag%n,aw%n),rwf(ag%n,aw%n))
  allocate(img(at%n,ag%n),imgall(at%n,ag%n,as%n))
  allocate(swfall(ag%n,aw%n,as%n),rwfall(ag%n,aw%n,as%n))
  allocate(slow_in(az%n,ax%n),rays(at%n,ag%n))
  allocate(ss(at%n,ag%n),aa(at%n,ag%n),jj(at%n,ag%n),mm(at%n,ag%n))
  allocate(s0(at%n,ar%n),a0(at%n,ar%n),j0(at%n,ar%n))

  !! . . Read in input files
  call sep_read(rwfall)
  call sreed("swf",swfall,size(swfall)*8)
  call sreed("slow",slow_in,size(slow_in)*4)
  call sreed("rays",rays,size(rays)*8)

  !! . . Take slowness model from Cartesian to Rays
  call cc2rc_init(az,ax,at,ag)
  call cc2rc_(rays,slow_in,ss,.false.)

  !! . . Calculate reference parameters
  call doc('./lloyd.rs90')
  call rweab_init(at,ag,nrefs,nrefa,nrefj,niter)
  call rweab(rays,ss,aa,jj,s0,a0,j0,mm)

  !! . . Output reference parameterws
write(0,*) 'REFERENCE PARAMETERS'
write(0,*) 'ss',minval(ss),maxval(ss)
write(0,*) 'aa',minval(aa),maxval(aa)
write(0,*) 'jj',minval(jj),maxval(jj)
write(0,*) 's0',minval(s0),maxval(s0)
write(0,*) 'a0',minval(a0),maxval(a0)
write(0,*) 'j0',minval(j0),maxval(j0)
write(0,*) " "

  !! . . Initialize module
  call wemig_init(ag,at,aw,ar,method); imgall=0.

  !! . . Loop over shot points
  do ii=1,as%n
     img=0.;rwf=0.;swf=0.;
     rwf=rwfall(:,:,ii)
     swf=swfall(:,:,ii)
     call wemig(rwf,swf,img,ss,aa,jj,mm,s0,a0,j0,ii)
     imgall(:,:,ii)=img
  end do

  !! . . Output to header file
  call houaxis(at,1)
  call houaxis(ag,2)
  call houaxis(as,3)
  call to_history("esize",4)
  call sep_write(imgall)

  call sep_close()
  deallocate(rwf,swf,img,rwfall,swfall,imgall)
end program NewRWEsp
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
