module rwe_wemig
  use sep
  use util_
  use fft

  implicit none
  integer, private :: nt,ng,nw,nr
  real,    private :: dt,dg,dw,dr
  real,    private :: ot,og,ow,or

  real,    private :: pi
  real,    private :: c1,c2
  real,    private :: sixth
  integer, private :: nloop
  integer, private :: method

  complex, allocatable, private :: lk(:),rk(:),ck(:)
  complex, allocatable, private :: a(:),b(:),c(:),u(:)
  real,                 private :: kmu,    knu,    kro 
  real,    allocatable, private ::  mu(  :  ),nu(  :  ),ro(  :  )
  real,    allocatable, private ::  m0(:,:,:),n0(:,:,:),r0(:,:,:)
  real,                 private :: okg,dkg

  complex, allocatable, private :: wfl(:),wtt(:)
  real,    allocatable, private :: msk(:),mtt(:)
  real,    allocatable, private :: dadt(:,:),djdt(:,:),m1(:,:)

  real,    allocatable, private :: tap(:)
  integer                       :: ntap
contains
  !----------------------------------------------------------------  
  subroutine wemig_init(ag,at,aw,ar,method_in)
    type(myaxis)     :: ag,at,aw,ar
    integer, intent(in) :: method_in
    pi=acos(-1.)
    method=method_in
    ng=ag%n;  nt=at%n;  nw=aw%n;          nr=ar%n
    dg=ag%d;  dt=at%d;  dw=aw%d * 2.*pi;  dr=ar%d
    og=ag%o;  ot=at%o;  ow=aw%o * 2.*pi;  or=ar%o
    call from_param("c1",c1,0.5)
    call from_param("c2",c2,0.)
    call from_param("sixth",sixth,0.122996)
    call from_param("nloop",nloop,3)
    allocate(lk(ng),rk(ng),ck(ng))
    allocate(a(ng),b(ng-1),c(ng-1),u(ng))
    allocate( dadt(nt,nr),djdt(nt,nr),m1(nt,nr) )
    kmu = 1./ dt         !! 1/ dt
    knu = 1./(2.*dg**2)  !! 1/( 2 dx^2)
    kro = 1./(dt*dg**2)  !! 1/(dt dx^2)
    allocate(m0(nt,ng,nr),n0(nt,ng,nr),r0(nt,ng,nr))
    allocate(mu(   ng   ),nu(   ng   ),ro(   ng   ))
    okg=  -pi/    dg
    dkg=2.*pi/(ng*dg)
    allocate(wfl(ng),wtt(ng),msk(ng),mtt(ng) )
    msk=1.; mtt=1.
    call from_param("ntap",ntap,1)
    allocate(tap(ng)); tap=1.
    call rwetaper_init()
write(0,*) 'INITIALIZED WE MIGRATION KERNAL'
  end subroutine wemig_init
  !----------------------------------------------------------------
  subroutine rwetaper_init()
    integer :: itap,jtap
    real    :: pi
    pi=acos(-1.)
    if(ntap>1) then
       do itap=1,ntap
          jtap = abs(ntap-itap)
          tap(   itap  ) = cos(pi/2.* jtap/(ntap-1))
       end do
       do itap=1,ntap
          jtap = abs(ntap-itap)
          tap(ng-itap+1) = cos(pi/2.* jtap/(ntap-1))
       end do
    end if
  end subroutine rwetaper_init
  !----------------------------------------------------------------
  subroutine rwetaper(dax)
    complex, pointer :: dax(:)
    dax = dax * tap
  end subroutine rwetaper
  !----------------------------------------------------------------
  !! WE migration driver
  subroutine wemig(rwf,swf,img,ss,aa,jj,mm,s0,a0,j0,ishot)
    integer :: iw,it,ig,ir,ishot
    real    :: w,sarg,rarg,saxmax
    real    :: ss(:,:),aa(:,:),jj(:,:),mm(:,:),s0(:,:),a0(:,:),j0(:,:)
    real,    pointer :: img(:,:)
    complex, pointer :: rax(:),sax(:),rwf(:,:),swf(:,:)
    sarg=-1. !! Source wavefield direction
    rarg=1.  !! Receiver wavefield direction
    
    !! . . Precompute required quantities
    dadt(1:nt-1,:) = ( a0(2:nt,:)-a0(1:nt-1,:) ) / dt
    djdt(1:nt-1,:) = ( j0(2:nt,:)-j0(1:nt-1,:) ) / dt
    m1  (1:nt-1,:) = ( log(j0(2:nt,:))-log(j0(1:nt-1,:))&
                      -log(a0(2:nt,:))+log(a0(1:nt-1,:)))/dt
    dadt(nt,:) = dadt(nt-1,:)
    djdt(nt,:) = djdt(nt-1,:)
    m1  (nt,:) = m1  (nt-1,:)

    !! Reference paremters
write(0,*) "dadt",minval(dadt),maxval(dadt)
write(0,*) "djdt",minval(djdt),maxval(djdt)
write(0,*) "m1  ",minval(  m1),maxval(  m1)
write(0,*) " "
    call wemig_psc_coef(ss,aa,jj,s0,a0,j0)
    call wemig_coef_report()

    do iw=1,nw; !! Frequency Loop        
       w=ow+(iw-1)*dw;
       write(0,*) "PSC",iw,nw,ishot,maxval(abs(img))
       rax => rwf(:,iw)
       do it=1,nt 
          wtt=rax; rax=0.
          do ir=1,nr !! Receiver wavefield
             wfl=wtt
             call wemig_phs(wfl,w,it,ir,s0,a0,j0,rarg)
             call wemig_ssf(wfl,w,it,ir,ss,aa,jj,s0,a0,j0,rarg)
             if (method .eq. 3) call wemig_fds(wfl,w,it,ir,rarg)
             call wemig_ref(wfl,  it,ir,mm,rax) 
          end do
          call rwetaper(rax)

          sax => swf(:,iw) 
          wtt=sax; sax=0.
          do ir=1,nr !! Source wavefield
             wfl=wtt
             call wemig_phs(wfl,w,it,ir,s0,a0,j0,sarg)
             call wemig_ssf(wfl,w,it,ir,ss,aa,jj,s0,a0,j0,sarg)
             if (method .eq. 3) call wemig_fds(wfl,w,it,ir,sarg)
             call wemig_ref(wfl,  it,ir,mm,sax)
          end do
          call rwetaper(sax)

          img(it,:)=img(it,:)+real(rax*conjg(sax))

       end do
    end do
  end subroutine wemig
  !----------------------------------------------------------------
  subroutine wemig_ref(wfl,it,ir,mm,dax)
    complex, pointer :: dax(:)
    complex          :: wfl(:)
    real             :: mm(:,:)
    integer          :: it,ir,ig,ii 

    where( nint(mm(it,:))==ir) 
       mtt=1.
    elsewhere
       mtt=0.
    end where

    do ii=1,nloop
       msk(1)=mtt(1)
       do ig=1+1,ng-1
          msk(ig) = 0.5*( mtt(ig-1)+mtt(ig+1) )
       end do
       msk(ng)=mtt(ng-1)
       mtt=msk
    end do

    dax = dax + msk*wfl
  end subroutine wemig_ref
  !----------------------------------------------------------------  
  !! Fourier-domain phase shift
  subroutine wemig_phs(dax,w,it,ir,s0,a0,j0,dir)
    complex :: dax(:)
    real    :: s0(:,:),a0(:,:),j0(:,:)
    real    :: w,kg,arg,dir
    integer :: ig,it,ir
    complex :: ikt

    call fth(.false.,.false.,dax)
    do ig=1,ng
       kg=okg+(ig-1)*dkg
       arg=1. - ( kg / (w*s0(it,ir)*j0(it,ir)) )**2 - m1(it,ir)**2/4
       if (arg < 0.0 ) then
          ikt = cmplx(   w*s0(it,ir)*a0(it,ir)*sqrt(-arg),0.)
       else
          ikt = cmplx( m1(it,ir)/2 ,dir*w*s0(it,ir)*a0(it,ir)*sqrt(+arg)   )
       end if
       dax(ig) = dax(ig) * cexp( ikt*(-dt))
    end do
    call fth( .true.,.false.,dax)

  end subroutine wemig_phs
  !----------------------------------------------------------------
  !! split-step Fourier
  subroutine wemig_ssf(dax,w,it,ir,ss,aa,jj,s0,a0,j0,dir) 
    complex :: dax(:)
    real    :: ss(:,:),aa(:,:),jj(:,:)
    real    :: s0(:,:),a0(:,:),j0(:,:)
    real    :: w,dir
    integer :: it,ir
    dax = dax * cexp(cmplx(0.,- dt * dir * & 
    ( (ss(it,:)-s0(it,ir) ) * &
               (w*a0(it,ir) - c1*m1(it,ir)**2/(4*w*s0(it,ir)**2*a0(it,ir))) + &
    (  aa(it,:)-a0(it,ir) ) * &
               (w*s0(it,ir) - m1(it,ir)/(4*w*s0(it,ir)*a0(it,ir)**3)*dadt(it,ir)) +&
    (  jj(it,:)-j0(it,ir) ) * &
               (              m1(it,ir)*djdt(it,ir)/(4*j0(it,ir)**2*w*s0(it,ir)*a0(it,ir))))))
    dax = dax * cexp( cmplx( - dt * ( dadt(it,ir)*(  aa(it,:)-a0(it,ir) )/a0(it,ir)**2 - &
                                      djdt(it,ir)*(  jj(it,:)-j0(it,ir) )/j0(it,ir)**2  ), 0. ))
  end subroutine wemig_ssf
  !----------------------------------------------------------------
  subroutine wemig_psc_coef(ss,aa,jj,s0,a0,j0)
    integer :: it,ir
    real    :: ss(:,:),aa(:,:),jj(:,:)
    real    :: s0(:,:),a0(:,:),j0(:,:)
    m0 = 1.
    r0 = 0. 
    do it=1,nt
       do ir=1,nr
          n0(it,:,ir) = &
          (-( ss(it,:) - s0(it,ir) ) * c1*a0(it,ir)/(j0(it,ir)**2*s0(it,ir)**2) &
           -( aa(it,:) - a0(it,ir) )               /(j0(it,ir)**2*s0(it,ir)   ) &
           +( jj(it,:) - j0(it,ir) ) *    a0(it,ir)/(j0(it,ir)**3*s0(it,ir)   ) )
       end do
    end do
    m0 = m0 * kmu
    n0 = n0 * knu
    r0 = r0 * kro
  end subroutine wemig_psc_coef
  !----------------------------------------------------------------  
  subroutine wemig_coef_report()
    write(0,*) minval(m0),"< mu <",maxval(m0)
    write(0,*) minval(n0),"< nu <",maxval(n0)
    write(0,*) minval(r0),"< ro <",maxval(r0)
  end subroutine wemig_coef_report
  !----------------------------------------------------------------  
  !! finite-differences solver
  subroutine wemig_fds(dax,w,it,ir,dir)
    real                  :: w,dir
    integer               :: ig,it,ir
    complex, dimension(:) :: dax
    mu = m0(it,:,ir) * dir
    nu =-n0(it,:,ir) / w 
    ro = r0(it,:,ir) * dir / w**2

    lk = cmplx(ro+sixth*mu, nu)
    rk = cmplx(ro+sixth*mu,-nu)
    ck = cmplx(         mu, 0.)

    !!             |                      |
    !!     b       |          a           |     c
    !!  ro + i nu  |  mu - 2(ro + i nu )  |  ro + i nu
    !!             |                      |
    a = ck - 2.*lk  
    b = lk(1  :ng-1)
    c = lk(1+1:ng  )

    do ig=2,ng-1
       u(ig)=           rk(ig) *(dax(ig-1)+dax(ig+1)) & 
       +    (ck(ig)-2.0*rk(ig))* dax(ig)
       !!             |                      | 
       !!  ro - i nu  |  mu - 2 (ro - i nu)  | ro - i nu
       !!             |                      | 
    end do
    ig= 1; u(ig)=(ck(ig)-2.0*rk(ig))*dax(ig) + rk(ig)*dax(ig+1)
    ig=ng; u(ig)=(ck(ig)-2.0*rk(ig))*dax(ig) + rk(ig)*dax(ig-1)

    call thr(a,b,c,u,ng)
    dax = u
  end subroutine wemig_fds
  !----------------------------------------------------------------
  !! tridiagonal solver
  subroutine thr(a,b,c,v,n)
    integer :: n
    complex :: a(n),b(n-1),c(n-1),v(n)
    integer :: i

    b(1)=b(1)/a(1)
    v(1)=v(1)/a(1)
    do i=2,n
       a(i)=a(i)-c(i-1)*b(i-1)
       if (i<n) then 
          b(i)=b(i)/a(i)
       end if
       v(i)=(v(i)-c(i-1)*v(i-1))/a(i)
    end do
    do i=n-1,1,-1
       v(i)=v(i)-b(i)*v(i+1)
    end do
  end subroutine thr
  !----------------------------------------------------------------
  subroutine attribute(x)
    complex :: x(:)
    write(0,*) minval( real(x)),maxval( real(x))
    write(0,*) minval(aimag(x)),maxval(aimag(x))
  end subroutine attribute
  !----------------------------------------------------------------
end module rwe_wemig
