module rwe_ab
  use sep 
  use util_
  use lloyd_mod
  implicit none
  integer, private :: nt,ng
  real,    private :: ot,og
  real,    private :: dt,dg
  integer, private :: nrefs,nrefa,nrefj,niter
contains
  subroutine rweab_init(at,ag,nrefs_in,nrefa_in,nrefj_in,niter_in)
    type(myaxis)  :: at,ag
    integer       :: nrefs_in,nrefa_in,nrefj_in,niter_in
    nt=at%n;    ng=ag%n
    dt=at%d;    dg=ag%d 
    ot=at%o;    og=ag%o
    nrefs=nrefs_in;nrefa=nrefa_in;nrefj=nrefj_in;niter=niter_in
    write(0,*) 'TAU  ',nt,dt,ot
    write(0,*) 'GAMMA',ng,dg,og
  end subroutine rweab_init
  !###########################################
  subroutine rweab(rays,ss,aa,jj,s0,a0,j0,mm)
    character(len=128)   :: name
    integer              :: nref,it,ig,is,ia,ij,ii
    integer, allocatable :: ix(:)
    real                 :: gx,gz
    real                 :: ss(:,:),aa(:,:),jj(:,:),mm(:,:)
    real                 :: s0(:,:),a0(:,:),j0(:,:)
    real,    allocatable :: ms(:,:),ma(:,:),mj(:,:)
    real,    allocatable :: x(:,:),z(:,:),stodo(:),atodo(:),jtodo(:)
    real,    allocatable :: near_s(:),near_a(:),near_j(:)
    complex              :: rays(:,:)
    nref=nrefs*nrefa*nrefj
    allocate(  x(nt,ng),z(nt,ng))
    allocate(  ix(1    ))
    allocate(  ms(nt,ng),ma(nt,ng),mj(nt,ng))
    allocate(  near_s(nref),near_a(nref),near_j(nref) )
    allocate(  stodo(ng),atodo(ng),jtodo(ng) )

    x =  real(rays)
    z = aimag(rays)
    do it=1,nt-1 !! . . ALPHA
       do ig=1,ng
          gx = (x(it+1,ig)-x(it,ig))/dt
          gz = (z(it+1,ig)-z(it,ig))/dt
          aa(it,ig) = sqrt(gx*gx+gz*gz)
       end do
    end do
    aa(nt,:) = aa(nt-1,:)

    do it=1,nt !! . . Jacobian
       do ig=1+1,ng-1
          gx = (x(it,ig+1)-x(it,ig-1))/2./dg
          gz = (z(it,ig+1)-z(it,ig-1))/2./dg
          jj(it,ig) = sqrt(gx*gx+gz*gz)
       end do
    end do
    jj(:,1)  = jj(:,2)
    jj(:,ng) = jj(:,ng-1)
    jj(1,:)  = jj(2,:)

    mm=1.; ms=1.; ma=1.; mj=1.
    !! compute reference velocities
    do it=1,nt
       near_s=0.; near_a=0.;near_j=0.; stodo=0.;atodo=0.;jtodo=0.
       stodo = ss(it,:)
       atodo = aa(it,:)
       jtodo = jj(it,:)

       call init_lloyd()
       call lloyd_go(stodo,near_s,nrefs,niter)
       call init_lloyd()
       call lloyd_go(atodo,near_a,nrefa,niter)
       call init_lloyd()
       call lloyd_go(jtodo,near_j,nrefj,niter)

       !! . . Make sure no zeros
       if (nrefs .gt. 1) then
          do is=2,nrefs
             if( near_s(is) .eq. 0. ) near_s(is)=near_s(is-1)
          end do
       end if
       if (nrefs .gt. 1) then
          do ia=2,nrefa
             if( near_a(ia) .eq. 0. ) near_a(ia)=near_a(ia-1)
          end do
       end if
       if (nrefj .gt. 1) then
          do ij=2,nrefj
             if( near_j(ij) .eq. 0. ) near_j(ij)=near_j(ij-1)
          end do
       end if

       ii=1
       do is=1,nrefs
          do ia=1,nrefa
             do ij=1,nrefj
                s0(it,ii)=near_s(is)
                a0(it,ii)=near_a(ia)
                j0(it,ii)=near_j(ij)
                ii=ii+1
             end do
          end do
       end do
       !! mask for a
       do ia=1,ng 
          ix=minloc( abs( ss(it,ia) - near_s ) )
          ms(it,ia)=ix(1)
          ix=minloc( abs( aa(it,ia) - near_a ) )
          ma(it,ia)=ix(1)
          ix=minloc( abs( jj(it,ia) - near_j ) )
          mj(it,ia)=ix(1)
       end do
    end do

    !! reference medium mask
    ii=1
    do is=1,nrefs
       do ia=1,nrefa
          do ij=1,nrefj
             where(ms==is .and. ma==ia .and. mj==ij)
                mm=ii
             end where
             ii=ii+1
          end do
       end do
    end do
    deallocate(x,z,ix,ms,ma,mj,near_s,near_a,near_j,stodo,atodo,jtodo)
  end subroutine rweab
end module rwe_ab








