module cc2rc
  implicit none
  integer, private :: nz,nx,nt,ng
  real,    private :: dz,dx,dt,dg
  real,    private :: oz,ox,ot,og
contains
  subroutine cc2rc_init(az,ax,at,ag) 
    type(myaxis) ::     az,ax,at,ag
    nz=az%n;    nx=ax%n;    nt=at%n;    ng=ag%n
    dz=az%d;    dx=ax%d;    dt=at%d;    dg=ag%d 
    oz=az%o;    ox=ax%o;    ot=at%o;    og=ag%o
  end subroutine cc2rc_init
!##############################################
  subroutine cc2rc(rays,mapCC,mapRC,adj)
    implicit none
    logical          :: adj
    real,    pointer :: mapCC(:,:)
    real,    pointer :: mapRC(:,:)
    complex, pointer :: rays(:,:)

    integer      :: ix,iz,it,ig,jx,jz
    real         :: x,z,xmin,xmax,zmin,zmax
    real         :: fx,fz,w00,w01,w10,w11, sumw
    mapCC=0.
    mapRC=0.
    xmin = ox;    xmax = ox + (nx-1)*dx
    zmin = oz;    zmax = oz + (nz-1)*dz
    do ig=1,ng
       do it=1,nt
          z=aimag(rays(it,ig))
          x= real(rays(it,ig))
          z = min(z,zmax)
          z = max(z,zmin)
          x = min(x,xmax)
          x = max(x,xmin)
          iz=floor((z-oz)/dz)+1
          ix=floor((x-ox)/dx)+1

          iz = min(iz  ,nz)
          iz = max(iz  , 1  )
          jz = min(iz+1,nz)

          ix = min(ix  ,nx)
          ix = max(ix  , 1  )
          jx = min(ix+1,nx)

          fz= (z-oz-(iz-1)*dz)
          fx= (x-ox-(ix-1)*dx)
          fz = max(0., fz)
          fz = min(1., fz)
          fx = max(0., fx)
          fx = min(1., fx)

          !! 00   01       --> x
          !!    .         |
          !! 10   11      v
          !!               z
          w00=(1.-fx)*(1.-fz)
          w10=(   fx)*(1.-fz)
          w01=(1.-fx)*(   fz)
          w11=(   fx)*(   fz)

          !! normalize the sum of the weights
          sumw = w00+w10+w01+w11
          w00 = w00 / sumw
          w10 = w10 / sumw
          w01 = w01 / sumw
          w11 = w11 / sumw

          if(adj) then
             mapCC(iz,ix) = mapCC(iz,ix) + w00 * mapRC(it,ig)
             mapCC(jz,ix) = mapCC(jz,ix) + w10 * mapRC(it,ig)
             mapCC(iz,jx) = mapCC(iz,jx) + w01 * mapRC(it,ig)
             mapCC(jz,jx) = mapCC(jz,jx) + w11 * mapRC(it,ig)
          else
             mapRC(it,ig) = mapRC(it,ig) + w00 * mapCC(iz,ix)
             mapRC(it,ig) = mapRC(it,ig) + w10 * mapCC(jz,ix)
             mapRC(it,ig) = mapRC(it,ig) + w01 * mapCC(iz,jx)
             mapRC(it,ig) = mapRC(it,ig) + w11 * mapCC(jz,jx)
          end if

       end do
    end do
    call deallocate(0)
  end subroutine cc2rc
end module cc2rc






