program RWEab_
  use sep
  use util_
  use lloyd_mod
  implicit none

  integer, allocatable :: ix(:)
  complex, allocatable :: rays(:,:)
  real,    allocatable :: x(:,:),z(:,:)
  real,    allocatable :: h1(:,:),h2(:,:)
  real,    allocatable :: ss(:,:),aa(:,:),bb(:,:)
  real,    allocatable :: ma(:,:),mb(:,:)
  real,    allocatable :: atodo(:),btodo(:)  
  complex, allocatable :: ab(:,:)
  real,    allocatable :: mm(:,:)
  real,    allocatable :: near_a(:),near_b(:)

  type(myaxis) ::at,ag,ar
  integer      ::it,ig,niter
  real         ::gx,gz
  character(len=128):: name
  integer          :: naref, nbref
  real             :: mina,maxa,dela
  real             :: minb,maxb,delb
  real             :: a0,b0
  integer          :: ia,ib,ii

  call sep_init()
  call initpar()
  call doc('./lloyd.rs90')
  call from_param("naref",naref,1);  write(0,*) "naref=",naref
  call from_param("nbref",nbref,1);  write(0,*) "nbref=",nbref
  call hinaxis(at,1)
  call hinaxis(ag,2)
  call to_history("esize",4)
  call to_history("n3",5)
  call from_param("niter",niter)

  ar%n=naref*nbref
  ar%o=0.
  ar%d=1.
  call aouaxis(at,1,'abref')
  call aouaxis(ar,2,'abref')
  call auxputch("esize",'i',8,'abref')
  allocate(rays(at%n,ag%n))
  allocate(  h1(at%n,ag%n))
  allocate(  h2(at%n,ag%n))
  allocate(   x(at%n,ag%n))
  allocate(   z(at%n,ag%n))
  allocate(  ix(1        ))
  allocate(  ss(at%n,ag%n))
  allocate(  aa(at%n,ag%n))
  allocate(  bb(at%n,ag%n))
  allocate(  ab(at%n,naref*nbref))
  allocate(  mm(at%n,ag%n))
  allocate(  ma(at%n,ag%n))
  allocate(  mb(at%n,ag%n))
  allocate(  near_a(naref),near_b(nbref) )
  allocate(  atodo(ag%n),btodo(ag%n) )
  !----------------------------------------------------------------
  call sep_read(rays)
  z=aimag(rays)
  x= real(rays)
  call sreed("slo",ss,size(ss)*4)
  !----------------------------------------------------------------

  !! h1 == alpha
  do it=1,at%n-1
     do ig=1,ag%n
        gx = (x(it+1,ig)-x(it,ig))/at%d
        gz = (z(it+1,ig)-z(it,ig))/at%d
        h1(it,ig) = sqrt(gx*gx+gz*gz)
     end do
  end do
  h1(at%n,:) = h1(at%n-1,:)

  !! h2 == J
  do it=1,at%n
     do ig=1+1,ag%n-1
        gx = (x(it,ig+1)-x(it,ig-1))/2./ag%d
        gz = (z(it,ig+1)-z(it,ig-1))/2./ag%d
        h2(it,ig) = sqrt(gx*gx+gz*gz)
     end do
  end do
  h2(:,1)    = h2(:,2)
  h2(:,ag%n) = h2(:,ag%n-1)
  h2(1,:)    = h2(2,:)


  aa=ss*h1
  bb=h1/h2
  mm=1.
  ma=1.
  mb=1.

  !! compute reference velocities
  do it=1,at%n
     near_a=0.; near_b=0.; atodo=0.;btodo=0.;

     call init_lloyd()
     atodo = aa(it,:)
     call lloyd_go(atodo,near_a,naref,niter)

     call init_lloyd()
     btodo = log10(bb(it,:))
     call lloyd_go(btodo,near_b,nbref,niter)
     near_b = 10**(near_b) 

     !! . . Make sure no zeros
     if (naref .gt. 1) then
        do ia=2,naref
           if( near_a(ia) .eq. 0. ) near_a(ia)=near_a(ia-1)
        end do
     end if
     if (nbref .gt. 1) then
        do ib=2,nbref
           if( near_b(ib) .eq. 0. ) near_b(ib)=near_b(ib-1)
        end do
     end if

     !     write(0,*) it,at%n
     !     write(0,*) near_a
     !     write(0,*) near_b
     ii=1
     do ia=1,naref
        do ib=1,nbref
           ab(it,ii)=cmplx(near_a(ia),near_b(ib))
           ii=ii+1
        end do
     end do

     !! mask for a

     do ia=1,ag%n 
        ix=minloc( abs( aa(it,ia) - near_a ) )
        ma(it,ia)=ix(1)
        ix=minloc( abs( bb(it,ia) - near_b ) )
        mb(it,ia)=ix(1)
     end do

  end do

  !! reference medium mask
  ii=1
  do ia=1,naref
     do ib=1,nbref
        where(ma==ia .and. mb==ib)
           mm=ii
        end where
        ii=ii+1
     end do
  end do

  !! write output
  call sep_close()
  call sep_write(aa)
  call sep_write(bb)
  call sep_write(mm)
  call sep_write(ma)
  call sep_write(mb)
  call sep_write(ab,'abref')

!  call exit()
end program RWEab_









