program TPSM 
  !! Teleseismic Phase-shift migration
  !! Jeff Shragge February, 2003
  !! Teleseismic Phaseshift Migration
  use sep
  use newgazdag

  implicit none
  integer                              :: n1,n2,n3,esize,i1,i2,ii,pad2
  integer                              :: m1,m2,imode
  integer, dimension(:),    allocatable:: modestodo
  real                                 :: dz,dt,dx
  real                                 :: o1,o2,d1,d2
  real                                 :: p1,p2,xdir
  real,    dimension(:,:,:),allocatable:: modl,data
  complex, dimension(:,:),allocatable  :: cmodl,cdata
  real,    dimension(:)  ,allocatable  :: avel,bvel,uvel,dvel
  logical                              :: adj,add

  call sep_init()
  call from_history( "n1", n1         ) ! Num Time Samples
  call from_history( "n2", n2         ) ! Num Offsets
  call from_history( "n3", n3         ) ! Num Offsets
  call from_history( "o1", o1         ) ! Time offset
  call from_history( "o2", o2         ) ! Space offset
  call from_history( "d1", d1         ) ! Temporal sampling rate
  call from_history( "d2", d2         ) ! Spatial sampling rate
  call from_history( "esize",esize,4  ) ! Length of data
  call from_param  ( "dt",dt          ) ! Data time sample step in model
  call from_history( "p1",p1          ) ! Slowness of incident plane wave
  call from_history( "p2",p2          ) ! Slowness of incident plane wave
  call from_param  ( "adj",adj,.true. ) ! Adjoint call
  call from_param  ( "dz" ,dz,1.      )
  call from_param  ( "dx" ,dx,5.      )
  call to_history  ( "n1", n1, "model"  )
  call to_history  ( "n2", n2, "model"  )
  call to_history  ( "n3", 7 , "model"  )
  call to_history  ( "o1", o1, "model"  )
  call to_history  ( "o2", o2, "model"  )
  call to_history  ( "o3", 1 , "model"  )
  call to_history  ( "d2", d2, "model"  )
  call to_history  ( "d3", 1 , "model"  )

  if (adj) then
     call to_history('d1',dz,"model") !! Write the proper depth step in the output header file
  else
     call to_history('d1',dt,"model") !! Write the proper depth step in the output header file
  end if
  call hclose() 

  m1=pad2(n1)
  m2=pad2(n2)
  allocate ( avel(m1), bvel(m1), uvel(m1), dvel(m1) )
  allocate( cdata(m1,m2)  , cmodl(m1,m2)  )
  allocate( data(n1,n2,n3), modl(n1,n2,7) )
  allocate( modestodo(7) )
  call sep_read( avel,"avel" )
  call sep_read( bvel,"bvel" )

  !! Ensure we have real data
  if(esize.ne.4) call erexit('esize !=4')

  !! Read in Data or Model
  if (adj) then
     call sep_read( data )
  else
     call sep_read( modl )
  end if
  modestodo(1)=1; modestodo(2)=2; modestodo(3)=3; modestodo(4)=4 
  modestodo(5)=5; modestodo(6)=6; modestodo(7)=7

  do imode=1,7
     !! Read in Data or Model
     if (adj) then
        if      (modestodo(imode)==1) then
           cdata(1:n1,1:n2)=cmplx(data(1:n1,1:n2,3),0.) 
           uvel=avel;dvel=0.*avel;xdir=1.
        else if (modestodo(imode)==2) then
           cdata(1:n1,1:n2)=cmplx(data(1:n1,1:n2,1),0.) 
           uvel=bvel;dvel=0.*avel;xdir=atan2(-p2,p1)/abs(atan2(-p2,p1))
        else if (modestodo(imode)==3) then
           cdata(1:n1,1:n2)=cmplx(data(1:n1,1:n2,3),0.) 
           uvel=avel;  dvel=avel; xdir=1.
        else if (modestodo(imode)==4) then
           cdata(1:n1,1:n2)=cmplx(data(1:n1,1:n2,1),0.) 
           uvel=bvel;  dvel=avel;  xdir=atan2(-p2,p1)/abs(atan2(-p2,p1))
        else if (modestodo(imode)==5) then
           cdata(1:n1,1:n2)=cmplx(data(1:n1,1:n2,3),0.) 
           uvel=bvel;  dvel=bvel; xdir=1.
        else if (modestodo(imode)==6) then
           cdata(1:n1,1:n2)=cmplx(data(1:n1,1:n2,1),0.) 
           uvel=bvel;  dvel=bvel;  xdir=-atan2(-p2,p1)/abs(atan2(-p2,p1))
        else if (modestodo(imode)==7) then
           cdata(1:n1,1:n2)=cmplx(data(1:n1,1:n2,2),0.) 
           uvel=bvel;  dvel=bvel;  xdir=-atan2(-p2,p1)/abs(atan2(-p2,p1))
        end if
     else
        cmodl(1:n1,1:n2)=cmplx(modl(1:n1,1:n2,modestodo(imode)),0.)
        if      (modestodo(imode)==1) then
           uvel=avel;dvel=0.*avel;xdir=1.
        else if (modestodo(imode)==2) then
           uvel=bvel;dvel=0.*avel;xdir=atan2(p2,p1)/abs(atan2(p2,p1))
        else if (modestodo(imode)==3) then
           uvel=avel;  dvel=avel; xdir=1.
        else if (modestodo(imode)==4) then
           uvel=bvel;  dvel=avel;  xdir=atan2(p2,p1)/abs(atan2(p2,p1))
        else if (modestodo(imode)==5) then
           uvel=avel;  dvel=bvel;  xdir=1.
        else if (modestodo(imode)==6) then
           uvel=bvel;  dvel=bvel;  xdir=-atan2(p2,p1)/abs(atan2(p2,p1))
        else if (modestodo(imode)==7) then
           uvel=bvel;  dvel=bvel;  xdir=-atan2(p2,p1)/abs(atan2(p2,p1))
        end if
     end if
     !! Do phase-shift migration
     call newgazdag1_init(n1,m1,m2,dt,d2,dz,p1,p2,xdir,modestodo(imode),dx)
     call newgazdag1(adj,.false.,cdata,cmodl,uvel,dvel)
     if (adj) then
        modl(1:n1,1:n2,modestodo(imode))=real(cmodl(1:n1,1:n2));cmodl=0.;cdata=0.;
     else
        if      (modestodo(imode)==1) then
           data(1:n1,1:n2,3)=data(1:n1,1:n2,3)+real(cdata(1:n1,1:n2)) 
        else if (modestodo(imode)==2)  then
           data(1:n1,1:n2,1)=data(1:n1,1:n2,1)+real(cdata(1:n1,1:n2)) 
        else if (modestodo(imode)==3)  then
           data(1:n1,1:n2,3)=data(1:n1,1:n2,3)+real(cdata(1:n1,1:n2)) 
        else if (modestodo(imode)==4)  then
           data(1:n1,1:n2,1)=data(1:n1,1:n2,1)+real(cdata(1:n1,1:n2)) 
        else if (modestodo(imode)==5)  then
           data(1:n1,1:n2,3)=data(1:n1,1:n2,3)+real(cdata(1:n1,1:n2)) 
        else if (modestodo(imode)==6)  then
           data(1:n1,1:n2,1)=data(1:n1,1:n2,1)+real(cdata(1:n1,1:n2)) 
        else if (modestodo(imode)==7)  then
           data(1:n1,1:n2,2)=data(1:n1,1:n2,2)+real(cdata(1:n1,1:n2)) 
        end if
        cdata=0.; cmodl=0.
     end if
  end do

  !! Write out data or model
  if (adj) then
     call sep_write(modl,"model")
  else
     call sep_write(data)
  end if

end program TPSM

