PROGRAM LAPLACE2D
  !     ===========================================================
  !     2-dimensionnal Laplace's equation
  !     ==========================================================
  use SEP
  implicit none
  INTEGER, PARAMETER         :: nx=64,nz=64
  INTEGER, PARAMETER         :: XMIN = 2, XMAX = nx-1
  INTEGER, PARAMETER         :: YMIN = 2, YMAX = nz-1

  INTEGER, PARAMETER         :: LXMIN = XMIN-1, LXMAX = XMAX-1
  INTEGER, PARAMETER         :: LYMIN = YMIN,   LYMAX = YMAX
  INTEGER, PARAMETER         :: RXMIN = XMIN+1, RXMAX = XMAX+1
  INTEGER, PARAMETER         :: RYMIN = YMIN,   RYMAX = YMAX
  INTEGER, PARAMETER         :: UXMIN = XMIN,   UXMAX = XMAX
  INTEGER, PARAMETER         :: UYMIN = YMIN-1, UYMAX = YMAX-1
  INTEGER, PARAMETER         :: DXMIN = XMIN,   DXMAX = XMAX
  INTEGER, PARAMETER         :: DYMIN = YMIN+1, DYMAX = YMAX+1
  REAL, DIMENSION(1:nx,1:nz) :: PHIOLD,PHINEW
  REAL                       :: MAXDELTA
  REAL, PARAMETER            :: CONVERGENCE = 0.001
  INTEGER                    :: ITERATIONS = 0
  INTEGER                    :: I,J,ii

  !! . . Output file setup
  call sep_init()
  call from_history("n1",nz             ) 
  call from_history("n2",nx             )  
  call from_history("d1",dz             ) 
  call from_history("d2",dx             )  
  call from_history("o1",oz             ) 
  call from_history("o2",ox             )  
  call to_history("n1",nz)
  call to_history("n2",nx)
  call to_history("o1",0)
  call to_history("o2",0)
  call to_history("d1",1)
  call to_history("d2",1)
  call hclose()


  !! . . Output file setup
  PHINEW(1,1:nz)     = (/ (ii,ii=1,nz) /)
  PHINEW(nx,1:nz) = (/ (ii,ii=1,nz) /)
  PHINEW(1:nx,1)     = 1
  PHINEW(1:nx,nz) = nx
  PHINEW(XMIN:XMAX,YMIN:YMAX) = 0

 
  JACOBI_ITERATION: DO

     ITERATIONS = ITERATIONS + 1 
     PHIOLD = PHINEW        

     PHINEW(XMIN:XMAX,YMIN:YMAX)=0.25*(PHIOLD(LXMIN:LXMAX,LYMIN:LYMAX)+ \
       PHIOLD(RXMIN:RXMAX,RYMIN:RYMAX)+PHIOLD(UXMIN:UXMAX,UYMIN:UYMAX)+ \
       PHIOLD(DXMIN:DXMAX,DYMIN:DYMAX))

     MAXDELTA=MAXVAL(ABS(PHINEW(XMIN:XMAX,YMIN:YMAX)-PHIOLD(XMIN:XMAX,YMIN:YMAX)))

     IF (MAXDELTA < CONVERGENCE) EXIT JACOBI_ITERATION

  END DO JACOBI_ITERATION

!  print *,' ' 
!  print *,'*******************************'
!  print *,'     2-D LAPLACES EQUATION '
!  print *,' PROBLEM SIZE: X, Y =', nx,nz
!  print *,'*******************************'
!  print *,' CONVERGENCE LIMIT = ', CONVERGENCE
!  print *,  ITERATIONS,' ITERATIONS'
!  print *,' FINAL RESULT = '
!  print *,'*******************************'
!  DO I=1,nx
!     DO J=1,nz
!        print *,'PHI(',I,',',J,')= ',PHINEW(I,J)
!     END DO
!  END DO
  
  call sep_write(phinew)

  STOP
END PROGRAM LAPLACE2D



