module exact_mod
  use sep 
  use igrad1
  implicit none
  integer,private                         :: nx,nz,nw
  real   ,private                         :: dx,dz,dw,ox,oz,ow,ds
contains
  subroutine exact_mod_init(nz_in,nx_in,nw_in,dz_in,dx_in,dw_in,oz_in,ox_in,ow_in,ds_in)
    integer                           :: nz_in,nx_in,nw_in
    real                              :: dz_in,dx_in,dw_in,ox_in,oz_in,ow_in,ds_in
    nz=nz_in;nx=nx_in;nw=nw_in
    dz=dz_in;dx=dx_in;dw=dw_in
    oz=oz_in;ox=ox_in;ow=ow_in;ds=ds_in
  end subroutine exact_mod_init
  !####################################################################
  subroutine get_gradx(adj,add,data,PoverP_dx,eps)
    integer                               :: iw,iz,stat
    real                                  :: eps
    real,    dimension(:),allocatable     :: ctemp,rtemp
    real,    dimension(:,:,:)             :: PoverP_dx
    complex, dimension(:),allocatable     :: ctmp
    complex, dimension(:,:,:)             :: data
    logical                               :: adj,add
    allocate(ctemp(nx),rtemp(nx),ctmp(nx) )
    do iw=1,nw
       do iz=1,nz
          stat=igrad1_lop(adj,add,real( data(iz,:,iw)),rtemp )
          stat=igrad1_lop(adj,add,aimag(data(iz,:,iw)),ctemp )
          ctmp=cmplx(rtemp,ctemp)
          PoverP_dx(iz,:,iw)=aimag(ctmp/(data(iz,:,iw)+eps))   
       end do
    end do
    deallocate(ctemp,rtemp,ctmp)
  end subroutine get_gradx
  !######################################################################
  subroutine get_gradz(adj,add,data,PoverP_dz,eps)
    integer                               :: iw,ix,stat
    real                                  :: eps
    real,    dimension(:),allocatable     :: ctemp,rtemp
    real,    dimension(:,:,:)             :: PoverP_dz
    complex, dimension(:),allocatable     :: ctmp
    complex, dimension(:,:,:)             :: data
    logical                               :: adj,add
    allocate( ctemp(nz),rtemp(nz),ctmp(nz) )
    do iw=1,nw 
       do ix=1,nx
         stat=igrad1_lop(adj,add,real( data(:,ix,iw)),rtemp )
         stat=igrad1_lop(adj,add,aimag(data(:,ix,iw)),ctemp )
         !ctmp(1:nz)=cmplx(rtemp(1:nz),ctemp(1:nz))
         PoverP_dz(:,ix,iw)=aimag( cmplx(rtemp,ctemp) / (data(:,ix,iw)+eps) )
       end do
    end do
    deallocate(ctemp,rtemp,ctmp)
  end subroutine get_gradz
  !######################################################################
  subroutine Run_exact(Rayout,adj,add,eps,maxnum,x,z,xtodo,ztodo)
    integer                                :: ix,iz,iw,ip,maxnum,iix,iiz,cnt,i2
    integer, dimension(:)    ,allocatable  :: xtmp,ztmp
    real                                   :: eps,dx_step,dz_step,tval1,tval2,tval3
    real,    dimension(:)                  :: xtodo,ztodo,x,z
    real,    dimension(:,:,:),allocatable  :: PoverP_dx,PoverP_dz,xray,zray
    real,    dimension(:,:,:),allocatable  :: d2_x,d2_z,d3_x,d3_z
    complex, dimension(:,:,:)              :: Rayout
    complex, dimension(:,:,:),allocatable  :: data
    logical                                :: adj,add  
    allocate( PoverP_dx(nz,nx,nw),PoverP_dz(nz,nx,nw), data(nz,nx,nw))
    allocate( d2_x(nz,nx,nw),d2_z(nz,nx,nw),d3_x(nz,nx,nw),d3_z(nz,nx,nw) )
    allocate( xtmp(size(xtodo)), ztmp(size(ztodo)) )
    allocate( xray(maxnum,size(xtodo),nw), zray(maxnum,size(xtodo),nw))
    call sep_read( data )
    !! . . Compute Gradients
    call get_gradx(adj,add,data,PoverP_dx,eps)
    call get_gradz(adj,add,data,PoverP_dz,eps)
    loop_freq: do iw=1,nw
       write(0,*) 'WORKING ON FREQUENCY NUM',iw,' OF ',nw
       xray(1,:,iw)=xtodo
       zray(1,:,iw)=ztodo
       loop_points: do ip=1,size(xtodo)
          write(0,*) 'WORKING ON RAY LOC',ip,' OF ',size(xtodo)
          xtmp=minloc(abs(x-xray(1,ip,iw))); iix=xtmp(1)  
          ztmp=minloc(abs(z-zray(1,ip,iw))); iiz=ztmp(1)
          do cnt=2,maxnum ! . . New Step  
             if (iix .ge. size(x)-2*ceiling(abs(ds/dx)) &
            .or. iix .le. 2                             &
            .or. iiz .ge. size(z)-2*ceiling(abs(ds/dx)) &
            .or. iiz .le. 2                             ) then
                xray(cnt:,ip,iw)=xray(cnt-1,ip,iw)
                zray(cnt:,ip,iw)=zray(cnt-1,ip,iw)
                continue
             else
                !! . . Calculate new step in X
                dx_step=-ds*PoverP_dx(iiz,iix,iw)/sqrt(PoverP_dx(iiz,iix,iw)**2+PoverP_dz(iiz,iix,iw)**2+eps)
                dz_step=-ds*PoverP_dz(iiz,iix,iw)/sqrt(PoverP_dx(iiz,iix,iw)**2+PoverP_dz(iiz,iix,iw)**2+eps)
                xray(cnt,ip,iw)=xray(cnt-1,ip,iw)+dx_step
                zray(cnt,ip,iw)=zray(cnt-1,ip,iw)+dz_step
             end if
          end do 
       end do loop_points
    end do loop_freq
    Rayout=cmplx(xray,zray)
  end subroutine Run_exact
end module exact_mod
