module exact_r2_mod
  use sep 
  use igrad1
  implicit none
  integer,private                         :: nx,nz,nw
  real   ,private                         :: dx,dz,dw,ox,oz,ow,dtau,mxint
contains
  subroutine exact_mod_init(nz_in,nx_in,nw_in,dz_in,dx_in,dw_in,oz_in,ox_in,ow_in,dtau_in,mxint_in)
    integer                            :: nz_in,nx_in,nw_in
    real                               :: dz_in,dx_in,dw_in,ox_in,oz_in,ow_in,dtau_in,mxint_in
    nz=nz_in; nx=nx_in; nw=nw_in; dz=dz_in; dtau=dtau_in; dx=dx_in 
    oz=oz_in; ox=ox_in; ow=ow_in; dw=dw_in; mxint=mxint_in;
  end subroutine exact_mod_init
  !####################################################################
  subroutine get_gradx(adj,add,data,PoverP_dx,eps)
    integer                               :: iw,iz,stat
    real                                  :: eps
    real,    dimension(:),allocatable     :: ctemp,rtemp
    real,    dimension(:,:)               :: PoverP_dx
    complex, dimension(:),allocatable     :: ctmp
    complex, dimension(:,:)               :: data
    logical                               :: adj,add
    allocate(ctemp(nx),rtemp(nx),ctmp(nx) )
    do iz=1,nz
       stat=igrad1_lop(adj,add,real( data(iz,:)),rtemp )
       stat=igrad1_lop(adj,add,aimag(data(iz,:)),ctemp )
       ctmp=cmplx(rtemp,ctemp)
       PoverP_dx(iz,:)=aimag(ctmp/(data(iz,:)+eps))   
    end do
    deallocate(ctemp,rtemp,ctmp)
  end subroutine get_gradx
  !######################################################################
  subroutine get_gradz(adj,add,data,PoverP_dz,eps)
    integer                               :: iw,ix,stat
    real                                  :: eps
    real,    dimension(:),allocatable     :: ctemp,rtemp
    real,    dimension(:,:)               :: PoverP_dz
    complex, dimension(:),allocatable     :: ctmp
    complex, dimension(:,:)               :: data
    logical                               :: adj,add
    allocate( ctemp(nz),rtemp(nz),ctmp(nz) )
    do ix=1,nx
       stat=igrad1_lop(adj,add,real( data(:,ix)),rtemp )
       stat=igrad1_lop(adj,add,aimag(data(:,ix)),ctemp )
       ctmp=cmplx(rtemp,ctemp)
       PoverP_dz(:,ix)=aimag(ctmp/(data(:,ix)+eps))
    end do
    deallocate(ctemp,rtemp,ctmp)
  end subroutine get_gradz
!!$  !######################################################################
!!$  subroutine Run_exact_interp(vel,data,Rayout,adj,add,eps,maxnum,x,z,xtodo,ztodo,it)
!!$    integer                                :: ix,iz,iw,ip,maxnum,iix,iiz,cnt,i2,it,i1
!!$    integer                                :: AllocateStatus
!!$    real                                   :: eps,dx_step,dz_step,mxdist,ds
!!$    real,    dimension(:)                  :: xtodo,ztodo,x,z,xin,zin
!!$    real,    dimension(:,:)                :: vel
!!$    real,    dimension(:,:),allocatable    :: PoverP_dx,PoverP_dz,bigx,bigz
!!$    real,    dimension(:,:),allocatable    :: xray,zray
!!$    complex, dimension(:,:)                :: Rayout
!!$    complex, dimension(:,:)                :: data
!!$    logical                                :: adj,add  
!!$    allocate( PoverP_dx(nz,nx),PoverP_dz(nz,nx), xin(size(x)), zin(size(z)) )
!!$    allocate( xray(maxnum,size(xtodo)) , zray(maxnum,size(xtodo)) )
!!$    xray(1,1:size(xtodo))=xtodo;    zray(1,1:size(xtodo))=ztodo
!!$    xin=x; zin=z;    mxdist=mxint*sqrt(dx**2+dz**2)  
!!$
!!$    call get_gradx(adj,add,data,PoverP_dx,eps)
!!$    call get_gradz(adj,add,data,PoverP_dz,eps)
!!$    
!!$    loop_steps:  do cnt=2,maxnum ! . . New Step  
!!$       write(0,*) 'Working on step ',cnt,' of ', maxnum ,it
!!$       loop_points: do ip=1,size(xray,2)
!!$          call get_min_loc(xin,xray,iix,cnt-1,ip,0.)
!!$          call get_min_loc(zin,zray,iiz,cnt-1,ip,0.)
!!$          ds=vel(iiz,iix)*dtau
!!$          if ( iix .ge. nx-2*ceiling(abs(ds/dx))  &
!!$          .or. iix .le. 2*   ceiling(abs(ds/dx))  &
!!$          .or. iiz .ge. nz-2*ceiling(abs(ds/dz))  &
!!$          .or. iiz .le. 2*   ceiling(abs(ds/dx))  ) then
!!$             xray(cnt:,ip)=xray(cnt-1,ip)
!!$             zray(cnt:,ip)=zray(cnt-1,ip)
!!$             continue
!!$          else  
!!$             call get_min_loc(xin,xray,iix,cnt-1,ip,0.)
!!$             call get_min_loc(zin,zray,iiz,cnt-1,ip,0.)
!!$             dx_step=-ds*PoverP_dx(iiz,iix)/sqrt(PoverP_dx(iiz,iix)**2+PoverP_dz(iiz,iix)**2+eps)
!!$             dz_step=-ds*PoverP_dz(iiz,iix)/sqrt(PoverP_dx(iiz,iix)**2+PoverP_dz(iiz,iix)**2+eps)
!!$             call get_min_loc(xin,xray,iix,cnt,ip,dx_step)
!!$             call get_min_loc(zin,zray,iiz,cnt,ip,dz_step)
!!$             xray(cnt,ip)=xray(cnt-1,ip)+dx_step
!!$             zray(cnt,ip)=zray(cnt-1,ip)+dz_step
!!$          end if
!!$       end do loop_points
!!$       !! . . Check to see distance between points
!!$       allocate( bigx(maxnum,2*size(xray,2)), bigz(maxnum,2*size(zray,2)) )
!!$       it=1
!!$       bigx(1:cnt,1)=xray(1:cnt,1); bigz(1:cnt,1)=zray(1:cnt,1)
!!$       do i1=2,size(xray,2)
!!$          it=it+1
!!$          if ( sqrt( (xray(cnt,i1)-xray(cnt,i1-1))**2 + (zray(cnt,i1)-zray(cnt,i1-1))**2).gt.mxdist) then
!!$             bigx(1:cnt,it)=0.5*(xray(1:cnt,i1)+xray(1:cnt,i1-1))
!!$             bigz(1:cnt,it)=0.5*(zray(1:cnt,i1)+zray(1:cnt,i1-1))
!!$             bigx(1:cnt,it+1)=xray(1:cnt,i1)
!!$             bigz(1:cnt,it+1)=zray(1:cnt,i1)
!!$             it=it+1
!!$          else
!!$             bigx(1:cnt,it)=xray(1:cnt,i1)
!!$             bigz(1:cnt,it)=zray(1:cnt,i1) 
!!$          end if
!!$       end do
!!$       if (it .gt. size(xray,2)) then
!!$          deallocate(xray,zray,stat=AllocateStatus)
!!$          allocate( xray(maxnum,it), zray(maxnum,it))
!!$          xray(1:cnt,1:it)=bigx(1:cnt,1:it)
!!$          zray(1:cnt,1:it)=bigz(1:cnt,1:it)
!!$          deallocate(bigx,bigz)
!!$       else
!!$          xray(1:cnt,1:it)=bigx(1:cnt,1:it)
!!$          zray(1:cnt,1:it)=bigz(1:cnt,1:it)
!!$          deallocate(bigx,bigz)
!!$       end if
!!$    end do loop_steps
!!$    Rayout(1:maxnum,1:it)=cmplx(xray,zray) 
!!$    deallocate( PoverP_dx,PoverP_dz)
!!$    deallocate( xray,zray)
!!$  end subroutine Run_exact_interp
!!$ !######################################################################
  subroutine get_min_loc(x,xray,iix,cnt,ip,dx_step)
    integer                                     :: iix,ip,cnt
    integer, dimension(:),allocatable           :: xtmp
   real,    dimension(:,:)                     :: xray
    real,    dimension(:)                       :: x
    real                                        :: dx_step
   allocate( xtmp(size(x)))
    xtmp=minloc(abs(x-(xray(cnt,ip)+dx_step))); 
    iix=xtmp(1)  
    deallocate(xtmp)
  end subroutine get_min_loc
 !######################################################################
  subroutine Run_exact_nointerp(vel,data,Rayout,adj,add,eps,maxnum,x,z,xtodo,ztodo)
    integer                                :: ix,iz,iw,ip,maxnum,iix,iiz,cnt,i2,i1
    integer, dimension(:),allocatable      :: xtmp,ztmp
    real                                   :: eps,dx_step,dz_step,ds
    real,    dimension(:)                  :: xtodo,ztodo,x,z
    real,    dimension(:,:)                :: vel
    real,    dimension(:,:),allocatable    :: PoverP_dx,PoverP_dz,xray,zray
    complex, dimension(:,:)                :: Rayout,data
    logical                                :: adj,add  
    allocate( PoverP_dx(nz,nx),PoverP_dz(nz,nx),xray(maxnum,size(xtodo)),zray(maxnum,size(xtodo)))
    allocate(xtmp(size(x)), ztmp(size(z)) )
    xray(1,1:size(xtodo))=xtodo;    zray(1,1:size(xtodo))=ztodo
    !! . . Compute Gradients
    call get_gradx(adj,add,data,PoverP_dx,eps)
    call get_gradz(adj,add,data,PoverP_dz,eps)

    loop_steps:  do cnt=2,maxnum ! . . New Step  
       write(0,*) 'Working on step ',cnt,' of ', maxnum ,size(xray,2)
       loop_points: do ip=1,size(xray,2)
          call get_min_loc(x,xray,iix,cnt-1,ip,0.)
          call get_min_loc(z,zray,iiz,cnt-1,ip,0.)
          ds=vel(iiz,iix)*dtau
          if ( iix .ge. nx-2*ceiling(abs(ds/dx))  &
          .or. iix .le. 2*   ceiling(abs(ds/dx))  &
          .or. iiz .ge. nz-2*ceiling(abs(ds/dz)) ) then !&
         !! .or. iiz .le. 2*   ceiling(abs(ds/dz))  ) then
             xray(cnt:,ip)=xray(cnt-1,ip)
             zray(cnt:,ip)=zray(cnt-1,ip)
             continue
          else  
             dx_step=-ds*PoverP_dx(iiz,iix)/sqrt(PoverP_dx(iiz,iix)**2+PoverP_dz(iiz,iix)**2+eps)
             dz_step=-ds*PoverP_dz(iiz,iix)/sqrt(PoverP_dx(iiz,iix)**2+PoverP_dz(iiz,iix)**2+eps)
             xray(cnt,ip)=xray(cnt-1,ip)+dx_step
             zray(cnt,ip)=zray(cnt-1,ip)+dz_step
          end if
       end do loop_points
    end do loop_steps
    Rayout=0.
    Rayout(1:maxnum,1:size(xray,2))=cmplx(xray,zray) 
    deallocate( PoverP_dx,PoverP_dz,xray,zray,xtmp)
  end subroutine Run_exact_nointerp
 !######################################################################
end module exact_r2_mod
