program IQM2_
  !! Compute image quality maps
  !! Paul Sava (paul@sep.stanford.edu)
  !! 10/14/02

  !!  input: z-ax-mx-my-ro (ADCIG)
  !! output: z-   mx-my-ro (image quality map)

  use sep
  use util_
  use iqtransf2_
  implicit none

  type(myaxis) :: az,amx,amy,aax,aro
  real, dimension(:,:,:), allocatable :: adcig
  real, dimension(:,:),   allocatable :: iqmap

  integer :: method
  logical :: verb

  integer :: iz,imx,imy,iax, iro
  integer :: xwin,zwin

  call sep_init()

  call from_param("method",method,0)
  call from_param("verb",verb,.false.)

  call hinaxis(az ,1)
  call hinaxis(aax,2)
  call hinaxis(amx,3)
  call hinaxis(amy,4)
  call hinaxis(aro,5)

  call houaxis(az,1)
  call houaxis(amx,2)
  call houaxis(amy,3)
  call houaxis(aro,4)

  call to_history("n5",1 )
  call to_history("o5",0.)
  call to_history("d5",1.)
  call to_history("label5"," ")

  call from_param("xwin",xwin,0)
  call from_param("zwin",zwin,0)
  if(amx%n==1) xwin=0

  call iqinit(amx%n,aax%n,az%n,xwin,zwin,verb)

  allocate(adcig(az%n,aax%n,amx%n)); adcig=0
  allocate(iqmap(az%n,      amx%n)); iqmap=0.

  write(0,*) "Method=",method
  do iro=1,aro%n
     if(verb) write (0,*) "ro=",iro
     do imy=1,amy%n
        if(verb) write (0,*) "imy=",imy
        call sep_read(adcig)

        select case(method)
        case(3)
           if(verb) write(0,*) "semblance"
           call to_history("title","semblance")
           call se(adcig,iqmap)
        case(2)
           if(verb) write(0,*) "differential semblance"
           call to_history("title","differential semblance")
           call ds(adcig,iqmap)
        case(1)
           if(verb) write(0,*) "stack intensity"
           call to_history("title","stack intensity")
           call si(adcig,iqmap)
        case default 
           if(verb) write(0,*) "stack power"
           call to_history("title","stack power")
           call sp(adcig,iqmap)
        end select

        do iz=1,zwin
           iqmap(iz,:) = iqmap(zwin+1,:)
        end do
        do iz=az%n-zwin+1,az%n
           iqmap(iz,:) = iqmap(az%n-zwin,:) 
        end do

        do imx=1,xwin
           iqmap(:,imx) = iqmap(:,xwin+1)
        end do
        do imx=amx%n-xwin+1,amx%n
           iqmap(:,imx) = iqmap(:,amx%n-xwin) 
        end do

        call sep_write(iqmap)
     end do
  end do

  call sep_close()
  deallocate(adcig,iqmap)

  call exit(0)
end program IQM2_













