program TopCut_
  !! cut wavefield on topography
  !!
  use sep
  use util_

  implicit none
  real,    pointer :: wfld(:,:),mask(:,:)
  real,    pointer :: topo(:)  ,data(:)

  type(myaxis) :: ax,az,at
  integer      :: ix,iz,it
  real         :: xx,zz,tt

  logical      :: linear
  real         :: zo,fo

  call sep_init()
  call from_param("linear",linear,.true.)

  call hinaxis(az,1)
  call hinaxis(ax,2)
  call hinaxis(at,3)

  call houaxis(ax,1)
  call houaxis(at,2)
  call to_history("n3",1)

  call to_history("esize",4)
  call sep_close()

  allocate(wfld(az%n,ax%n))
  allocate(topo(ax%n))
  allocate(mask(az%n,ax%n))
  allocate(data(ax%n))

  write(0,*) "create topography mask"
  call sreed("topo",topo,size(topo)*4)

  if(linear) then
     mask=0.
     do ix=1,ax%n
        zz=topo(ix)
        iz=floor((zz-az%o)/az%d+1)
        iz=maxval((/   1,iz/))
        iz=minval((/az%n,iz/))

        zo = az%o + (iz-1)*az%d
        fo = (zz-zo)/az%d

        mask(iz  ,ix) = (1.-fo)
        mask(iz+1,ix) =     fo
     end do
  else
     mask=0.
     do ix=1,ax%n
        zz=topo(ix)
        iz=floor((zz-az%o)/az%d+1)     
        iz=maxval((/   1,iz/))
        iz=minval((/az%n,iz/))

        mask(iz,ix) = 1.
     end do
  end if

  write(0,*) "create data"
  do it=1,at%n
     write(0,*) it,at%n
     call sep_read(wfld)

     data=0
     do iz=1,az%n
        data = data + wfld(iz,:)*mask(iz,:)
     end do

     call sep_write(data)
  end do

  deallocate(wfld,data,topo,mask)

  call exit (0)
end program TOPCUT_




