!! weighted smoothing
program WSmo2_
  use sep
  use utilities
  implicit none

  type(paxis) :: a1,a2,a3
  integer     :: i1,i2,i3
  integer     :: j1,j2,j3
  integer     :: k1,k2,k3
  real, dimension(:,:,:), pointer :: din,dou  !! data
  real, dimension(:,:,:), pointer :: wgh      !! weight
  logical :: verb   !! verbose
  integer :: rect1, rect2, rect3, nrepeat,irepeat
  integer :: pow
  integer :: kk,nn

  call sep_init()
  call from_param("verb",verb,.false.)
  call from_param("rect1",rect1,0)
  call from_param("rect2",rect2,0)
  call from_param("rect3",rect3,0)
  call from_param("repeat",nrepeat,1)
  call from_param("pow",pow,1)

  call hinaxis(a1,1); a1%l=" "; call houaxis(a1,1)
  call hinaxis(a2,2); a2%l=" "; call houaxis(a2,2)
  call hinaxis(a3,3); a3%l=" "; call houaxis(a3,3)
  call to_history("n4",1)

  if(verb) write(0,*) "allocate"
  allocate(din(a1%n,a2%n,a3%n))
  allocate(dou(a1%n,a2%n,a3%n))
  allocate(wgh(a1%n,a2%n,a3%n))

  if(verb) write(0,*) "read din"
  call sep_read(din)

  if(verb) write(0,*) "read wgh"
  call sep_read(wgh)
  wgh = wgh / maxval(wgh)
  wgh = wgh **pow
  if(verb) write(0,*) minval(wgh), maxval(wgh)

  nn = (a3%n-2*rect3)*(a2%n-2*rect2)*(a1%n-2*rect1)

  do irepeat=1,nrepeat
     if(verb) write(0,*) "smooth pass",irepeat
     dou=0.

     kk=0
     do       i3=1+rect3,a3%n-rect3
        do    i2=1+rect2,a2%n-rect2
           do i1=1+rect1,a1%n-rect1
              if(mod(kk,1000)==0) write(0,*) kk,nn

              j1=i1-rect1
              k1=i1+rect1

              j2=i2-rect2
              k2=i2+rect2

              j3=i3-rect3
              k3=i3+rect3

              dou(i1,i2,i3) = sum( wgh(j1:k1,j2:k2,j3:k3)   & 
              *                    din(j1:k1,j2:k2,j3:k3) ) &
              /               sum( wgh(j1:k1,j2:k2,j3:k3) )

              kk=kk+1
           end do
        end do
     end do

     if(verb) write(0,*) "update boundaries"
     do i1=1,rect1
        dou(i1,:,:) = dou(     rect1+1,:,:)
     end do
     do i1=a1%n-rect1+1,a1%n
        dou(i1,:,:) = dou(a1%n-rect1  ,:,:) 
     end do

     do i2=1,rect2
        dou(:,i2,:) = dou(:,     rect2+1,:)
     end do
     do i2=a2%n-rect2+1,a2%n
        dou(:,i2,:) = dou(:,a2%n-rect2  ,:) 
     end do

     do i3=1,rect3
        dou(:,:,i3) = dou(:,:,     rect3+1)
     end do
     do i3=a3%n-rect3+1,a3%n
        dou(:,:,i3) = dou(:,:,a3%n-rect3  ) 
     end do

     din = dou
  end do

  if(verb) write(0,*) "write dou"
  call sep_write(dou)

  call exit(0)
end program WSmo2_








