!! acoustic modeling
!! constant density
!! no boundary condition

program aWFv_
  use sep
  implicit none

  logical :: verb
  integer :: nz,nx,nt,nw, nsz, nsx
  integer :: iz,ix,it,iw, isz, isx
  real    :: dz,dx,dt,dw, dsz, dsx
  real    :: oz,ox,ot,ow, osz, osx
  real    :: energy
  integer :: bndx,bndz
  real    :: idx,idz,idt2
  real    :: c0,c1,c2
  integer :: iop,nop

  !! wavefield
  character(len=128)            :: wfld_
  real, dimension(:,:), pointer :: up,um,uo,ud

  !! source
  real, dimension(:),   pointer :: sou, wvl 
  integer :: izsou,ixsou

  !! velocity: v
  character(len=128)            :: vp_
  real, dimension(:,:), pointer :: vp, vp2

  !! density
  character(len=128)            :: ro_
  real, dimension(:,:), pointer :: ro

  real    :: delp
  integer :: jsnap



  !----------------------------------------------------------------
  real, dimension(:), pointer :: bndz0,bndz1,bndx0,bndx1
  real, dimension(:), pointer :: xtaper,ztaper
  real                        :: xtapco,ztapco


  !----------------------------------------------------------------  
  nop = 2

  vp_ ='vp'
  ro_ ='ro'
  wfld_='wfld'

  call sep_init()
  call from_param("verb",verb,.false.)
  call from_param("jsnap",jsnap,20)

  call from_param("bndz",bndz,nop); if(bndz<nop) bndz=nop
  call from_param("bndx",bndx,nop); if(bndx<nop) bndx=nop

  call from_history("n1",nw)
  call from_param("nt",nt,nw)
  call from_param("dt",dt,.0001)

  call from_aux(vp_,"n1",nz,1 )
  call from_aux(vp_,"d1",dz,1.)
  call from_aux(vp_,"o1",oz,0.)

  call from_aux(vp_,"n2",nx,1 )
  call from_aux(vp_,"d2",dx,1.)
  call from_aux(vp_,"o2",ox,0.)

  !! source location
  call from_param("nsz",nsz,1 )
  call from_param("dsz",dsz,1.)
  call from_param("osz",osz,0.)

  call from_param("nsx",nsx,1 )
  call from_param("dsx",dsx,1.)
  call from_param("osx",osx,0.)

  !! taper coefficients
  call from_param("xtapco",xtapco,0.0002)
  call from_param("ztapco",ztapco,0.0002)



  !----------------------------------------------------------------
  !! output files
  call to_history("n1",nx)
  call to_history("d1",dx)
  call to_history("o1",ox)

  call to_history("n2",nt)
  call to_history("d2",dt)
  call to_history("o2",0.)

  call to_history("n3",nsx)
  call to_history("d3",dsx)
  call to_history("o3",osx)

  call to_history("n4",nsz)
  call to_history("d4",dsz)
  call to_history("o4",osz)

  call auxputch("n1",'i',nz,trim(wfld_))
  call auxputch("d1",'f',dz,trim(wfld_))
  call auxputch("o1",'f',oz,trim(wfld_))

  call auxputch("n2",'i',nx,trim(wfld_))
  call auxputch("d2",'f',dx,trim(wfld_))
  call auxputch("o2",'f',ox,trim(wfld_))

  call auxputch("n3",'i',nt/jsnap,trim(wfld_))
  call auxputch("d3",'f',dt*jsnap,trim(wfld_))
  call auxputch("o3",'f',      0.,trim(wfld_))

  call auxputch("n4",'i',nsx,trim(wfld_))
  call auxputch("d4",'f',dsx,trim(wfld_))
  call auxputch("o4",'f',osx,trim(wfld_))

  call auxputch("n5",'i',nsz,trim(wfld_))
  call auxputch("d5",'f',dsz,trim(wfld_))
  call auxputch("o5",'f',osz,trim(wfld_))
  !----------------------------------------------------------------
  call sep_close()

  !! expand domain
  nz = nz + 2*bndz
  nx = nx + 2*bndx

  allocate(up(nz,nx),um(nz,nx),uo(nz,nx),ud(nz,nx))
  allocate( ro(nz,nx))
  allocate( vp(nz,nx), vp2(nz,nx))
  allocate(sou(nt), wvl(nt))



  !----------------------------------------------------------------
  !! material
  call sep_read( vp(1+bndz:nz-bndz,1+bndx:nx-bndx), vp_)
  do iz=1,bndz
     vp(   iz  ,:) = vp( 1+bndz,:)
     vp(nz-iz+1,:) = vp(nz-bndz,:)
  end do
  do ix=1,bndx
     vp(:,   ix  ) = vp(:, 1+bndx)
     vp(:,nx-ix+1) = vp(:,nx-bndx)
  end do

  call sep_read( ro(1+bndz:nz-bndz,1+bndx:nx-bndx), ro_)
  do iz=1,bndz
     ro(   iz  ,:) = ro( 1+bndz,:)
     ro(nz-iz+1,:) = ro(nz-bndz,:)
  end do
  do ix=1,bndx
     ro(:,   ix  ) = ro(:, 1+bndx)
     ro(:,nx-ix+1) = ro(:,nx-bndx)
  end do

  vp2 = vp*vp

  idx = 1./dx**2
  idz = 1./dz**2
  idt2= 1./dt**2
  !----------------------------------------------------------------
  !! source derivative (displacement to acceleration)
  sou=0.
  wvl=0.
  call sep_read(sou(1:nw))
  do iw=1+1,nw-1
     wvl(iw) = (-2.*sou(iw) + sou(iw+1) + sou(iw-1))
  end do
  wvl = wvl *idt2
  !----------------------------------------------------------------
  !! laplacian coefficients
  c0 = -30./12.
  c1 = +16./12.
  c2 = - 1./12.
  !----------------------------------------------------------------
  do isz=1,nsz;    izsou = bndz + 1+((osz+(isz-1)*dsz) - oz)/dz
     do isx=1,nsx; ixsou = bndx + 1+((osx+(isx-1)*dsx) - ox)/dx
        write(0,*) "sou ",(isz-1)*nsx+isx,"/",nsx*nsz," @",osz+(isz-1)*dsz,osx+(isx-1)*dsx

        sou = wvl / ro(izsou,ixsou)
       
        !! MAIN loop
        up=0;um=0;uo=0;ud=0;
        do it=1,nt










!!$           !! 2nd order laplacian
!!$           do iz=1+1,nz-1
!!$              do ix=1+1,nx-1
!!$                 ud(iz,ix) = &
!!$                 (-2.*uo(iz,ix) + uo(iz  ,ix-1) + uo(iz  ,ix+1))/dx**2 + &
!!$                 (-2.*uo(iz,ix) + uo(iz-1,ix  ) + uo(iz+1,ix  ))/dz**2
!!$              end do
!!$           end do

           !! 4th order laplacian
           do iz=1+nop,nz-nop
              do ix=1+nop,nx-nop
                 ud(iz,ix) = &
                 c0* uo(iz,  ix  ) * (idx + idz)        + &
                 c1*(uo(iz  ,ix-1) + uo(iz  ,ix+1))*idx + &
                 c2*(uo(iz  ,ix-2) + uo(iz  ,ix+2))*idx + &
                 c1*(uo(iz-1,ix  ) + uo(iz+1,ix  ))*idz + &
                 c2*(uo(iz-2,ix  ) + uo(iz+2,ix  ))*idz
              end do
           end do

           ud= ud *vp2

           !! acceleration source
           ud(izsou,ixsou) = &
           ud(izsou,ixsou) - sou(it)

           !! time step
           up = 2*uo - um + ud * dt**2
           um =   uo
           uo =   up

           if(mod(it,jsnap)==0) then
              call sep_write(uo(bndz+1:nz-bndz,bndx+1:nx-bndx),wfld_)
           end if
           call sep_write(uo(1+bndz,1+bndx:nx-bndx)) 
           if(verb) write(0,*) it,sum(uo)
        end do
     end do
  end do

  call exit(0)
end program AWFv_








