program CalculateRays
  use sep
  use igrad1
  use util_
  use interp1

  implicit none
  type(myaxis) :: az,ax,ar
  character(len=128)                  :: file  
  integer                             :: iz,ix,itau,ip,nt,esize2,ii,iix,iiz,stat,iflag
  real                                :: ansy,dy1,dy2,dxnew,dznew,dxold,dzold,ds
  real                                :: dPot,oldpot,zshift,eps
  real,    dimension(:  ), allocatable:: tmpx,tmpz,rayin,xline,zline,y,y1,y2,y12
  real,    dimension(:,:), allocatable:: Pot_xy,xray,zray,Pot_tg,gradx,gradz,crszx
  complex, dimension(:  ), allocatable:: Curve
  complex, dimension(:,:), allocatable:: rays
  file="rayin"

  call sep_init()
  call hinaxis(az,1)
  call hinaxis(ax,2)
  call ainaxis(ar,1,file)
  call houaxis(ar,2)
  call from_param("nt",nt,100)
  call from_param("ds",ds,0.01)
  call from_aux(file,"esize",esize2)
  call to_history("n1",nt)
  call to_history("d1",ds)
  call to_history("o1",0.)
  call to_history("esize",8)
  call hclose()


  !! . . Allocation
  allocate(Pot_xy(az%n,ax%n),tmpx(ax%n),tmpz(az%n),xline(ax%n),zline(az%n),crszx(az%n,ax%n))
  allocate(xray(nt,ar%n),zray(nt,ar%n),rays(nt,ar%n),Pot_tg(nt,ar%n))
  allocate(gradx(az%n,ax%n),gradz(az%n,ax%n),y(4),y1(4),y2(4),y12(4))
  zshift=100000;  dPot=1./(nt-1); eps=.00000001;
  Pot_tg(1,:)=1.
  xline=(/ (ii,ii=0,ax%n-1) /)*ax%d+ax%o
  zline=(/ (ii,ii=0,az%n-1) /)*az%d+az%o+zshift

  !! . . Input data files
  call sep_read( Pot_xy )
  if (esize2 .eq. 4) then
     allocate( rayin(ar%n) )
     call sreed( "rayin",rayin,size(rayin)*4)
     xray(1,:)=(/ (ix,ix=0,ar%n-1) /)*ar%d+ar%o
     zray(1,:)=rayin+zshift
     deallocate( rayin )
  else
     allocate( Curve(ar%n) )
     call sreed( "rayin",Curve,size(Curve)*8)
     xray(1,:)=real (Curve) 
     zray(1,:)=aimag(Curve)+zshift
     deallocate( Curve )
  end if

  !! . . Calculate 3 derivative fields
  do iz=1,az%n !! . . X-derivative
     stat=igrad1_lop(.false.,.false.,Pot_xy(iz,:),tmpx )
     gradx(iz,:)=tmpx/ax%d; tmpx=0.
  end do
  do ix=1,ax%n !! . . Z-derivative
     stat=igrad1_lop(.false.,.false.,Pot_xy(:,ix),tmpz )
     gradz(:,ix)=tmpz/az%d; tmpz=0.
  end do
  do iz=1,az%n !! . . Cross-derivative
     stat=igrad1_lop(.false.,.false.,gradz(iz,:),tmpx )
     crszx(iz,:)=tmpx/ax%d; tmpx=0.
  end do
  deallocate( tmpx , tmpz )

  !! . . Calculate phase rays
  do ip=1,ar%n
     write(0,*) 'WORKING ON RAY ', ip ,' OF ', ar%n
     do itau=2,nt

        !! . . Set fields to old values
        dxold=xray(itau-1,ip); dzold=zray(itau-1,ip)
        oldpot=Pot_tg(itau-1,ip); iflag=0

        !! . . Get 4 indicies of corner cells
        iix=floor(  ( xray(itau-1,ip)-ax%o ) / ax%d + 1.5 )
        iiz=floor( -( az%o-zray(itau-1,ip)+zshift ) / az%d + 1.5 )
        if (iix .lt. 1) then
           iix=1
        elseif (iix .gt. ax%n-1) then
           iix=ax%n-1
        end if
        if (iiz .lt. 1) then
           iiz=1
        elseif (iiz .gt. az%n-1) then
           iiz=az%n-1
        end if
        ii=0
        !! . . Loop until reach next potential
        do while (iflag .eq. 0)

           !! . . Solved by numerical integration
           dxnew=dxold-ds*gradx(iiz,iix)/sqrt(gradx(iiz,iix)**2+gradz(iiz,iix)**2+eps) 
           dznew=dzold-ds*gradz(iiz,iix)/sqrt(gradx(iiz,iix)**2+gradz(iiz,iix)**2+eps)

           !! . . Evaluate Potential
           y=(/Pot_xy(iiz+1,iix),Pot_xy(iiz+1,iix+1),Pot_xy(iiz,iix+1),Pot_xy(iiz,iix)/)
           y1 =(/ gradz(iiz+1,iix),gradz(iiz+1,iix+1),gradz(iiz,iix+1),gradz(iiz,iix) /)
           y2 =(/ gradx(iiz+1,iix),gradx(iiz+1,iix+1),gradx(iiz,iix+1),gradx(iiz,iix) /)
           y12=(/ crszx(iiz+1,iix),crszx(iiz+1,iix+1),crszx(iiz,iix+1),crszx(iiz,iix) /)
           call bcuint(y,y1,y2,y12,zline(iiz),zline(iiz+1),xline(iix),xline(iix+1),&
           dznew,dxnew,ansy,dy1,dy2)

           if ( (oldpot - ansy) .gt. dPot) then
              iflag=1
           else
              dxold=dxnew; dzold=dznew; ii=ii+1
           end if

           if (ii .gt. 10*nt ) then
              write(0,*) ii
              iflag=1
           end if

        end do

        !! Interpolate to find exact location
        xray(itau,ip)=-(ansy-oldpot)/(( ansy-oldpot )/(dxnew-dxold+eps))+dxold
        zray(itau,ip)=-(ansy-oldpot)/(( ansy-oldpot )/(dznew-dzold+eps))+dzold
        Pot_tg(itau,ip)=Pot_tg(itau-1,ip)-dPot
     end do
  end do
  rays=cmplx(xray,zray-zshift) 

  !! output and deallocate
  call sep_write(rays)
  deallocate(xline,zline,y,y1,y2,y12,Pot_xy,xray,zray,Pot_tg,gradx,gradz,crszx)

  call exit() 
end program CalculateRays
