program CalculateRays3D
  use sep
  use igrad1
  use util_
  use interp1

  implicit none
  type(myaxis) :: az,ax,ay,apx,apy
  character(len=128)                :: file  
  integer                           :: iz,ix,iy,itau,ipx,ipy,nt,ii,iix,iiy,nr
  integer                           :: iiz,stat,iflag
  real                              :: ansy,dxnew,dynew,dznew,dxold,dyold,dzold,ds
  real                              :: dPot,oldpot,zshift,eps,K
  real, dimension(:    ),allocatable:: tmpx,tmpz,tmpy,rayin,xline,yline,zline,WW,DD
  real, dimension(:,:,:),allocatable:: Pot_xyz,xray,yray,zray,Pot_tpp,gradx,grady,gradz,rays
  real, dimension(:,:  ),allocatable:: zsurf
  logical                           :: verbose
  file="rayin"
  call sep_init()
  call from_param("nt",nt,100)
  call from_param("ds",ds,0.01)
  call from_param("verbose",verbose,.false.)
  call hinaxis(az,1)
  call hinaxis(ax,2)
  call hinaxis(ay,3)
  call ainaxis(apx,1,file)
  call ainaxis(apy,2,file)
  call to_history("n1",nt)
  call to_history("d1",ds)
  call to_history("o1",0.)
  call to_history("n2",apx%n*apy%n)
  call to_history("d2",1)
  call to_history("o2",0)
  call to_history("n3",3)
  call to_history("d3",1)
  call to_history("o3",0)
  call to_history("esize",4)
  call to_history("off1",apx%o)
  call to_history("off2",apy%o)
  call to_history("dist1",apx%d)
  call to_history("dist2",apy%d)
  call to_history("nray1",apx%n)
  call to_history("nray2",apy%n)
  call hclose() 

  !! . . Parameters
  nr=apx%n*apy%n
  zshift=100000;   
  eps=.00000001;
  dPot=1./(nt-1); 
  write(0,*) 'Read in Parameters'

  !! . . Cartesian space allocation
  allocate( xline(ax%n), yline(ay%n), zline(az%n) )
  xline=(/ (ii,ii=0,ax%n-1) /)*ax%d+ax%o
  yline=(/ (ii,ii=0,ay%n-1) /)*ay%d+ay%o
  zline=(/ (ii,ii=0,az%n-1) /)*az%d+az%o+zshift
  write(0,*) 'Cartesian Space'
  call printaxis(ax)
  call printaxis(ay)

  !! . . Potential Function allocation
  allocate( Pot_xyz(az%n,ax%n,ay%n) )
  allocate( gradx(az%n,ax%n,ay%n), grady(az%n,ax%n,ay%n), gradz(az%n,ax%n,ay%n) )

  !! . . Weighting allocation
  allocate( DD(8),WW(8) )

  !! . . Ray space allocation
  allocate( Pot_tpp(nt,apx%n,apy%n),rays(nt,nr,3) )
  allocate( xray(nt,apx%n,apy%n), yray(nt,apx%n,apy%n),zray(nt,apx%n,apy%n) )
  write(0,*) 'Ray Space'
  call printaxis(apx)
  call printaxis(apy)
  write(0,*) 'Allocated arrays'

  !! . . Input data files
  call sep_read( Pot_xyz )
  write(0,*) 'Read in 3-D Potential'

  !! . . Calculate 3 derivative fields
  allocate( tmpx(ax%n), tmpy(ay%n), tmpz(az%n) )
  do ix=1,ax%n !! . . Z-derivative
     do iy=1,ay%n
        stat=igrad1_lop(.false.,.false.,Pot_xyz(:,ix,iy),tmpz )
        gradz(:,ix,iy)=tmpz/az%d; tmpz=0.
     end do
  end do
  do iz=1,az%n !! . . X-derivative
     do iy=1,ay%n
        stat=igrad1_lop(.false.,.false.,Pot_xyz(iz,:,iy),tmpx )
        gradx(iz,:,iy)=tmpx/ax%d; tmpx=0.
     end do
  end do
  do ix=1,ax%n !! . . Y-derivative
     do iz=1,az%n
        stat=igrad1_lop(.false.,.false.,Pot_xyz(iz,ix,:),tmpy )
        grady(iz,ix,:)=tmpy/ay%d; tmpy=0.
     end do
  end do
  deallocate( tmpx , tmpz, tmpy )
  write(0,*) 'Calculated Derivative Fields'

  !! . . Read in Ray locations
  allocate( zsurf(apx%n,apy%n) )
  call sreed( "rayin", zsurf, size(zsurf)*4 )
  do iy=1,apy%n
     xray(1,:,iy)=(/ (ix,ix=0,apx%n-1) /)*apx%d+apx%o
  end do
  do ix=1,apx%n
     yray(1,ix,:)=(/ (iy,iy=0,apy%n-1) /)*apy%d+apy%o
  end do
  zray(1,:,:)=zsurf+zshift
  deallocate( zsurf )
  write(0,*) 'Read in Data files'

  !! . . Calculate gradient rays
  Pot_tpp(1,1:apx%n,1:apy%n)=1.
  do ipx=1,apx%n 
     do ipy=1,apy%n
        write(0,*) 'WORKING ON RAY ', ipx,ipy ,' OF ', nr
        do itau=2,nt 

           !! . . Set fields to old values
           dxold=xray(itau-1,ipx,ipy) 
           dyold=yray(itau-1,ipx,ipy)
           dzold=zray(itau-1,ipx,ipy)
           oldpot=Pot_tpp(itau-1,ipx,ipy)
           if (verbose) write(0,*)'OLD VALUES'
           if (verbose) write(0,*) dzold,dxold,dyold,oldpot

           ii=0 !! . . Loop until reach next potential
           iflag=0
           do while (iflag .eq. 0)

              !! . . Get indicies of lower corner cell
              iix=floor(  ( dxold-ax%o          ) / ax%d + 1.5 )
              iiy=floor(  ( dyold-ay%o          ) / ay%d + 1.5 )
              iiz=floor(  ( dzold-az%o-zshift   ) / az%d + 1.5 )
              if (verbose)  write(0,*) 'Z VALUES'
              if (verbose)  write(0,*) az%o,zshift,dzold,-(az%o+zshift-dzold)/az%d, iiz
              if (iix .lt. 1     ) iix=1
              if (iix .gt. ax%n-1) iix=ax%n-1
              if (iiy .lt. 1     ) iiy=1
              if (iiy .gt. ay%n-1) iiy=ay%n-1
              if (iiz .lt. 1)      iiz=1
              if (iiz .gt. az%n-1) iiz=az%n-1

              !! . . Solved by numerical integration
              K=sqrt(gradx(iiz,iix,iiy)**2+grady(iiz,iix,iiy)**2+gradz(iiz,iix,iiy)**2)
              if (K .eq. 0. ) then
                 dxnew=dxold
                 dynew=dyold
                 dznew=dzold-ds
              else 
                 dxnew=dxold-ds*gradx(iiz,iix,iiy)/K
                 dynew=dyold-ds*grady(iiz,iix,iiy)/K
                 dznew=dzold-abs(ds*gradz(iiz,iix,iiy)/K)
              end if
              if (verbose) write(0,*) 'STEPS - New and old'
              if (verbose) write(0,*) 'DZ: ',dznew,dzold
              if (verbose) write(0,*) 'DX: ',dxnew,dxold
              if (verbose) write(0,*) 'DY: ',dynew,dyold
              if (verbose) write(0,*) gradz(iiz,iix,iiy)/K,gradx(iiz,iix,iiy)/K,grady(iiz,iix,iiy)/K
              if (verbose) write(0,*) iiz,iix,iiy

              !! . . Find distances for linear interpolation
              DD(1)=sqrt((dznew-zline(iiz  ))**2+(dxnew-xline(iix  ))**2+(dynew-yline(iiy  ))**2)
              DD(2)=sqrt((dznew-zline(iiz+1))**2+(dxnew-xline(iix  ))**2+(dynew-yline(iiy  ))**2)
              DD(3)=sqrt((dznew-zline(iiz  ))**2+(dxnew-xline(iix+1))**2+(dynew-yline(iiy  ))**2)
              DD(4)=sqrt((dznew-zline(iiz  ))**2+(dxnew-xline(iix  ))**2+(dynew-yline(iiy+1))**2)
              DD(5)=sqrt((dznew-zline(iiz+1))**2+(dxnew-xline(iix+1))**2+(dynew-yline(iiy  ))**2)
              DD(6)=sqrt((dznew-zline(iiz+1))**2+(dxnew-xline(iix  ))**2+(dynew-yline(iiy+1))**2)
              DD(7)=sqrt((dznew-zline(iiz  ))**2+(dxnew-xline(iix+1))**2+(dynew-yline(iiy+1))**2)
              DD(8)=sqrt((dznew-zline(iiz+1))**2+(dxnew-xline(iix+1))**2+(dynew-yline(iiy+1))**2)

              do ix=1,8
                 WW(ix)=(DD(ix)**-2) / ( DD(1)**-2+DD(2)**-2+DD(3)**-2+DD(4)**-2+DD(5)**-2+&
                 DD(6)**-2+DD(7)**-2+ DD(8)**-2 )
              end do

              ansy=WW(1)*Pot_xyz(iiz  ,iix  ,iiy  )+&
              WW(2)*Pot_xyz(iiz+1,iix  ,iiy  )+&
              WW(3)*Pot_xyz(iiz  ,iix+1,iiy  )+&
              WW(4)*Pot_xyz(iiz  ,iix  ,iiy+1)+&
              WW(5)*Pot_xyz(iiz+1,iix+1,iiy  )+&
              WW(6)*Pot_xyz(iiz+1,iix  ,iiy+1)+&
              WW(7)*Pot_xyz(iiz  ,iix+1,iiy+1)+&
              WW(8)*Pot_xyz(iiz+1,iix+1,iiy+1)

              if (DD(1) .eq. 0) ansy=Pot_xyz(iiz,iix,iiy)
              if (verbose) write(0,*) 'POTENTIAL UPDATE'
              if (verbose) write(0,*) ii,oldpot, ansy,oldpot-ansy,dPot            
              if ( (oldpot - ansy) .gt. dPot) then
                 iflag=1
              else
                 dxold=dxnew; dyold=dynew; dzold=dznew; ii=ii+1
              end if

              if (ii .gt. 5000 ) then
                 iflag=1
              end if
           end do

           !! Interpolate to find exact location
!           xray(itau,ipx,ipy)= (ansy-oldpot)/(( ansy-oldpot )/(dxnew-dxold))+dxold
!           yray(itau,ipx,ipy)= (ansy-oldpot)/(( ansy-oldpot )/(dynew-dyold))+dyold
           xray(itau,ipx,ipy)=-(ansy-oldpot)/(( ansy-oldpot )/(dxnew-dxold))+dxold
           yray(itau,ipx,ipy)=-(ansy-oldpot)/(( ansy-oldpot )/(dynew-dyold))+dyold
           zray(itau,ipx,ipy)=-(ansy-oldpot)/(( ansy-oldpot )/(dznew-dzold))+dzold
           Pot_tpp(itau,ipx,ipy)=Pot_tpp(itau-1,ipx,ipy)-dPot
         !  write(0,*) 'New Ray Coordinates - Z,X,Y'
         !  write(0,*)    zray(itau,ipx,ipy),xray(itau,ipx,ipy),yray(itau,ipx,ipy)
         !  write(0,*) 'New Potential'
         !  write(0,*) Pot_tpp(itau,ipx,ipy),dPot
        end do
     end do
  end do
  write(0,*) 'Calculated all ray steps'

  rays(:,: ,1)=reshape( zray , (/nt,ax%n*ay%n/) )-zshift
  rays(:,: ,2)=reshape( xray , (/nt,ax%n*ay%n/) )
  rays(:,: ,3)=reshape( yray , (/nt,ax%n*ay%n/) )
  rays(nt,:,1)=(az%n-1)*az%d+az%o
  write(0,*) 'Reshaped rays for output'

  !! output and deallocate
  call sep_write(rays)
  deallocate(xline,yline,zline,Pot_xyz,Pot_tpp,xray,yray,zray,gradx,grady,gradz,rays)
  write(0,*) 'Wrote output and deallocated'
  call exit() 
end program CalculateRays3D
