program Potential2D
  use sep
  use igrad2ew
  use cgstep_mod
  use solver_reg_mod
  use mask1
  use weight
  use laplac2
  use igrad1
  implicit none
  integer                               :: nz,iix,iiz,ii,esize,iz,stat
  integer                               :: n1,n2,niter,jj,nx,ix
  integer                               :: iizb,iizt,xpad
  real                                  :: xmin,xmax,zmin,zmax,dfill,ifill
  real                                  :: dx,dz,depth,ox,oz,eps
  real,    dimension(:  ), allocatable  :: xcoord,zcoord,xin,zin,fillin
  real,    dimension(:,:), allocatable  :: data,model,wei
  real   , dimension(:  ), pointer      :: wtt
  real   , dimension(:  ), allocatable  :: testinput,dzdg
  complex, dimension(:  ), allocatable  :: Surface  
  real   , dimension(:  ), allocatable  :: Potential,nulldata,tmp
  logical, dimension(:  ), pointer      :: msk
  logical, dimension(:,:), allocatable  :: mfixed
  call sep_init()
  call from_history("n1",nx)
  call from_history("o1",ox)
  call from_history("d1",dx)
  call from_param("nz",nz,100)
  call from_param("depth",depth,0.)
  call from_history("esize",esize)
  call from_param("niter",niter,100)
  call from_param("xpad",xpad,20)
  call from_param("eps",eps,1.)
  n1=nz
  n2=nx+2*xpad
  allocate( Surface(nx),data(n1,n2),model(n1,n2) )
  allocate( xin(nx),zin(nx),wtt(n1*n2) )
  allocate( xcoord(n2), zcoord(n2), mfixed(n1,n2) )
  allocate( Potential(n1*n2) , msk(n1*n2), wei(n1,n2) )
  allocate( nulldata(n1*n2) ,  dzdg(n2),tmp(n2) );
  nulldata = 0.;Potential=0.;

  if (esize .eq. 4) then
     call sep_read( zin )
     zin=zin+depth
     xin=(/ (ii,ii=0,nx-1) /)*dx+ox
  else 
     call sep_read( Surface )
     xin=real(Surface)
     zin=aimag(Surface)+depth
  end if

  !! . .Fix Z-coords
  zcoord(xpad+1:nx+xpad)=zin
  zcoord(1:xpad)=zin(1)
  zcoord(nx+xpad+1:n2)=zin(nx) 
  oz=maxval(zcoord)
  dz=oz/(real(n1)-1.)

  !! . . Fix X-coords
  dx=(xin(nx)-xin(1))/real(nx-1.)
  ox=xin(1)-xpad*dx
  xcoord(1:xpad)=(/ (ix,ix=0,xpad-1) /)*dx+ox
  xcoord(xpad+1:nx+xpad)=xin
  xcoord(nx+xpad+1:n2)= (/ (ix,ix=nx+xpad+1,n2) /)*dx+ox

  mfixed=.true.
  do ii=1,n2
     !! . . Top Layer
     iix =floor( (xcoord(ii)-ox)/dx + 1.5 )
     iizt=floor( (oz-zcoord(ii))/dz + 1.5 )
     if (iix .lt. 1) then 
        iix=1
     elseif (iizt .lt.  1) then 
        iiz=1
     elseif (iix  .gt. n2) then 
        iix=n2
     elseif (iizt .gt. n1) then 
        iiz=n1
     end if

     data  (1:iizt,iix)=1.
     mfixed(1:iizt,iix)=.false.

     dfill=1./real(n1-iizt+1)
     ifill=1.
     do iz=iizt+1,n1-1
        data(iz,ii)= 1.-ifill*dfill
        ifill=ifill+1.
     end do
  end do

  data  (n1,:)=0.;
  mfixed(n1, 1)=.false.
  mfixed(n1,n2)=.false.
   
  !! Estimate Curvature for weight
 ! call igrad_init(zcoord)
  stat=igrad1_lop(.false.,.false.,zcoord,tmp)
  stat=igrad1_lop(.true. ,.true. ,dzdg  ,tmp)
  do ix=1,n2
     wei(:,ix)=1.!!-eps*dzdg(ix)/2./dx/dx
  end do
!write(0,*) wei(1,:)

write(0,*) 'Constructed Weight Matrix'
!write(0,*) wei(1:3,1:3)

  !! Reshape matricies for input to CG  
  Potential = reshape(data  ,(/n1*n2/))
  msk       = reshape(mfixed,(/n1*n2/)) 
  wtt       = reshape(wei,(/n1*n2/))
    
  !! CG Step Solver
  write(0,*) '# of iterations: ',niter
  call weight_init(wtt)
  call mask1_init(msk)
  call laplac2_init (n1,n2)
  call solver_reg( m=Potential, d=nulldata, Fop=laplac2_lop ,&
                   stepper=cgstep, niter=niter, m0=Potential,&
                   Jop=mask1_lop,Aop=laplac2_lop,naop=n1*n2 ,&
                   eps=1.,verb=.true.,Wop=weight_lop )

  !! Reshape Output
  model=reshape(Potential,(/n1,n2/) )






  !! Output file
  call to_history("n1",n1)
  call to_history("n2",n2)
  call to_history("o1",oz-depth)
  call to_history("o2",ox)
  call to_history("d1",-dz)
  call to_history("d2",dx)
  call to_history("esize",4)
  call hclose()

  call sep_write(model)

  call laplac2_close()
  call cgstep_close ()
  
  call exit()
end program Potential2D
