program NewPot2D
  use sep
  use cgstep_mod
  use laplac2
  use igrad1
  use solver_reg_mod
  use mask1
  use igrad2ew
  use scaled_identity
  use Newpot2D_

  implicit none
  integer                               :: nt,ng,ii,stat,it
  integer                               :: niter1,niter2,jj,ig
  real                                  :: dg,dt,og,ot,d1out,o1out
  real                                  :: gx,gz
  real,    dimension(:,:), allocatable  :: Xdata,Zdata,Xmodel,Zmodel
  real,    dimension(:,:), allocatable  :: JAC,JAC1
  complex, dimension(:  ), allocatable  :: Surf1,Surf2 
  complex, dimension(:,:), allocatable  :: Output 
  real   , dimension(:  ), allocatable  :: Xnd,Znd,PX,PZ,WJ
  logical, dimension(:  ), pointer      :: Xmsk,Zmsk
  logical, dimension(:,:), allocatable  :: Xfix,Zfix
  call sep_init()
  call from_history("n1",ng)
  call from_history("o1",og)
  call from_history("d1",dg)
  call from_param("nt",nt,100)
  call from_param("niter1",niter1,100)
  call from_param("niter2",niter2,100)

  allocate( Surf1(ng),Surf2(ng)        )
  allocate( Xdata(nt,ng),Xmodel(nt,ng) )
  allocate( Zdata(nt,ng),Zmodel(nt,ng) )
  allocate( Xfix (nt,ng),Zfix (nt,ng)  )
  allocate( PX(nt*ng), PZ(nt*ng), Xmsk(nt*ng), Zmsk(nt*ng) )
  allocate( Xnd(nt*ng),Znd(nt*ng),Output(nt,ng))
  allocate( JAC(nt,ng),WJ(nt*ng), JAC1(nt,ng) )
  Xnd = 0.; Znd = 0.
  PX  = 0.; PZ  = 0.

  call sep_read(Surf1)
  call sreed  ("Surf2",Surf2,size(Surf2)*8)

  Xdata( 1,1:ng) =  real(Surf1)
  Zdata( 1,1:ng) = aimag(Surf1)
  Xdata(nt,1:ng) =  real(Surf2)
  Zdata(nt,1:ng) = aimag(Surf2)

  do ig=1,ng
     Xdata(2:nt-1,ig)= Xdata(1,ig)+&
          (Xdata(nt,ig)-Xdata(1,ig))/dble(nt-1.)&
          *dble((/ (ii,ii=2,nt-1) /))
     Zdata(2:nt-1,ig)= Zdata(1,ig)+&
          (Zdata(nt,ig)-Zdata(1,ig))/dble(nt-1.)&
          *dble((/ (ii,ii=2,nt-1) /))
  end do

  call Calculate_Jacobian_init(nt,ng,dg)
  call Calculate_Jacobian(Xdata,Zdata,Jac)

write(0,*) 'The MAX value is: ', log10(maxval((JAC)))
write(0,*) 'The MIN value is: ', log10(minval((JAC)))

  !! Fix in appropriate locations
  Xfix = .true.
  Zfix = .true.
  !! Fixed along top
  Xfix(1,1:ng) = .false.
  Zfix(1,1:ng) = .false.

  !! Fixed at basement corner points
  Xfix(nt,  1) = .false.
  Zfix(nt,  1) = .false.
  Xfix(nt, ng) = .false.
  Zfix(nt, ng) = .false.

  !! . . Call the laplacian filler
  PX   = reshape(Xdata  ,(/nt*ng/))
  PZ   = reshape(Zdata  ,(/nt*ng/))
  Xmsk = reshape(Xfix   ,(/nt*ng/)) 
  Zmsk = reshape(Zfix   ,(/nt*ng/)) 
  WJ   = reshape(JAC    ,(/nt*ng/))


  write(0,*) '# of iterations: ',niter1
  call mask1_init(Xmsk)
  call mask1_init(Zmsk)
  call laplac2_init (nt,ng)
  call igrad2ew_init(nt,ng)
  call scaled_identity_init(WJ,nt*ng,1)

  call solver_reg( m=PX, d=Xnd, Fop=laplac2_lop, stepper=cgstep,&
  niter=niter1, m0=PX, Jop=mask1_lop, Aop=laplac2_lop,naop=nt*ng,eps=0.,&
  verb=.true. )

  call solver_reg( m=PZ, d=Znd, Fop=laplac2_lop, stepper=cgstep,&
  niter=niter1, m0=PZ, Jop=mask1_lop, Aop=laplac2_lop,naop=nt*ng,eps=0.,&
  verb=.true. )


  Xmodel=reshape(PX,(/nt,ng/) )
  Zmodel=reshape(PZ,(/nt,ng/) )

  Output=cmplx(Xmodel,Zmodel)


  call Calculate_Jacobian(Xdata,Zdata,Jac)
  write(0,*) log10(JAC(1:5,1:5))
write(0,*) 'The MAX value is: ', log10(maxval((JAC)))
write(0,*) 'The MIN value is: ', log10(minval((JAC)))


  d1out=dble(1./(nt-1)) * ( aimag(Surf2(1) - Surf1(1)) )
  o1out=aimag(Surf1(1))

!  call to_history("n1",nt,"Jac.H")
!  call to_history("n2",ng,"Jac.H")
!  call to_history("d1",d1out,"Jac.H")
!  call to_history("d2",dg,"Jac.H")
!  call to_history("o1",o1out,"Jac.H")
!  call to_history("o2",og,"Jac.H")
!  call to_history("esize",4,"Jac.H")
!
!  call srite("Jac.H",JAC,size(JAC)*4)

  call to_history("n1",nt)
  call to_history("n2",ng)
  call to_history("o1",o1out)
  call to_history("o2",og)
  call to_history("d1",d1out )
  call to_history("d2",dg)
  call to_history("esize",8)
  call hclose()

  call sep_write(Output)

  call laplac2_close()
  call cgstep_close ()
  call igrad1_close()
  call exit()
end program NewPot2D
