program Potential3D
!! Create Potential Field for Feathered Coordinate System
  use sep
  use cgstep_mod
  use laplac3
  use solver_reg_mod
  use mask1
  use util_
  use Idbvip2d

  implicit none
  type(myaxis) :: at,ag,an,az,ax,ay
  integer      :: nz,nx,ny,nt,ng,nn
  integer      :: iz,ix,iy,it,ig,in
  real         :: ox,oy,oz,dz,dx,dy
  real         :: ot,og,on,dt,dg,dn
  real         :: oxmin,oxmax,oymin,oymax
  integer      :: ii,jj,stat,ndp
  integer      :: niter,n1,n2,n3,nip
  real         :: dfill,ifill,depth
  
  real (kind=8),dimension(:),allocatable:: xcartd,ycartd,gind,nind
  real (kind=8),dimension(:),allocatable:: Ptind,Ptoutd
  real   , dimension(:    ), allocatable:: Potential,nulldata,xin,yin,zin
  real   , dimension(:    ), allocatable:: xt
  real (kind=8), dimension(:,:), allocatable:: tmp
  real,    dimension(:,:,:), allocatable:: data,model
  real,    dimension(:,:  ), allocatable:: Ptout
  complex, dimension(:,:  ), allocatable:: tin
  logical, dimension(:    ), pointer    :: msk
  logical, dimension(:,:,:), allocatable:: mfixed

  call sep_init()
  call from_param("nt",nt,100)
  call from_param("ot",ot,0.)
  call from_param("depth",depth,1000.)
  call from_param("niter",niter,20)
  call from_param("nx",nx,ng)
  call from_param("ny",ny,nn)

  call from_param("oxmin",oxmin,0.)
  call from_param("oxmax",oxmax,1000.)
  call from_param("oymin",oymin,0.)
  call from_param("oymax",oymax,1000.)
 
  nz = nt
  ox =  oxmin
  dx = (oxmax-oxmin)/(nx-1)
  oy =  oymin
  dy = (oymax-oymin)/(ny-1)
  write(0,*) 'READ IN PARAMETERS'

  !! Set up input dimensions (Ray Space)
  call hinaxis(ag,1);call hinaxis(an,2);
  at%n=nt;at%o=ot;at%d=(depth-ot)/real(nt-1)
  nn = an%n; ng = ag%n
  !! Set up output dimensions (Physical Space)
  az%n = nz; n1=nz; az%o = ot; az%d = at%d
  ax%n = nx; n2=nx; ax%d = dx; ax%o = ox
  ay%n = ny; n3=ny; ay%o = oy; ay%d = dy
write(0,*) 'T-axis',at%n,at%o,at%d
write(0,*) 'Z-axis',az%n,az%o,az%d
write(0,*) 'X-axis',ax%n,ax%o,ax%d
write(0,*) 'Y-axis',ay%n,ay%o,ay%d
  call houaxis(az,1);call houaxis(ax,2);call houaxis(ay,3)
  call to_history("esize",4)

  !! . . Output Interpolated Sheet
  call auxputch("n1","i",nx,"Interp.surf.H")
  call auxputch("n2","i",ny,"Interp.surf.H")
  call auxputch("o1","r",ox,"Interp.surf.H")
  call auxputch("o2","r",oy,"Interp.surf.H")
  call auxputch("d1","r",dx,"Interp.surf.H")
  call auxputch("d2","r",dy,"Interp.surf.H")
  call auxputch("n3","i",2, "Interp.surf.H")
  call auxputch("esize","i",4,"Interp.surf.H")
  call hclose()
  
  !! . . Pad data space
  allocate( mfixed(n1,n2,n3),msk(n1*n2*n3),tmp(ng,nn))
  allocate( xin(nx),yin(ny),tin(ng,nn) )
  allocate( data(n1,n2,n3),model(n1,n2,n3) )
  allocate( Potential(n1*n2*n3),nulldata(n1*n2*n3),xt(nx))
  allocate( Ptind(ng*nn) , Ptoutd(n2*n3) , Ptout(n2,n3), zin(nz) )
  allocate( xcartd(nx*ny), ycartd(nx*ny),gind(ng*nn),nind(ng*nn) )

  !! Read in the feathered grid
  call sep_read(   tin  )
  tmp  = dble( real(tin))
  gind = reshape( tmp , (/ng*nn/) )
  tmp  = dble(aimag(tin))
  nind = reshape( tmp , (/ng*nn/) )
  do in = 1,nn
  	Ptind( (in-1)*ng+1:in*ng ) = dble( in )-0.5
  end do
  write(0,*) 'READ IN DATA'
  
  !! Cartesian output grid
  xt = dble( (/ (ii,ii=0,nx-1) /)*ax%d+ax%o )
  ix=1
  do ii=1,ny
    xcartd( (ii-1)*nx+1:ii*nx ) = xt
    ycartd( (ii-1)*nx+1:ii*nx ) = (ix-1)*ay%d + ay%o
	ix = ix + 1 
  end do
  write(0,*) 'SET UP GRID'

  !! call the interpolation function  
  nip = nx * ny
  ndp = ng * nn
  call idbvip(1,ndp,gind,nind,Ptind,nip,xcartd,ycartd,Ptoutd)

  Ptout = reshape(Ptoutd, (/n2,n3/) )

  !! Top Layer
  data(1,:,:) = Ptout

  !! Bottom Layer
  do in=1,ny
    data(n1,:,in) = real( (in-1)*(nn-1)/real(ny-1) )+1.
  end do
  
  !! Bottom Layer
  do ii=1,n2
     do jj=1,n3
        !! . . Top Layer
!	zin=(/(ii,ii=0,nz-1)/)/real(nz-1)*(data(nz,ii,jj)-data(1,ii,jj)) &
!		+ data(1,ii,jj)
!	data(:,ii,jj) = zin
	data(:,ii,jj)=(/(ii,ii=0,nz-1)/)/&
                real(nz-1)*(data(nz,ii,jj)-data(1,ii,jj)) &
		+ data(1,ii,jj)
!	data(:,ii,jj) = zin

     end do
  end do
  call srite("Interp.surf.H",Ptout,size(Ptout)*4)
  call srite("Interp.surf.H",data(n1,:,:),size(data(n1,:,:))*4)

 write(0,*) 'FIXED SURFACE POINTS'

  mfixed=.true.
  mfixed( 1,:,:)=.false.
  mfixed(n1,:,:)=.false.
  write(0,*) 'SET UP INITIAL DATA MATRIX WITH MASK'

  !! . . Call the laplacian filler
  Potential = reshape(data  ,(/n1*n2*n3/))
  msk       = reshape(mfixed,(/n1*n2*n3/)) 
!  deallocate(mfixed,xin,yin,zin,data,Ptind,Ptoutd,zin,xcartd,ycartd)

  write(0,*) 'STARTING LAPLACIAN SOLVER'
  write(0,*) 'NUMBER OF ITERATIONS: ',niter
  call mask1_init(msk)
  call laplac3_init (n1,n2,n3)
  call solver_reg( m=Potential, d=nulldata, &
                   Fop=laplac3_lop, stepper=cgstep,&
                   niter=niter, m0=Potential,& 
                   Jop=mask1_lop,Aop=laplac3_lop,&
                   naop=n1*n2*n3,eps=0.,verb=.true.)

  model=reshape(Potential,(/n1,n2,n3/) )

  write(0,*) 'OUTPUTING MODEL FILE'
  call sep_write(model)
  call laplac3_close()
  call cgstep_close ()
  deallocate(model,Potential,msk,nulldata)
  call exit()
end program Potential3D
