program main
!***********************************************************************
!
!! BIVAR_PRB runs tests on the BIVAR bivariate scattered data routines.
!
  use Idbvip2d
  use SEP
  implicit none
  integer, parameter :: nd = 20
  integer, parameter :: ni = 5
  integer, parameter :: nx = 5
  integer, parameter :: ny = 4
  integer, parameter :: nz = nx
  integer, parameter :: niwk = 31 * nd + nx * ny
  integer, parameter :: nwk = 6 * nd
  real ( kind = 8 ) error
  integer i
 
  integer md
  real ( kind = 8 ) temp
  real ( kind = 8 ), dimension ( nd ) :: xd = (/ &
    0.632000D+00, 0.295109D+00, 0.578769D+00, 0.140772D+00, 0.985549D+00, &
    0.209717D+00, 0.893022D+00, 0.928088D+00, 0.766803D+00, 0.819351D+00, &
    0.959191D+00, 0.516312D+00, 0.052801D+00, 0.626174D+00, 0.275320D+00, &
    0.659645D+00, 0.396438D+00, 0.204824D+00, 0.906341D+00, 0.880356D+00 /)
  real ( kind = 8 ) xi(nx)
  real ( kind = 8 ), dimension ( nd ) :: yd = (/ &
      0.918675D+00, 0.821677D+00, 0.962992D+00, 0.477774D+00, 0.056257D+00, &
      0.654656D+00, 0.377358D+00, 0.263626D+00, 0.706323D+00, 0.584658D+00, &
      0.154619D+00, 0.986449D+00, 0.197552D+00, 0.924616D+00, 0.788099D+00, &
      0.886828D+00, 0.945136D+00, 0.643341D+00, 0.335303D+00, 0.416050D+00 /)
  real ( kind = 8 ) yi(ni)
  real ( kind = 8 ) yi2(ny)
  real ( kind = 8 ) zd(nd)
  real ( kind = 8 ) zi(ni)

  integer iwk(niwk)
  integer j
  real ( kind = 8 ) zi2(nz,ny)
  real ( kind = 8 ) wk(nwk)


!! Test 1
  write ( *, '(a)' ) ' '
  write ( *, '(a)' ) 'BIVAR_PRB'
  write ( *, '(a)' ) '  Tests for BIVAR, which interpolates scattered '
  write ( *, '(a)' ) '  (X,Y) data.'

  
  write ( *, '(a)' ) ' '
  write ( *, '(a)' ) 'TEST01'
  write ( *, '(a)' ) '  IDBVIP accepts scattered Z(X,Y) data, and interpolates '
  write ( *, '(a)' ) '  values of Z at new points.'
  write ( *, '(a)' ) ' '
!
!  Set up the Z data.
!
  do i = 1, nd
    zd(i) = exp ( xd(i) ) * sin ( yd(i) )
  end do
!
!  Set the points at which we want interpolated values.
!
  do i = 1, ni
    xi(i) = real ( i - 1, kind = 8 ) / real ( ni - 1, kind = 8 )
    yi(i) = 1.0D+00 - xi(i)
  end do
!
!  Call IDBVIP to estimate the value of Z at the points (XI,YI).
!
  md = 1
 
  call idbvip ( md, nd, xd, yd, zd, ni, xi, yi, zi )
 
  write ( *, '(a)' ) ' '
  write ( *, '(a)' ) &
    '   X            Y             Z             Exact Z       Error'
  write ( *, '(a)' ) ' '

  do i = 1, ni
    temp = exp ( xi(i) ) * sin ( yi(i) )
    error = zi(i) - temp
    write ( *, '(5g14.6)' ) xi(i), yi(i), zi(i), temp, error
  end do
  
 
  write ( *, '(a)' ) ' '
  write ( *, '(a)' ) 'TEST02'
  write ( *, '(a)' ) '  IDSFFT accepts scattered Z(X,Y) data, and interpolates'
  write ( *, '(a)' ) '  values of Z on a regular grid of points.'
  write ( *, '(a)' ) ' '
!
!  Set up the Z data
!
  do i = 1, nd
    zd(i) = exp ( xd(i) ) * sin ( yd(i) )
  end do
!
!  Set up the points at which we want interpolated values.
!
  do i = 1, nx
    xi(i) = real ( i - 1, kind = 8 ) / real ( nx - 1, kind = 8 )
  end do
 
  do j = 1, ny
    yi2(j) = real ( j - 1, kind = 8 ) / real ( ny - 1, kind = 8 )
  end do
!
!  Call IDSFFT to estimate the value of Z.
!
  md = 1
 
  call idsfft ( md, nd, xd, yd, zd, nx, ny, nz, xi, yi2, zi2 )!, iwk, wk )
 
  write ( *, '(a)' ) ' '
  write ( *, '(a)' ) &
    '   X            Y             Z             Exact Z       Error'
  write ( *, '(a)' ) ' '

  do i = 1, nx
    do j = 1, ny
      temp = exp ( xi(i) ) * sin ( yi2(j) )
      error = zi2(i,j) - temp
      write ( *, '(5g14.6)' ) xi(i), yi2(j), zi2(i,j), temp, error
    end do
    write ( *, '(a)' ) ' '
  end do
  write ( *, '(a)' ) ' '
  write ( *, '(a)' ) 'BIVAR_PRB'
  write ( *, '(a)' ) '  Normal end of execution.'

  write ( *, '(a)' ) ' '

	call exit()
end program main