!$=head1 NAME
!$
!$igrad2 - 2D gradient operator
!$
!$=head1 SYNOPSIS
!$
!$C<ierr=igrad2_init(n1,n2)>
!$
!$C<ierr=igrad2_lop(adj,add,xx,yy)>
!$
!$=head1 PARAMETERS
!$
!$=over 4
!$
!$=item n1,n2  - integer
!$
!$      dimensions of dataset
!$
!$=item adj,add,xx,yy -
!$
!$      standard operator interface
!$
!$=back
!$
!$=head1 DESCRIPTION
!$
!$ Calculate 2-D gradient
!$
!$=head1 SEE ALSO
!$
!$L<igrad1>
!$
!$=head1 LIBRARY
!$
!$B<geef90>
!$
!$=cut
!$
module igrad3  
! 2-D gradient with adjoint,  r= grad( p)
  use adj_mod
  implicit none
  integer,private :: n1,n2,n3 
  contains
  subroutine igrad3_init ( n1_in,n2_in,n3_in )
    integer            :: n1_in, n2_in, n3_in
    n1 = n1_in
    n2 = n2_in
    n3 = n3_in
  end subroutine
  function igrad3_lop ( adj, add, p, r) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: p,r 
    call adjnull (adj,add,p,r )
    call igrad3_lop2(adj,p,r )
    stat=0
  end function 
  subroutine igrad3_lop2(adj,p,r)
    logical,intent(in) :: adj 
    real, dimension (n1, n2, n3)  :: p 
    real, dimension (n1,n2,n3,3)  :: r 
    integer i,j,k
    do i= 1, n1-1  
      do j= 1, n2-1  
         do k=1, n3-1
            if ( adj) then
               p(i+1,j  ,k  )  =		p(i+1,j  ,k  )  +  r(i,j,k,1) 
               p(i  ,j  ,k  )  =                p(i  ,j  ,k  )  -  r(i,j,k,1) 
               p(i  ,j+1,k  )  =                p(i  ,j+1,k  )  +  r(i,j,k,2) 
               p(i  ,j  ,k  )  =                p(i  ,j  ,k  )  -  r(i,j,k,2)
               p(i  ,j  ,k+1)  =                p(i  ,j  ,k+1)  +  r(i,j,k,3)
               p(i  ,j  ,k  )  =                p(i  ,j  ,k  )  -  r(i,j,k,3)
            else
               r(i,j,k,1)  =   r(i,j,k,1)  +  ( p(i+1,j,k) - p(i,j,k)) 
               r(i,j,k,2)  =   r(i,j,k,2)  +  ( p(i,j+1,k) - p(i,j,k))
               r(i,j,k,3)  =   r(i,j,k,3)  +  ( p(i,j,k+1) - p(i,j,k))
            end if
         end do
      end do
    end do
  end subroutine 
  subroutine igrad2_close()
  end subroutine 
end module 
