module interp1
  use SEP
  implicit none

  integer,private :: nx,nz
contains

  subroutine interp1_init(nx_in,nz_in)
    integer :: nx_in,nz_in
    nx=nx_in;  nz=nz_in



  end subroutine interp1_init


  subroutine hunt(xx,x,jlo)
    integer, intent(inout) :: jlo
    real,    intent(in   ) ::x
    real,dimension(:) :: xx
    integer :: n,inc,jhi,jm
    logical :: ascnd
    n=size(xx)
    ascnd = (xx(n) >= xx(1))
    if (jlo <=0 .or. jlo > n) then
       jlo=0
       jhi=n+1
    else
       inc=1
       if (x >= xx(jlo) .eqv. ascnd) then
          do 
             jhi=jlo + inc
             if(jhi > n) then 
                jhi=n+1
                exit
             else
                if (x < xx(jhi) .eqv. ascnd) exit
                jlo=jhi
                inc=inc+inc
             end if
          end do
       else
          jhi=jlo
          do
             jlo=jhi-inc
             if(jlo < 1) then
                jlo=0
                exit
             else
                if (x >= xx(jlo) .eqv. ascnd) exit
                jhi=jlo
                inc=inc+inc
             end if
          end do
       end if
    end if
    do
       if (jhi-jlo <= 1) then
          if (x ==xx(n)) jlo=n-1
          if (x ==xx(1)) jlo=1
          exit
       else
          jm=(jhi+jlo)/2
          if (x >=xx(jm) .eqv. ascnd) then
             jlo=jm
          else
             jhi=jm
          end if
       end if
    end do
  end subroutine hunt


  subroutine spline(x,y,yp1,ypn,y2)
    real, dimension(:) :: x,y
    real               :: yp1,ypn
    real, dimension(:) :: y2
    integer :: n
    real, dimension(size(x)) :: a,b,c,r
    n=size(x)
    c(1:n-1)=x(2:n)-x(1:n-1)
    r(1:n-1)=6.0*((y(2:n)-y(1:n-1))/c(1:n-1))
    r(2:n-1)=r(2:n-1)-r(1:n-2)
    a(2:n-1)=c(1:n-2)
    b(2:n-1)=2.0*(c(2:n-1)+a(2:n-1))
    b(1)=1.0
    b(n)=1.0
    if (yp1 > 0.99e30) then
       r(1)=0.0
       c(1)=0.0
    else
       r(1)=(3.0/(x(2)-x(1))*((y(2)-y(1))/x(2)-x(1))-yp1)
       c(1)=0.5
    end if
    if(ypn > .99e30) then
       r(n)=0.0
       a(n)=0.0
    else 
       r(n)=(-3.0/(x(n)-x(n-1))*((y(n)-y(n-1))/x(n)-x(n-1))-ypn)
       a(n)=0.5
    end if
    call tridag(a(2:n),b(1:n),c(1:n-1),r(1:n),y2(1:n))
  end subroutine spline

  function locate(xx,x)
    real,dimension(:),intent(in)::xx
    real, intent (in)::x
    integer :: locate
    integer :: n,jl,jm,ju
    logical :: ascnd
    n=size(xx)
    ascnd = (xx(n) >= xx(1))
    jl=0
    ju=n+1
    do
       if (ju-jl <=1) exit
       jm=(ju+jl)/2
       if ( ascnd .eqv. ( x>=xx(jm))) then
          jl=jm
       else
          ju=jm
       end if
    end do
    if (x == xx(1)) then
       locate=1
    else if (x==xx(n)) then
       locate = n-1
    else
       locate=jl
    end if
  end function locate


  function splint(xa,ya,y2a,x)
    real,dimension(:), intent(in) :: xa,ya,y2a
    real,intent(in) :: x
    real :: splint
    integer:: khi,klo,n
    real :: a,b,h
    n=size(xa)
    klo=max(min(locate(xa,x),n-1),1)
    khi=klo+1
    h=xa(khi)-xa(klo)
!!    if (h==0.0) call nrerror('bad xa input in splint')
    a=(xa(khi)-x)/h
    b=(x-xa(klo))/h
    splint=a*ya(klo)+b*ya(khi)+((a**3-a)*y2a(klo)+(b**3-b)*y2a(khi))*(h**2)/6.0
  end function splint

  subroutine tridag(a,b,c,r,u)
    real,dimension(:), intent(in) :: a,b,c,r
    real,dimension(:), intent(out):: u
    real,dimension(size(b)) :: gam
    integer :: n,j
    real :: bet
    n=size(b)
    bet=b(1)
    u(1)=r(1)/bet
    do j=2,n
       gam(j)=c(j-1)/bet
       bet=b(j)-a(j-1)*gam(j)
       u(j)=u(j)-gam(j+1)*u(j+1)
    end do
    do j=n-1,1,-1
       u(j)=u(j)-gam(j+1)*u(j+1)
    end do
  end subroutine tridag

  subroutine bcuint(y,y1,y2,y12,x1l,x1u,x2l,x2u,x1,x2,ansy,ansy1,ansy2)
    real,dimension(4), intent(in) :: y,y1,y2,y12
    real,intent(in) ::x1l,x1u,x2l,x2u,x1,x2
    real,intent(out)::ansy,ansy1,ansy2
    integer :: i
    real :: t,u
    real,dimension(4,4) :: c
    call bcucof(y,y1,y2,y12,x1u-x1l,x2u-x2l,c)
    t=(x1-x1l)/(x1u-x1l)
    u=(x2-x2l)/(x2u-x2l)
    ansy=0.0
    ansy1=0.0
    ansy2=0.0
    do i=4,1,-1
       ansy =t*ansy+((c(i,4)*u+c(i,3))*u+c(i,2))+c(i,1)
       ansy2=t*ansy2+(3.0*c(i,4)*u+2.0*c(i,3))*u+c(i,2)
       ansy1=u*ansy1+(3.0*c(4,i)*t+2.0*c(3,i))*u+c(2,i)
    end do
    ansy1=ansy1/(x1u-x1l)
    ansy2=ansy2/(x2u-x2l)
  end subroutine bcuint

  subroutine bcucof(y,y1,y2,y12,d1,d2,c)
    real, intent(in) :: d1,d2
    real, dimension(4), intent(in) :: y,y1,y2,y12
    real, dimension(4,4), intent(out) :: c
    real, dimension(16) :: x
    real, dimension(16,16) :: wt
    DATA wt /1,0,-3,2,4*0,-3,0,9,-6,2,0,-6,4,&
       8*0,3,0,-9,6,-2,0,6,-4,10*0,9,-6,2*0,-6,4,2*0,3,-2,6*0,-9,6,&
       2*0,6,-4,4*0,1,0,-3,2,-2,0,6,-4,1,0,-3,2,8*0,-1,0,3,-2,1,0,-3,&
       2,10*0,-3,2,2*0,3,-2,6*0,3,-2,2*0,-6,4,2*0,3,-2,0,1,-2,1,5*0,&
       -3,6,-3,0,2,-4,2,9*0,3,-6,3,0,-2,4,-2,10*0,-3,3,2*0,2,-2,2*0,&
       -1,1,6*0,3,-3,2*0,-2,2,5*0,1,-2,1,0,-2,4,-2,0,1,-2,1,9*0,-1,2,&
       -1,0,1,-2,1,10*0,1,-1,2*0,-1,1,6*0,-1,1,2*0,2,-2,2*0,-1,1/
    x(1:4)=y
    x(5:8)=y1*d1
    x(9:12)=y2*d2
    x(13:16)=y12*d1*d2
    x=matmul(wt,x)
    c=reshape(x,(/4,4/),order=(/2,1/))
  end subroutine bcucof

end module interp1






