module laplac3
! Laplacian operator in 2-D
  use adj_mod
  use igrad3
  implicit none
  logical,   parameter, private   :: T = .true. 
  logical,   parameter, private   :: F = .false. 
  real ,private,  dimension (:), allocatable  :: tmp 
  contains
  subroutine laplac3_init ( m1,m2,m3 )
    integer m1, m2, m3
    if (.not. allocated(tmp)) then
      allocate(tmp ( m1*m2*m3*3)) 
    end if
    call igrad3_init (m1, m2, m3)
  end subroutine
  function laplac3_lop ( adj, add, x, y) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: x,y 
    call adjnull (adj,add,x,y )
    call laplac3_lop2(adj,x,y )
    stat=0
  end function 
  subroutine laplac3_lop2(adj,x,y)
    logical,intent(in) :: adj 
    real, dimension (:)  :: x 
    real, dimension (:)  :: y 
    integer stat1, stat2
    if ( adj) then
      stat1 = igrad3_lop (  F,  F, y, tmp)   ! tmp = grad y
      stat2 = igrad3_lop (  T,  T, x, tmp)   ! x = x + grad' tmp
    else
      stat1 = igrad3_lop (  F,  F, x, tmp)   ! tmp = grad x
      stat2 = igrad3_lop (  T,  T, y, tmp)   ! y = y + grad' tmp
    end if
  end subroutine 
  subroutine laplac3_close()
    deallocate( tmp )
  end subroutine 
end module 
