module laplacf
! Laplacian operator in 2-D (on 3D filter)
  use adj_mod
  use laplac
  implicit none

  logical, parameter,     private               :: T = .true. 
  logical, parameter,     private               :: F = .false. 
  integer, dimension (:), allocatable, private  :: m
  integer,                             private  :: nlag 
  contains
  subroutine laplacf_init ( m_in,nlag_in )
    integer            :: m_in(:), nlag_in
    allocate (m(size(m_in)))
    m = m_in
    nlag = nlag_in
    call laplac_init (m)
  end subroutine

  function laplacf_lop ( adj, add, x, y) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: x,y 
    call adjnull (adj,add,x,y )
    call laplacf_lop2(adj,x,y )
    stat=0
  end function 

  subroutine laplacf_lop2(adj,x,y)
    logical,intent(in) :: adj
    real, dimension (:)  :: x,y 
    real, dimension (:), allocatable :: temp
    real, dimension (:,:),allocatable  :: x2,y2
    integer stat1, ilag, m123

    m123 = product(m)
    allocate (x2(nlag,m123),y2(nlag,m123),temp(nlag*m123))
    if ( adj) then
       y2 = reshape(y,(/nlag,m123/))
       do ilag=1,nlag
          stat1 = laplac_lop (  adj,  F, x2(ilag,:), y2(ilag,:))
       end do
       temp = reshape(x2,(/nlag*m123/))
       x = x + temp
    else
       x2 = reshape(x,(/nlag,m123/))
       do ilag=1,nlag
          stat1 = laplac_lop (  adj,  F, x2(ilag,:), y2(ilag,:))
       end do
       temp = reshape(y2,(/nlag*m123/))
       y = y + temp
    end if
    deallocate(x2,y2,temp)
  end subroutine laplacf_lop2
  subroutine laplacf_close()
  end subroutine 
end module 

