module laplacf3  
! Laplacian operator in 2-D (on 3D filter)
  use adj_mod
  use laplac3
  implicit none

  logical,   parameter, private   :: T = .true. 
  logical,   parameter, private   :: F = .false. 
  integer ,             private   :: m1, m2, m3, nlag 
  contains
  subroutine laplacf3_init ( m1_in,m2_in,m3_in,nlag_in )
    integer m1_in, m2_in, m3_in,nlag_in
    m1 = m1_in
    m2 = m2_in
    m3 = m3_in
    nlag = nlag_in
    call laplac3_init (m1, m2, m3)
  end subroutine
  function laplacf3_lop ( adj, add, x, y) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: x,y 
    call adjnull (adj,add,x,y )
    call laplacf3_lop2(adj,x,y )
    stat=0
  end function 
  subroutine laplacf3_lop2(adj,x,y)
    logical,intent(in) :: adj
    real, dimension (:)  :: x,y 
    real, dimension (:), allocatable :: temp
    real, dimension (:,:),allocatable  :: x2,y2
    integer stat1, ilag, m123

    m123 = m1*m2*m3
    allocate (x2(nlag,m123),y2(nlag,m123),temp(nlag*m123))
    if ( adj) then
       y2 = reshape(y,(/nlag,m123/))
       do ilag=1,nlag
          stat1 = laplac3_lop (  adj,  F, x2(ilag,:), y2(ilag,:))
       end do
       temp = reshape(x2,(/nlag*m123/))
       x = x + temp
    else
       x2 = reshape(x,(/nlag,m123/))
       do ilag=1,nlag
          stat1 = laplac3_lop (  adj,  F, x2(ilag,:), y2(ilag,:))
       end do
       temp = reshape(y2,(/nlag*m123/))
       y = y + temp
    end if
    deallocate(x2,y2,temp)
  end subroutine laplacf3_lop2
  subroutine laplacf3_close()
  end subroutine 
end module 

