program SourceProp
  use sep
  use util_
  use sourceprop_wemig

  implicit none
  type(myaxis) :: ax,az,aw,ar,as,ah

  integer :: iz,ii,nref,method,nh
  character(len=128)   :: name
  real                 :: forward,dr
  real,    allocatable,dimension(:,:)     :: saa,sbb,smm,sa0,sb0
  real,    pointer    ,dimension(:,:,:)   :: imgx      
  real,    allocatable,dimension(:,:,:,:) :: sall
  real,    pointer    ,dimension(:,:,:,:) :: imgallx 
  complex, allocatable,dimension(:,:)     :: sab
  complex, pointer    ,dimension(:,:,:)   :: swfin,swf 
  complex, allocatable,dimension(:,:,:)   :: srefall
  complex, pointer    ,dimension(:,:,:,:) :: swfall  
  logical              :: verbose

  call sep_init()
  call from_param("method",method,0)    !! propagation type
  call from_param("forward",forward,0.) !! Forward scattering option
  call from_param("nref",nref,2)        !! # reference velocities
  call from_param("nh",nh,1)            !! # image offset in x
  call from_param("verbose",verbose,.false.)
  call hinaxis(ax,1)
  call hinaxis(aw,2) 
  call hinaxis(as,3)

  name="sabm";
  call ainaxis(az,1,name)
  name="sabref"
  call ainaxis(ar,2,name)

  if(nref .eq. 1 ) ar%n=1

  if (verbose) then
     write(0,*) ax%n,ax%o,ax%d
     write(0,*) az%n,az%o,az%d
     write(0,*) aw%n,aw%o,aw%d
     write(0,*) ar%n,ar%o,ar%d
     write(0,*) as%n,as%o,as%d
  end if

  !! . . Create offset axis
  !! . . Output Image
  dr=ax%d
  call houaxis(az,1)
  call houaxis(ax,2)
  call houaxis(as,3)
  call to_history("n4",nh)
  call to_history("d4",dr)
  call to_history("o4",-(nh-1)/2.*dr)
  call to_history("esize",4)

  !! . . Output Wavefields
  name="swfout"
  call aouaxis(aw,1,name)
  call aouaxis(ax,2,name)
  call aouaxis(as,3,name)
  call aouaxis(az,4,name)
  call auxputch("esize","i",8,name)
  call auxputch("label1","s","frequency",name)
  call auxputch("label2","s","midpoint",name)
  call auxputch("label3","s","shot",name)
  call auxputch("label4","s","depth",name)

  !! . . Allocate input wavefields 
  allocate( swfin(ax%n,aw%n,as%n) )

  !! . . Allocate arrays for global coefficients
  allocate(swfall(ax%n,aw%n,as%n,az%n))
  allocate(imgallx(az%n,ax%n,as%n,nh))

  !! . . Allocate arrays for individual shots
  allocate(swf(ax%n,az%n,aw%n))
  allocate(imgx(az%n,ax%n,nh))

  !! . . Coordinate system variable initialization
  allocate(sall(az%n,ax%n,as%n,5) )
  allocate(srefall(az%n,ar%n,as%n) )
  allocate(saa(az%n,ax%n),sbb(az%n,ax%n))
  allocate(smm(az%n,ax%n))
  allocate(sab(az%n,ax%n),sa0(az%n,ar%n))
  allocate(sb0(az%n,ar%n))   

  !! . . Read in the coefficients
  call sreed("sabm",sall,size(sall)*4)
  call sreed("sabref",srefall,size(srefall)*8)

  !! . . Read in RWF and SWF
  call sep_read(swfin)
  swfall(:,:,:,1)=swfin

  !! . . Initialize module and image space
  call wemig_init(ax,az,aw,ar,method,forward,nh); imgallx=0.;

  do ii=1,as%n   !! . . Loop over shot points
     imgx=0.; swf=0.;saa=0.;sbb=0.;sa0=0.;sb0=0.;swf=0.;
!     swf=swfin(:,:,ii)
     swf(:,1,:)=swfall(:,:,ii,1)
     saa=sall(:,:,ii,1); sbb=sall(:,:,ii,2); smm=sall(:,:,ii,3)
     sa0=real(srefall(:,:,ii)); sb0=aimag(srefall(:,:,ii))

     if (verbose) then
        write(0,*) "saa",minval(saa),maxval(saa)
        write(0,*) "sbb",minval(sbb),maxval(sbb)
        write(0,*) "smm",minval(smm),maxval(smm)
        write(0,*) "sa0",minval(sa0),maxval(sa0)
        write(0,*) "sb0",minval(sb0),maxval(sb0)
     end if

     call wemig(swf,imgx,saa,sbb,smm,sa0,sb0,ii)
     do iz=1,az%n
        swfall (:,:,ii,iz)=swf(:,iz,:); !! Output wavefield
     end do
     imgallx(:,:,ii,:)=imgx !! Output autocorr image

  end do !! Shot-point loop
  call sep_write(imgallx)

  call srite(name,swfall,size(swfall)*8)

  call sep_close()
  call exit()
end program SourceProp
