program RCabcd2_
  use sep
  use utils_

  implicit none

  complex, pointer :: rays(:,:)
  real,    pointer :: x(:,:),z(:,:)
  real,    pointer :: h1(:,:),h2(:,:),h21(:,:),h12(:,:)
  real,    pointer :: m1(:,:),m2(:,:)
  real,    pointer :: a(:,:),b(:,:),c(:,:),d(:,:)

  type(myaxis) ::at,ag
  integer      ::it,ig,j
  real         ::gx,gz
  real         ::eps,peps

  call sep_init()
  call hinaxis(at,1)
  call hinaxis(ag,2)
  call to_history("esize",4)
  call to_history("n3",4)
  call from_param("peps",peps,0.1)
  call sep_close()
  allocate(rays(at%n,ag%n))
  allocate(  h1 (at%n,ag%n),h2 (at%n,ag%n), h21(at%n,ag%n), h12(at%n,ag%n)  )
  allocate(  m1 (at%n,ag%n),m2 (at%n,ag%n))
  allocate(   x(at%n,ag%n))
  allocate(   z(at%n,ag%n))
  allocate(   a(at%n,ag%n),b(at%n,ag%n),c(at%n,ag%n),d(at%n,ag%n))

  call sep_read(rays)
  z=aimag(rays)
  x= real(rays)

  !! h1 is alpha: apply deviative along ray
  do it=2,at%n-1
     do ig=1,ag%n
        gx = (x(it+1,ig)-x(it-1,ig))/(2.*at%d)
        gz = (z(it+1,ig)-z(it-1,ig))/(2.*at%d)
        h1(it,ig) = sqrt(gx*gx+gz*gz)
     end do
  end do
  h1(1,:)    = h1(2,:)
  h1(at%n,:) = h1(at%n-1,:)
  h1(:,1)    = h1(:,2)
  h1(:,ag%n) = h1(:,ag%n-1)

  !! h2 is Jacobian: apply derivate across ray
  do it=1,at%n
     do ig=2,ag%n-1
        gx = (x(it,ig+1)-x(it,ig-1))/(2.*ag%d)
        gz = (z(it,ig+1)-z(it,ig-1))/(2.*ag%d)
        h2(it,ig) = sqrt(gx*gx+gz*gz)
     end do
  end do
  h2(:,1)    = h2(:,2)
  h2(:,ag%n) = h2(:,ag%n-1)
  h2(1,:)    = h2(2,:)
  h2(at%n,:) = h2(at%n-1,:)

  !! m1 = d(h2/h1)/dt = ( d(h2)/dt * h1 - h2 * d(h1)/dt )/ h1^2 
  h21=h2/h1
  h12=h1/h2
  do it=2,at%n-1
     do ig=1,ag%n
        !        m1(it,ig) = (h21(it+1,ig)-h21(it-1,ig))/(2.*at%d)
        m1(it,ig) &
        = h1(it,ig)* (h2(it+1,ig)-h2(it-1,ig))/(2.*at%d) &
        + h2(it,ig)* (h1(it+1,ig)-h1(it-1,ig))/(2.*at%d)
     end do
  end do
  m1(1,:)    = m1(2,:)
  m1(at%n,:) = m1(at%n-1,:)
  m1(:,1)    = m1(:,2)
  m1(:,ag%n) = m1(:,ag%n-1)

  !! m2 = d(h1/h2)/dg = ( d(h1)/dt * h2 - h1 * d(h2)/dt )/ h2^2 
  do it=1,at%n
     do ig=2,ag%n-1
        !        m2(it,ig) = (h12(it,ig+1)-h12(it,ig-1))/(2.*ag%d)
        m2(it,ig) &
        = h2(it,ig)* (h1(it,ig+1)-h1(it,ig-1))/(2.*ag%d) &
        + h1(it,ig)* (h2(it,ig+1)-h2(it,ig-1))/(2.*ag%d)
     end do
  end do
  m2(:,1)    = m2(:,2)
  m2(:,ag%n) = m2(:,ag%n-1)
  m2(1,:)    = m2(2,:)
  m2(at%n,:) = m2(at%n-1,:)

  eps=peps*( maxval(h1) + minval(h1) )/2.
  where( h1 < eps)
     h1=eps
  end where

  !! . . Force conditions on Jacobian
  eps=peps*( maxval(h2) + minval(h2) )/2.
  where( h2 < eps)
     h2=eps
  end where

  a= 1./(h1*h1)
  d= 1./(h2*h2)
  b= m1*a/(h1*h2) 
  c= m2*d/(h1*h2)

  call sep_write(a)
  call sep_write(b)
  call sep_write(c)
  call sep_write(d)

  call exit(0)
end program RCabcd2_









