module RWE2D_wemig
  use sep
  use util_
  use fft
  use geometry2D
  use sep_timers_mod

  implicit none
  integer, private :: nz,nx,nw,nr,nloop,nref,ntap
  real,    private :: dz,dx,dw,oz,ox,ow,okx,dkx,forward,eps,pi
  logical, private :: kin,verbose

  integer, allocatable,private :: allmask(:)
  real,    allocatable,private :: msk(:),mtt(:),tap(:),ikz2d(:)
  real,    allocatable,private :: allrefs(:,:),allfields(:,:)
  complex, allocatable,private :: swfl(:),swtt(:),rwfl(:),rwtt(:),rdax(:),sdax(:)

contains

  !----------------------------------------------------------------  
  subroutine wemig_init(ax,az,aw,ar,forward_in,kin_in,nref_in,verbose_in)
    type(myaxis)     :: ax,az,aw,ar
    integer          ::                               nref_in
    real             ::             forward_in
    logical          ::                        kin_in,        verbose_in
    real             :: pi
    pi=acos(-1.)
    forward=forward_in
    nz=az%n;    nx=ax%n;    nw=aw%n
    dz=az%d;    dx=ax%d;    dw=aw%d * 2.*pi
    oz=az%o;    ox=ax%o;    ow=aw%o * 2.*pi

    nref=nref_in

    call from_param("ntap",ntap,5)
    call from_param("nloop",nloop,3)
    okx=  -pi/    dx
    dkx=2.*pi/(nx*dx)

    kin=kin_in
    verbose=verbose_in;    
    eps=0.0000001

    allocate(rdax(nx),sdax(nx))
    allocate(ikz2d(nx),swfl(nx))
    allocate(swtt(nx),rwfl(nx))
    allocate(rwtt(nx),msk(nx));
    allocate(mtt(nx),tap(nx))
    allocate(allmask  (nx)   )
    allocate(allfields(nx,10) ) 
    allocate(allrefs(nref,10))
    call rwetaper_init()

     msk=1.;mtt=1.;tap=1.;ikz2d=1.
  end subroutine wemig_init

  !----------------------------------------------------------------

  !! WE migration driver
  subroutine wemig(rwf,swf,Pimg,iss,cutfields,refs,cutmask)
    integer          :: iw,iz,ix,ir,iss,cutmask(:,:),t1,t2,t3,t4,t5,t6
    real             :: rarg,saxmax,sarg,w
    real             :: Pimg(:,:),refs(:,:,:),cutfields(:,:,:)
    complex,pointer  :: rwf(:,:),swf(:,:),rax(:),sax(:)
!    logical          :: logic

 !   logic=init_sep_timers()
 !   logic=setup_next_timer("WLoop",t1)
 !   logic=setup_next_timer("PHS",t2)
 !   logic=setup_next_timer("SSF",t3)
 !   logic=setup_next_timer("REF",t4)

    !! . . Forward Scattering Option
    sarg=-1.; !! if ( forward .ne. 0. ) sarg=1.
    rarg= 1.
    write(0,*) 'STARTING MIGRATION'
    write(0,*) 'USING REFERENCE MAX OF :',nref

    do iw=1,nw  !! Frequency Loop           
!       call start_timer_num(t1)

       w=ow+(iw-1)*dw
       rax => rwf(:,iw)
       sax => swf(:,iw)
       saxmax=maxval(cabs(sax))
       do iz=2,nz

 !         call start_timer_num(t3)
          call wemig_ssfboth(sax,rax,w,cutmask(:,iz),cutfields(:,:,iz),&
                              refs(:,:,iz),rarg,sarg)
  !        call stop_timer_num(t3)

          rwtt=rax;rax=0. 
          swtt=sax;sax=0.

          call fth(.false.,.false.,swtt)
          call fth(.false.,.false.,rwtt)

	  do ir=1,nref  
             if (refs(ir,4,iz) .lt. 0.000000001) cycle 
             rwfl=rwtt
             swfl=swtt

   !          call start_timer_num(t2)
             call wemig_phsboth(swfl,rwfl,w,iz,ir,refs(:,:,iz),rarg,sarg)
    !         call stop_timer_num(t2)

             call fth( .true.,.false.,swfl)
             call fth( .true.,.false.,rwfl)

     !        call start_timer_num(t4)
             call wemig_refboth(swfl,rwfl,iz,ir,cutmask(:,iz),sax,rax)
      !       call stop_timer_num(t4)

          end do 

          call rwetaper(rax) 
          call rwetaper(sax)

          Pimg(iz,:)=Pimg(iz,:)+real(rax*conjg(sax))/&
                                  max( maxval(real(sax*conjg(sax))) ,0.01 )

       end do !! . . z-loop

       write(0,*) "SSF",iw,nw,iss,maxval(abs(Pimg(:,:)))
 !      call stop_timer_num(t1)

    end do  !! . . w-loop
    write(0,*) 'FINISHED OFF SHOT!'

!    call print_timers()
  end subroutine wemig

  !----------------------------------------------------------------
  !! split-step Fourier
  subroutine wemig_ssfboth(sdax,rdax,w,allmask,allfields,allrefs,rdir,sdir) 
    integer :: ix,allmask(:)
    real    :: w,sdir,rdir,allfields(:,:),allrefs(:,:)
    complex :: sdax(:),rdax(:)

    do ix=1,nx
       ikz2d(ix)=-dz*(w**2*allrefs(allmask(ix),4 )*&
       (allfields(ix,4 )-allrefs(allmask(ix),4))-&
       allrefs(allmask(ix),10)*(allfields(ix,10)-allrefs(allmask(ix),10)))/&
       sqrt(abs(allrefs(allmask(ix),4)**2*w**2-allrefs(allmask(ix),10)**2)+eps ) 
    end do

    rdax = rdax*cmplx( cos(ikz2d) ,rdir*sin(ikz2d) )
    sdax = sdax*cmplx( cos(ikz2d) ,sdir*sin(ikz2d) )

  end subroutine wemig_ssfboth
  !----------------------------------------------------------------  

  !! Fourier-domain phase shift
  subroutine wemig_phsboth(sdax,rdax,w,iz,ir,allrefs,rdir,sdir)
    integer :: ix,iz,ir
    real    :: w,kx,rdir,sdir,carg,carg1,allrefs(:,:),pkz
    complex :: kz,rdax(:),sdax(:)

    carg1= (allrefs(ir,4)* w)**2 - allrefs(ir,10)**2
 
    do ix=1,nx
       kx=okx+(ix-1)*dkx

       carg = carg1 - (allrefs(ir,5)*kx)**2 
       
       if ( carg .lt. 0) then

          pkz = -abs(dz*sqrt(-carg))
          rdax(ix)=rdax(ix)*exp(pkz)
          sdax(ix)=sdax(ix)*exp(pkz)   

       else

          pkz =-dz*( sqrt(carg)-kx*allrefs(ir,1) )
          rdax(ix)=rdax(ix)*cmplx(cos(pkz),rdir*sin(pkz))
          sdax(ix)=sdax(ix)*cmplx(cos(pkz),sdir*sin(pkz))

       end if
    end do

  end subroutine wemig_phsboth


   !----------------------------------------------------------------

  subroutine wemig_refboth(swfl,rwfl,iz,ir,allmask,sdax,rdax)
    complex :: sdax(:),rdax(:),swfl(:),rwfl(:)
    integer :: iz,ir,ix,ii,allmask(:)

    mtt=0.
    where( allmask .eq. ir) 
       mtt=1.
    end where

    do ii=1,nloop !! Smoothing Iterations
       msk(1 ) = mtt(1   )
       msk(nx) = mtt(nx-1)
       do ix=2,nx-1
          msk(ix)=( mtt(ix-1)+mtt(ix+1) )/2.
       end do
       mtt=msk
    end do

    sdax = sdax + msk*swfl
    rdax = rdax + msk*rwfl

  end subroutine wemig_refboth

  !----------------------------------------------------------------

  subroutine rwetaper_init()
    integer :: itap,jtap,ix
    real :: pi
    pi=acos(-1.)

    if (ntap .gt. 1) then

       do itap=1,ntap
          jtap = abs(ntap-itap)
          tap(   itap) = cos(pi/2.* jtap/(ntap-1))
       end do

       do itap=1,ntap
          jtap = abs(ntap-itap)
          tap(nx-itap+1) = cos(pi/2.* jtap/(ntap-1))
       end do

    end if

  end subroutine rwetaper_init

  !----------------------------------------------------------------

  subroutine rwetaper(dax)
    complex, pointer  :: dax(:)
    dax = dax * tap
  end subroutine rwetaper

  !----------------------------------------------------------------

end module  RWE2D_wemig
