module RWE3D_wemig
  use sep
  use util_
  use fft
  use lloyd_mod

  implicit none
  integer, private :: nz,nx,ny,nw,nr
  real,    private :: dz,dx,dy,dw
  real,    private :: oz,ox,oy,ow

  integer          :: ntap
  integer, private :: nloop,nref

  real   , private :: forward,eps
  real   , private :: okx,dkx,oky,dky
  logical, private :: kin,verbose

  integer, allocatable, private :: iregion(:),nstart(:)

  real,    allocatable, private :: scales(:),Vel(:,:,:),have_zeros(:)
  real,    allocatable,dimension(:,:),private :: g11,g12,g13,g22,g23,g33
  real,    allocatable,dimension(:,:),private :: iG11,iG12,iG13,iG22,iG23,iG33
  real,    allocatable,dimension(:,:),private :: m11,m12,m13,m22,m23,m33
  real,    allocatable,dimension(:,:),private :: gdetr,n1,n2,n3,slice
  real,    allocatable,dimension(:,:),private :: msk,mtt,tap,vslice
  complex, allocatable,dimension(:,:),private :: wfl,wtt,ikz2d

contains
  !----------------------------------------------------------------  
  subroutine wemig_init(ax,ay,az,aw,ar,forward_in,kin_in,nref_in,verbose_in)
    type(myaxis)     :: ax,az,ay,aw,ar
    integer          ::                                  nref_in
    real             ::             forward_in
    logical          ::                           kin_in,        verbose_in
    real             :: pi

    pi=acos(-1.)

    forward=forward_in
    nz=az%n
    nx=ax%n
    ny=ay%n
    nw=aw%n

    dz=az%d
    dx=ax%d
    dy=ay%d
    dw=aw%d * 2.*pi

    oz=az%o
    ox=ax%o
    oy=ay%o
    ow=aw%o * 2.*pi

    nref=nref_in
    
    nloop=3
    ntap=1
    okx=  -pi/    dx
    dkx=2.*pi/(nx*dx)
    oky=  -pi/    dy
    dky=2.*pi/(ny*dy)

    kin=kin_in
    verbose=verbose_in
    eps=0.0001;

    allocate(ikz2d(nx,ny))
    allocate(wfl(nx,ny))
    allocate(wtt(nx,ny))
    allocate(msk(nx,ny)); msk=1.
    allocate(mtt(nx,ny)); mtt=1.
    allocate(tap(nx,ny)); tap=1.

    !! Global for reference call
    allocate(scales(10)); scales=0.
    allocate(slice(nx*ny,10)); slice=0.
    allocate(iregion(nx*ny)); iregion=0;
    allocate(nstart(10));nstart=0.;
    allocate(vslice(nx,ny) );vslice=0.;
    allocate(have_zeros(10));have_zeros=0.;

    !! For geometry
    allocate(g11(nx,ny),g12(nx,ny),g13(nx,ny),g22(nx,ny),g23(nx,ny),g33(nx,ny))
    allocate(iG11(nx,ny),iG12(nx,ny),iG13(nx,ny),iG22(nx,ny),iG23(nx,ny),iG33(nx,ny))
    allocate(m11(nx,ny),m12(nx,ny),m13(nx,ny),m22(nx,ny),m23(nx,ny),m33(nx,ny))
    allocate(gdetr(nx,ny),n1(nx,ny),n2(nx,ny),n3(nx,ny) )
    g11=0.;g12=0.;g13=0.;g22=0.;g23=0.;g33=0.;
    iG11=0.;iG12=0.;iG13=0.;iG22=0.;iG23=0.;iG33=0.;
    m11=0.;m12=0.;m13=0.;m22=0.;m23=0.;m33=0.;
    gdetr=0.;n1=0.;n2=0.;n3=0.;

!    call rwetaper_init()
  end subroutine wemig_init
!!$  !----------------------------------------------------------------
!!$  subroutine rwetaper_init()
!!$    integer :: itap,jtap
!!$    real    :: pi
!!$
!!$    pi=acos(-1.)
!!$
!!$    if(ntap>1) then
!!$       do itap=1,ntap
!!$          jtap = abs(ntap-itap)
!!$          tap(   itap  ) = cos(pi/2.* jtap/(ntap-1))
!!$       end do
!!$       do itap=1,ntap
!!$          jtap = abs(ntap-itap)
!!$          tap(nx-itap+1) = cos(pi/2.* jtap/(ntap-1))
!!$       end do
!!$    end if
!!$  end subroutine rwetaper_init
!!$
!!$  subroutine rwetaper(dax)
!!$    complex, pointer :: dax(:,:)
!!$    dax = dax * tap
!!$  end subroutine rwetaper
  !----------------------------------------------------------------
  !! WE migration driver
  subroutine wemig(rwf,swf,Pimg,rays,Vel,ishot)
    complex, pointer :: rwf (:,:,:)
    complex, pointer :: swf (:,:,:)
    real             :: Pimg(:,:,:)
    real             :: rays(:,:,:,:)
    real             :: Vel(:,:,:)
    integer :: iw,iz,ix,iy,ir,ishot
    real    :: w,sarg
    complex, pointer :: rax(:,:)
    complex, pointer :: sax(:,:)
    integer, allocatable :: allmask(:,:),mask(:,:,:)
    real,    allocatable :: allfields(:,:,:),ray(:,:,:,:)
    real,    allocatable :: allrefs(:,:)
    real,    allocatable :: refs(:,:,:),fields(:,:,:,:)
    allocate(mask(nx,ny,nz),refs(nref,10,nz),fields(nx,ny,10,nz) )
    allocate(allfields(nx,ny,10),ray(3,nx,ny,3) )
    allocate(allrefs  (nref,10) ,allmask(nx,ny) );
    allrefs=0.; allmask=0; allfields=0.; ray=0.;
    mask=0; refs=0.;fields=0.;
    sarg=-1.; if ( forward .ne. 0. ) sarg=1.

    call init_lloyd()

    do iw=1,nw  !! Frequency Loop           
       w=ow+(iw-1)*dw
       !rax => rwf(:,:,iw)
       sax => swf(:,:,iw)

       do iz=2,nz

          if (iw .eq. 1) then !! Get reference parameters
             if (iz .ne. nz ) then
                call get_fields(rays(iz-1:iz+1,:,:,:),Vel(iz,:,:),allfields)
             else 
                call get_fields(rays(iz-2:iz  ,:,:,:),Vel(iz,:,:),allfields)
                fields(:,:,:,1)=fields(:,:,:,2)
                refs(:,:,1)=refs(:,:,2)
                mask(:,:,1)=mask(:,:,2)
             end if
             call get_references(allfields,allrefs,allmask)
             fields(:,:,:,iz)=allfields
             mask(:,:,iz)=allmask
             refs(:,:,iz)=allrefs
          else
             allfields = fields(:,:,:,iz)
             allmask   = mask  (:,:,iz  )
             allrefs   = refs  (:,:,iz  )
          end if


          !          !! . . Do receiver wavefield
          !          wtt=raxve
          !          rax=0. 
          !          do ir=1,nr  
          !             wfl=wtt 
          !             call wemig_phs(wfl,w,iz,ir,a0,b0,c0,d0,1.)
          !             call wemig_ssf(wfl,w,iz,ir,a,a0,1.)
          !             call wemig_ref(wfl,  iz,ir,mm,rax)
          !          end do
          !call rwetaper(rax)

          !! . . Do source wavefield
          wtt=sax
          sax=0. 
          do ir=1,nref  
             wfl=wtt

             !! Do not do useless reference sets,
             if (allrefs(ir,4) .eq. 0. ) write(0,*) 'ZERO?'
             if (allrefs(ir,4) .le. 10e-10) cycle 

             call wemig_phs(wfl,w,iz,ir,allrefs,sarg)
             call wemig_ssf(wfl,w,iz,ir,allfields,allrefs,sarg)       
             if (nref .eq. 1 ) then
                sax = wfl
             else
                call wemig_ref(wfl,  iz,ir,allmask,sax)
             end if

          end do
          !call rwetaper(sax)

       !Pimg(iz,:,:) = Pimg(iz,:,:) + real( rax * conjg(sax) )
       Pimg(iz,:,:) = Pimg(iz,:,:) + real( sax )
      
       !write(0,*) 'iz',iz,maxval(Pimg(iz,:,:))
       end do !! . . z-loop
    if ( iw .eq. 1 ) then
        call coeff_report(fields,refs)
    end if
    write(0,*) "SSF",iw,nw,ishot,maxval(abs(Pimg(:,:,:)))
    end do  !! . . w-loop

  end subroutine wemig
  !----------------------------------------------------------------
  subroutine coeff_report(fields,refs)
    real  ::  fields(:,:,:,:),refs(:,:,:)
    write(0,*) 'METRIC COEFFICIENTS'
    write(0,*) "g11 min/max", minval(g11),maxval(g11)
    write(0,*) "g12 min/max", minval(g12),maxval(g12)
    write(0,*) "g13 min/max", minval(g13),maxval(g13)
    write(0,*) "g22 min/max", minval(g22),maxval(g22)
    write(0,*) "g23 min/max", minval(g23),maxval(g23)
    write(0,*) "g33 min/max", minval(g33),maxval(g33)
    write(0,*) " "

    write(0,*) 'INVERSE METRIC COEFFICIENTS'
    write(0,*) "G11 min/max", minval(iG11),maxval(iG11)
    write(0,*) "G12 min/max", minval(iG12),maxval(iG12)
    write(0,*) "G13 min/max", minval(iG13),maxval(iG13)
    write(0,*) "G22 min/max", minval(iG22),maxval(iG22)
    write(0,*) "G23 min/max", minval(iG23),maxval(iG23)
    write(0,*) "G33 min/max", minval(iG33),maxval(iG33)
    write(0,*) " "


    write(0,*) "COEFFICIENT RANGES"
    write(0,*) "a1 min/max", minval(fields(:,:,1,:)),maxval(fields(:,:,1,:))
    write(0,*) "a2 min/max", minval(fields(:,:,2,:)),maxval(fields(:,:,2,:))
    write(0,*) "a3 min/max", minval(fields(:,:,3,:)),maxval(fields(:,:,3,:))
    write(0,*) "a4 min/max", minval(fields(:,:,4,:)),maxval(fields(:,:,4,:))
    write(0,*) "a5 min/max", minval(fields(:,:,5,:)),maxval(fields(:,:,5,:))
    write(0,*) "a6 min/max", minval(fields(:,:,6,:)),maxval(fields(:,:,6,:))
    write(0,*) "a7 min/max", minval(fields(:,:,7,:)),maxval(fields(:,:,7,:))
    write(0,*) "a8 min/max", minval(fields(:,:,8,:)),maxval(fields(:,:,8,:))
    write(0,*) "a9 min/max", minval(fields(:,:,9,:)),maxval(fields(:,:,9,:))
    write(0,*) "a10 min/max", minval(fields(:,:,10,:)),maxval(fields(:,:,10,:))
    write(0,*) " "

    write(0,*) "REFERENCE PARAMETERS"
    write(0,*) "b1 min/max", minval(refs(:,1,:)),maxval(refs(:,1,:))
    write(0,*) "b2 min/max", minval(refs(:,2,:)),maxval(refs(:,2,:))
    write(0,*) "b3 min/max", minval(refs(:,3,:)),maxval(refs(:,3,:))
    write(0,*) "b4 min/max", minval(refs(:,4,:)),maxval(refs(:,4,:))
    write(0,*) "b5 min/max", minval(refs(:,5,:)),maxval(refs(:,5,:))
    write(0,*) "b6 min/max", minval(refs(:,6,:)),maxval(refs(:,6,:))
    write(0,*) "b7 min/max", minval(refs(:,7,:)),maxval(refs(:,7,:))
    write(0,*) "b8 min/max", minval(refs(:,8,:)),maxval(refs(:,8,:))
    write(0,*) "b9 min/max", minval(refs(:,9,:)),maxval(refs(:,9,:))
    write(0,*) "b10 min/max", minval(refs(:,10,:)),maxval(refs(:,10,:))
    
  end subroutine coeff_report

  !----------------------------------------------------------------

  subroutine get_fields(ray,vslice,allfields)
    real, dimension(:,:    ),intent(in ) :: vslice
    real, dimension(:,:,:,:),intent(in ) :: ray
    real, dimension(:,:,:  ),intent(out) :: allfields
    g11=0.;g12=0.;g13=0.;g22=0.;g23=0.;g33=0.;
    m11=0.;m12=0.;m13=0.;m22=0.;m23=0.;m33=0.;
    allfields=0.;

    call calc_g11(ray,g11)
    where (g11 .lt. 10e-4) g11=0

    call calc_g12(ray,g12)
    where (g12 .lt. 10e-4) g12=0

    call calc_g13(ray,g13)
    where (g13 .lt. 10e-4) g13=0

    call calc_g22(ray,g22)
    where (g22 .lt. 10e-4) g22=0

    call calc_g23(ray,g23)
    where (g23 .lt. 10e-4) g23=0

    call calc_g33(ray,g33)
    where (g33 .lt. 10e-4) g33=0

    call calc_determinant(g11,g12,g13,g22,g23,g33,gdetr)
   
 	call calc_inverse(g11,g12,g13,g22,g23,g33,iG11,iG12,iG13,iG22,iG23,iG33,gdetr)
 	
    m11=gdetr*iG11;
    m12=gdetr*iG12;
    m13=gdetr*iG13;
    m22=gdetr*iG22;
    m23=gdetr*iG23;
    m33=gdetr*iG33;
    call calc_n(m11,m12,m13,n1)
    call calc_n(m12,m22,m23,n2)
    call calc_n(m13,m23,m33,n3)

    !! Non-stationary coefficient fields
    allfields(:,:,1) = iG13/iG33
    allfields(:,:,2) = iG23/iG33
    allfields(:,:,3) = n3/(2.*m33)
    allfields(:,:,4) = sqrt(1./(vslice**2*iG33))
    allfields(:,:,5) = sqrt(iG11/iG33 - (iG13/iG33)**2)
    allfields(:,:,6) = sqrt(iG22/iG33 - (iG23/iG33)**2)
    allfields(:,:,7) = 2.*iG12/iG33 - 2.*iG13*iG23/(iG33**2)
    allfields(:,:,8) = n1/m33 - m13*n3/(m33**2)
    allfields(:,:,9) = n2/m33 - m23*n3/(m33**2)
    allfields(:,:,10)= n3/m33
 
    !! Set kinematic stuff to zero
    if (kin) then
       allfields(:,:,8)=0.
       allfields(:,:,9)=0.
    end if

  end subroutine get_fields

  !----------------------------------------------------------------
  
  subroutine calc_inverse(g11,g12,g13,g22,g23,g33,iG11,iG12,iG13,iG22,iG23,iG33,gdetr)
	real,dimension(:,:),intent(in ) :: g11,g12,g13,g22,g23,g33,gdetr
	real,dimension(:,:),intent(out) :: iG11,iG12,iG13,iG22,iG23,iG33
	
	iG11 = ( g22*g33-g23*g23 ) / gdetr**2
	iG12 = ( g13*g23-g12*g33 ) / gdetr**2
	iG13 = ( g12*g23-g13*g22 ) / gdetr**2
	iG22 = ( g11*g33-g13*g13 ) / gdetr**2
	iG23 = ( g13*g12-g11*g23 ) / gdetr**2
	iG33 = ( g11*g22-g12*g12 ) / gdetr**2
	
  end subroutine calc_inverse

  !----------------------------------------------------------------

  subroutine calc_determinant(g11,g12,g13,g22,g23,g33,gdetr)
    real, dimension(:,:),intent(in ) :: g11,g12,g13,g22,g23,g33
    real, dimension(:,:),intent(out) :: gdetr
  
   !	gdetr= sqrt( &
   !		g33*(g11*g22 - g12**2) * &
   ! 	(1.+(g11*g23**2+g22*g13**2-2.*g12*g13*g23))/&
   ! 	(g33*(g11*g22-g12**2)))
    gdetr = sqrt(+g11*g22*g33 &
    			 -g11*g23**2  &
    			 -g12**2*g33  &
    		  +2.*g12*g23*g13 &
    		  	 -g22*g13**2 )	
  end subroutine
  
  !----------------------------------------------------------------

  subroutine get_references(allfields,allrefs,allmask)
    integer :: iref,ii
    real,   intent(in ) :: allfields(:,:,:)
    real,   intent(out) :: allrefs(:,:)
    integer,intent(out) :: allmask(:,:)
    !! allfields (in)- Input fields to be used (nx,ny,10)
    !! scales (tmp)  - Scaling function (10)
    !! slice (tmp)   - 1D array used to get refs (nx*ny,10)
    !! iregion (tmp) - What region has the best attribute set (nx*ny)
    !! nstart (tmp)  - Number of regions (nref)
    !! allmask(out)  - Output mask (nx,ny)
    !! allrefs (out) - Reference values in (nref,10)
    scales=0.; slice=0.; iregion=0;
    nstart = (/ 1,1,1,1,1,1,1,1,1,1 /); 

    slice=reshape(allfields , (/nx*ny,10/) )

    do ii=1,10 !! Get new scaling function
       scales(ii) = sum( abs( slice(:,ii) ) )  / real(nx*ny)
    end do

    where (scales .eq. 0.) 
       have_zeros = 0.
    elsewhere
       have_zeros = 1. 
    end where

    do ii=1,10 !! Rescale Data
       if ( have_zeros(ii) .ne. 0.) then
          slice(:,ii)=slice(:,ii)/scales(ii)
       else
          slice(:,ii)=0. 
       end if
    end do

    iref=lloyd_go(slice,allrefs,nref,iregion,nstart) !! Call ND Lloyd's algorithm

    do ii=1,10  !! rescaling 
       if ( have_zeros(ii) .ne. 0.) then
          allrefs(:,ii)=allrefs(:,ii)*scales(ii)
       else
          allrefs(:,ii)=0. 
       end if
    end do

    if(iref < nref) then !! Mark unnecessary reference sets
       allrefs(iref+1:,:)=-1.
    end if

    allmask=reshape(iregion,(/nx,ny/) ) !! Mask

  end subroutine get_references

  !----------------------------------------------------------------

  subroutine wemig_ref(wfl,iz,ir,allmask,dax)
    complex, pointer :: dax(:,:)
    complex          :: wfl(:,:)
    integer          :: iz,ir,ix,iy,ii,allmask(:,:)

    where( allmask .eq. ir) 
       mtt=1.
    elsewhere
       mtt=0.
    end where

    do ii=1,nloop !! Smoothing Iterations

       msk(1,:) = mtt(1,:)
       msk(:,1) = mtt(:,1)

       do ix=2,nx-1
          do iy=2,ny-1
             msk(ix,iy)=( mtt(ix-1,iy)+mtt(ix+1,iy)+mtt(ix,iy-1)+mtt(ix,iy+1) )/4.
          end do
       end do

       msk(nx,:)=mtt(nx-1,:)
       msk(:,ny)=mtt(:,ny-1)

       mtt=msk
    end do
    dax = dax + msk*wfl
  !  write(0,*) minval(abs(wfl)),maxval(abs(wfl))
  end subroutine wemig_ref
  !----------------------------------------------------------------  
  !! Fourier-domain phase shift
  subroutine wemig_phs(dax,w,iz,ir,allrefs,dir)
    integer :: ix,iy,iz,ir
    real    :: w,kx,ky,dir,allrefs(:,:)
    complex :: kz,carg,dax(:,:)

    !! Initial Fourier Transform (x,y) -> (kx,ky)
    call ft1axis(.false.,.true. ,dax)
    call ft2axis(.false.,.false.,dax)

    !! Loop over all kx and ky
    do ix=1,nx
       kx=okx+(ix-1)*dkx
       do iy=1,ny
          ky=oky+(iy-1)*dky

          !! Evaluate the complex exponential argument
          carg = &
            allrefs(ir,4 )**2 *  w**2  &
          - allrefs(ir,5 )**2 * kx**2  &
          - allrefs(ir,6 )**2 * ky**2  &
          - allrefs(ir,7 )*kx*ky  &
          + allrefs(ir,8 )*cmplx(0.,1.)*kx  &
          + allrefs(ir,9 )*cmplx(0.,1.)*ky  & 
          - allrefs(ir,10)**2
 
        !  if (kx .eq. 0. .and. ky .eq. 0. .and. iz .eq. 2) then
        !     write(0,*) 150**2*carg/w**2,1./allrefs(ir,4)
        !  end if

          if ( aimag(carg) .ne. 0) then !! Complex Check

             if ( real(carg) .lt. 0) then
                kz = (-allrefs(ir,1)*kx-allrefs(ir,2)*ky+allrefs(ir,3)*cmplx(0.,1.)+csqrt(-carg))
                dax(ix,iy) = dax(ix,iy) * cexp( cmplx( -dz*real(kz),0. ) )
             else
                kz = (-allrefs(ir,1)*kx-allrefs(ir,2)*ky+allrefs(ir,3)*cmplx(0.,1.)+csqrt(+carg))
                dax(ix,iy) = dax(ix,iy) * cexp( cmplx( 0., -dz*dir*(aimag(kz)) ))
             end if
          else
             if ( real(carg) .lt. 0) then
                kz = (-allrefs(ir,1)*kx-allrefs(ir,2)*ky+allrefs(ir,3)*cmplx(0.,1.)+sqrt(-real(carg)))
                dax(ix,iy) = dax(ix,iy) * cexp( cmplx( -dz*abs(real(kz)),-dz*dir*aimag(kz) ) )
             else
                kz = (-allrefs(ir,1)*kx-allrefs(ir,2)*ky+allrefs(ir,3)*cmplx(0.,1.)+sqrt(+real(carg)))
                dax(ix,iy) = dax(ix,iy) * cexp( cmplx( -dz*abs(aimag(kz)), -dz*dir*real(kz)  ))
             end if

          end if  !! End Complex check
         ! if (kx .eq. 0. .and. ky .eq. 0. .and. iz .eq. 2) then
         !    write(0,*) kz*1500./w,dz
        !  end if
 
       end do
    end do

    !! Do Inverse Fourier transform (kx,ky) -> (x,y)
    call ft2axis( .true.,.false.,dax)
    call ft1axis( .true.,.true. ,dax)
    !   write(0,*) "PHS",minval(abs(dax)),maxval(abs(dax))

  end subroutine wemig_phs

  !----------------------------------------------------------------
  !! split-step Fourier
  subroutine wemig_ssf(dax,w,iz,ir,allfields,allrefs,dir) 
    complex :: dax(:,:)
    real    :: w,dir,allfields(:,:,:),allrefs(:,:)
    integer :: iz,ir

    !! Had -1 on complex before
    ikz2d =(cmplx(0., 1.)*allrefs(ir,3)*( allfields(:,:,3 ) -allrefs(ir,3 ) ))  &
                       + &
                       (  allrefs(ir,4) * w**2  * ( allfields(:,:,4 ) -allrefs(ir,4 ) ))/   &
                              sqrt( allrefs(ir,4)**2 * w**2 - allrefs(ir,10)**2 ) !!&
!!                       -  &
  !!                        allrefs(ir,10)*(  allfields(:,:,10) -allrefs(ir,10) ) /   &
    !!                          sqrt( allrefs(ir,4)**2 * w**2 - allrefs(ir,10)**2 )
!        ikz2d = cmplx(0.,-1.)*( &
!           (cmplx(0., 1.)*( allfields(:,:,3 ) -allrefs(ir,3 ) )) +  &
!          (w**2         *( allfields(:,:,4 ) -allrefs(ir,4 ) ))/   &
!                    (sqrt( allrefs(ir,4)*w**2-allrefs(ir,10) )) -  &
!                        (  allfields(:,:,10) -allrefs(ir,10) ) /   &
!                    (sqrt( allrefs(ir,4)*w**2-allrefs(ir,10) )) )
!   dax = dax * exp(-dz*abs(real(ikz2d))) * &
!                (cos(-dir*dz* aimag(ikz2d))+cmplx(0.,1.)*sin(-dz*dir*aimag(ikz2d)) )
!    dax = dax * cexp( cmplx( -dz*(real(ikz2d)),-dz*dir*(aimag(ikz2d)) ) )

    dax = dax * cexp( cmplx( -dz*abs(aimag(ikz2d)),-dz*dir*real(ikz2d)  ))

 !   write(0,*) allrefs(ir,4)*w**2-allrefs(ir,10),sqrt( allrefs(ir,4)*w**2-allrefs(ir,10) )
 !   write(0,*) "SSF",minval(abs(dax)),maxval(abs(dax))
   end subroutine wemig_ssf


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine calc_g11(ray,g11)
    real,intent(in ) :: ray(:,:,:,:)
    real,intent(out) :: g11(:,:)
    g11(2:nx-1,2:ny-1) = &
           ( ( ray(2,3:nx,2:ny-1,1)-ray(2,1:nx-2,2:ny-1,1) ) / (2.*dx) )**2 + &
           ( ( ray(2,3:nx,2:ny-1,2)-ray(2,1:nx-2,2:ny-1,2) ) / (2.*dx) )**2 + &
           ( ( ray(2,3:nx,2:ny-1,3)-ray(2,1:nx-2,2:ny-1,3) ) / (2.*dx) )**2 
    !! Side Boundaries
    g11(1 ,:) = g11(   2,:) 
    g11(nx,:) = g11(nx-1,:)
    g11(:,1 ) = g11(:,2   ) 
    g11(:,ny) = g11(:,ny-1)
 end subroutine calc_g11

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine calc_g22(ray,g22)
    real,intent(in ) :: ray(:,:,:,:)
    real,intent(out) :: g22(:,:)
    g22(2:nx-1,2:ny-1) = &
           ( ( ray(2,2:nx-1,3:ny,1)-ray(2,2:nx-1,1:ny-2,1) ) / (2.*dy) )**2 + &
           ( ( ray(2,2:nx-1,3:ny,2)-ray(2,2:nx-1,1:ny-2,2) ) / (2.*dy) )**2 + &
           ( ( ray(2,2:nx-1,3:ny,3)-ray(2,2:nx-1,1:ny-2,3) ) / (2.*dy) )**2 
    !! Side Boundaries
    g22(1 ,:) = g22(   2,:) 
    g22(nx,:) = g22(nx-1,:)
    g22(:,1 ) = g22(:,2   ) 
    g22(:,ny) = g22(:,ny-1)
  end subroutine calc_g22

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine calc_g33(ray,g33)
    real,intent(in ) :: ray(:,:,:,:)
    real,intent(out) :: g33(:,:)
    g33(2:nx-1,2:ny-1) = &
         ( ( ray(3,2:nx-1,2:ny-1,1)-ray(1,2:nx-1,2:ny-1,1) ) / (2.*dz) )**2 + &
         ( ( ray(3,2:nx-1,2:ny-1,2)-ray(1,2:nx-1,2:ny-1,2) ) / (2.*dz) )**2 + &
         ( ( ray(3,2:nx-1,2:ny-1,3)-ray(1,2:nx-1,2:ny-1,3) ) / (2.*dz) )**2 

    !! Side Boundaries
    g33(1 ,:) = g33(   2,:) 
    g33(nx,:) = g33(nx-1,:)
    g33(:,1 ) = g33(:,2   ) 
    g33(:,ny) = g33(:,ny-1)

  end subroutine calc_g33

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine calc_g12(ray,g12)
    real,intent(in ) :: ray(:,:,:,:)
    real,intent(out) :: g12(:,:)
    g12(2:nx-1,2:ny-1) = &
         ( ray(2,3:nx  ,2:ny-1,1)-ray(2,1:nx-2,2:ny-1,1) ) * &
         ( ray(2,2:nx-1,3:ny  ,1)-ray(2,2:nx-1,1:ny-2,1) ) / (4.*dx*dy) + &
         ( ray(2,3:nx  ,2:ny-1,2)-ray(2,1:nx-2,2:ny-1,2) ) * &
         ( ray(2,2:nx-1,3:ny  ,2)-ray(2,2:nx-1,1:ny-2,2) ) / (4.*dx*dy) + &
         ( ray(2,3:nx  ,2:ny-1,3)-ray(2,1:nx-2,2:ny-1,3) ) * &
         ( ray(2,2:nx-1,3:ny  ,3)-ray(2,2:nx-1,1:ny-2,3) ) / (4.*dx*dy) 
     !! Side boundaries
    g12(1 ,:) = g12(   2,:) 
    g12(nx,:) = g12(nx-1,:)
    g12(:,1 ) = g12(:,2   ) 
    g12(:,ny) = g12(:,ny-1)
  end subroutine calc_g12

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine calc_g13(ray,g13)
    real,intent(in ) :: ray(:,:,:,:)
    real,intent(out) :: g13(:,:)
    g13(2:nx-1,2:ny-1) = &
         ( ray(2,3:nx  ,2:ny-1,1)-ray(2,1:nx-2,2:ny-1,1) ) * &
         ( ray(3,2:nx-1,2:ny-1,1)-ray(1,2:nx-1,2:ny-1,1) ) / (4.*dx*dz) + &
         ( ray(2,3:nx  ,2:ny-1,2)-ray(2,1:nx-2,2:ny-1,2) ) * &
         ( ray(3,2:nx-1,2:ny-1,2)-ray(1,2:nx-1,2:ny-1,2) ) / (4.*dx*dz) + &
         ( ray(2,3:nx  ,2:ny-1,3)-ray(2,1:nx-2,2:ny-1,3) ) * &
         ( ray(3,2:nx-1,2:ny-1,3)-ray(1,2:nx-1,2:ny-1,3) ) / (4.*dx*dz) 
    !! Side boundaries
    g13(1 ,:) = g13(   2,:) 
    g13(nx,:) = g13(nx-1,:)
    g13(:,1 ) = g13(:,2   ) 
    g13(:,ny) = g13(:,ny-1)

  end subroutine calc_g13

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine calc_g23(ray,g23)
    real,intent(in ) :: ray(:,:,:,:)
    real,intent(out) :: g23(:,:)

    g23(2:nx-1,2:ny-1) = &
         ( ray(2,2:nx-1,3:ny  ,1)-ray(2,2:nx-1,1:ny-2,1) ) * &
         ( ray(3,2:nx-1,2:ny-1,1)-ray(1,2:nx-1,2:ny-1,1) ) / (4.*dy*dz) + &
         ( ray(2,2:nx-1,3:ny  ,2)-ray(2,2:nx-1,1:ny-2,2) ) * &
         ( ray(3,2:nx-1,2:ny-1,2)-ray(1,2:nx-1,2:ny-1,2) ) / (4.*dy*dz) + &
         ( ray(2,2:nx-1,3:ny  ,3)-ray(2,2:nx-1,1:ny-2,3) ) * &
         ( ray(3,2:nx-1,2:ny-1,3)-ray(1,2:nx-1,2:ny-1,3) ) / (4.*dy*dz)       
    !! Side boundaries
    g23(1 ,:) = g23(   2,:) 
    g23(nx,:) = g23(nx-1,:)
    g23(:,1 ) = g23(:,2   ) 
    g23(:,ny) = g23(:,ny-1)

  end subroutine calc_g23

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine calc_n(m11,m12,m13,n)
    real, intent(in) :: m11(:,:),m12(:,:),m13(:,:)
    real, intent(out):: n(:,:)

    n(2:nx-1,2:ny-1) = &
        ( m11(3:nx  ,2:ny-1) - m11(1:nx-2,2:ny-1) )/ (2.*dx) + &
        ( m12(2:nx-1,3:ny  ) - m12(2:nx-1,1:ny-2) )/ (2.*dy) 
        !! Oops ( m13(2:nx-1,2:ny-1) - m13(2:nx-1,2:ny-1) )/ (2.*dz) 

    !! Side boundaries
    n(1 ,:) = n(   2,:) 
    n(nx,:) = n(nx-1,:)
    n(:,1 ) = n(:,2   ) 
    n(:,ny) = n(:,ny-1)

  end subroutine calc_n


end module  RWE3D_wemig
