program SP_elastic
  use sep
  use util_
  use NewRWEwemig

  implicit none
  type(myaxis) :: ax,az,aw,ar,as

  integer :: method
  integer  :: iz,ii
  character(len=128)   :: name
  real,    allocatable :: aa(:,:),bb(:,:),mm(:,:)
  complex, allocatable :: ab(:,:)
  real,    allocatable :: a0(:,:),b0(:,:)
  complex, pointer     :: rwfall(:,:,:) !! receiver wavefield
  complex, pointer     :: swfall(:,:,:) !! source   wavefield
  real,    pointer     :: imgall(:,:,:) !! image
  complex, pointer     :: rwf(:,:) !! receiver wavefield
  complex, pointer     :: swf(:,:) !! source   wavefield
  real,    pointer     :: img(:,:) !! image

  call sep_init()
  call from_param("method",method,0)

  call hinaxis(ax,1)
  call hinaxis(aw,2)
  call hinaxis(as,3)

  name="abm"
  call ainaxis(az,1,name)
  name="abref"
  call ainaxis(ar,2,name)
  if(method==0) ar%n=1   !! pure finite-differences

  write(0,*) ax%n,ax%o,ax%d
  write(0,*) az%n,az%o,az%d
  write(0,*) aw%n,aw%o,aw%d
  write(0,*) ar%n,ar%o,ar%d
  write(0,*) as%n,as%o,as%d

  call houaxis(az,1)
  call houaxis(ax,2)
  call houaxis(as,3)

  call to_history("esize",4)

  allocate(swf(ax%n,aw%n))
  allocate(rwf(ax%n,aw%n))
  allocate(img(az%n,ax%n))
  allocate(swfall(ax%n,aw%n,as%n))
  allocate(rwfall(ax%n,aw%n,as%n))
  allocate(imgall(az%n,ax%n,as%n))

  ! . . Coordinate system variables
  allocate(aa(az%n,ax%n))
  allocate(bb(az%n,ax%n))
  allocate(mm(az%n,ax%n))
  call sreed("abm",aa,size(aa)*4)
  call sreed("abm",bb,size(bb)*4)
  call sreed("abm",mm,size(mm)*4)
  write(0,*) "aa",minval(aa),maxval(aa)
  write(0,*) "bb",minval(bb),maxval(bb)
  write(0,*) "mm",minval(mm),maxval(mm)

  ! . . Coordinate system reference variables
  allocate(ab(az%n,ax%n))
  allocate(a0(az%n,ar%n))
  allocate(b0(az%n,ar%n))
  call sreed("abref",ab,size(ab)*8)
  a0= real(ab)
  b0=aimag(ab)
  write(0,*) "a0",minval(a0),maxval(a0)
  write(0,*) "b0",minval(b0),maxval(b0)

  call wemig_init(ax,az,aw,ar,method)
  call sep_read(rwfall)
  call sreed("swf",swfall,size(swfall)*8)
  imgall=0.

  ! . . Loop over shot points
  do ii=1,as%n
     img=0.;rwf=0.;swf=0.;
     rwf=rwfall(:,:,ii)
     swf=swfall(:,:,ii)
     call wemig(rwf,swf,img,aa,bb,mm,a0,b0,ii)
     imgall(:,:,ii)=img
  end do

  call sep_write(imgall)

  call sep_close()
  deallocate(rwf,swf,img,rwfall,swfall,imgall)
end program NewRWEsp
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
