program RWE3D
  use sep
  use util_
  use RWE3D_wemig

  implicit none
  type(myaxis) :: ax,ay,az,aw,ar,as

  integer :: ix,iy,iz,ii,iw
  integer :: nref
  real    :: forward
  real,    allocatable,dimension(:,:,:,:):: raysin  !! Input Rayfield
  !  real,    allocatable,dimension(:,:,:,:,:):: raysin  !! Input Rayfield
  !  real,    allocatable,dimension(:,:,:,:)  :: rays    !! Single-shot rays
  real,    allocatable,dimension(:,:,:)    :: Vel     !! Velocity model  
  complex, allocatable,dimension(:,:,:,:)  :: rwfin,swfin !! input S&R wflds
  complex, pointer,dimension(:,:,:)    :: rwf,swf     !! S&R wfld
  real,    allocatable,dimension(:,:,:):: Pimg        !! Single shot-image 
  logical                              :: verbose,kin
  character(len=128) :: name

  !! Geometry
  integer, allocatable :: mask(:,:,:)
  real,    allocatable :: refs(:,:,:), fields(:,:,:,:)
  INTEGER DATE_TIME (8)
  CHARACTER (LEN = 10) BIG_BEN (3)


  call sep_init(SOURCE)
  call getch_add_string("head=/dev/null")
  call getch_add_string("noheader=y")
  call set_no_putch()
  call sep_begin_prog()

  call from_param("forward",forward,0.) !! Forward scattering option
  call from_param("nref",nref,1)        !! # reference velocities
  call from_param("verbose",verbose,.false.)
  call from_param("kinematic",  kin,.true.)
  name="rays"

  !! Read in input dimensions
  call from_aux("rwf","n1",ax%n)
  call from_aux("rwf","n2",ay%n)
  call from_aux("rwf","n3",aw%n)
  call from_aux("rwf","n4",as%n)
  call from_aux("rwf","d1",ax%d)
  call from_aux("rwf","d2",ay%d)
  call from_aux("rwf","d3",aw%d)
  call from_aux("rwf","d4",as%d)
  call from_aux("rwf","o1",ax%o)
  call from_aux("rwf","o2",ay%o)
  call from_aux("rwf","o3",aw%o)
  call from_aux("rwf","o4",as%o)
  call from_aux("rays","n1",az%n)
  call from_aux("rays","d1",az%d)
  call from_aux("rays","o1",az%o)

  write(0,*) 'DEPTH',az%n,az%o,az%d
  write(0,*) 'X1   ',ax%n,ax%o,ax%d
  write(0,*) 'X2   ',ay%n,ay%o,ay%d
  write(0,*) 'SHOTS',as%n,as%o,as%d
  write(0,*) 'FREQ ',aw%n,aw%o,aw%d

  !! . . Create offset axis
  call auxputch("n1","i",az%n,"image")
  call auxputch("n2","i",ax%n,"image")
  call auxputch("n3","i",ay%n,"image")
  call auxputch("n4","i",as%n,"image")
  call auxputch("d1","r",az%d,"image")
  call auxputch("d2","r",ax%d,"image")
  call auxputch("d3","r",ay%d,"image")
  call auxputch("d4","r",as%d,"image")
  call auxputch("o1","r",az%o,"image")
  call auxputch("o2","r",ax%o,"image")
  call auxputch("o3","r",ay%o,"image")
  call auxputch("o4","r",as%o,"image")
  call auxputch("esize","i",4,"image")

  allocate( rwfin(ax%n,ay%n,aw%n,as%n) , swfin(ax%n,ay%n,aw%n,as%n) )
  allocate(swf(ax%n,ay%n,aw%n),rwf(ax%n,ay%n,aw%n),Pimg(az%n,ax%n,ay%n) )
  allocate(raysin(az%n,ax%n,ay%n,3) )
  allocate( Vel(az%n,ax%n,ay%n) )
  allocate(mask(ax%n,ay%n,az%n),refs(nref,10,az%n),fields(ax%n,ay%n,10,az%n) )

  call sep_read(raysin,"rays")
  call sep_read(Vel,"vel")
  call sep_read(rwfin,"rwf")
  call sep_read(swfin,"swf")

  write(0,*) 'VELOCITY MIN/MAX', minval(Vel),maxval(Vel)
  write(0,*) 'RWF MIN/MAX', minval(abs(rwfin)),maxval(abs(rwfin))
  write(0,*) 'SWF MIN/MAX', minval(abs(swfin)),maxval(abs(swfin))

  !! . . Calculate Geometry
  CALL DATE_AND_TIME (BIG_BEN (1), BIG_BEN (2), BIG_BEN (3), DATE_TIME)

  write(0,*) 'START GEOMETRY: ',BIG_BEN(2)
  call geometry_init(az%n,ax%n,ay%n,az%d,ax%d,ay%d,kin,nref,verbose)
  call geometry_calc(raysin,Vel,fields,refs,mask)
  call geometry_close()
  CALL DATE_AND_TIME (BIG_BEN (1), BIG_BEN (2), BIG_BEN (3), DATE_TIME)
  write(0,*) 'END GEOMETRY: ',BIG_BEN(2)

  deallocate( raysin, Vel )

  !! . . Initialize module and image space
  call wemig_init(ax,ay,az,aw,ar,forward,kin,nref,verbose)

  do ii=1,as%n   !! . . Loop over shot points

     Pimg=0.;rwf=0.;swf=0.;
     rwf  = rwfin(:,:,:,ii)
     swf  = swfin(:,:,:,ii)
     call wemig(rwf,swf,Pimg,ii,fields,refs,mask)
     call sep_write(Pimg,"image")

  end do !! . . End Shot-point loop

  call sep_end_prog()
  call exit()
end program RWE3D
