program RWEconmap
  use SEP
  use conmap
  use prepconmap
  implicit none

  integer                                    :: nx,nt,k,nsamp
  real                                       :: dxin,oxin,depthin
  real(KIND=8)                               :: dx  ,ox  ,depth  
  real(KIND=8)                               :: xcent,zcent
  complex     , dimension(:),allocatable     :: topozin
  real(KIND=8), dimension(:),allocatable     :: points,tline

  complex        ,dimension(:,:),allocatable :: zout
  complex(KIND=8),dimension(:),allocatable   :: z,topoz,endpoints
  complex,dimension(:),allocatable   :: bndout
  call sep_init()
  call from_history("n1",nx         ) !! total # of topo samples 
  call from_history("d1",dxin       ) !! total dist b/w topo samples
  call from_history("o1",oxin       ) !! offset of first topo sample
  call from_param("depth",depthin   ) !! Depth extension of grid
  call from_param("nt"  ,nt         ) !! number of depth steps
  allocate( tline(nx),points(nx),endpoints(5),z(4*nx+1) )
  allocate( zout(nt,nx),topoz(nx),topozin(nx),bndout(4*nx+1) )
  dx=dxin; ox=oxin; depth=depthin;

  !! . . Conformal mapping centers + desamp
  nsamp=2; zcent=depth; 

  call sep_read ( topozin )
  topoz  = topozin
  tline  = real(topozin)
  points = aimag(topozin)

  xcent=(real(topoz(nx))-real(topoz(1)))/2. + ox

  !! . . Generate appropriate boundary points
  call prepconmap_init(nx,dx,ox,nsamp,xcent,zcent,depth)
  call prepconmap_main(tline,points,endpoints,z)

  !! . . Writeout boundary file  
  call auxputch("n1","i",4*nx,"border.H")
  call auxputch("d1","r",1  ,"border.H")
  call auxputch("o1","r",0   ,"border.H")
  call auxputch("esize","i",8,"border.H")
  bndout=z
  call sep_write( bndout,"border.H",4*nx+1,8)

  !! . . Generate conformal map
  call conmap_init(nx,dx,ox,nsamp,depth,nt)
  call conmap_main(z,topoz,endpoints,zout)

  call hclose()

  call sep_write(zout)
  call exit()
end program RWEconmap
