module  prepconmap
  use sep 
  implicit none
  integer     , private :: nsamp,nx
  real(KIND=8), private :: xcent,zcent,dx,ox,depth
contains
  !-------------------------------------------------------------------
  subroutine prepconmap_init(nxin,dxin,oxin,nsampin,xcentin,zcentin,depthin)
    integer        :: nxin,nsampin
    real(KIND=8)   :: xcentin,zcentin,dxin,oxin,depthin
    nx=nxin;dx=dxin;ox=oxin;nsamp=nsampin;xcent=xcentin
    zcent=zcentin;depth=depthin
  end subroutine prepconmap_init
  !-------------------------------------------------------------------
  subroutine prepconmap_main(tline,points,endpoints,z)
    integer                                  :: i,j,cnt
    integer,dimension(:),allocatable         :: nsline,endline
    real(KIND=8    )                         :: dsl,dsr,dsy,dsx
    real(KIND=8),dimension(:)                :: tline,points
    real(KIND=8),dimension(:),allocatable    :: yvals,endvals
    complex(KIND=8), dimension(:),intent(out):: z,endpoints
    allocate( nsline(nsamp),yvals(nsamp) )

    !! . . Interpolate top points
    do i=1,nx-nsamp,nsamp
       nsline=(/ (j,j=i,i+nsamp-1) /)
       dsy=points(nsline(nsamp))-points(nsline(1))
       dsx=tline (nsline(nsamp))-tline (nsline(1))
       yvals=dsy/dsx*(nsline-nsline(1))*dx+points(nsline(1))
       z(nsline)=cmplx(tline(nsline),yvals)
    end do

    !! . . Deal with end of line issues
    allocate( endline(nx+1-i), endvals(nx+1-i) )
    endline=(/ (j,j=i,nx) /)
    dsy=points(endline(size(endline)))-points(endline(1))
    dsx=tline (endline(size(endline)))-tline (endline(1))
    endvals=dsy/dsx*(endline-endline(1))*dx+points(endline(1))
    z(endline)=cmplx(tline(endline),endvals)
    deallocate (endline,endvals)

    !! . . Right side
    dsr=2.*(depth-points(nx))/(nx-1)
    do i=1,nx
       z(nx+i)=cmplx(tline(nx)+dx,dble(i-1)*dsr+points(nx))
    end do
    !! . . Bottom points
    cnt=1
    do i=nx,1,-1
       z(2*nx+cnt)=conjg(z(i))+cmplx(0.,2.*depth)
       cnt=cnt+1
    end do
    !! . . Left side
    dsl=2.*(depth-points(1))/(nx-1)
    cnt=1
    do i=nx,1,-1
       z(3*nx+cnt)=cmplx(tline(1)-dx,dble(i-1)*dsl+points(1))
       cnt=cnt+1
    end do
    endpoints(1) = z(nx+1  )
    endpoints(2) = z(2*nx  )
    endpoints(3) = z(3*nx+1)
    endpoints(4) = z(4*nx  )  
    endpoints(5) = cmplx(xcent,zcent)
    z(4*nx+1)=endpoints(5)
  end subroutine prepconmap_main
  !-------------------------------------------------------------------
end module prepconmap
