!!-----------------------------------------------------------
!! Program Rec2PS_baz
!!
!! Takes a three component recording for one event at all stations
!! along array rotates it either from 
!!         [P SH SV] --> [Easting Northing Vertical] (forward process) 
!!         [Easting Northing Vertical] --> [P SV SH] (adjoint process)
!!
!!   Input:
!!
!!   forward: model in order  [SV SH P]
!!   Adjoint: data  in order  [Northing Easting Vertical]
!!
!!  Needs:
!!     aaz = azimuth of the array clockwise from an N-S orientation (param degrees)
!!       p = slowness of event (header - s/km) 
!!     baz = backazimuth of the event (relative to N=0 deg) (header)
!!   alpha = P-wave velocity at the surface (parameter km/s)
!!    beta = S-wave velocity at the surface (parameter km/s)
!!
!!-----------------------------------------------------------
program Rec2PS_baz 
  use R2PS_baz
  use SEP
  implicit none 
  integer                                 :: n1,n2,n3		
  real, dimension(:,:,:), allocatable     :: modl,data
  real	                                  :: alpha,beta,p1,p2,p,baz,pi,aaz
  logical                                 :: adj,add
  call sep_init()
  call from_param("alpha",alpha         )
  call from_param("beta", beta          )
  call from_history("n1"   ,n1          )
  call from_history("n2"   ,n2          )
  call from_history("n3"   ,n3          )
  call from_history("p "   ,p           )
  call from_history("baz"  ,baz         )
  call from_param("adj"    ,adj,  .true.)  
  call from_param("aaz"    ,aaz         ) 
  allocate( modl(n1,n2,n3), data(n1,n2,n3) )
  pi=acos(-1.)  

  p1=-p*cos( baz/180*pi + aaz/180*pi )
  p2=-p*sin( baz/180*pi + aaz/180*pi )

  call to_history("p1",p1)
  call to_history("p2",p2)
  call hclose()

  !! . . Read in File
  if (adj) then
     call sep_read( data )
  else 
     call sep_read( modl )
  end if

  !! . . Initialize
  call R2PS_baz_init(n1,n2,n3,alpha,beta,p1,p2)
  
  !! . . Rotation subroutine
  call R2PS_baz(adj,.false.,data,modl)

  !! . . Output file
  if (adj) then
     call sep_write(modl) 
  else
     call sep_write(data)
  end if
  deallocate(data,modl)
end program Rec2PS_baz


