program NewPot2D
  !! This program does post-coordinate system generation
  !! Smoothing using Laplacian regularization operator
  use sep
  use cgstep_mod
  use laplac2
  use igrad1
  use solver_reg_mod
  use mask1
  use NewPot2D_

  implicit none
  integer                               :: nt,ng,ii,stat,it
  integer                               :: niter,jj,ig
  real                                  :: dg,dt,og,ot,d1out,o1out
  real,    dimension(:,:), allocatable  :: Xdata,Zdata,Xmodel,Zmodel
  real,    dimension(:,:), allocatable  :: Jac
  complex, dimension(:,:), allocatable  :: Surf 
  complex, dimension(:,:), allocatable  :: Output 
  real   , dimension(:  ), allocatable  :: Xnd,Znd,PX,PZ,WJ
  logical, dimension(:  ), pointer      :: Xmsk,Zmsk
  logical, dimension(:,:), allocatable  :: Xfix,Zfix
  call sep_init()
  call from_history("n1",nt)
  call from_history("o1",ot)
  call from_history("d1",dt)
  call from_history("n2",ng)
  call from_history("o2",og)
  call from_history("d2",dg)

  call from_param("niter",niter,100) 

  allocate( Surf(nt,ng))
  allocate( Xdata(nt,ng),Xmodel(nt,ng) )
  allocate( Zdata(nt,ng),Zmodel(nt,ng) )
  allocate( Xfix (nt,ng),Zfix (nt,ng)  )
  allocate( PX(nt*ng), PZ(nt*ng), Xmsk(nt*ng), Zmsk(nt*ng) )
  allocate( Xnd(nt*ng),Znd(nt*ng),Output(nt,ng))
  allocate( Jac(nt,ng),WJ(nt*ng) )
  Xnd = 0.; Znd = 0.
  PX  = 0.; PZ  = 0.

  call sep_read(Surf)

  Xdata =  real(Surf)
  Zdata = aimag(Surf)

  call NewPot2D_init(nt,ng,dg,dt)
  
  !! Fix in appropriate locations
  Xfix = .true.
  Zfix = .true.

  !! Fixed along top
  Xfix(1,1:ng) = .false.
  Zfix(1,1:ng) = .false.

  !! . . Call the laplacian filler
  PX   = reshape(Xdata  ,(/nt*ng/))
  PZ   = reshape(Zdata  ,(/nt*ng/))
  Xmsk = reshape(Xfix   ,(/nt*ng/)) 
  Zmsk = reshape(Zfix   ,(/nt*ng/)) 
  WJ   = reshape(JAC    ,(/nt*ng/))


  write(0,*) '# of iterations: ',niter
  call mask1_init  (Xmsk)
  call mask1_init  (Zmsk)
  call laplac2_init(nt,ng)
  call solver_reg  ( m=PX, d=Xnd, Fop=laplac2_lop, &
                   stepper=cgstep, niter=niter, &
                   m0=PX,Jop=mask1_lop,Aop=laplac2_lop,&
                   naop=nt*ng,eps=10., verb=.true. )

  call solver_reg  ( m=PZ, d=Znd, Fop=laplac2_lop, &
                   stepper=cgstep, niter=niter, &
                   m0=PZ,Jop=mask1_lop,Aop=laplac2_lop,&
                   naop=nt*ng,eps=10., verb=.true. )
 
  Xmodel=reshape(PX,(/nt,ng/) )
  Zmodel=reshape(PZ,(/nt,ng/) )
  Output=cmplx(Xmodel,Zmodel)

  call Calculate_Jacobian(Xdata,Zdata,Jac)
  write(0,*) minval(Jac), maxval(Jac)

  call Calculate_Jacobian(Xmodel,Zmodel,Jac)
  write(0,*) minval(Jac), maxval(Jac)

  d1out=dble(1./(nt-1)) * ( aimag(Surf(nt,1) - Surf(1,1)) )
  o1out=aimag(Surf(1,1))

  !! Output Jacobian
  call to_history("n1",nt,"Jac.H")
  call to_history("n2",ng,"Jac.H")
  call to_history("d1",d1out,"Jac.H")
  call to_history("d2",dg,"Jac.H")
  call to_history("o1",o1out,"Jac.H")
  call to_history("o2",og,"Jac.H")
  call to_history("esize",4,"Jac.H")
  call srite("Jac.H",Jac,size(Jac)*4)

  !! Output Rays
  call to_history("n1",nt)
  call to_history("n2",ng)
  call to_history("o1",o1out)
  call to_history("o2",og)
  call to_history("d1",d1out )
  call to_history("d2",dg)
  call to_history("esize",8)
  call hclose()
  call sep_write(Output)

  !! Clean up
  call laplac2_close()
  call cgstep_close ()
  call exit()
end program NewPot2D
