module NewPot2D_
  use SEP
  use igrad1
  implicit none
  integer, private  :: nt,ng
  real   , private  :: ot,og,dt,dg
contains

  subroutine NewPot2D_init(nt_in,ng_in,dg_in,dt_in)
    integer:: nt_in,ng_in
    real   :: dg_in,dt_in
    nt=nt_in;ng=ng_in;dg=dg_in;dt=dt_in
  end subroutine NewPot2D_init

  subroutine Calculate_Jacobian(Xdata,Zdata,Jac)  
    integer                    :: it,ig
    real                       :: gx,gz
    real, dimension(:,:)       :: Xdata,Zdata,Jac
   
    do it=1,nt
       do ig=2,ng-1
          gx = (Xdata(it,ig+1)-Xdata(it,ig-1))/2./dg
          gz = (Zdata(it,ig+1)-Zdata(it,ig-1))/2./dg
          Jac(it,ig) = sqrt(gx*gx+gz*gz)
       end do
    end do
    Jac(:, 1) = Jac(:,   2)
    Jac(:,ng) = Jac(:,ng-1) 
    Jac(1, :) = Jac(2,   :)
  end subroutine Calculate_Jacobian

  !####################################################################
  subroutine get_gradx(data,deriv)
    integer                        :: iz,stat
    real,dimension(:,:)            :: data, deriv
    real,dimension(:  ),allocatable:: tmp
    allocate( tmp(ng) ); deriv=0.
    do iz=1,nt
       stat=igrad1_lop(.false.,.false.,data(iz,:),tmp )
       deriv(iz,:)=tmp
    end do
    deallocate(tmp)
  end subroutine get_gradx

   !####################################################################
  subroutine get_gradz(data,deriv)
    integer                        :: ix,stat
    real,dimension(:,:)            :: data, deriv
    real,dimension(:  ),allocatable:: tmp
    allocate( tmp(nt) )

    do ix=1,ng
       stat=igrad1_lop(.false.,.false.,data(:,ix),tmp )
       deriv(:,ix)=tmp
    end do

    deallocate(tmp)
  end subroutine get_gradz
    !####################################################################


  subroutine Ray_calc(Xdata,Zdata,rays)
    integer                             :: ip,itau
    real                                :: dx_step,dz_step,ds,K
    real   ,dimension(:,:),allocatable  :: dxdt,dzdt,dxdg,dzdg,xray,zray
    real   ,dimension(:,:),allocatable  :: deriv
    real   ,dimension(:,:)              :: Xdata,Zdata
    complex,dimension(:,:)              :: rays
    allocate(dxdt(nt,ng),dzdt(nt,ng))
    allocate(dxdg(nt,ng),dzdg(nt,ng))
    allocate(xray(nt,ng),zray(nt,ng))
    allocate( deriv(nt,ng) )
write(0,*) nt,ng

    call get_gradx(Xdata,deriv);dxdg=deriv
    call get_gradx(Zdata,deriv);dzdg=deriv
    call get_gradz(Xdata,deriv);dxdt=deriv
    call get_gradz(Zdata,deriv);dzdt=deriv

    xray(1,:)=Xdata(1,:)
    zray(1,:)=Zdata(1,:)
    do itau=2,nt
!!       write(0,*) 'Working on step ',itau, maxnum
       do ip=1,ng
!          iix=floor( (xray(itau-1,ip)-xray(1,ip))/dt + 1.5 )
!          iiz=floor( (zray(itau-1,ip)-zray(1,ip))/dz + 1.5 )
!
!          if (iix .lt. 1) then
!             iix=1
!          elseif (iix .gt. nvx) then
!             iix=nwx
!          end if
!
!          if (iiz .lt. 1) then
!             iiz=1
!          elseif (iiz .gt. nvz) then
!             iiz=nvz
!          end if
!
          ds=dt
          dx_step =  dxdt(itau,ip) - dzdg(itau,ip)
          dz_step =  dxdg(itau,ip) + dzdt(itau,ip)
          K=sqrt(dx_step**2+dz_step**2+0.0000001)
write(0,*) ds*dx_step/K,ds*dz_step/K
          xray(itau,ip)=xray(itau-1,ip)+ds*dx_step/K
          zray(itau,ip)=zray(itau-1,ip)+ds*dz_step/K
       end do
!write(0,*) 'Indicies: ',iix,iiz,itau
write(0,*) 'Prev.sta: ',xray(itau-1,ip-1),zray(itau-1,ip-1)
write(0,*) 'Add ons : ',-ds*dx_step/K,-ds*dz_step/K
    end do

    rays=cmplx(xray,zray)
    deallocate(dxdt,dxdg,dzdt,dzdg,xray,zray)
  end subroutine Ray_calc



end module NewPot2D_
