!$=head1 NAME
!$
!$igrad2ew - 2D gradient operator
!$
!$=head1 SYNOPSIS
!$
!$C<ierr=igrad2ew_init(n1,n2)>
!$
!$C<ierr=igrad2ew_lop(adj,add,xx,yy)>
!$
!$=head1 PARAMETERS
!$
!$=over 4
!$
!$=item n1,n2  - integer
!$
!$      dimensions of dataset
!$
!$=item adj,add,xx,yy -
!$
!$      standard operator interface
!$
!$=back
!$
!$=head1 DESCRIPTION
!$
!$ Calculate 2-D gradient
!$
!$=head1 SEE ALSO
!$
!$L<igrad1>
!$
!$=head1 LIBRARY
!$
!$B<geef90>
!$
!$=cut
!$
module igrad2ew  
! 2-D gradient with adjoint,  r= grad( p)
  use adj_mod
  implicit none 
  integer, private :: n1 
  integer, private :: n2 
  contains
  subroutine igrad2ew_init ( n1_in,n2_in )
    integer    :: n1_in 
    integer    :: n2_in 
    n1 = n1_in 
    n2 = n2_in
  end subroutine
  function igrad2ew_lop ( adj, add, p, r) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: p,r 
    call adjnull (adj,add,p,r )
    call igrad2ew_lop2(adj,p,r )
    stat=0
  end function 
  subroutine igrad2ew_lop2(adj,p,r)
    logical,intent(in) :: adj 
    real, dimension (n1, n2)  :: p 
    real, dimension (n1,n2,2)  :: r 
    integer i,j
      do j= 1, n2-1  
        if ( adj) then
          p(:  ,j+1)  =   p(:  ,j+1)  +  r(:,j,2) 
          p(:  ,j  )  =   p(:  ,j  )  -  r(:,j,2)
        else
          r(:,j,2)  =   r(:,j,2)  +  ( p(:,j+1) - p(:,j))
        end if
      end do
  end subroutine 
  subroutine igrad2ew_close()
  end subroutine 
end module 
