module RWEwemig
  use sep
  use fft
  implicit none
  integer, private :: nz,nx,nw
  real,    private :: dz,dx,dw
  real,    private :: oz,ox,ow

  real,    private :: pi
  real,    private :: c1,c2
  real,    private :: sixth
  integer, private :: method

  complex, allocatable, private :: lk(:),rk(:),ck(:)
  complex, allocatable, private :: a(:),b(:),c(:),u(:)
  real,                 private :: kmu,   knu,    kro 
  real,    allocatable, private :: mu(  :),nu(  :),ro(  :)
  real,    allocatable, private :: m0(:,:),n0(:,:),r0(:,:)
  real,                 private :: okx,dkx

contains
  !----------------------------------------------------------------  
  subroutine wemig_init(nz_in,nx_in,nw_in,dz_in,dx_in,dw_in,oz_in,ox_in,ow_in,method_in)
    integer, intent(in) :: nz_in,nx_in,nw_in
    real,    intent(in) :: dz_in,dx_in,dw_in
    real,    intent(in) :: oz_in,ox_in,ow_in
    integer, intent(in) :: method_in

    pi=acos(-1.)

    method=method_in

    nz=nz_in
    nx=nx_in
    nw=nw_in

    dz=dz_in
    dx=dx_in
    dw=dw_in * 2.*pi

    oz=oz_in
    ox=ox_in
    ow=ow_in * 2.*pi

    call from_param("c1",c1,0.47)
    call from_param("c2",c2,0.37)
    call from_param("sixth",sixth,0.122996)

    allocate(lk(nx),rk(nx),ck(nx))
    allocate(a(nx),b(nx-1),c(nx-1),u(nx))

    kmu = 1./ dz         !! 1/ dz
    knu = 1./(2.*dx**2)  !! 1/( 2 dx^2)
    kro = 1./(dz*dx**2)  !! 1/(dz dx^2)

    allocate(m0(nz,nx),n0(nz,nx),r0(nz,nx))
    allocate(mu(   nx),nu(   nx),ro(   nx))

    okx=  -pi/    dx
    dkx=2.*pi/(nx*dx)

  end subroutine wemig_init 
  !----------------------------------------------------------------
  !! WE migration driver
  subroutine wemig(rwf,swf,img,aa,bb,a0,b0)
    complex, pointer :: rwf(:,:)
    complex, pointer :: swf(:,:)
    real,    pointer :: img(:,:)
    real    :: aa(:,:),bb(:,:)
    real    :: a0(:),  b0(:)
    integer :: iw,iz,ix
    real    :: w,sarg,rarg
    complex, pointer :: rax(:)
    complex, pointer :: sax(:)
    sarg=-1.
    rarg= 1.
    if     (method==2) then 
      ! call wemig_ffd_coef(aa,bb,a0,b0)
      ! call wemig_coef_report()
       do iw=1,nw
          write(0,*) "FFD",iw,nw, minval(img),maxval(img)
          w=ow+(iw-1)*dw; 
          rax => rwf(:,iw)
          sax => swf(:,iw) 
          do iz=1,nz 
             call wemig_phs(rax,w,iz,a0,b0,rarg)
             call wemig_ssf(rax,w,iz,aa,a0,rarg)
             !call wemig_fds(rax,w,iz      ,rarg)
             call wemig_phs(sax,w,iz,a0,b0,sarg)
             call wemig_ssf(sax,w,iz,aa,a0,sarg)
             !call wemig_fds(sax,w,iz      ,sarg)
             img(iz,:) = img(iz,:) + real(rax*conjg(sax))
          end do
       end do
    else if(method==1) then
     !  call wemig_psc_coef(aa,bb,a0,b0)
     !  call wemig_coef_report()
       do iw=1,nw
          write(0,*) "PSC",iw,nw, minval(img),maxval(img)
          w=ow+(iw-1)*dw;
          rax => rwf(:,iw)
          sax => swf(:,iw)
          do iz=1,nz
             call wemig_phs(rax,w,iz,a0,b0,rarg)
             call wemig_ssf(rax,w,iz,aa,a0,rarg)
           !  call wemig_fds(rax,w,iz      ,rarg)

             call wemig_phs(sax,w,iz,a0,b0,sarg)
             call wemig_ssf(sax,w,iz,aa,a0,sarg)
         !    call wemig_fds(sax,w,iz      ,sarg)

             img(iz,:)=img(iz,:)+real(conjg(rax)*sax)
          end do
       end do
    else
       call wemig_xfd_coef(aa,bb,a0,b0)
       call wemig_coef_report()
       do iw=1,nw
          write(0,*) "XFD",iw,nw, minval(img),maxval(img)
          w=ow+(iw-1)*dw;
          rax => rwf(:,iw)
          sax => swf(:,iw)
          do iz=1,nz
             call wemig_ssh(rax,w,iz,aa,rarg)
             call wemig_fds(rax,w,iz   ,rarg)
             call wemig_ssh(sax,w,iz,aa,sarg)
             call wemig_fds(sax,w,iz   ,sarg)
             img(iz,:) = img(iz,:) + real(conjg(rax)*sax)
          end do
       end do
    end if

  end subroutine wemig
  !----------------------------------------------------------------
  !! Fourier-domain phase shift
  subroutine wemig_phs(dax,w,iz,a0,b0,dir)
    complex :: dax(:)
    real    :: a0(:),b0(:)
    real    :: w,kx,arg,dir
    integer :: ix,iz
    complex :: ikz

    call fth(.false.,.false.,dax)
    do ix=1,nx
       kx=okx+(ix-1)*dkx
       arg=1. - ( (b0(iz)*kx) / (w*a0(iz)) )**2
       if (arg < 0.0 ) then
          ikz = cmplx(   w*a0(iz)*sqrt(-arg),0.)
       else
          ikz = cmplx(0.,dir*w*a0(iz)*sqrt(+arg)   )
      end if
       dax(ix) = dax(ix) * cexp( ikz*(-dz))
    end do
    call fth(.true.,.false.,dax)

  end subroutine wemig_phs
  !----------------------------------------------------------------
  !! space-domain phase shift
  subroutine wemig_ssh(dax,w,iz,aa,dir)
    complex :: dax(:)
    real    :: aa(:,:)
    real    :: w,dir
    integer :: iz
    dax = dax * cexp( cmplx(0.,w*aa(iz,:)*(-dz*dir)) )
  end subroutine wemig_ssh
  !----------------------------------------------------------------
  !! split-step Fourier
  subroutine wemig_ssf(dax,w,iz,aa,a0,dir) 
    complex :: dax(:)
    real    :: aa(:,:)
    real    :: a0(:)
    real    :: w,dir
    integer :: iz
    dax=dax*cexp( cmplx(0.,w*(aa(iz,:)-a0(iz))*(-dz*dir)) )
  end subroutine wemig_ssf
  !----------------------------------------------------------------
  subroutine wemig_xfd_coef(aa,bb,a0,b0)
    real    :: aa(:,:),bb(:,:)
    real    :: a0(:),  b0(:)
    integer :: iz

    write(0,*) "compute XFD coefficients"
    m0 = 1.
    n0 = c1 * bb**2 / aa
    r0 = c2 * bb**2 / aa**2

    m0 = m0 * kmu
    n0 = n0 * knu
    r0 = r0 * kro
  end subroutine wemig_xfd_coef
  !----------------------------------------------------------------
  subroutine wemig_psc_coef(aa,bb,a0,b0)
    real    :: aa(:,:),bb(:,:)
    real    :: a0(:),  b0(:)
    integer :: iz

    write(0,*) "compute PSC coefficients"
    do iz=1,nz
       n0(iz,:) = -c1 * (b0(iz)/a0(iz))**2 *(aa(iz,:)-a0(iz)) &
       +                 b0(iz)/a0(iz)     *(bb(iz,:)-b0(iz))
       r0(iz,:) = 3*c2* (b0(iz)/a0(iz))**2 
    end do
    m0 = 1.

    m0 = m0 * kmu
    n0 = n0 * knu
    r0 = r0 * kro

  end subroutine wemig_psc_coef
  !----------------------------------------------------------------
  subroutine wemig_ffd_coef(aa,bb,a0,b0)
    real    :: aa(:,:),bb(:,:)
    real    :: a0(:),  b0(:)
    integer :: iz
    real,    allocatable :: d1(:,:),d2(:,:)
    real    :: tt

    allocate(d1(nz,nx),d2(nz,nx))
    tt = abs(maxval(b0))
    do iz=1,nz
       d1(iz,:)=a0(iz)*((b0(iz)/tt)/a0(iz))**2 &
       * (             ( bb(iz,:)  /b0(iz))**2*a0(iz)/aa(iz,:)     - 1. )
       d2(iz,:)=a0(iz)*((b0(iz)/tt)/a0(iz))**4 &
       * (             ( bb(iz,:)  /b0(iz))**4*(a0(iz)/aa(iz,:))**3- 1. )
    end do

    where(abs(d1)<epsilon(d1)) 
       d1=epsilon(d1)
    end where

    m0 =      1    
    n0 = c1 * d1    * tt**2
    r0 = c2 * d2/d1 * tt**2

    m0 = m0 * kmu
    n0 = n0 * knu
    r0 = r0 * kro

    deallocate(d1,d2)
  end subroutine wemig_ffd_coef
  !----------------------------------------------------------------
  subroutine wemig_coef_report()

    write(0,*) minval(m0),"< mu <",maxval(m0)
    write(0,*) minval(n0),"< nu <",maxval(n0)
    write(0,*) minval(r0),"< ro <",maxval(r0)

  end subroutine wemig_coef_report
  !----------------------------------------------------------------  
  !! finite-differences solver
  subroutine wemig_fds(dax,w,iz,dir)
    real                  :: w,dir
    integer               :: iz
    complex, dimension(:) :: dax
    integer               :: ix

    mu = m0(iz,:)
    nu = n0(iz,:) / w
    ro = r0(iz,:) / w**2

    lk = cmplx(ro+sixth*mu, nu)
    rk = cmplx(ro+sixth*mu,-nu)
    ck = cmplx(         mu, 0.)

    !!             |                      |
    !!     b       |          a           |     c
    !!  ro + i nu  |  mu - 2(ro + i nu )  |  ro + i nu
    !!             |                      |
    a = ck - 2.*lk  
    b = lk(1  :nx-1)
    c = lk(1+1:nx  )

    do ix=2,nx-1
       u(ix)=           rk(ix) *(dax(ix-1)+dax(ix+1)) & 
       +    (ck(ix)-2.0*rk(ix))* dax(ix)
       !!             |                      | 
       !!  ro - i nu  |  mu - 2 (ro - i nu)  | ro - i nu
       !!             |                      | 
    end do
    ix= 1; u(ix)=(ck(ix)-2.0*rk(ix))*dax(ix) + rk(ix)*dax(ix+1)
    ix=nx; u(ix)=(ck(ix)-2.0*rk(ix))*dax(ix) + rk(ix)*dax(ix-1)

    call thr(a,b,c,u,nx)
    dax = u
  end subroutine wemig_fds
  !----------------------------------------------------------------
  !! tridiagonal solver
  subroutine thr(a,b,c,v,n)
    integer :: n
    complex :: a(n),b(n-1),c(n-1),v(n)
    integer :: i

    b(1)=b(1)/a(1)
    v(1)=v(1)/a(1)
    do i=2,n
       a(i)=a(i)-c(i-1)*b(i-1)
       b(i)=b(i)/a(i)
       v(i)=(v(i)-c(i-1)*v(i-1))/a(i)
    end do
    do i=n-1,1,-1
       v(i)=v(i)-b(i)*v(i+1)
    end do
  end subroutine thr
  !----------------------------------------------------------------
  subroutine attribute(x)
    complex :: x(:)
    write(0,*) minval( real(x)),maxval( real(x))
    write(0,*) minval(aimag(x)),maxval(aimag(x))
  end subroutine attribute
  !----------------------------------------------------------------
end module RWEwemig
