program SP_2D_VOCIGs
  use sep
  use util_
  use SP_2D_VOCIGs_wemig

  implicit none
  type(myaxis) :: ax,az,aw,ar,as,ah

  integer :: method,nhx,nhz,rl,rr,sl,sr
  integer :: ix,iz,ii,ih,ihz,nref,iw
  character(len=128)   :: name
  real                 :: forward,dr
  real,    allocatable,dimension(:,:,:,:) :: rall,sall
  complex, allocatable,dimension(:,:,:)   :: rrefall,srefall
  real,    allocatable,dimension(:,:)     :: saa,sbb,smm
  real,    allocatable,dimension(:,:)     :: raa,rbb,rmm
  real,    allocatable,dimension(:,:)     :: sa0,sb0,ra0,rb0
  complex, allocatable,dimension(:,:)     :: sab,rab
  complex, pointer    ,dimension(:,:,:)   :: rwfin,swfin     !! input S&R wflds
  complex, pointer    ,dimension(:,:,:)   :: rwf,swf         !! S&R wfld
  complex, pointer    ,dimension(:,:,:,:) :: rwfall,swfall   !! full S&R wfld
  real,    pointer    ,dimension(:,:,:)   :: imgx,imgz    !! image with offsets
  real,    pointer    ,dimension(:,:,:,:) :: imgallx,imgallz !! all shots image
  logical              :: verbose
  call sep_init(SOURCE)
  call getch_add_string("head=/dev/null")
  call getch_add_string("noheader=y")
  call set_no_putch()

  call from_param("method",method,0)    !! propagation type
  call from_param("forward",forward,0.) !! Forward scattering option
  call from_param("nref",nref,1)        !! # reference velocities
  call from_param("nhx",nhx,1)            !! # image offset in x
  call from_param("nhz",nhz,1)            !! # image offset in z
  call from_param("verbose",verbose,.false.)

  call from_aux("data","n1",ax%n)
  call from_aux("data","o1",ax%o)
  call from_aux("data","d1",ax%d)
  call from_aux("data","n2",aw%n)
  call from_aux("data","o2",aw%o)
  call from_aux("data","d2",aw%d)
  call from_aux("data","n3",as%n)
  call from_aux("data","o3",as%o)
  call from_aux("data","d3",as%d)

  name="rabm"
  call ainaxis(az,1,name)
  name="rabref"
  call ainaxis(ar,2,name)
  name="image"

  if(nref .eq. 1 ) ar%n=1
  if(method==0) ar%n=1   !! pure finite-differences

  if (verbose) then
     write(0,*) ax%n,ax%o,ax%d
     write(0,*) az%n,az%o,az%d
     write(0,*) aw%n,aw%o,aw%d
     write(0,*) ar%n,ar%o,ar%d
     write(0,*) as%n,as%o,as%d
  end if

  call sep_begin_prog()
  !! . . Create offset axis
  dr=ax%d

  call auxputch("n1","i",az%n,name)
  call auxputch("d1","r",az%d,name)
  call auxputch("o1","r",az%o,name)
  call auxputch("n2","i",ax%n,name)
  call auxputch("d2","r",ax%d,name)
  call auxputch("o2","r",ax%o,name)  
  call auxputch("n3","i",as%n,name)
  call auxputch("d3","r",as%d,name)
  call auxputch("o3","r",as%o,name)  
  call auxputch("n4","i",nhx,name)
  call auxputch("d4","r",dr,name)
  call auxputch("o4","r",-(nhx-1)/2.*dr,name)  
  call auxputch("esize","i",4,name)
  call auxputch("label1","s","depth",name)
  call auxputch("label2","s","midpoint",name)
  call auxputch("label3","s","shot_number",name)
  call auxputch("label4","s","offsets",name)


  !! . . Allocate input wavefields 
  allocate( rwfin(ax%n,aw%n,as%n) , swfin(ax%n,aw%n,as%n) )

  !! . . Allocate arrays for global coefficients
  allocate(swfall(ax%n,az%n,aw%n,as%n))
  allocate(rwfall(ax%n,az%n,aw%n,as%n))
  allocate(imgallx(az%n,ax%n,as%n,nhx))
  allocate(imgallz(az%n,ax%n,as%n,nhz))

  !! . . Allocate arrays for individual shots
  allocate(swf(ax%n,az%n,aw%n))
  allocate(rwf(ax%n,az%n,aw%n))
  allocate(imgx(az%n,ax%n,nhx))
  allocate(imgz(az%n,ax%n,nhz))

  !! . . Coordinate system variable initialization
  allocate(rall(az%n,ax%n,as%n,5) , sall(az%n,ax%n,as%n,5) )
  allocate(rrefall(az%n,ar%n,as%n),srefall(az%n,ar%n,as%n) )
  allocate(raa(az%n,ax%n))
  allocate(rbb(az%n,ax%n),rmm(az%n,ax%n))
  allocate(saa(az%n,ax%n),sbb(az%n,ax%n))
  allocate(smm(az%n,ax%n),rab(az%n,ax%n))
  allocate(ra0(az%n,ar%n),rb0(az%n,ar%n))
  allocate(sab(az%n,ax%n),sa0(az%n,ar%n))
  allocate(sb0(az%n,ar%n))   


  !! . . Read in the coefficients
  call sreed("rabm",rall,size(rall)*4)
  call sreed("sabm",sall,size(sall)*4)
  call sreed("rabref",rrefall,size(rrefall)*8)
  call sreed("sabref",srefall,size(srefall)*8)

  !! . . Read in RWF and SWF
!  call sep_read(rwfin,"data")
  call sreed("data",rwfin,size(rwfin)*8)
  rwfall(:,1,:,:)=rwfin
  call sreed("swf",swfin,size(swfin)*8)
  swfall(:,1,:,:)=swfin

  !! . . Initialize module and image space
write(0,*) 'INITiNG'
  call wemig_init(ax,az,aw,ar,method,forward,nhx)
write(0,*) 'UNiniting'

  imgallx=0.;  imgallz=0.

  do ii=1,as%n   !! . . Loop over shot points
     imgx=0.;imgz=0.;rwf=0.;swf=0.;raa=0.;saa=0.;rbb=0.;sbb=0.;
write(0,*) 'initialized 1/2 arrays'     
     ra0=0.;rb0=0.;sa0=0.;sb0=0.;rwf=0.;swf=0.;
write(0,*) 'initialized all arrays'
     rwf=rwfall(:,:,:,ii)
     swf=swfall(:,:,:,ii)
     raa=rall(:,:,ii,1);        saa=sall(:,:,ii,1)
     rbb=rall(:,:,ii,2);        sbb=sall(:,:,ii,2)
     rmm=rall(:,:,ii,3);        smm=sall(:,:,ii,3)
     ra0= real(rrefall(:,:,ii));sa0= real(srefall(:,:,ii))
     rb0=aimag(rrefall(:,:,ii));sb0=aimag(srefall(:,:,ii))

     if (verbose) then
        write(0,*) "raa",minval(raa),maxval(raa)
        write(0,*) "rbb",minval(rbb),maxval(rbb)
        write(0,*) "rmm",minval(rmm),maxval(rmm)
        write(0,*) "saa",minval(saa),maxval(saa)
        write(0,*) "sbb",minval(sbb),maxval(sbb)
        write(0,*) "smm",minval(smm),maxval(smm)
        write(0,*) "ra0",minval(ra0),maxval(ra0)
        write(0,*) "rb0",minval(rb0),maxval(rb0)
        write(0,*) "sa0",minval(sa0),maxval(sa0)
        write(0,*) "sb0",minval(sb0),maxval(sb0)
     end if

     !! . . Image event
!    call sep_close()
write(0,*) 'INTO PROPAGATION LOOP'
     call wemig(rwf,swf,imgx,raa,rbb,rmm,ra0,rb0,saa,sbb,smm,sa0,sb0,ii)
write(0,*) 'OUT OF PROPAGATION LOOP'
     imgallx(:,:,ii,:)=imgx

 !    !! . . VOCIG Imaging Condition
 !    do ix=1,ax%n !! . . X-LOOP
 !       do iz=1,az%n !! . . Z-LOOP
 !          do ihz=1,nhz !! . . Hz-loop
 !             if (iz-(nhz+1)/2+ihz .ge. 1 .and. iz-(nhz+1)/2+ihz .le. az%n .and.&
 !                 iz+(nhz+1)/2-ihz .ge. 1 .and. iz+(nhz+1)/2-ihz .le. az%n       ) then
 !                do iw=1,aw%n
 !                   imgz(iz,ix,ihz)=imgz(iz,ix,ihz)+&
 !                   (real( rwf(ix,iz-(nhz+1)/2+ihz,iw)&
 !                   *conjg(swf(ix,iz+(nhz+1)/2-ihz,iw))))
 !                end do
 !             end if
 !          end do !! . .VOCIG Hz-LOOP
 !       end do !! . . VOCIG Z-LOOP
 !    end do !! VOCIG x-loop
 !
 !     imgallz(:,:,ii,:)=imgz

  end do !! Shot-point loop

  call sep_write(imgallx,"image")
write(0,*) 'OUTPUT IMAGE'
 ! call srite(name,imgallz,size(imgallz)*4)

  call sep_end_prog()
  call exit()

end program SP_2D_VOCIGs
