program SP2D
  use sep
  use util_
  use SP2D_wemig

  implicit none
  type(myaxis) :: ax,az,aw,ar,as,ah

  integer :: method,nhx 
  integer :: ii,nref,iw
  character(len=128)   :: name
  real                 :: forward,dr,divide,eps
  real,    pointer,dimension(:,:,:,:) :: rall,sall
  complex, pointer,dimension(:,:,:)   :: rrefall,srefall
  real,    pointer,dimension(:,:)     :: saa,sbb,smm
  real,    pointer,dimension(:,:)     :: raa,rbb,rmm
  real,    pointer,dimension(:,:)     :: sa0,sb0,ra0,rb0
  complex, pointer,dimension(:,:)     :: sab,rab
  complex, pointer,dimension(:,:,:)   :: rwfin,swfin!! input S&R wflds
  complex, pointer,dimension(:,:)     :: rwf,swf    !! S&R wfld
  complex, pointer,dimension(:,:,:)   :: rwfall,swfall   !! full S&R wfld
  real,    pointer,dimension(:,:,:)   :: img   !! image with offsets
  real,    pointer,dimension(:,:,:,:) :: imgall !! all shots image
  logical              :: verbose,same,single
  call sep_init(SOURCE)
  call getch_add_string("head=/dev/null")
  call getch_add_string("noheader=y")
  call set_no_putch()
  call sep_begin_prog()

  call from_param("method",method,0)    !! propagation type
  call from_param("forward",forward,0.) !! Forward scattering option
  call from_param("nref",nref,1)        !! # reference velocities
  call from_param("nhx",nhx,1)            !! # image offset in x
  call from_param("verbose",verbose,.false.)
  call from_param("divide",divide,0.)
  call from_param("eps",eps,0.001)
  call from_param("same",same,.true.)
  call from_param("single",single,.true.)

  call from_aux("data","n1",ax%n)
  call from_aux("data","o1",ax%o)
  call from_aux("data","d1",ax%d)
  call from_aux("data","n2",aw%n)
  call from_aux("data","o2",aw%o)
  call from_aux("data","d2",aw%d)
  call from_aux("data","n3",as%n)
  call from_aux("data","o3",as%o)
  call from_aux("data","d3",as%d)

  name="rabm"
  call ainaxis(az,1,name)
  name="rabref"
  call ainaxis(ar,2,name)
  name="image"

  if (same .eq. .true.) then
     write(0,*) 'USING SAME SOURCE  AND RECEIVER GRIDS'
  else
     write(0,*) 'USING DIFFERENT GRIDS'
  end if

  if(nref .eq. 1 ) ar%n=1

  if(method==0) ar%n=1   !! pure finite-differences

  if (verbose) then
     write(0,*) ax%n,ax%o,ax%d
     write(0,*) az%n,az%o,az%d
     write(0,*) aw%n,aw%o,aw%d
     write(0,*) ar%n,ar%o,ar%d
     write(0,*) as%n,as%o,as%d
  end if

  !! . . Create offset axis
  dr=ax%d

  call auxputch("n1","i",az%n,name)
  call auxputch("d1","r",az%d,name) 
  call auxputch("o1","r",az%o,name) 
  call auxputch("n2","i",ax%n,name) 
  call auxputch("d2","r",ax%d,name) 
  call auxputch("o2","r",ax%o,name)  
  call auxputch("n3","i",as%n,name) 
  call auxputch("d3","r",as%d,name)  
  call auxputch("o3","r",as%o,name)   
  call auxputch("n4","i",nhx,name)
  call auxputch("d4","r",dr,name)
  call auxputch("o4","r",-(nhx-1)/2.*dr,name)  
  call auxputch("esize","i",4,name)
  call auxputch("label1","s","depth",name)
  call auxputch("label2","s","midpoint",name)
  call auxputch("label3","s","shot_number",name)
  call auxputch("label4","s","offsets",name)


  !! . . Allocate input wavefields 
  write(0,*) ax%n,az%n,aw%n,as%n

  !! . . Allocate arrays for global coefficients
  allocate(swfall(ax%n,aw%n,as%n))
  allocate(rwfall(ax%n,aw%n,as%n))
  allocate(imgall(az%n,ax%n,as%n,nhx))

  !! . . Allocate arrays for individual shots
  allocate(swf(ax%n,aw%n))
  allocate(rwf(ax%n,aw%n))
  allocate(img(az%n,ax%n,nhx))

  !! . . Coordinate system variable initialization
  if (single .eq. .false.) then
     allocate(rall(az%n,ax%n,as%n,3),rrefall(az%n,ar%n,as%n))
  else 
     allocate(rall(az%n,ax%n,1,3),rrefall(az%n,ar%n,1))
  end if

  if (same .eq. .false.) then
     if (single .eq. .false.) then
        allocate(sall(az%n,ax%n,as%n,3),srefall(az%n,ar%n,as%n))
     else 
        allocate(sall(az%n,ax%n,1,3),srefall(az%n,ar%n,1))
     end if
  end if

  !! . . Sub arrays for grid
  allocate(raa(az%n,ax%n),ra0(az%n,ar%n),rb0(az%n,ar%n))
  allocate(rbb(az%n,ax%n),rmm(az%n,ax%n),rab(az%n,ax%n))
  if (same .eq. .false.) then
     allocate(saa(az%n,ax%n),sbb(az%n,ax%n),smm(az%n,ax%n))
     allocate(sab(az%n,ax%n),sa0(az%n,ar%n),sb0(az%n,ar%n))   
  end if

  !! . . Read in the coefficients
  call sreed("rabm",rall,size(rall)*4)
  call sreed("rabref",rrefall,size(rrefall)*8)
  if (same .eq. .false. ) then
     call sreed("sabm",sall,size(sall)*4)
     call sreed("sabref",srefall,size(srefall)*8)
  end if

  !! . . Read in RWF and SWF
  call sreed("data",rwfall,size(rwfall)*8)
  call sreed("swf",swfall,size(swfall)*8)

  write(0,*) 'MIN/MAX DATA: ',minval(real(rwfall)),maxval(real(swfall))
  write(0,*) 'MIN/MAX SORC: ',minval(real(swfall)),maxval(real(swfall))

  !! . . Initialize module and image space
  call wemig_init(ax,az,aw,ar,method,forward,nhx,divide,eps)

  imgall=0.

  do ii=1,as%n   !! . . Loop over shot points
     img=0.;rwf=0.;swf=0.;raa=0.;rbb=0.;ra0=0.;rb0=0.;
     rwf=rwfall(:,:,ii)
     swf=swfall(:,:,ii)

     if (single .eq. .false.) then
        raa=rall(:,:,ii,1);        
        rbb=rall(:,:,ii,2);        
        rmm=rall(:,:,ii,3);        
        ra0= real(rrefall(:,:,ii));
        rb0=aimag(rrefall(:,:,ii));
     else
        raa=rall(:,:,1,1);        
        rbb=rall(:,:,1,2);        
        rmm=rall(:,:,1,3);        
        ra0= real(rrefall(:,:,1));
        rb0=aimag(rrefall(:,:,1));
     end if

     if (same .eq. .false.) then
        saa=0.;sbb=0.;sa0=0.;sb0=0.;
        if (single .eq. .false.) then
           saa=sall(:,:,ii,1)
           sbb=sall(:,:,ii,2)
           smm=sall(:,:,ii,3)
           sa0= real(srefall(:,:,ii))
           sb0=aimag(srefall(:,:,ii))
        else
           saa=sall(:,:,1,1)
           sbb=sall(:,:,1,2)
           smm=sall(:,:,1,3)
           sa0= real(srefall(:,:,1))
           sb0=aimag(srefall(:,:,1))
        end if
     end if

     if (verbose) then
        write(0,*) "raa",minval(raa),maxval(raa)
        write(0,*) "rbb",minval(rbb),maxval(rbb)
        write(0,*) "rmm",minval(rmm),maxval(rmm)
        write(0,*) "ra0",minval(ra0),maxval(ra0)
        write(0,*) "rb0",minval(rb0),maxval(rb0)
        if (same .eq. .false.) then
           write(0,*) "saa",minval(saa),maxval(saa)
           write(0,*) "sbb",minval(sbb),maxval(sbb)
           write(0,*) "smm",minval(smm),maxval(smm)
           write(0,*) "sa0",minval(sa0),maxval(sa0)
           write(0,*) "sb0",minval(sb0),maxval(sb0)
        end if
     end if

     !! . . Image event
     if (same .eq. .false.) then
        call wemig(rwf,swf,img,raa,rbb,rmm,ra0,rb0,saa,sbb,smm,sa0,sb0,ii)
     else 
        call wemig(rwf,swf,img,raa,rbb,rmm,ra0,rb0,raa,rbb,rmm,ra0,rb0,ii)
     end if
     imgall(:,:,ii,:)=img

  end do !! Shot-point loop

  call sep_write(imgall,"image")

  call sep_end_prog()
  call exit()

end program SP2D
