module SP_2D_VOCIGs_wemig
  use sep
  use util_
  use fft

  implicit none
  integer, private :: nz,nx,nw,nr,nhx,nhz,ih
  real,    private :: dz,dx,dw
  real,    private :: oz,ox,ow
  real,    private :: pi
  real,    private :: c1,c2
  real,    private :: sixth,divide,eps
  integer, private :: nloop
  integer, private :: method
  real   , private :: forward
  complex, allocatable, private :: lk(:),rk(:),ck(:)
  complex, allocatable, private :: a(:),b(:),c(:),u(:)
  real,                 private :: kmu,    knu,    kro 
  real,    allocatable, private ::  mu(  :  ),nu(  :  ),ro(  :  )
  real,    allocatable, private ::  m0(:,:,:),n0(:,:,:),r0(:,:,:)
  real,                 private :: okx,dkx

  complex, allocatable, private :: wfl(:),wtt(:)
  real,    allocatable, private :: msk(:),mtt(:)

  real,    allocatable, private :: tap(:)
  integer                       :: ntap
contains
  !----------------------------------------------------------------  
  subroutine wemig_init(ax,az,aw,ar,method_in,forward_in,nhx_in,divide_in,eps_in)
    type(myaxis)     :: ax,az,aw,ar
    integer, intent(in) :: method_in,nhx_in
    real   , intent(in) :: forward_in,divide_in,eps_in
    pi=acos(-1.)
    divide=divide_in
    eps=eps_in
    nhx=nhx_in
    method=method_in
    forward=forward_in
    nz=az%n
    nx=ax%n
    nw=aw%n
    dz=az%d
    dx=ax%d
    dw=aw%d * 2.*pi
    oz=az%o
    ox=ax%o
    ow=aw%o * 2.*pi
    nr=ar%n

    call from_param("c1",c1,0.5)
    call from_param("c2",c2,0.)
    call from_param("sixth",sixth,0.122996)
    call from_param("nloop",nloop,3)

    allocate(lk(nx),rk(nx),ck(nx))
    allocate(a(nx),b(nx-1),c(nx-1),u(nx))

    kmu = 1./ dz         !! 1/ dz
    knu = 1./(2.*dx**2)  !! 1/( 2 dx^2)
    kro = 1./(dz*dx**2)  !! 1/(dz dx^2)

    allocate(m0(nz,nx,nr),n0(nz,nx,nr),r0(nz,nx,nr))
    allocate(mu(   nx   ),nu(   nx   ),ro(   nx   ))

    okx=  -pi/    dx
    dkx=2.*pi/(nx*dx)

    allocate(wfl(nx))
    allocate(wtt(nx))
    allocate(msk(nx)); msk=1.
    allocate(mtt(nx)); mtt=1.

    call from_param("ntap",ntap,1)
    allocate(tap(nx)); tap=1.
    call rwetaper_init()
  end subroutine wemig_init
  !----------------------------------------------------------------
  subroutine rwetaper_init()
    integer :: itap,jtap
    real    :: pi

    pi=acos(-1.)

    if(ntap>1) then
       do itap=1,ntap
          jtap = abs(ntap-itap)
          tap(   itap  ) = cos(pi/2.* jtap/(ntap-1))
       end do
       do itap=1,ntap
          jtap = abs(ntap-itap)
          tap(nx-itap+1) = cos(pi/2.* jtap/(ntap-1))
       end do
    end if
  end subroutine rwetaper_init

  subroutine rwetaper(dax)
    complex, pointer :: dax(:)
    dax = dax * tap
  end subroutine rwetaper
  !----------------------------------------------------------------
  !! WE migration driver
  subroutine wemig(rwf,swf,imgx,raa,rbb,rmm,ra0,rb0,saa,sbb,smm,sa0,sb0,ishot)
    complex, pointer:: rwf (:,:)
    complex, pointer:: swf (:,:)
    real,    pointer:: imgx(:,:,:)
    real,    pointer:: raa(:,:),rbb(:,:),rmm(:,:)
    real,    pointer:: ra0(:,:),rb0(:,:)
    real,    pointer:: saa(:,:),sbb(:,:),smm(:,:)
    real,    pointer:: sa0(:,:),sb0(:,:)
    integer :: iw,iz,ix,ir,ishot,iih,ihx
    real    :: w,sarg,rarg,saxmax
    complex, pointer :: rax(:)
    complex, pointer :: sax(:)

    sarg=-1.
    if ( forward .ne. 0. ) sarg=1. !! Forward scattering 
    rarg=1.

    if   (method==3) then 
       call wemig_psc_coef(raa,rbb,ra0,rb0)
       call wemig_coef_report()
       do iw=1,nw  !! Frequency Loop           
          w=ow+(iw-1)*dw
          write(0,*) "PSC",iw,nw,ishot,maxval(abs(img(:,:,:)))
          rax => rwf(:,iw)
          sax => swf(:,iw)
          saxmax=maxval(cabs(sax))

          do iz=2,nz 
             !! . . Do receiver wavefield
             wtt=rax
             rax=0. 
             do ir=1,nr  
                wfl=wtt 
                call wemig_phs(wfl,w,iz,ir,ra0,rb0,rarg)
                call wemig_ssf(wfl,w,iz,ir,raa,ra0,rarg)
                call wemig_fds(wfl,w,iz,ir,rarg)
                call wemig_ref(wfl,  iz,ir,rmm,rax)
             end do
             call rwetaper(rax)

             !! . . Do source wavefield
             wtt=sax
             sax=0.
             do ir=1,nr  
                wfl=wtt
                call wemig_phs(wfl,w,iz,ir,sa0,sb0,sarg)
                call wemig_ssf(wfl,w,iz,ir,saa,sa0,sarg)
                call wemig_fds(wfl,w,iz,ir,sarg)
                call wemig_ref(wfl,  iz,ir,smm,sax)
             end do
             call rwetaper(sax)
         
             if (nhx .eq. 1) then
                call image_condition       (sax,rax,img,iz,saxmax,divide,eps)
             else
                call image_condition_offset(sax,rax,img,iz,saxmax,divide,eps)
             end if
             
          end do !! . . z-loop
       end do  !! . . w-loop

    else if(method==2) then

       call wemig_ffd_coef(raa,rbb,ra0,rb0)
       call wemig_coef_report()
       do iw=1,nw;         
          w=ow+(iw-1)*dw;
          write(0,*) "FFD",iw,nw,ishot,maxval(abs(img))
          rax => rwf(:,iw)
          sax => swf(:,iw) 
          saxmax=maxval(cabs(sax))
          do iz=1,nz
             !! . . Do receiver wavefield
             wtt=rax
             rax=0.
             do ir=1,nr 
                wfl=wtt
                call wemig_phs(wfl,w,iz,ir,ra0,rb0,rarg)
                call wemig_ssf(wfl,w,iz,ir,raa,ra0,rarg)
                call wemig_fds(wfl,w,iz,ir,rarg)
                call wemig_ref(wfl,  iz,ir,rmm,rax) 
             end do
             call rwetaper(rax)

             !! . . Do source wavefield
             wtt=sax
             sax=0.
             do ir=1,nr
                wfl=wtt
                call wemig_phs(wfl,w,iz,ir,sa0,sb0,sarg)
                call wemig_ssf(wfl,w,iz,ir,saa,sa0,sarg)
                call wemig_fds(wfl,w,iz,ir,sarg)
                call wemig_ref(wfl,  iz,ir,smm,sax)
             end do
             call rwetaper(sax)
             
             if (nhx .eq. 1) then
                call image_condition       (sax,rax,img,iz,saxmax,divide,eps)
             else
                call image_condition_offset(sax,rax,img,iz,saxmax,divide,eps)
             end if
             
           end do !! . . z-loop
       end do !! . . w-loop

!!$    else if(method==1) then
!!$       do iw=1,nw;          
!!$          w=ow+(iw-1)*dw
!!$          write(0,*) "SSF",iw,nw,maxval(abs(img))
!!$          rax => rwf(:,iw)
!!$          sax => swf(:,iw) 
!!$          do iz=1,nz
!!$
!!$             !! . . Do recevier wavefield
!!$             wtt=rax
!!$             rax=0.
!!$             do ir=1,nr
!!$                wfl=wtt
!!$                call wemig_phs(wfl,w,iz,ir,ra0,rb0,rarg)
!!$                call wemig_ssf(wfl,w,iz,ir,raa,ra0,rarg)
!!$                call wemig_ref(wfl,  iz,ir,rmm,rax)
!!$             end do
!!$             call rwetaper(rax)
!!$
!!$             !! . . Do source wavefield
!!$             wtt=sax
!!$             sax=0.
!!$             do ir=1,nr
!!$                wfl=wtt
!!$                call wemig_phs(wfl,w,iz,ir,sa0,sb0,sarg)
!!$                call wemig_ssf(wfl,w,iz,ir,saa,sa0,sarg)
!!$                call wemig_ref(wfl,  iz,ir,smm,sax)
!!$             end do
!!$             call rwetaper(sax)
!!$             !! . . Imaging Condition
!!$             if (nh .eq. 1) then !! . . Zero subsurface offset
!!$                img(iz,:,(nh+1)/2 )=img(iz,:,ih+(nh+1)/2) + &
!!$                real( rax* conjg(sax) )
!!$             end if
!!$             if (nh .gt. 1) then !! . . Multiple subsurface offset
!!$                do ix=1,nx
!!$                   iih=1
!!$                   do ih=ix-(nh-1)/2,ix+(nh-1)/2
!!$                      if (ih .ge. 1 .and. ih .le. nx) then
!!$                         img(iz,ix,iih) = img(iz,ix,iih) + &
!!$                         real( rax(ih) * conjg( sax(ix+(nh-1)/2-iih+1) ) )
!!$                      else
!!$                         img(iz,ix,iih) = img(iz,ix,iih)
!!$                      end if
!!$                      iih=iih+1
!!$                   end do
!!$                end do
!!$             end if
!!$
!!$          end do !! . . z-loop
!!$       end do  !! . . w-loop
!!$    else
!!$       call wemig_xfd_coef(raa,rbb)
!!$       call wemig_coef_report()
!!$       do iw=1,nw;          
!!$          w=ow+(iw-1)*dw
!!$          write(0,*) "XFD",iw,nw,maxval(abs(img))
!!$          rax => rwf(:,iw)
!!$          sax => swf(:,iw) 
!!$          do iz=1,nz
!!$             !! . . Do receiver wavefield
!!$             call wemig_ssh(rax,w,iz,raa,rarg)
!!$             call wemig_fds(rax,w,iz,1,rarg)
!!$             call rwetaper(rax)
!!$
!!$             !! . . Do source wavefield
!!$             call wemig_ssh(sax,w,iz,saa,sarg)
!!$             call wemig_fds(sax,w,iz,1,sarg)
!!$             call rwetaper(sax)
!!$
!!$              !! . . Imaging Condition
!!$             if (nh .eq. 1) then !! . . Zero subsurface offset
!!$                img(iz,:,(nh+1)/2 )=img(iz,:,ih+(nh+1)/2) + &
!!$                  real( rax* conjg(sax) )
!!$             end if
!!$             if (nh .gt. 1) then !! . . Multiple subsurface offset
!!$                do ix=1,nx
!!$                   iih=1
!!$                   do ih=ix-(nh-1)/2,ix+(nh-1)/2
!!$                      if (ih .ge. 1 .and. ih .le. nx) then
!!$                         img(iz,ix,iih) = img(iz,ix,iih) + &
!!$                         real( rax(ih) * conjg( sax(ix+(nh-1)/2-iih+1) ) )
!!$                      else
!!$                         img(iz,ix,iih) = img(iz,ix,iih)
!!$                      end if
!!$                      iih=iih+1
!!$                   end do
!!$                end do
!!$             end if
!!$
!!$          end do !! . . z-loop
!!$       end do  !! . . w-loop
    end if !! . . Prop type loop

  end subroutine wemig
  !----------------------------------------------------------------
  subroutine wemig_ref(wfl,iz,ir,mm,dax)
    complex, pointer :: dax(:)
    complex          :: wfl(:)
    real             :: mm(:,:)
    integer          :: iz,ir,ix,ii

    where( nint(mm(iz,:))==ir) 
       mtt=1.
    elsewhere
       mtt=0.
    end where

    do ii=1,nloop
       msk(1)=mtt(1)
       do ix=1+1,nx-1
          msk(ix) = 0.5*( mtt(ix-1)+mtt(ix+1) )
       end do
       msk(nx)=mtt(nx-1)
       mtt=msk
    end do

    dax = dax + msk*wfl
  end subroutine wemig_ref
  !----------------------------------------------------------------  
  !! Fourier-domain phase shift
  subroutine wemig_phs(dax,w,iz,ir,a0,b0,dir)
    complex :: dax(:)
    real    :: a0(:,:),b0(:,:)
    real    :: w,kx,arg,dir
    integer :: ix,iz,ir
    complex :: ikz

    call fth(.false.,.false.,dax)
    do ix=1,nx
       kx=okx+(ix-1)*dkx
       arg=1. - ( (b0(iz,ir)*kx) / (w*a0(iz,ir)) )**2
       if (arg < 0.0 ) then
          ikz = cmplx(   w*a0(iz,ir)*sqrt(-arg),0.)
       else
          ikz = cmplx(0.,dir*w*a0(iz,ir)*sqrt(+arg)   )
       end if
       dax(ix) = dax(ix) * cexp( ikz*(-dz))
    end do
    call fth( .true.,.false.,dax)

  end subroutine wemig_phs
  !----------------------------------------------------------------
  !! space-domain phase shift
  subroutine wemig_ssh(dax,w,iz,aa,dir)
    complex :: dax(:)
    real    :: aa(:,:)
    real    :: w,dir
    integer :: iz

    dax = dax * cexp(cmplx(0.,-w*aa(iz,:) * dz * dir))
  end subroutine wemig_ssh
  !----------------------------------------------------------------
  !! split-step Fourier
  subroutine wemig_ssf(dax,w,iz,ir,aa,a0,dir) 
    complex :: dax(:)
    real    :: aa(:,:)
    real    :: a0(:,:)
    real    :: w,dir
    integer :: iz,ir

    dax = dax * cexp(cmplx(0.,-w*(aa(iz,:)-a0(iz,ir)) * dz * dir))
  end subroutine wemig_ssf
  !----------------------------------------------------------------
  subroutine wemig_xfd_coef(aa,bb)
    real    :: aa(:,:),bb(:,:)
    integer :: iz

    write(0,*) "compute XFD coefficients"
    m0(:,:,1) = 1.
    n0(:,:,1) =-c1 * bb**2 / aa
    r0(:,:,1) = c2 * bb**2 / aa**2

    m0 = m0 * kmu
    n0 = n0 * knu
    r0 = r0 * kro
  end subroutine wemig_xfd_coef
  !----------------------------------------------------------------
  subroutine wemig_psc_coef(aa,bb,a0,b0)
    real    :: aa(:,:),bb(:,:)
    real    :: a0(:,:),b0(:,:)
    integer :: iz,ir
    real    :: tt
    real    :: boa

    write(0,*) "compute PSC coefficients"
    tt = abs(maxval(b0))

    m0 = 1 / tt**2
    do iz=1,nz
       do ir=1,nr
          boa = (b0(iz,ir)/tt)/a0(iz,ir)

          n0(iz,:,ir) = c1*(aa(iz,:)/a0(iz,ir)-1.)-(bb(iz,:)/b0(iz,ir) - 1.)
          n0(iz,:,ir) = n0(iz,:,ir) * a0(iz,ir) * boa**2
          r0(iz,:,ir) = 3.*c2 * boa**2

       end do
    end do

    m0 = m0 * kmu
    n0 = n0 * knu
    r0 = r0 * kro
    write(0,*) "done PSC coefficients"

  end subroutine wemig_psc_coef
  !----------------------------------------------------------------
  subroutine wemig_ffd_coef(aa,bb,a0,b0)
    real    :: aa(:,:),bb(:,:)
    real    :: a0(:,:),b0(:,:)
    integer :: iz,ir
    real    :: tt
    real    :: boa
    real,    allocatable :: d1(:,:,:),d2(:,:,:)

    allocate(d1(nz,nx,nr),d2(nz,nx,nr))

    write(0,*) "compute FFD coefficients"
    tt = abs(maxval(b0))

write(0,*) 'nz'
    do iz=1,nz
      do ir=1,nr
          boa = (b0(iz,ir)/tt)/a0(iz,ir)
          d1(iz,:,ir)=a0(iz,ir)*boa**2*((bb(iz,:)/b0(iz,ir))**2*(a0(iz,ir)/aa(iz,:))   -1.)
          d2(iz,:,ir)=a0(iz,ir)*boa**4*((bb(iz,:)/b0(iz,ir))**4*(a0(iz,ir)/aa(iz,:))**3-1.)
       end do
    end do

!    where(abs(d1)<epsilon(d1)) 
!       d1=epsilon(d1)
!    end where
write(0,*) 'CHECK 2'
    m0 =      d1    / tt**2
    n0 =-c1 * d1**2
    r0 = c2 * d2
write(0,*) 'CHECK 3'

    m0 = m0 * kmu
    n0 = n0 * knu
    r0 = r0 * kro
write(0,*) 'Finished FFD computation'

    deallocate(d1,d2)
  end subroutine wemig_ffd_coef
  !----------------------------------------------------------------
  subroutine wemig_coef_report()

    write(0,*) minval(m0),"< mu <",maxval(m0)
    write(0,*) minval(n0),"< nu <",maxval(n0)
    write(0,*) minval(r0),"< ro <",maxval(r0)

  end subroutine wemig_coef_report
  !----------------------------------------------------------------  
  !! finite-differences solver
  subroutine wemig_fds(dax,w,iz,ir,dir)
    real                  :: w,dir
    integer               :: iz,ir
    complex, dimension(:) :: dax
    integer               :: ix

    mu = m0(iz,:,ir) * dir
    nu =-n0(iz,:,ir) / w
    ro = r0(iz,:,ir) * dir / w**2

    lk = cmplx(ro+sixth*mu, nu)
    rk = cmplx(ro+sixth*mu,-nu)
    ck = cmplx(         mu, 0.)

    !!             |                      |
    !!     b       |          a           |     c
    !!  ro + i nu  |  mu - 2(ro + i nu )  |  ro + i nu
    !!             |                      |
    a = ck - 2.*lk  
    b = lk(1  :nx-1)
    c = lk(1+1:nx  )

    do ix=2,nx-1
       u(ix)=           rk(ix) *(dax(ix-1)+dax(ix+1)) & 
       +    (ck(ix)-2.0*rk(ix))* dax(ix)
       !!             |                      | 
       !!  ro - i nu  |  mu - 2 (ro - i nu)  | ro - i nu
       !!             |                      | 
    end do
    ix= 1; u(ix)=(ck(ix)-2.0*rk(ix))*dax(ix) + rk(ix)*dax(ix+1)
    ix=nx; u(ix)=(ck(ix)-2.0*rk(ix))*dax(ix) + rk(ix)*dax(ix-1)

    call thr(a,b,c,u,nx)
    dax = u
  end subroutine wemig_fds
  !----------------------------------------------------------------
  !! tridiagonal solver
  subroutine thr(a,b,c,v,n)
    integer :: n
    complex :: a(n),b(n-1),c(n-1),v(n)
    integer :: i

    b(1)=b(1)/a(1)
    v(1)=v(1)/a(1)
    do i=2,n
       a(i)=a(i)-c(i-1)*b(i-1)
       if (i<n) then 
          b(i)=b(i)/a(i)
       end if
       v(i)=(v(i)-c(i-1)*v(i-1))/a(i)
    end do
    do i=n-1,1,-1
       v(i)=v(i)-b(i)*v(i+1)
    end do
  end subroutine thr
  !----------------------------------------------------------------
  subroutine attribute(x)
    complex :: x(:)
    write(0,*) minval( real(x)),maxval( real(x))
    write(0,*) minval(aimag(x)),maxval(aimag(x))
  end subroutine attribute
  !----------------------------------------------------------------
  subroutine image_condition_offset(sax,rax,img,iz,saxmax,divide,eps)
    complex :: sax(:),rax(:)
    real    :: saxmax,eps,divide
    real    :: img(:,:,:)
    integer :: ix,ihx,iz

    do ix=1,nx !! . . x-LOOP
       do ihx=1,nhx !! . . HX-loop
          if (ix-(nhx+1)/2+ihx .ge. 1 .and. ix-(nhx+1)/2+ihx .le. nx .and.&
               ix+(nhx+1)/2-ihx .ge. 1 .and. ix+(nhx+1)/2-ihx .le. nx       ) then
             
             if (divide .eq. 1.) then
                img(iz,ix,ihx)=img(iz,ix,ihx)+ &
                     real( rax(ix-(nhx+1)/2+ihx)*conjg(sax(ix+(nhx+1)/2-ihx)))/&
                     real( conjg(sax(ix+(nhx+1)/2-ihx))*sax(ix+(nhx+1)/2-ihx)+eps*saxmax )
             else
                img(iz,ix,ihx)=img(iz,ix,ihx)+ &
                     real( rax(ix-(nhx+1)/2+ihx)* &
                     conjg(sax(ix+(nhx+1)/2-ihx)))
             end if
          end if
       end do
    end do
  end subroutine image_condition_offset
!----------------------------------------------------------------
 subroutine image_condition(sax,rax,img,iz,saxmax,divide,eps)
    complex :: sax(:),rax(:)
    real    :: saxmax,eps,divide
    real    :: img(:,:,:)
    integer :: iz

    if (divide .eq. 1.) then
       img(iz,:,1)=img(iz,:,1)+real(rax*conjg(sax))/real(sax*conjg(sax)+eps*saxmax)
    else
       img(iz,:,1)=img(iz,:,1)+real(rax*conjg(sax))
    end if

  end subroutine image_condition
  !----------------------------------------------------------------
                                                                           
end module  SP_2D_VOCIGs_wemig
