module conmap    
  use sep
  use conmap_utils
  implicit none
  integer     , private :: nsamp,nx,nt
  real(KIND=8), private :: dx,ox,depth
contains
  !-----------------------------------------------------------------------

  subroutine conmap_init(nxin,dxin,oxin,nsampin,depthin,ntin)
    integer        :: nxin,nsampin,ntin
    real(KIND=8)   :: dxin,oxin,depthin
    nx=nxin;dx=dxin;ox=oxin;nsamp=nsampin
    depth=depthin;nt=ntin
  end subroutine conmap_init

  !-------------------------------------------------------------------

  subroutine conmap_main(z,topoz,endpoints,zout)
    integer                              :: nmap,mm,n,j,k,ng,m,cnt
    integer, dimension(:),allocatable    :: index1,xtmp
    real(KIND=8)                         :: angler,dgam,dtau,anglers
    real(KIND=8)                         :: xlen,perim,glen
    real(KIND=8),dimension(:),allocatable::a,b,c,as,bs,cs,yline,gdist,gline
    complex(KIND=8)              :: z1 ,z2 ,z3 ,zrot1 ,zto0 ,zto1 ,zrot2
    complex(KIND=8)              :: z1s,z2s,z3s,zrot1s,zto0s,zto1s,zrot2s
    complex(KIND=8)              :: zint,zi,za,zwa,zc1,zc2,zc3,zc4
    complex(KIND=8),dimension(:),allocatable:: zend,zr,zn,zzz,zin,znt,zsave
    complex(KIND=8),dimension(:)         :: z,endpoints,topoz,zget
    complex        ,dimension(:,:)       :: zout
    nmap=4*nx;ng=nx
    n=nmap;  mm=ng*nt
    allocate( zend(4),zr(4),zn(7),index1(4),zzz(4),znt(7) )
    allocate( a(nmap+1),b(nmap+1),c(nmap+1))
    allocate( zin( ng*nt ), zget(ng)  )
    allocate( as(nmap+1),bs(nmap+1),cs(nmap+1),xtmp(1) )
    allocate( yline(nt),gdist(ng),zsave(nmap+1),gline(ng) )
    zend=0.d0;zr=0.d0;zn=0.d0;zzz=0.d0;znt=0.d0;a=0.d0;b=0.d0;c=0.d0
    zin=0.d0;zout=0.;as=0.d0;bs=0.d0;cs=0.d0
    yline=0.d0;gdist=0.d0;zsave=0.d0;gline=0.d0;index1=0;xtmp=0
    zzz=endpoints(1:4)

    !! . . TESTS - Winding and index order
    call wind_test(z,nmap); 
    do j=1,4
       xtmp=minloc(cdabs(zzz(j)-z)); index1(j)=xtmp(1)
    end do
    call index_test(index1)
    write(0,*) 'Did tests'

    !! . . JORDAN REGION --> UHP --> DISK
    !! . . Forward transformation: f
    call invers(z,a,b,c,z1,z2,z3,zrot1,zto0,zto1,angler,zrot2,nmap,n)
    zend=z(index1);
    z1s=z1; z2s=z2; z3s=z3;         !! Fix 3 points
    zrot1s=zrot1;   zrot2s=zrot2;   !! Fix rotation angles
    zto0s=zto0;     zto1s=zto1;     !! Compute image points
    anglers=angler;                 !! Angle coeff
    as=a; bs=b; cs=c                !! Output mapping coeffs 

    !! . . Save image of topo bnd points on UC
    zsave=z; write(0,*) 'Computed mapping Jordan region to disk'

    !! . . DISK --> UHP --> SQUARE
    call get_length(zend,xlen) !! . . Find length of conformal rect
    call get_coeffs(z,zn,xlen) !! . . Rect mapping coefficients (new z)
    do j=1,4
       xtmp=minloc(cdabs(zn(j)-z)); index1(j)=xtmp(1);
    end do

    !! . . Forward transformation: g
    call invers(z,a,b,c,z1,z2,z3,zrot1,zto0,zto1,angler,zrot2,nmap,n)
    zr=z(index1);                           !! Define 4 corner points
    za=(zr(3)-zr(2))/(zr(3)-zr(1))          !! Transform corner
    zwa=(zend(3)-zend(2))/(zend(3)-zend(1)) 
    zc1=zwa*zend(1)-za*zend(2);     
    zc2=za*zend(2)*zr(1)-zwa*zend(1)*zr(2)
    zc3=zwa-za;                  
    zc4=za*zr(1)-zwa*zr(2)

    !! Save image of rect bnd points on UC
    zsave=(zsave*zc4-zc2)/(-zc3*zsave+zc1);
    write(0,*) 'Computed mapping from square to disk'

    !! Image of topo bndry points
    call eval(zsave,a,b,c,z1,z2,z3,zrot1,zto0,zto1,angler,zrot2,nmap,n)
    gline=dreal(zsave(1:ng))
write(0,*) zsave(110:120)

    !do j=3,ng-1
    !  if (gline(j) .gt. gline(j-1) .and. gline(j) .gt. gline(j-2) ) then
    !      gline(j)= gline(j-1) + (gline(j-1)-gline(j+1))/2.
    !      write(0,*) 'BAD J:',j
    !   end if
    !end do
    write(0,*) 'Computed image of topography on rectangle'

    !! . . SETUP OUTPUT GRID
    dgam=xlen/dble(ng+1);
    dtau=1.d0/dble(nt-1); 
    yline=(/(k,k=nt-1,0,-1)/)*dtau

    do j=1,ng
       zin(1+(j-1)*nt:j*nt)=dcmplx(gline(j),yline/2.d0+.5d0)
    end do
    write(0,*) 'Set up output grid'
    write(0,*) 'nt',nt,'ng',ng
    write(0,*) 'dt',dtau,'dg',dgam

    !! . . SQUARE --> DISK
    call invert(zin,a,b,c,z1,z2,z3,zrot1,zto0,zto1,angler,zrot2,mm,n) 
    zin(1:mm)=(zin(1:mm)*zc1+zc2)/(zc3*zin(1:mm)+zc4)
    write(0,*) 'Mapped grid from rectangle to disk'

    !! . . DISK --> JORDAN REGION
    call eval(zin,as,bs,cs,z1s,z2s,z3s,zrot1s,zto0s,zto1s,anglers,zrot2s,mm,n)
    zout=reshape(zin,(/nt,ng/))
    zout(1,:)=topoz
    write(0,*) 'Mapped Grid from disk to Jordan region'  

    !! . . Output to file
    call write_out(nt,ng,dble(ng-1)*dx/(2.*xlen)/dble(nt-1),dble(dx),dtau*dble(nt-1)/2.,xlen)
    deallocate( zend,zr,zn,index1,zzz,znt,a,b,c,zin,zget )
    deallocate( as,bs,cs,xtmp,yline,gdist,zsave,gline )
 end subroutine conmap_main
end module conmap
