program SPelas
  use sep
  use util_
  use SPwemig

  implicit none
  type(myaxis) :: ax,az,aw,ar,as,ah

  integer :: method,nh
  integer  :: iz,ii,nref
  character(len=128)   :: name
  real                 :: forward,dr
  real,    allocatable :: rall(:,:,:,:),sall(:,:,:,:)
  complex, allocatable :: rrefall(:,:,:),srefall(:,:,:)
  real,    allocatable :: saa(:,:),sbb(:,:),smm(:,:)
  complex, allocatable :: sab(:,:)
  real,    allocatable :: sa0(:,:),sb0(:,:)
  real,    allocatable :: raa(:,:),rbb(:,:),rmm(:,:)
  complex, allocatable :: rab(:,:)
  real,    allocatable :: ra0(:,:),rb0(:,:)
  complex, pointer     :: rwfall(:,:,:) !! receiver wavefield
  complex, pointer     :: swfall(:,:,:) !! source   wavefield
  real,    pointer     :: imgall(:,:,:,:) !! image
  complex, pointer     :: rwf(:,:) !! receiver wavefield
  complex, pointer     :: swf(:,:) !! source   wavefield
  real,    pointer     :: img(:,:,:) !! image with offsets

  call sep_init()
  call from_param("method",method,0)    !! propagation type
  call from_param("forward",forward,0.) !! Forward scattering option
  call from_param("nref",nref,1)        !! # reference velocities
  call from_param("nh",nh,1)            !! # image offset
  call hinaxis(ax,1)
  call hinaxis(aw,2) 
  call hinaxis(as,3)

  name="rabm"
  call ainaxis(az,1,name)
  name="rabref"
  call ainaxis(ar,2,name)

  if(nref .eq. 1 ) ar%n=1
  if(method==0) ar%n=1   !! pure finite-differences

  write(0,*) ax%n,ax%o,ax%d
  write(0,*) az%n,az%o,az%d
  write(0,*) aw%n,aw%o,aw%d
  write(0,*) ar%n,ar%o,ar%d
  write(0,*) as%n,as%o,as%d

  !! . . Create offset axis
  dr=ax%d
  call houaxis(az,1)
  call houaxis(ax,2)
  call houaxis(as,3)
  call to_history("n4",nh)
  call to_history("d4",dr)
  call to_history("o4",-(nh-1)/2.*dr)
  call to_history("esize",4)

  !! . . Allocate arrays for global coefficients
  allocate(swfall(ax%n,aw%n,as%n))
  allocate(rwfall(ax%n,aw%n,as%n))
  allocate(imgall(az%n,ax%n,as%n,nh))
  allocate(rall(az%n,ax%n,as%n,5) , sall(az%n,ax%n,as%n,5) )
  allocate(rrefall(az%n,ar%n,as%n),srefall(az%n,ar%n,as%n) )

  !! . . Allocate arrays for individual shots
  allocate(swf(ax%n,aw%n),rwf(ax%n,aw%n))
  allocate(img(az%n,ax%n,nh),raa(az%n,ax%n))
  allocate(rbb(az%n,ax%n),rmm(az%n,ax%n))
  allocate(saa(az%n,ax%n),sbb(az%n,ax%n))
  allocate(smm(az%n,ax%n),rab(az%n,ax%n))
  allocate(ra0(az%n,ar%n),rb0(az%n,ar%n))
  allocate(sab(az%n,ax%n),sa0(az%n,ar%n))
  allocate(sb0(az%n,ar%n))   

  !! . . Read in the coefficients
  call sreed("rabm",rall,size(rall)*4)
  call sreed("sabm",sall,size(sall)*4)
  call sreed("rabref",rrefall,size(rrefall)*8)
  call sreed("sabref",srefall,size(srefall)*8)

  !! . . Read in RWF and SWF
  call sep_read(rwfall)
  call sreed("swf",swfall,size(swfall)*8)

  !! . . Initialize module and image space
  call wemig_init(ax,az,aw,ar,method,forward,nh)
  imgall=0.

  do ii=1,as%n   !! . . Loop over shot points
     img=0.;rwf=0.;swf=0.;raa=0.;saa=0.;rbb=0.;sbb=0.;
     ra0=0.;rb0=0.;sa0=0.;sb0=0.;rwf=0.;swf=0.;
     rwf=rwfall(:,:,ii)
     swf=swfall(:,:,ii)
     raa=rall(:,:,ii,1);        saa=sall(:,:,ii,1)
     rbb=rall(:,:,ii,2);        sbb=sall(:,:,ii,2)
     rmm=rall(:,:,ii,3);        smm=sall(:,:,ii,3)
     ra0= real(rrefall(:,:,ii));sa0= real(srefall(:,:,ii))
     rb0=aimag(rrefall(:,:,ii));sb0=aimag(srefall(:,:,ii))

     write(0,*) "raa",minval(raa),maxval(raa)
     write(0,*) "rbb",minval(rbb),maxval(rbb)
     write(0,*) "rmm",minval(rmm),maxval(rmm)
     write(0,*) "saa",minval(saa),maxval(saa)
     write(0,*) "sbb",minval(sbb),maxval(sbb)
     write(0,*) "smm",minval(smm),maxval(smm)
     write(0,*) "ra0",minval(ra0),maxval(ra0)
     write(0,*) "rb0",minval(rb0),maxval(rb0)
     write(0,*) "sa0",minval(sa0),maxval(sa0)
     write(0,*) "sb0",minval(sb0),maxval(sb0)

     !! . . Image event
     call wemig(rwf,swf,img,raa,rbb,rmm,ra0,rb0,saa,sbb,smm,sa0,sb0,ii)
     imgall(:,:,ii,:)=img

  end do

  call sep_write(imgall)

  call sep_close()
  call exit()
end program SPelas
