module  prepconmap
  use sep 
  implicit none
  integer     , private :: nsamp,nx
  real(KIND=8), private :: xcent,zcent,dx,ox,depth
contains
  !-------------------------------------------------------------------
  subroutine prepconmap_init(nxin,dxin,oxin,nsampin,xcentin,zcentin,depthin)
    integer        :: nxin,nsampin
    real(KIND=8)   :: xcentin,zcentin,dxin,oxin,depthin
    nx=nxin;dx=dxin;ox=oxin;nsamp=nsampin;xcent=xcentin
    zcent=zcentin;depth=depthin
  end subroutine prepconmap_init
  !-------------------------------------------------------------------
  subroutine prepconmap_main(tline,points,endpoints,z)
    integer                                  :: i,j,cnt
    integer,dimension(:),allocatable         :: nsline,endline
    real(KIND=8)                             :: dsl,dsr,dsy,dsx
    real(KIND=8),dimension(:)                :: tline,points
    real(KIND=8),dimension(:),allocatable    :: yvals,endvals
    complex(KIND=8), dimension(:),intent(out):: z,endpoints
    allocate( nsline(nsamp),yvals(nsamp) )
    yvals=0.d0
    write(0,*) points(nx)
    !! . . Interpolate top points
    if (nsamp .gt. 1) then
       do i=1,nx-nsamp,nsamp
          nsline=(/ (j,j=i,i+nsamp-1) /)
          dsy=points(nsline(nsamp))-points(nsline(1))
          dsx=tline (nsline(nsamp))-tline (nsline(1))
          yvals=dsy/dsx*(nsline-nsline(1))*dx+points(nsline(1))
          z(nsline)=dcmplx(tline(nsline),yvals)
       end do
       !! . . Deal with end of line issues
       allocate( endline(nx+1-i), endvals(nx+1-i) )
       endline=(/ (j,j=i,nx) /)
       dsy=points(endline(size(endline)))-points(endline(1))
       dsx=tline (endline(size(endline)))-tline (endline(1))
       endvals=dsy/dsx*(endline-endline(1))*dx+points(endline(1))
       z(endline)=dcmplx(tline(endline),endvals)
       deallocate (endline,endvals)
    else 
       z(1:nx)=dcmplx(tline(1:nx),points(1:nx))
    end if

    !! . . Right side
    dsr=2.d0*(depth-points(nx))/dble(nx-1)
    do i=1,nx
       z(nx+i)=dcmplx(tline(nx)+20.d0*dx,dble(i-1)*dsr+points(nx))
    end do

    !! . . Bottom points
    cnt=1
    do i=nx,1,-1
       z(2*nx+cnt)=dconjg(z(i))+dcmplx(0.d0,2.d0*depth)
       cnt=cnt+1
    end do
    !! . . Left side
    dsl=2.d0*(depth-points(1))/dble(nx-1)
    cnt=1
    do i=nx,1,-1
       z(3*nx+cnt)=dcmplx(tline(1)-20.d0*dx,dble(i-1)*dsl+points(1))
       cnt=cnt+1
    end do
    endpoints(1) = z(nx+1  )
    endpoints(2) = z(2*nx  )
    endpoints(3) = z(3*nx+1)
    endpoints(4) = z(4*nx  )  
    endpoints(5) = dcmplx(xcent,zcent)
    z(4*nx+1)=endpoints(5)
    write(0,*) 'ENDPOINTS'
    write(0,*) endpoints(1:4)
    write(0,*) 'CONFORMAL CENTER'
    write(0,*) endpoints(5)
  end subroutine prepconmap_main
  !-------------------------------------------------------------------
end module prepconmap
