program Conmap    
  !! invers calculates the coefficients for mapping a given 
  !!     Jordan curve to the disk 
  !! invert calculates the mapping
  !! eval evaluates the points for a mapping
  use sep
  use CMap
  implicit none
  integer                              :: nmap,nt,mm,n,j,k,ng,m,cnt
  integer, dimension(:),allocatable    :: index1,xtmp
  real, dimension(:),allocatable       :: topo
  real(KIND=8)                         :: angler,dgam,dtau,anglers
  real(KIND=8)                         :: xlen,perim,glen
  real(KIND=8),dimension(:),allocatable:: a,b,c,as,bs,cs,yline,gdist,gline
  complex,dimension(:),allocatable     :: points,endpoints
  complex,dimension(:,:),allocatable   :: zout
  complex(KIND=8)                      :: z1 ,z2 ,z3 ,zrot1 ,zto0 ,zto1 ,zrot2
  complex(KIND=8)                      :: z1s,z2s,z3s,zrot1s,zto0s,zto1s,zrot2s
  complex(KIND=8)                      :: zint,zi,za,zwa,zc1,zc2,zc3,zc4
  complex(KIND=8),dimension(:),allocatable:: zend,zr,zn,z,zzz,zin,ztopo,znt,zsave

  call sep_init()
  call from_history("n1",nmap); ng=nmap/4
  call from_param("nt",nt,50);  n=nmap;  mm=ng*nt
  allocate( zend(4),zr(4),zn(7),index1(4),zzz(4),znt(7) )
  allocate( z(nmap+1),a(nmap+1),b(nmap+1),c(nmap+1))
  allocate( points(nmap),endpoints(5), zin( ng*nt ), zout(nt,ng) )
  allocate( as(nmap+1),bs(nmap+1),cs(nmap+1),xtmp(1) )
  allocate( yline(nt),gdist(ng),zsave(nmap+1),gline(ng) )

  !! . . INPUT
  call sep_read(endpoints,"endpoints"); zzz=dcmplx(endpoints(1:4))
  call sep_read ( points ); z(1:nmap)=dcmplx(points)
  z(nmap+1)=dcmplx( endpoints(5) ) 

!  glen    =1./cdabs(z(1)-z(nmap))
!  gdist(1)=1./cdabs(z(1)-z(nmap))
!  do j=2,ng
!     glen=glen+1./cdabs(z(j)-z(j-1))
!     gdist(j)=gdist(j-1)+1./cdabs(z(j)-z(j-1))
!     write(0,*) j,1./cdabs(z(j)-z(j-1))
!  end do
!  glen=glen+1./cdabs(z(ng+1)-z(ng))
!  gdist=gdist/glen

  !! . . TESTS - Winding and index order
  call wind_test(z,nmap); 
  do j=1,4
     xtmp=minloc(cdabs(zzz(j)-z)); index1(j)=xtmp(1)
  end do
  call index_test(index1)

  !! . . JORDAN REGION --> DISK
  call invers(z,a,b,c,z1,z2,z3,zrot1,zto0,zto1,angler,zrot2,nmap,n)
  zend=z(index1);z1s=z1;z2s=z2;z3s=z3;zrot1s=zrot1;zto0s=zto0;
  zto1s=zto1;anglers=angler;zrot2s=zrot2;as=a;bs=b;cs=c 
  zsave=z;

  !! . . DISK --> SQUARE
  call get_length(zend,xlen) !! . . Find length of conformal rect
  call get_coeffs(z,zn,xlen) !! . . Mapping coefficients (new z)
  do j=1,4
     xtmp=minloc(cdabs(zn(j)-z)); index1(j)=xtmp(1);
  end do

  !! . . INVERSE MAPPING POINTS
  call invers(z,a,b,c,z1,z2,z3,zrot1,zto0,zto1,angler,zrot2,nmap,n)
  zr(1:4)=z(index1(1:4)); za=(zr(3)-zr(2))/(zr(3)-zr(1))
  zwa=(zend(3)-zend(2))/(zend(3)-zend(1))
  zc1=zwa*zend(1)-za*zend(2);  zc2=za*zend(2)*zr(1)-zwa*zend(1)*zr(2)
  zc3=zwa-za;                  zc4=za*zr(1)-zwa*zr(2)
  zsave=(zsave*zc4-zc2)/(-zc3*zsave+zc1)

  !! . . Image of bndry points
  call eval(zsave,a,b,c,z1,z2,z3,zrot1,zto0,zto1,angler,zrot2,nmap,n)
  gline=real(zsave(1:ng))

  !! . . SETUP OUTPUT GRID
  dgam=xlen/dble(ng+1);
  dtau=1./dble(nt-1); 
  yline=(/(k,k=0,nt-1)/)*dtau
  do j=1,ng
     !zin( 1+(j-1)*nt:j*nt )=dcmplx(dble(j)*dgam,yline) !! Regular spaced
     zin(1+(j-1)*nt:j*nt)=dcmplx(gline(j),yline/2+.5)
!     zin(1+(j-1)*nt:j*nt)=dcmplx(xlen*gdist(j),yline/2+.5)
  end do

  !! . . SQUARE --> DISK
  call invert(zin,a,b,c,z1,z2,z3,zrot1,zto0,zto1,angler,zrot2,mm,n) 
  zin(1:mm)=(zin(1:mm)*zc1+zc2)/(zc3*zin(1:mm)+zc4)

  !! . . DISK --> JORDAN REGION
  call eval(zin,as,bs,cs,z1s,z2s,z3s,zrot1s,zto0s,zto1s,anglers,zrot2s,mm,n)
  zout=reshape(zin,(/nt,ng/))
  zout(nt,1:ng)=points(1:ng)

  !! . . Output to file
  call write_out(zout,nt,ng)
  call exit()

end program Conmap
