module CMap
  use sep 
  implicit none
  real(KIND=8), parameter, private ::   acc =1.d-12
  real(KIND=8), parameter, private ::   accr=1.d+12
  real(KIND=8), parameter, private ::   r1  =1.25
  real(KIND=8), parameter, private ::   r2  =.25
contains

  !-------------------------------------------------------------------------
  subroutine invers(z,a,b,c,z1,z2,z3,zrot1,zto0,zto1,angler,zrot2,mm,n)
    !! computes map from region to disk and finds inverse image of vertices.
    !! z(n+1) mapped to 0, z(1) mapped to 1
    integer                            :: mm,n,k,j,i
    real(KIND=8)                       :: angler,aa,bb,cc
    real(KIND=8), dimension(:)         :: a,b,c
    complex(KIND=8)                    :: z1,z2,z3,zrot1,zto0,zto1,zrot2
    complex(KIND=8), dimension(:)      :: z
    real(KIND=8)                       :: b1,oma,omar,azt,zl
    real(KIND=8)                       :: ar,pia,x,y,rr1,rr2,a0,c2,a1,c1
    real(KIND=8)                       :: xl,a2,pi,tpi
    real(KIND=8)                       :: xlzw,xzw,yzw,test,xw,xwr,xt,xwt
    complex(KIND=8)                    :: zi,z1rc,z2r,zt,zangc,zrota,ztip
    complex(KIND=8)                    :: zc1,zwr,zw,zwi,zwt,zto0c
    pi  =acos(-1.)
    tpi =2.*pi

    !! Map complement of arc of circle through z(1),z(2),z(3) to UHP
    zi=dcmplx(0.,1.)
    z1=z(1)
    z2=z(2)
    z3=z(3)
    zrot1=(z(1)-z(2))/(z(2)-z(3))
    z(4:n+1)=zi*cdsqrt(zrot1*(z(4:n+1)-z(3))/(z(4:n+1)-z(1)))
    z(1)=dcmplx(-1.d+20,0.)
    z(2)=dcmplx(-1.    ,0.)
    z(3)=dcmplx(0.     ,0.)

    do j=4,n-2,2 !! Pull back curve
       z1rc=dconjg(1./z(j))
       z2r=1./z(j+1)
       b1=-dimag(z1rc*z2r)/dimag(z1rc+z2r)
       zt=z2r+b1
       pia=-datan2(dimag(zt),dreal(zt))
       a(j)=pia/pi
       aa=a(j)
       oma=1.-aa 
       omar=1./oma 
       azt=cdabs(zt)
       xl=(aa**aa)*(oma**oma) 
       b(j)=1./(xl*azt)
       c(j)=b1*b(j)
       ar=1./aa 
       zangc=cdexp(dcmplx(0.,-pia)) 
       zrota=zangc*zangc
       x=xl*dreal(zangc) 
       y=-xl*dimag(zangc) 
       ztip=dcmplx(x,y) 
       rr1=1./(r1*xl) 
       rr2=r2*y 
       a0=aa-oma 
       c2=-2.*a0/3.
       a1=aa*oma/2.
       c1=4.*a1
       a2=c2*a1 
       zc1=dcmplx(0.,dsqrt(c1)) 
       do k=j+2,n+1
          zwr=c(j)+b(j)/z(k)
          xlzw=cdabs(zwr)
          if (xlzw.lt.rr1) then
             zt=1.+(a0+(a1+a2*zwr)*zwr)*zwr
             call newton(zwr,aa,zt,zrota)
             go to 8
          endif
          zw=1./zwr
          xzw=dreal(zw)
          yzw=dimag(zw)
          if (cdabs(zw-ztip).lt.rr2) then
             zwi=cdsqrt(zw/ztip-1.)
             if((dreal(zwi).lt.1.d-4).and.(dimag(zwi).lt.0.)) zwi=-zwi
             zt=(zc1+c2*zwi)*zwi
             call newt1(zwi,aa,c1,zt,zrota)
          elseif (xlzw.gt.accr) then
             zt=-oma
          elseif (datan2(yzw,xzw).ge.pia) then
             zwt=(zangc*zw)**omar
             zt=-oma+zwt
             test=cdabs(zwt)
             test=test*test
             if(test.lt.acc) go to 8 
             zt=zt*zwr
             call newton(zwr,aa,zt,zrota)
          else
             zwt=zw**ar
             zt=aa+zwt
             test=cdabs(zwt)
             test=test*test
             if(test.lt.acc) go to 8
             zt=zt*zwr
             call newton(zwr,aa,zt,zrota)
          endif
8         z(k)=zt
       end do
       z(j+1)=dcmplx(0.,0.)
       z(j-1)=dcmplx(aa-1.,0.)
       zwr=c(j)+b(j)/z(j)
       zw=1./zwr
       if (cdabs(zw-ztip).lt.rr2) then
          zwi=dcmplx(0.d0,dsqrt(dreal(1.d0-zw/ztip)))
          zt=(zc1+c2*zwi)*zwi
          call newt1(zwi,aa,c1,zt,zrota)
       elseif (cdabs(zwr).gt.accr) then
          zt=-oma
       else
          zwt=(zangc*zw)**omar
          zt=-oma+zwt
          test=cdabs(zwt)
          test=test*test
          if (test.lt.acc) go to 28
          zt=zt*zwr
          call newton(zwr,aa,zt,zrota)
       endif
28     z(j)=zt
       do k=1,j-2
          xw=dreal(z(k))
          xwr=c(j)+b(j)/xw
          xlzw=dabs(xwr)
          if (xlzw.lt.rr1) then
             xt=1.+(a0+(a1+a2*xwr)*xwr)*xwr
             call rnewt(xwr,aa,xt) 
             go to 18
          endif
          xw=1./xwr
          if (xlzw**omar.gt.accr) then
             xt=-oma
          elseif (xw.le.0.) then
             xwt=-(-xw)**omar
             xt=-oma+xwt
             test=xt*xt
             if (test.lt.acc) go to 18
             xt=xt*xwr
             call rnewt(xwr,aa,xt)
          else
             xwt=xw**ar
             xt=aa+xwt
             test=xwt*xwt
             if (test.lt.acc) go to 18
             xt=xt*xwr
             call rnewt(xwr,aa,xt)
          endif
18        z(k)=dcmplx(xt,0.)
       end do
    end do
    z(n-1)=dcmplx(0.,0.)
    !! Map circular sector to disk
    zrot2=1./z(n)-1./z(1)
    zto1=z(1)
    angler=pi/datan2(dimag(-zrot2),dreal(-zrot2))
    zto0=(z(n+1)*zrot2/(1.-z(n+1)/z(1)))**angler
    zto0c=dconjg(zto0)
    do j=2,n-2
       zt=(z(j)*zrot2/(1.-z(j)/z(1)))**angler
       z(j)=(zt-zto0)/(zt-zto0c)
    end do
    z(n-1)=zto0/zto0c
    zt=(z(n)*zrot2/(1.d0-z(n)/z(1)))**angler
    z(n)=(zt-zto0)/(zt-zto0c)
    z(1)=dcmplx(1.,0.)
    z(n+1)=dcmplx(0.,0.)
  end subroutine invers
  !-------------------------------------------------------------------------
  subroutine rnewt(xwr,aa,x)  
    !!    Version of newtons method (upper-half plane)
    !!    iterating on x/w  (xwr=1/w) -- REAL arithmetic
    !!    x=output and initial guess
    integer                                    :: i
    real(KIND=8)                               :: xwr,aa,x,xb1,xb2,xma,xmap
    real(KIND=8)                               :: xfract,xd,test
    xb1=-aa*xwr
    xb2=xb1+xwr
    do i=1,20
       xma=x+xb1
       xmap=x+xb2
       xfract=(xmap/xma)**aa
       xd=(xfract-xmap)*xma/x
       x=x+xd
       test=dabs(xd)
       if(test.lt.acc) exit
    end do
    x=x/xwr
  end subroutine rnewt

  !-------------------------------------------------------------------------

  subroutine invert(z,a,b,c,z1,z2,z3,zrot1,zto0,zto1,angler,zrot2,mm,n)
    !! Computes map from region to disk 
    integer                            :: mm,n,k,j,i
    real(KIND=8)                       :: angler,aa,bb,cc
    real(KIND=8), dimension(:)         :: a,b,c
    complex(KIND=8)                    :: z1,z2,z3,zrot1,zto0,zto1,zrot2
    complex(KIND=8), dimension(:)      :: z
    real(KIND=8)                       :: oma,omar,xl,pia,ar,x,y,rr1,rr2
    real(KIND=8)                       :: a0,a1,c2,a2,c1,xlzw,test,pi,tpi
    complex(KIND=8)                    :: zi,zd,zangc,zrota,ztip,zc1,zt
    complex(KIND=8)                    :: zwr,zw,zwi,zwrot,zwt,zto0c
    zi=dcmplx(0.,1.)
    pi  =acos(-1.)
    tpi =2.*pi
    !! Map complement of arc of circle through z1,z2,z3 to the upper half plane
    do j=1,mm
       zd=z(j)-z1
       if (cdabs(zd).lt.1.d-20) then
          z(j)=dcmplx(1.d+20,0.)
          cycle
       endif
       z(j)=zi*cdsqrt(zrot1*(z(j)-z3)/(z(j)-z1))
    end do
    !! Pull back points
    do j=4,n-2,2
       aa=a(j)
       oma=1.-aa 
       omar=1./oma 
       xl=(aa**aa)*(oma**oma) 
       pia=pi*aa 
       ar=1./aa 
       zangc=cdexp(dcmplx(0.,-pia)) 
       zrota=zangc*zangc
       x=xl*dreal(zangc) 
       y=-xl*dimag(zangc) 
       ztip=dcmplx(x,y) 
       rr1=r1*xl 
       rr2=r2*y 
       a0=aa-oma 
       a1=aa*oma/2. 
       c2=-2.*a0/3.
       a2=c2*a1
       c1=4.*a1 
       zc1=dcmplx(0.,dsqrt(c1)) 
       do k=1,mm
          if (cdabs(z(k)).lt.1.d-4) then 
             zt=-oma
             go to 8
          endif
          zwr=c(j)+b(j)/z(k)
          xlzw=1./cdabs(zwr)
          if (xlzw.gt.rr1) then
             zt=1.+(a0+(a1+a2*zwr)*zwr)*zwr
             call newton(zwr,aa,zt,zrota) 
             go to 8
          endif
          zw=1./zwr
          if (cdabs(zw-ztip).lt.rr2) then
             zwi=cdsqrt(zw/ztip-1.)
             if ((dreal(zwi).lt.1.d-4).and.(dimag(zwi).lt.0.d0)) then 
                zwi=-zwi
             end if
             zt=(zc1+c2*zwi)*zwi
             test=cdabs(zwi)
             test=test*test*test
             if (test.lt.acc) then
                if (dreal(zt).gt.0.) zt=-zt
                if (dimag(zt).lt.0.) zt=dconjg(zt)
                go to 8
             endif
             call newt1(zwi,aa,c1,zt,zrota) 
             go to 8
          endif
          zwrot=zangc*zw
          if (dimag(zwrot).gt.-1.d-3) then
             zwt=zwrot**omar
             zt=-oma+zwt
             test=cdabs(zwt)
             test=test*test
             if(test.lt.acc) go to 8 
             zt=zt*zwr
             call newton(zwr,aa,zt,zrota)
          else
             zwt=zw**ar
             zt=aa+zwt
             test=cdabs(zwt)
             test=test*test
             if (test.lt.acc) go to 8
             zt=zt*zwr
             call newton(zwr,aa,zt,zrota)
          endif
8         z(k)=zt
       end do
    end do
    zto0c=dconjg(zto0)
!    do j=1,mm
!       zt=(z(j)*zrot2/(1.-z(j)/zto1))**angler
!       z(j)=(zt-zto0)/(zt-zto0c)
!    end do
    z(1:mm)=((z(1:mm)*zrot2/(1.-z(1:mm)/zto1))**angler-zto0)/&
    ((z(1:mm)*zrot2/(1.-z(1:mm)/zto1))**angler-zto0c)  
  end subroutine invert
  !------------------------------------------------------------------------
  subroutine newton(zwr,aa,z,zrota)
    !!  test version of newtons method (upper-half plane)
    !!  iterating on z/w
    integer                          :: i
    real(KIND=8)                     :: aa,test
    complex(KIND=8)                  :: zwr,zrota,z1,z2,z,zr,zfr,z3,zfcn
    complex(KIND=8)                  :: zq,zd,zb1,zb2,zma,zmap,zw,zz
    zb1=-aa*zwr
    zb2=zb1+zwr
    do i=1,20
       zma=z+zb1
       zmap=z+zb2
       zr=zmap/zma
       zfr=zr**aa
       if (dimag(zfr).gt.0.) then
          if(dreal(zr).lt.0.) then
             zfr=zfr*zrota
          end if
       endif
       zd=(zfr-zmap)*zma/z
       z=z+zd
       test=cdabs(zd)
       if (test.lt.acc) exit
    end do
    z=z/zwr
  end subroutine newton
  !------------------------------------------------------------------------
  subroutine newt1(zwi,aa,c1,z,zrota)
    !!  test version of newtons method (upper-half plane)
    !!    iterating on z.  Used near tip of the slit.
    integer                          :: i
    real(KIND=8)                     :: aa,c1,oma,test
    complex(KIND=8)                  :: zwi,zrota,z1,z2,z,zr,zfr,z3,zfcn
    complex(KIND=8)                  :: zq,zd
    oma=1.-aa
    do i=1,20
       z1=1.-z/aa
       z2=1.+z/oma
       zr=z1/z2
       zfr=zr**aa
       if (dimag(zfr).gt.0.) then
          if (dreal(zr).lt.0.) then 
             zfr=zfr*zrota
          end if
          z3=cdsqrt(zfr*z2-1.)
          if(dreal(zr).gt.0.) then
             z3=-z3
          end if
       else
          z3=cdsqrt(zfr*z2-1.)
       endif
       zfcn=z3-zwi
       zq=z3/z
       zd=zq*c1*z1*zfcn/zfr
       z=z+zd
       test=cdabs(zd)
       if (test.lt.acc) exit
    end do
  end subroutine newt1
  !------------------------------------------------------------------------
  subroutine eval(z,a,b,c,z1,z2,z3,zrot1,zto0,zto1,angler,zrot2,mm,n)
    !!  computes map from open disk to region at vector z
    integer                                 :: mm,n,k,i,j,nk
    integer, dimension(:), allocatable      :: kline
    real(KIND=8)                            :: angler,angle
    real(KIND=8), dimension(:)              :: a,b,c
    real(KIND=8), dimension(:),allocatable  :: aa,bb,cc
    complex(KIND=8)                         :: z1,z2,z3,zrot1,zto0,zto1,zrot2
    complex(KIND=8)                         :: zi,zto0c,z1rot,zma,zmap,zr,zfr
    complex(KIND=8),dimension(:)            :: z
    complex(KIND=8),dimension(:),allocatable:: zrv,zfrv,zrota
    nk=(n-6)/2+1
    allocate( kline(nk),aa(nk),bb(nk),cc(nk),zrota(nk), zrv(mm),zfrv(mm)  )
    kline=(/ (k,k=n-2,4,-2) /)
    aa=a(kline)
    zrota=cdexp(dcmplx(0.,2.*dacos(-1.)*aa))
    bb=1./b(kline)
    cc=c(kline)*bb
    zi=dcmplx(0.,1.)
    zto0c=dconjg(zto0)
    angle=1./angler
    z1rot=zi**angle

    z(1:mm)=(z1rot*((zi*(zto0-z(1:mm)*zto0c)/(z(1:mm)-1.))**angle))/&
      (zrot2+(z1rot*((zi*(zto0-z(1:mm)*zto0c)/(z(1:mm)-1.))**angle))/zto1)

    do k=1,nk
      zrv=(z-aa(k))/((z-aa(k))+1.)
      zfrv=zrv**aa(k)
      where (dimag(zfrv).lt.0. .and. dreal(zrv) .lt.0.) zfrv=zfrv*zrota(k)
      z=((z-aa(k))+1.)*zfrv/(bb(k)-cc(k)*((z-aa(k))+1.)*zfrv)
    end do

    z(1:mm)=(z3-z1*(-z(1:mm)*z(1:mm)/zrot1))/(1.-(-z(1:mm)*z(1:mm)/zrot1))

    deallocate(kline,aa,bb,cc,zrota,zrv,zfrv)
  end subroutine eval

  !------------------------------------------------------------------------
 
  subroutine wind_test(z,n)
  integer                              :: j,n
  real(KIND=8)                         :: wind,dthet,pi
  complex(KIND=8)                      :: z0,zdo,zbo,zd,zb,ztemp
  complex(KIND=8),dimension(:)         :: z
  !! . . Test to ensure data is wound correctly
  pi=dacos(-1.)
  wind=0.
  z0=z(n+1)
  zdo=1./(z(1)-z0)
  zbo=zdo-1./(z(n)-z0)
  do j=1,n-1
     zd=1./(z(j+1)-z0)
     zb=zd-zdo
     dthet=dimag(cdlog(zb/zbo))
     wind=wind+dthet
     zbo=zb
     zdo=zd
  end do
  dthet=dimag(cdlog((1./(z(1)-z0)-zdo)/zbo))
  wind=wind+dthet
  wind=-wind/(2.*pi)
  if (dabs(wind-1.).gt.0.1) then
     write(0,*) 'Reversing input data'
     do j=1,n/2
        ztemp=z(j)
        z(j)=z(n-j)
        z(n-j)=ztemp
     end do
  endif
  end subroutine wind_test

  !------------------------------------------------------------------------
  
  subroutine index_test(index1)
    integer                         :: ind
    integer, dimension(:)           :: index1

    if (index1(1).gt.index1(2)) then
       ind=index1(1)
       index1(1)=index1(2)
       index1(2)=ind
    endif
    if (index1(2).gt.index1(3)) then
       ind=index1(2)
       index1(2)=index1(3)
       index1(3)=ind
    endif
    if (index1(3).gt.index1(4)) then
       ind=index1(3)
       index1(3)=index1(4)
       index1(4)=ind
    endif
    if (index1(1).gt.index1(2)) then
       ind=index1(1)
       index1(1)=index1(2)
       index1(2)=ind
    endif
    if (index1(2).gt.index1(3)) then
       ind=index1(2)
       index1(2)=index1(3)
       index1(3)=ind
    endif
    if (index1(1).gt.index1(2)) then
       ind=index1(1)
       index1(1)=index1(2)
       index1(2)=ind
    endif
  end subroutine index_test

  !------------------------------------------------------------------------

  subroutine get_length(zend,xlen)
    real(KIND=8)                          :: s,xk,ax,bx,tmp,xkk,xkp,xkkp,xlen
    complex(KIND=8)                       :: zr1,zr2
    complex(KIND=8), dimension(:)         :: zend
    zr1=(zend(1)-zend(3))/(zend(1)-zend(4))
    zr2=(zend(2)-zend(4))/(zend(2)-zend(3))
    s=dreal(zr1*zr2)
    xk=2.*s-1.+2*dsqrt(s*s-s)
    if(xk.gt.1.) then 
       xk=1./xk
    end if
    ax=1.
    bx=xk
    do while (dabs(ax-bx).gt.1.d-13)
       tmp=(ax+bx)/2.
       bx=dsqrt(ax*bx)
       ax=tmp
    end do
    xkk=ax
    xkp=dsqrt(1.-xk*xk)
    ax=1.
    bx=xkp
    do while (dabs(ax-bx).gt.1.d-13)
       tmp=(ax+bx)/2.
       bx=dsqrt(ax*bx)
       ax=tmp
    end do
    xkkp=ax
    xlen=2.*xkk/xkkp
  end subroutine get_length
  !------------------------------------------------------------------------
  subroutine get_coeffs(z,zn,xlen)
    integer                      :: n,ind,j,nj,k,nx
    real(KIND=8)                 :: xlen,xl
    complex(KIND=8)              :: z1r,z2r,z3r,z4r,zint,zi,zd,zz
    complex(KIND=8),dimension(:) :: z,zn
    z1r=dcmplx(0.,1.)
    z2r=dcmplx(0.,0.)
    z3r=dcmplx(xlen,0.)
    z4r=z3r+z1r
    zint=(z2r+z4r)/2.
    zi=dcmplx(0.,1.)
    nx=size(z)-1
    if (xlen.lt.1.) then
       zn(1)=dcmplx(0.,0.5)
       zn(2)=dcmplx(0.,0.)
       zn(3)=dcmplx(xlen,0.)
       zn(4)=zn(3)+zn(1)
       zn(5)=zn(3)+zi
       zn(6)=zi
       zint=zn(4)/2.
    else
       zn(1)=dcmplx(xlen/2.,0.)
       zn(2)=dcmplx(xlen,0.)
       zn(3)=zn(2)+zi
       zn(4)=zn(1)+zi
       zn(5)=zi
       zn(6)=dcmplx(0.,0.)
       zint=zn(3)/2.
    endif
    zn(7)=zn(1)
    ind=0
    do j=1,6
       xl=cdabs(zn(j+1)-zn(j))
       nj=nx*xl/(2.*xlen+2.)
       if (nj.lt.4) nj=4
       if (2*(nj/2).lt.nj) nj=nj+1    
       ind=ind+1
       zd=(zn(j+1)-zn(j))/float(nj)
       zz=zn(j)
       z(ind)=zn(j)
       do k=1,nj-1
          zz=zz+zd
          ind=ind+1
          z(ind)=zz
       end do
    end do
    z(ind+1)=zint
    zn(1:4)= (/zi,0.,dcmplx(xlen,0.),zn(3)/)
!    zn(1:4)= (/zi,0.,dcmplx(xlen,0.),zn(3)+zi/)
  end subroutine get_coeffs
  !-------------------------------------------------------
  subroutine write_out(zout,ny,nx)
  integer                 :: ny,nx
  complex, dimension(:,:) :: zout
  call to_history("n1",ny)
  call to_history("n2",nx)
  call to_history("d1",1 )
  call to_history("d2",1 )
  call to_history("o1",0 )
  call to_history("o2",0 )
  call to_history("esize",8)
  call hclose()
  call sep_write(zout)
  end subroutine write_out

end module CMap

