program TPSM 
  !! Teleseismic Phase-shift migration
  !! Jeff Shragge February, 2003
  !! Teleseismic Phaseshift Migration
  use sep
  use gazdag

  implicit none
  integer                              :: n1,n2,esize,i1,i2,ii,pad2
  integer                              :: m1,m2,qmode
  real                                 :: dz,slow,dt
  real                                 :: eps,o1,o2,d1,d2
  real                                 :: acrust,bcrust,amantle,bmantle,moho,b1
  real,    dimension(:,:),allocatable  :: modl,data
  complex, dimension(:,:),allocatable  :: cmodl,cdata
  logical                              :: adj,add

  call sep_init()
  call from_history( "n1", n1         ) ! Num Time Samples
  call from_history( "n2", n2         ) ! Num Offsets
  call from_history( "o1", o1         ) ! Time offset
  call from_history( "o2", o2         ) ! Space offset
  call from_history( "d1", d1         ) ! Temporal sampling rate
  call from_history( "d2", d2         ) ! Spatial sampling rate
  call from_param( "esize",esize      ) ! Length of data
  call from_param( "acrust",acrust    ) ! Crustal P-wave vel
  call from_param( "bcrust",bcrust    ) ! Crustal S-wave vel
  call from_param( "amantle",amantle  ) ! Mantle P-wave vel
  call from_param( "bmantle",bmantle  ) ! Mantle S-wave vel
  call from_param( "b1",b1            ) ! Moho depth
  call from_param( "dz",dz            ) ! Depth step in model
  call from_param( "dt",dt            ) ! Data time sample step in model
  call from_param( "slow",slow        ) ! Slowness of incident plane wave
  call from_param( "eps",eps          ) !
  call from_param( "qmode",qmode      ) ! Scattering mode
  call from_param( "adj",adj          ) ! Adjoint call
write(0,*) 'First dt', dt 
 if (adj) then
     call to_history('d1',dz) !! Write the proper depth step in the output header file
  else
     call to_history('d1',dt) !! Write the proper depth step in the output header file
  end if
write(0,*) 'Second dt',dt
  call hclose() 
 
  !! Ensure we have real data
  if(esize.ne.4) call erexit('esize !=4')
  !! Compute zero-padding numbers
  m1=pad2(n1)
  m2=pad2(n2)
  allocate( data(n1,n2), modl(n1,n2)  )
  allocate( cdata(m1,m2),cmodl(m1,m2) )
  !! Read in Data or Model
  if (adj) then
     call sep_read( data )
     cdata(1:n1,1:n2)=cmplx(data,0.) 
  else
     call sep_read(modl)
     cmodl(1:n1,1:n2)=cmplx(modl,0.)
  end if
  !! Do phase-shift migration
  call gazdag1_init(m1,m2,dt,d2,dz,slow,qmode,acrust,amantle,bcrust,bmantle,b1,n1)
  call gazdag1(adj,.false.,cdata,cmodl)

  !! Write out data or model
  if (adj) then
     modl=real(cmodl)
     call sep_write(modl)
  else
     data=real(cdata)
     call sep_write(data)
  end if

end program TPSM

