module ert_mod
  use sep 
  use igrad1
  implicit none
  integer,private:: ng,nx,nz,ntau
  real   ,private:: dz,dx,ox,oz,eps,dtau
  logical,private:: verbose
contains

  !######################################################################

  subroutine phaseray_init(ng_in,nx_in,nz_in,dx_in,dz_in,ox_in,oz_in,ntau_in,&
  dtau_in,eps_in,verbose_in)
    integer:: ng_in,nx_in,nz_in,ntau_in
    real   :: dz_in,dx_in,ox_in,oz_in
    real   :: eps_in, dtau_in
    logical:: verbose_in
    nx=nx_in;  dx=dx_in;  ox=ox_in;  
    nz=nz_in;  dz=dz_in;  oz=oz_in;
    ng=ng_in;  verbose=verbose_in
    ntau=ntau_in; eps=eps_in; dtau=dtau_in
  end subroutine phaseray_init

  !######################################################################

  subroutine phaserays(data,rays,x,z)
    integer            :: ip,it,iix,iiz
    real               :: dxo,dzo,K
    real               :: x(:),z(:),data(:,:)
    real   ,allocatable:: Gdx(:,:),Gdz(:,:),xr(:,:),zr(:,:)
    complex            :: rays(:,:)

    allocate(Gdx(nz,nx),Gdz(nz,nx))
    allocate(xr(ntau,ng),zr(ntau,ng))

    call get_gradx(data,Gdx)
    call get_gradz(data,Gdz)

    xr(1,:)=x;  zr(1,:)=z
    do it=2, ntau
       if (verbose) write(0,*) 'Working on step ',it, ntau
       do ip=1,ng

          iix=floor( (xr(it-1,ip)-ox)/dx + 1.5 )
          iiz=floor( (zr(it-1,ip)-oz)/dz + 1.5 )

          if (iix .lt. 1) then
             iix=1
          elseif (iix .gt. nx) then
             iix=nx
          end if

          if (iiz .lt. 1) then
             iiz=1
          elseif (iiz .gt. nz) then
             iiz=nz
          end if

          dxo=Gdx(iiz,iix)
          dzo=Gdz(iiz,iix)
          K=sqrt(dxo**2+dzo**2)
          xr(it,ip)=xr(it-1,ip)+dtau*dxo/K
          zr(it,ip)=zr(it-1,ip)+dtau*dzo/K
       end do
       if(verbose) write(0,*) 'Indicies: ',iix,iiz,it
       if(verbose) write(0,*) 'Prev.sta: ',xr(it-1,ip-1),zr(it-1,ip-1)
       if(verbose) write(0,*) 'Add ons : ',dtau*dxo/K,dtau*dzo/K
    end do
    rays=cmplx(xr,zr) 

    call get_boundary(xr,zr)

    deallocate(Gdx,Gdz,xr,zr)
  end subroutine phaserays
  !####################################################################

  subroutine get_boundary(xr,zr)
    integer :: at,ig,iters,iis1,iis2,it
    integer,allocatable :: iis(:)
    real :: xr(:,:),zr(:,:)
    real :: ds,dM,dP,s1f,s2f,stodo
    real, allocatable:: xb(:),zb(:),s(:),x(:,:),z(:,:)
    complex, allocatable::boundary(:)
    complex, allocatable ::mesh(:,:)
    allocate( xb(ng),zb(ng),iis(1),s(ng),boundary(ntau*2+ng*2) )
    allocate( x(ntau,ng),z(ntau,ng),mesh(ntau,ng) )
    !! Integrate the bottom surface,.

    s(1)=0.; ds=0.
    do ig=2,ng
       s(ig)=s(ig-1)+sqrt((xr(ntau,ig)-xr(ntau,ig-1))**2+&
                          (zr(ntau,ig)-zr(ntau,ig-1))**2 )   
    end do
    ds=s(ng)/real(ng-1);
    xb(1 )=xr(ntau,1 );   zb(1 )=zr(ntau,1 )
    xb(ng)=xr(ntau,ng);   zb(ng)=zr(ntau,ng)

    do ig=2,ng-1
       stodo=real(ig-1)*ds
       iis=minloc(abs(s-stodo))
       iis1=iis(1)
       if (iis1 .eq. 1) then
          iis2=2
       else if (iis1 .eq. ng) then
          iis2=ng-1
       else
          if (s(iis1) .lt. stodo ) then 
             iis2=iis1+1
          else 
             iis2=iis1-1
          end if
       end if
       s1f = (s(iis1)-stodo)/(s(iis1)-s(iis2) )
       s2f = (stodo-s(iis2))/(s(iis1)-s(iis2) )
       xb(ig)=s2f*xr(ntau,iis1)+s1f*xr(ntau,iis2)
       zb(ig)=s2f*zr(ntau,iis1)+s1f*zr(ntau,iis2)
    end do

    boundary(1:ng) = cmplx( xr(1,1:ng) , zr(1,1:ng) )
    boundary(ng+1:ng+ntau) = cmplx( xr(1:ntau,ng) , zr(1:ntau,ng) )
    do ig=1,ng
       boundary(ng+ntau+ig) = cmplx( xb(ng-ig+1),zb(ng-ig+1) )
    end do

    do ig=1,ntau
       boundary(2*ng+ntau+ig) = cmplx( xr(ntau-ig+1,1),zr(ntau-ig+1,1) )
    end do

    !! Blend the Z coord
    z(:,1   )=zr(:,1   );  z(:,ng  )=zr(:,ng  )
    z(1,:   )=zr(1,:   );  z(ntau,:)=zb
    do ig=1,ng
       z(:,ig) = z(ntau,ig)*  (/(it,it=0,ntau-1)/)/real(ntau-1)    +&
                    z(1,ig)*(-(/(it,it=0,ntau-1)/)/real(ntau-1)+1)
    end do
    
    !! Blend the X-coord
    x(:,1   )=xr(:,1   )
    x(:,ng  )=xr(:,ng  )
    x(1,:   )=xr(1,:   )
    x(ntau,:)=xb
    do it=1,ntau
       x(it,:) = x(it,ng)*  (/(ig,ig=0,ng-1)/)/real(ng-1)    +&
                 x(it, 1)*(-(/(ig,ig=0,ng-1)/)/real(ng-1)+1)
    end do

    mesh=cmplx(x,z)

    call sep_write(boundary,"boundary.H")
    call sep_write(mesh,"mesh.H")

    deallocate(xb,zb)
  end subroutine get_boundary

  !####################################################################
  subroutine get_gradx(data,Gdx)
    integer            :: iz,stat
    real               :: Gdx(:,:),data(:,:)
    do iz=1,nz
       stat=igrad1_lop(.false.,.false.,data(iz,:),Gdx(iz,:) )
    end do
  end subroutine get_gradx
  !######################################################################
  subroutine get_gradz(data,Gdz)
    integer            :: ix,stat
    real               :: Gdz(:,:),data(:,:)
    do ix=1,nx
       stat=igrad1_lop(.false.,.false.,data(:,ix),Gdz(:,ix) )
    end do
  end subroutine get_gradz

  !######################################################################
end module ert_mod
