module exact_r4_mod
  use sep 
  use igrad1
  implicit none
  integer     ,private   :: nx,nz,nw
  real        ,private   :: mxint,eps,mnint
  real        ,private   :: dtau,dx,dz,dw,ox,oz,ow,xsp,R
  logical     ,private   :: fromSP
contains
  subroutine exact_mod_init(nz_in,nx_in,nw_in,dz_in,dx_in,dw_in,oz_in,ox_in,&
        ow_in,dtau_in,mxint_in,fromSP_in,xsp_in,R_in,eps_in,mnint_in)
    integer        :: nz_in,nx_in,nw_in
    real           :: mxint_in,R_in,xsp_in,eps_in,mnint_in
    real           :: dtau_in,dz_in,dx_in,dw_in,ox_in,oz_in,ow_in
    logical        :: fromSP_in
    nz=nz_in; nx=nx_in; nw=nw_in; dz=dz_in; dtau=dtau_in; dx=dx_in
    R=R_in;xsp=xsp_in;mnint=mnint_in
    oz=oz_in; ox=ox_in; ow=ow_in; dw=dw_in; mxint=mxint_in
    fromSP=fromSP_in;eps=eps_in
  end subroutine exact_mod_init
  !####################################################################
  subroutine get_gradx(adj,add,data,PoverP_dx)
    integer                                   :: iz,stat
    real           ,dimension(:)  ,allocatable:: ctemp,rtemp
    real(KIND=8)   ,dimension(:,:)            :: PoverP_dx
    complex(kind=8),dimension(:)  ,allocatable:: ctmp
    complex        ,dimension(:,:)            :: data
    logical                                   :: adj,add
    allocate(ctemp(nx),rtemp(nx),ctmp(nx) )
    do iz=1,nz
       stat=igrad1_lop(adj,add,real( data(iz,:)),rtemp )
       stat=igrad1_lop(adj,add,aimag(data(iz,:)),ctemp )
       ctmp=cmplx(rtemp,ctemp)
       PoverP_dx(iz,:)=aimag(ctmp/(data(iz,:)+eps))   
    end do
    deallocate(ctemp,rtemp,ctmp)
  end subroutine get_gradx
  !######################################################################
  subroutine get_gradz(adj,add,data,PoverP_dz)
    integer                                   :: ix,stat
    real           ,dimension(:)  ,allocatable:: ctemp,rtemp
    real(kind=8)   ,dimension(:,:)            :: PoverP_dz
    complex(kind=8),dimension(:)  ,allocatable:: ctmp
    complex        ,dimension(:,:)            :: data
    logical                                   :: adj,add
    allocate( ctemp(nz),rtemp(nz),ctmp(nz) )
    do ix=1,nx
       stat=igrad1_lop(adj,add,real( data(:,ix)),rtemp )
       stat=igrad1_lop(adj,add,aimag(data(:,ix)),ctemp )
       ctmp=cmplx(rtemp,ctemp)
       PoverP_dz(:,ix)=aimag(ctmp/(data(:,ix)+eps))
    end do
    deallocate(ctemp,rtemp,ctmp)
  end subroutine get_gradz
  !######################################################################
  subroutine Run_exact_interp(vel,data,Rayout,adj,add,maxnum,x,z,xtodo,ztodo,it)
    integer                                :: ip,maxnum
    integer                                :: itau,it,iix,iiz
    integer     ,dimension(:),allocatable  :: xtmp,ztmp
    real(KIND=8)                           :: dx_step,dz_step,ds
    real(KIND=8),dimension(:)              :: xtodo,ztodo,x,z,xt1,zt1
    real(KIND=8),dimension(:,:)            :: vel
    real(KIND=8),dimension(:,:),allocatable:: PoverP_dx,PoverP_dz
    real(KIND=8),dimension(:,:),allocatable:: xray,zray
    real(KIND=8),dimension(:  ),allocatable:: extra_x,extra_z
    complex,     dimension(:,:)            :: Rayout
    complex,     dimension(:,:)            :: data
    logical                                :: adj,add  
    allocate(PoverP_dx(nz,nx),PoverP_dz(nz,nx))
    allocate( xtmp(1), ztmp(1) )
    allocate(xray( maxnum,nx),zray(maxnum,nx))
    xray(1,1:nx)=xtodo;  zray(1,1:nx)=ztodo
    call get_gradx(adj,add,data,PoverP_dx)
    call get_gradz(adj,add,data,PoverP_dz)

    do itau=2, maxnum
       write(0,*) 'Working on step ',itau, maxnum
       do ip=1, nx
          xtmp=minloc(abs(x-xray(itau-1,ip)));  iix=xtmp(1)
          ztmp=minloc(abs(z-zray(itau-1,ip)));  iiz=ztmp(1)
          ds=vel(iiz,iix)*dble(dtau)

          if (iix .ge. nx-2*ceiling(abs(ds/dx))) then
             iix= nx-2*ceiling(abs(ds/dx))
          else if (iix .le. 2*ceiling(abs(ds/dx)) ) then
             iix= 2*ceiling(abs(ds/dx))
          else if (iiz .ge. nz-2*ceiling(abs(ds/dz)) )  then
             iiz= nz-2*ceiling(abs(ds/dz))
          end if

          !! Add onto existing rays
          xray(itau,ip)=xray(itau-1,ip)-ds*PoverP_dx(iiz,iix)/&
          sqrt(PoverP_dx(iiz,iix)**2+PoverP_dz(iiz,iix)**2+eps)
          zray(itau,ip)=zray(itau-1,ip)-ds*PoverP_dz(iiz,iix)/&
          sqrt(PoverP_dx(iiz,iix)**2+PoverP_dz(iiz,iix)**2+eps)
       end do
    end do
    it=nx
    Rayout(1:maxnum,1:nx)=cmplx(xray,zray) 
    deallocate(PoverP_dx,PoverP_dz,xray,zray)
  end subroutine Run_exact_interp
 !######################################################################
  subroutine get_min_loc(x,xray,iix,itau,ix)
    integer                                     :: iix,ix,itau
    real(KIND=8),    dimension(:,:)             :: xray
    real(KIND=8),    dimension(:)               :: x
    real(KIND=8),    dimension(:),allocatable   :: tmp
    allocate( tmp(size(x)) )
    tmp = abs(x-xray(itau,ix))
    iix = minloc(tmp,1)
    deallocate(tmp)
  end subroutine get_min_loc
 !######################################################################
  subroutine write_on_baby(Rayout,maxnum,it,endang,begang)
    complex,  dimension(:,:)   :: Rayout
    integer                    :: it,maxnum
    real                       :: endang,begang
  call to_history("n1",maxnum          ) 
  call to_history("n2",nx              )  
  call to_history("d1",dtau            ) 
  call to_history("d2",dx              )  
  call to_history("o1",0.              ) 
  call to_history("o2",ox              )  
  call hclose()
  call sep_write(Rayout(1:maxnum,1:nx))
  end subroutine write_on_baby
 !######################################################################
end module exact_r4_mod
