module exact_r5_mod
  use sep 
  use igrad1
  implicit none
  integer     ,private   :: nx,nz,nw,nray
  real        ,private   :: mxint,eps,mnint,begang,endang
  real        ,private   :: dtau,dx,dz,dw,ox,oz,ow,xsp,R
  logical     ,private   :: fromSP
contains
  subroutine exact_mod_init(nz_in,nx_in,nw_in,dz_in,dx_in,dw_in,oz_in,ox_in,ow_in,dtau_in,mxint_in,&
               fromSP_in,xsp_in,R_in,eps_in,mnint_in,nray_in,begang_in,endang_in)
    integer        :: nz_in,nx_in,nw_in,nray_in
    real           :: mxint_in,R_in,xsp_in,eps_in,mnint_in,begang_in,endang_in
    real           :: dtau_in,dz_in,dx_in,dw_in,ox_in,oz_in,ow_in
    logical        :: fromSP_in
    nz=nz_in; nx=nx_in; nw=nw_in; dz=dz_in; dtau=dtau_in; dx=dx_in; R=R_in;xsp=xsp_in;mnint=mnint_in
    oz=oz_in; ox=ox_in; ow=ow_in; dw=dw_in; mxint=mxint_in; fromSP=fromSP_in;eps=eps_in;nray=nray_in
    begang=begang_in;endang=endang_in
  end subroutine exact_mod_init
  !####################################################################
  subroutine get_gradx(adj,add,data,PoverP_dx)
    integer                                   :: iw,iz,stat
    real           ,dimension(:)  ,allocatable:: ctemp,rtemp
    real(KIND=8)   ,dimension(:,:)            :: PoverP_dx
    complex(kind=8),dimension(:)  ,allocatable:: ctmp
    complex        ,dimension(:,:)            :: data
    logical                                   :: adj,add
    allocate(ctemp(nx),rtemp(nx),ctmp(nx) )
    do iz=1,nz
       stat=igrad1_lop(adj,add,real( data(iz,:)),rtemp )
       stat=igrad1_lop(adj,add,aimag(data(iz,:)),ctemp )
       ctmp=cmplx(rtemp,ctemp)
       PoverP_dx(iz,:)=aimag(ctmp/(data(iz,:)+eps))   
    end do
    deallocate(ctemp,rtemp,ctmp)
  end subroutine get_gradx
  !######################################################################
  subroutine get_gradz(adj,add,data,PoverP_dz)
    integer                                   :: iw,ix,stat
    real           ,dimension(:)  ,allocatable:: ctemp,rtemp
    real(kind=8)   ,dimension(:,:)            :: PoverP_dz
    complex(kind=8),dimension(:)  ,allocatable:: ctmp
    complex        ,dimension(:,:)            :: data
    logical                                   :: adj,add
    allocate( ctemp(nz),rtemp(nz),ctmp(nz) )
    do ix=1,nx
       stat=igrad1_lop(adj,add,real( data(:,ix)),rtemp )
       stat=igrad1_lop(adj,add,aimag(data(:,ix)),ctemp )
       ctmp=cmplx(rtemp,ctemp)
       PoverP_dz(:,ix)=aimag(ctmp/(data(:,ix)+eps))
    end do
    deallocate(ctemp,rtemp,ctmp)
  end subroutine get_gradz
  !######################################################################
  subroutine Run_exact_interp(vel,presol,data,Rayout,adj,add,maxnum,x,z,xtodo,ztodo,it)
    integer                                :: ix,iz,iw,ip,maxnum,iix,iiz,cnt,i2,it,i1
    integer                                :: AllocateStatus,ig
    real        ,dimension(:)              :: dang
    real(kind=8)                           :: dx_step,dz_step,mxdist,ds,pi,mnang
    real(KIND=8),dimension(:)              :: xtodo,ztodo,x,z
    real(kind=8),dimension(:,:)            :: vel
    real(kind=8),dimension(:,:),allocatable:: PoverP_dx,PoverP_dz
    real(KIND=8),dimension(:,:),allocatable:: xray,zray,bigx,bigz
    real(kind=8),dimension(:)  ,allocatable:: ang
    complex,     dimension(:,:)            :: Rayout
    complex,     dimension(:,:)            :: data,presol
    logical                                :: adj,add  
    allocate( PoverP_dx(nz,nx),PoverP_dz(nz,nx))
    allocate( xray(maxnum,size(xtodo)) , zray(maxnum,size(xtodo)) )
    xray(1,1:size(xtodo))=xtodo;    zray(1,1:size(xtodo))=ztodo
    pi=2.*asin(1.)
    mxdist=mxint*min(dx,dz)  
    mnang=mnint;  
    call get_gradx(adj,add,data,PoverP_dx)
    call get_gradz(adj,add,data,PoverP_dz)

    loop_steps:  do cnt=2,maxnum ! . . New Step  

       write(0,*) 'Working on step ',cnt,' of ', maxnum ,it
       loop_points: do ip=1,size(xray,2)
          call get_min_loc(x,xray,iix,cnt-1,ip,0.)
          call get_min_loc(z,zray,iiz,cnt-1,ip,0.)
          ds=-(vel(iiz,iix))*dble(dtau)
          if ( iix .ge. nx-2*ceiling(abs(ds/dx))  &
          .or. iix .le. 2*   ceiling(abs(ds/dx))  &
          .or. iiz .ge. nz-2*ceiling(abs(ds/dz))  &
          .or. iiz .le. 2*   ceiling(abs(ds/dz))  ) then
             xray(cnt:,ip)=xray(cnt-1,ip)
             zray(cnt:,ip)=zray(cnt-1,ip)
             continue
          else  
             dx_step=ds*PoverP_dx(iiz,iix)/sqrt(PoverP_dx(iiz,iix)**2+PoverP_dz(iiz,iix)**2+eps)
             dz_step=ds*PoverP_dz(iiz,iix)/sqrt(PoverP_dx(iiz,iix)**2+PoverP_dz(iiz,iix)**2+eps)
             xray(cnt,ip)=xray(cnt-1,ip)+dx_step
             zray(cnt,ip)=zray(cnt-1,ip)+dz_step
          end if
       end do loop_points

       !! . . Check to see distance between points
       allocate( bigx(maxnum,2*size(xray,2)), bigz(maxnum,2*size(zray,2)) )
       it=1
       bigx(1:cnt,1)=xray(1:cnt,1); bigz(1:cnt,1)=zray(1:cnt,1)
       do i1=2,size(xray,2)
          it=it+1
          if     ( sqrt( (xray(cnt,i1)-xray(cnt,i1-1))**2 + (zray(cnt,i1)-zray(cnt,i1-1))**2).gt.mxdist) then
             bigx(1:cnt,it)=0.5*(xray(1:cnt,i1)+xray(1:cnt,i1-1))
             bigz(1:cnt,it)=0.5*(zray(1:cnt,i1)+zray(1:cnt,i1-1))
             bigx(1:cnt,it+1)=xray(1:cnt,i1)
             bigz(1:cnt,it+1)=zray(1:cnt,i1)
             it=it+1
          elseif ( (sqrt( (xray(1,i1)-xray(1,i1-1))**2 + (zray(1,i1)-zray(1,i1-1))**2))/(2*pi*R) .lt.mnang) then
             bigx(1:cnt,it-1)=xray(1:cnt,i1)
             bigz(1:cnt,it-1)=zray(1:cnt,i1)
             it=it-1
          else
             bigx(1:cnt,it)=xray(1:cnt,i1)
             bigz(1:cnt,it)=zray(1:cnt,i1) 
          end if
       end do
       if (it .ne. size(xray,2)) then
          deallocate(xray,zray,stat=AllocateStatus)
          allocate( xray(maxnum,it), zray(maxnum,it))
          xray(1:cnt,1:it)=bigx(1:cnt,1:it)
          zray(1:cnt,1:it)=bigz(1:cnt,1:it)
          deallocate(bigx,bigz)
       else
          xray(1:cnt,1:it)=bigx(1:cnt,1:it)
          zray(1:cnt,1:it)=bigz(1:cnt,1:it)
          deallocate(bigx,bigz)
       end if


    end do loop_steps

    Rayout(1:maxnum,1:it)=cmplx(xray,zray) 

    allocate( ang(it),dang(it) )
    ang=atan( (xray(1,:)-xsp) / zray(1,:)) *180./pi
    do ig=1,it-1
       dang(ig)=-(ang(ig+1)-ang(ig))
    end do
    dang(it)=-(ang(it)-ang(it-1))
    call to_history("esize",4 ,"dang.H")
    call to_history("d1",   1.,"dang.H")
    call to_history("o1",   0.,"dang.H")
    call to_history("n1",   it,"dang.H")
    call sep_write(dang,"dang.H")
    deallocate( PoverP_dx,PoverP_dz)
    deallocate( xray,zray)
  end subroutine Run_exact_interp
 !######################################################################
  subroutine get_min_loc(x,xray,iix,cnt,ip,dx_step)
    integer                                     :: iix,ip,cnt
    integer, dimension(:),allocatable           :: xtmp
    real(KIND=8),    dimension(:,:)                     :: xray
    real(KIND=8),    dimension(:)                       :: x
    real                                        :: dx_step
   allocate( xtmp(size(x)))
    xtmp=minloc(abs(x-(xray(cnt,ip)+dx_step))); 
    iix=xtmp(1)  
    deallocate(xtmp)
  end subroutine get_min_loc
 !######################################################################
  subroutine write_on_baby(Rayout,maxnum,it,endang,begang)
    complex,  dimension(:,:)   :: Rayout
    integer                    :: it,maxnum
    real                       :: endang,begang
  call to_history("n1",maxnum                   ) 
  call to_history("n2",it                       )  
  call to_history("d1",dtau                     ) 
  call to_history("d2",(endang-begang)/(it+1)   )  
  call to_history("o1",0.                       ) 
  call to_history("o2",begang                   )  
  call hclose()
  call sep_write(Rayout(1:maxnum,1:it))
  end subroutine write_on_baby
 !######################################################################
  subroutine get_init_vect(x,z,ntodo)
    integer                                :: i2,iray
    integer,      dimension(:)             :: ntodo
    integer,      dimension(:),allocatable :: ixx,izz
    real(KIND=8), dimension(:)             :: x,z 
    allocate(ixx(nx),izz(nz) )
    ixx(1:nx) = (/ (i2,i2=1,nx) /) 
    izz(1:nz) = (/ (i2,i2=1,nz) /) 
    x=(ixx-1)*dx+ox  
    z=(izz-1)*dz+oz
    ntodo = (/ (iray, iray=0,nray-1) /)  
    deallocate(ixx,izz)
  end subroutine get_init_vect
 !######################################################################
  subroutine get_start_pos(wantP,xang,zang,x,z,ntodo)
    integer                           :: wantP
    integer,      dimension(:)        :: ntodo
    real(kind=8)                      :: pi
    real(kind=8), dimension(:)        :: xang,zang,x,z
    if (wantP .eq. 1 ) then
       xang= x
       zang= dz
       write(0,*) ' USING PLANE WAVES '
    else  
       pi=2.*asin(1.)
       xang=R*cos( (endang-begang)*pi/(180*(nray-1))*ntodo+begang*pi/180.0+pi/2.0) + xsp
       zang=R*sin( (endang-begang)*pi/(180*(nray-1))*ntodo+begang*pi/180.0+pi/2.0) 
    end if
  end subroutine get_start_pos
 !######################################################################
end module exact_r5_mod
