module exact_rest_mod
  use sep 
  use igrad1
  implicit none
  integer     ,private   :: nx,nz,nw
  real        ,private   :: mxint,eps,mnint
  real        ,private   :: dtau,dx,dz,dw,ox,oz,ow,xsp,R
  logical     ,private   :: fromSP
contains
  subroutine exact_mod_init(nz_in,nx_in,nw_in,dz_in,dx_in,dw_in,oz_in,ox_in,ow_in,dtau_in,mxint_in,fromSP_in,xsp_in,R_in,eps_in,mnint_in)
    integer        :: nz_in,nx_in,nw_in
    real           :: mxint_in,R_in,xsp_in,eps_in,mnint_in
    real           :: dtau_in,dz_in,dx_in,dw_in,ox_in,oz_in,ow_in
    logical        :: fromSP_in
    nz=nz_in; nx=nx_in; nw=nw_in; dz=dz_in; dtau=dtau_in; dx=dx_in; R=R_in;xsp=xsp_in;mnint=mnint_in
    oz=oz_in; ox=ox_in; ow=ow_in; dw=dw_in; mxint=mxint_in; fromSP=fromSP_in;eps=eps_in
  end subroutine exact_mod_init
  !####################################################################
  subroutine get_gradx(adj,add,data,PoverP_dx)
    integer                                   :: iw,iz,stat
    real           ,dimension(:)  ,allocatable:: ctemp,rtemp
    real(KIND=8)   ,dimension(:,:)            :: PoverP_dx
    complex(kind=8),dimension(:)  ,allocatable:: ctmp
    complex        ,dimension(:,:)            :: data
    logical                                   :: adj,add
    allocate(ctemp(nx),rtemp(nx),ctmp(nx) )
    do iz=1,nz
       stat=igrad1_lop(adj,add,real( data(iz,:)),rtemp )
       stat=igrad1_lop(adj,add,aimag(data(iz,:)),ctemp )
       ctmp=cmplx(rtemp,ctemp)
       PoverP_dx(iz,:)=aimag(ctmp/(data(iz,:)+eps))   
    end do
    deallocate(ctemp,rtemp,ctmp)
  end subroutine get_gradx
  !######################################################################
  subroutine get_gradz(adj,add,data,PoverP_dz)
    integer                                   :: iw,ix,stat
    real           ,dimension(:)  ,allocatable:: ctemp,rtemp
    real(kind=8)   ,dimension(:,:)            :: PoverP_dz
    complex(kind=8),dimension(:)  ,allocatable:: ctmp
    complex        ,dimension(:,:)            :: data
    logical                                   :: adj,add
    allocate( ctemp(nz),rtemp(nz),ctmp(nz) )
    do ix=1,nx
       stat=igrad1_lop(adj,add,real( data(:,ix)),rtemp )
       stat=igrad1_lop(adj,add,aimag(data(:,ix)),ctemp )
       ctmp=cmplx(rtemp,ctemp)
       PoverP_dz(:,ix)=aimag(ctmp/(data(:,ix)+eps))
    end do
    deallocate(ctemp,rtemp,ctmp)
  end subroutine get_gradz
  !######################################################################
  subroutine Run_exact_interp(vel,data,Rayout,adj,add,maxnum,x,z,xtodo,ztodo,it)
    integer                                :: ix,iz,iw,ip,maxnum,iix,iiz,cnt,i2,it,i1
    real        ,dimension(:)              :: dang
    real(kind=8)                           :: dx_step,dz_step,mxdist,ds,pi
    real(KIND=8),dimension(:)              :: xtodo,ztodo,x,z
    real(kind=8),dimension(:,:)            :: vel
    real(kind=8),dimension(:,:),allocatable:: PoverP_dx,PoverP_dz
    real(KIND=8),dimension(:,:),allocatable:: xray,zray,bigx,bigz
    real(kind=8),dimension(:)  ,allocatable:: ang
    complex,     dimension(:,:)            :: Rayout
    complex,     dimension(:,:)            :: data
    logical                                :: adj,add  
    allocate( PoverP_dx(nz,nx),PoverP_dz(nz,nx))
    allocate( xray(maxnum,size(xtodo)) , zray(maxnum,size(xtodo)) )
    xray(1,1:size(xtodo))=xtodo;    zray(1,1:size(xtodo))=ztodo
    pi=2*asin(1.) 
    mxdist=mxint*min(dx,dz)  

    call get_gradx(adj,add,data,PoverP_dx)
    call get_gradz(adj,add,data,PoverP_dz)

    loop_steps:  do cnt=2,maxnum ! . . New Step  
       write(0,*) 'Working on step ',cnt,' of ', maxnum ,it
       loop_points: do ip=1,size(xray,2)
          call get_min_loc(x,xray,iix,cnt-1,ip)
          call get_min_loc(z,zray,iiz,cnt-1,ip)
          ds=-vel(iiz,iix)*dble(dtau)
          if ( iix .ge. nx-2*ceiling(abs(ds/dx))  .or. iix .le. 2*   ceiling(abs(ds/dx))  &
          .or. iiz .ge. nz-2*ceiling(abs(ds/dz))  .or. iiz .le. 2*   ceiling(abs(ds/dz))  ) then
             xray(cnt:,ip)=xray(cnt-1,ip)
             zray(cnt:,ip)=zray(cnt-1,ip)
             continue
          else  
             xray(cnt,ip)=xray(cnt-1,ip)+&
             ds*PoverP_dx(iiz,iix)/sqrt(PoverP_dx(iiz,iix)**2+PoverP_dz(iiz,iix)**2+eps)
             zray(cnt,ip)=zray(cnt-1,ip)+&
             ds*PoverP_dz(iiz,iix)/sqrt(PoverP_dx(iiz,iix)**2+PoverP_dz(iiz,iix)**2+eps)
          end if
       end do loop_points
       !! . . Check to see distance between points
       allocate( bigx(maxnum,2*size(xray,2)), bigz(maxnum,2*size(zray,2)) )
       it=1
       bigx(1:cnt,1)=xray(1:cnt,1); bigz(1:cnt,1)=zray(1:cnt,1)
       do i1=2,size(xray,2)
          it=it+1
          if     ( sqrt( (xray(cnt,i1)-xray(cnt,i1-1))**2 + (zray(cnt,i1)-zray(cnt,i1-1))**2).gt.mxdist) then
             bigx(1:cnt,it)=(xray(1:cnt,i1)+xray(1:cnt,i1-1))/2.
             bigz(1:cnt,it)=(zray(1:cnt,i1)+zray(1:cnt,i1-1))/2.
             bigx(1:cnt,it+1)=xray(1:cnt,i1)
             bigz(1:cnt,it+1)=zray(1:cnt,i1)
             it=it+1
          elseif ( sqrt( (xray(1,i1)-xray(1,i1-1))**2 + (zray(1,i1)-zray(1,i1-1))**2)/(2.*pi*R) .lt. mnint &
           .or. sqrt( (xray(cnt,i1)-xray(cnt,i1-1))**2 + (zray(cnt,i1)-zray(cnt,i1-1))**2)  .lt. 1./mxdist ) then
             it=it-1
          else
             bigx(1:cnt,it)=xray(1:cnt,i1)
             bigz(1:cnt,it)=zray(1:cnt,i1) 
          end if
       end do

       if (it .ne. size(xray,2)) then
          deallocate(xray,zray)
          allocate( xray(maxnum,it), zray(maxnum,it))
          xray(1:cnt,1:it)=bigx(1:cnt,1:it)
          zray(1:cnt,1:it)=bigz(1:cnt,1:it)
          deallocate(bigx,bigz)
       else
          xray(1:cnt,1:it)=bigx(1:cnt,1:it)
          zray(1:cnt,1:it)=bigz(1:cnt,1:it)
          deallocate(bigx,bigz)
       end if

    end do loop_steps

    Rayout(1:maxnum,1:it)=cmplx(xray,zray) 
    allocate( ang(it),dang(it) )
    ang=atan( (xray(1,:)-xsp) / zray(1,:)) *180./pi
    dang(1:it-1)=ang(1:it-1)-ang(2:it)
    dang(it)=-(ang(it)-ang(it-1))
    call to_history("esize",4 ,"dang.H")
    call to_history("d1",   1.,"dang.H")
    call to_history("o1",   0.,"dang.H")
    call to_history("n1",   it,"dang.H")
    call sep_write(dang,"dang.H")
    deallocate( PoverP_dx,PoverP_dz)
    deallocate( xray,zray)

  end subroutine Run_exact_interp
 !######################################################################
  subroutine get_min_loc(x,xray,iix,cnt,ip)
    integer                                 :: iix,ip,cnt
    integer,     dimension(:)  ,allocatable :: xtmp
    real(KIND=8),dimension(:,:)             :: xray
    real(KIND=8),dimension(:)               :: x
   allocate( xtmp(size(x)))
    xtmp=minloc(abs(x-(xray(cnt,ip)))); 
    iix=xtmp(1)  
    deallocate(xtmp)
  end subroutine get_min_loc
 !######################################################################
  subroutine write_on_baby(Rayout,maxnum,it,endang,begang)
    complex,  dimension(:,:)   :: Rayout
    integer                    :: it,maxnum
    real                       :: endang,begang
  call to_history("n1",maxnum                   ) 
  call to_history("n2",it                       )  
  call to_history("d1",dtau                     ) 
  call to_history("d2",(endang-begang)/(it-1)   )  
  call to_history("o1",0.                       ) 
  call to_history("o2",begang                   )  
  call hclose()
  call sep_write(Rayout(1:maxnum,1:it))
  end subroutine write_on_baby
 !######################################################################
end module exact_rest_mod
