module exact_starting_mod
  use sep 
  use igrad1
  implicit none
  integer     ,private   :: nx,nz,nw
  real        ,private   :: mxint,eps,mnint
  real        ,private   :: dtau,dx,dz,dw,ox,oz,ow,xsp,R
  logical     ,private   :: fromSP
contains
  subroutine exact_mod_init(nz_in,nx_in,nw_in,dz_in,dx_in,dw_in,oz_in,ox_in,ow_in,dtau_in,mxint_in,fromSP_in,xsp_in,R_in,eps_in,mnint_in)
    integer        :: nz_in,nx_in,nw_in
    real           :: mxint_in,R_in,xsp_in,eps_in,mnint_in
    real           :: dtau_in,dz_in,dx_in,dw_in,ox_in,oz_in,ow_in
    logical        :: fromSP_in
    nz=nz_in; nx=nx_in; nw=nw_in; dz=dz_in; dtau=dtau_in; dx=dx_in; R=R_in;xsp=xsp_in;mnint=mnint_in
    oz=oz_in; ox=ox_in; ow=ow_in; dw=dw_in; mxint=mxint_in; fromSP=fromSP_in;eps=eps_in
  end subroutine exact_mod_init
  !####################################################################
  subroutine get_gradx(adj,add,data,PoverP_dx)
    integer                                   :: iw,iz,stat
    real           ,dimension(:)  ,allocatable:: ctemp,rtemp
    real(KIND=8)   ,dimension(:,:)            :: PoverP_dx
    complex(kind=8),dimension(:)  ,allocatable:: ctmp
    complex        ,dimension(:,:)            :: data
    logical                                   :: adj,add
    allocate(ctemp(nx),rtemp(nx),ctmp(nx) )
    do iz=1,nz
       stat=igrad1_lop(adj,add,real( data(iz,:)),rtemp )
       stat=igrad1_lop(adj,add,aimag(data(iz,:)),ctemp )
       ctmp=cmplx(rtemp,ctemp)
       PoverP_dx(iz,:)=aimag(ctmp/(data(iz,:)+eps))   
    end do
    deallocate(ctemp,rtemp,ctmp)
  end subroutine get_gradx
  !######################################################################
  subroutine get_gradz(adj,add,data,PoverP_dz)
    integer                                   :: iw,ix,stat
    real           ,dimension(:)  ,allocatable:: ctemp,rtemp
    real(kind=8)   ,dimension(:,:)            :: PoverP_dz
    complex(kind=8),dimension(:)  ,allocatable:: ctmp
    complex        ,dimension(:,:)            :: data
    logical                                   :: adj,add
    allocate( ctemp(nz),rtemp(nz),ctmp(nz) )
    do ix=1,nx
       stat=igrad1_lop(adj,add,real( data(:,ix)),rtemp )
       stat=igrad1_lop(adj,add,aimag(data(:,ix)),ctemp )
       ctmp=cmplx(rtemp,ctemp)
       PoverP_dz(:,ix)=aimag(ctmp/(data(:,ix)+eps))
    end do
    deallocate(ctemp,rtemp,ctmp)
  end subroutine get_gradz
  !######################################################################
  subroutine Run_exact_nointerp(vel,data,Rayout,adj,add,maxnum,x,z,xtodo,ztodo,ang)
    integer                                :: ix,iz,iw,ip,maxnum,iix,iiz,cnt,i2,it,i1
    integer                                :: ig
    real(kind=8),dimension(:)              :: ang
    real(kind=8)                           :: dx_step,dz_step,ds,pi
    real(KIND=8),dimension(:)              :: xtodo,ztodo,x,z
    real(kind=8),dimension(:,:)            :: vel
    real(KIND=8),dimension(:,:),allocatable:: xray,zray
    complex,     dimension(:,:)            :: Rayout
    complex,     dimension(:,:)            :: data
    logical                                :: adj,add  
    allocate( xray(maxnum,size(xtodo)) , zray(maxnum,size(xtodo)) )
    xray(1,1:size(xtodo))=xtodo;    zray(1,1:size(xtodo))=ztodo
    pi=2*asin(1.) 

    loop_steps:  do cnt=2,maxnum ! . . New Step  
       write(0,*) 'Working on step ',cnt,' of ', maxnum, size(xtodo)
       loop_points: do ip=1,size(xray,2)
          call get_min_loc(x,xray,iix,cnt-1,ip)
          call get_min_loc(z,zray,iiz,cnt-1,ip)
          ds=vel(iiz,iix)*dble(dtau)
          if ( iix .ge. nx-2*ceiling(abs(ds/dx))  &
          .or. iix .le. 2*   ceiling(abs(ds/dx))  &
          .or. iiz .ge. nz-2*ceiling(abs(ds/dz))  &
          .or. iiz .le. 2*   ceiling(abs(ds/dz))  ) then
             xray(cnt:,ip)=xray(cnt-1,ip)
             zray(cnt:,ip)=zray(cnt-1,ip)
             continue
          else  
             xray(cnt,ip)=xray(cnt-1,ip)+ds*sin( ang(ip) )
             zray(cnt,ip)=zray(cnt-1,ip)+ds*cos( ang(ip) )
          end if
       end do loop_points
     end do loop_steps
    Rayout=cmplx(xray,zray) 
    deallocate( xray,zray)
  end subroutine Run_exact_nointerp
 !######################################################################
  subroutine get_min_loc(x,xray,iix,cnt,ip)
    integer                                     :: iix,ip,cnt
    integer, dimension(:),allocatable           :: xtmp
    real(KIND=8),    dimension(:,:)                     :: xray
    real(KIND=8),    dimension(:)                       :: x
   allocate( xtmp(size(x)))
    xtmp=minloc(abs(x-(xray(cnt,ip)))); 
    iix=xtmp(1)  
    deallocate(xtmp)
  end subroutine get_min_loc
 !######################################################################
  subroutine write_on_baby(Rayout,maxnum,endang,begang,nray)
    complex,  dimension(:,:)   :: Rayout
    integer                    :: maxnum,nray
    real                       :: endang,begang
  call to_history("n1",maxnum                   ) 
  call to_history("n2",nray                       )  
  call to_history("d1",dtau                     ) 
  call to_history("d2",(endang-begang)/(nray-1)   )  
  call to_history("o1",0.                       ) 
  call to_history("o2",begang                   )  
  call hclose()
  call sep_write(Rayout)
  end subroutine write_on_baby
 !######################################################################
end module exact_starting_mod
