module fddtmrc_
  use fft
  use util_
  use tridiag

  implicit none 
  real, parameter, private :: sixth = 0.122996

  integer, private :: nz,nx,nw,nk
  real,    private :: dz,dx,dw,dk
  real,    private :: oz,ox,ow,ok
  real,    private :: pi

  complex, pointer, private :: av(:),bv(:),cv(:)
  complex, pointer, private :: b2a(:,:),s2a(:,:),d2a(:,:),c2a(:,:)
  complex, pointer, private :: kda(:,:),kca(:,:)
  complex, pointer, private :: nu(:,:),mu(:,:)
  complex, pointer, private :: ko(:,:)
contains
  !----------------------------------------------------------------
  subroutine fddtmrc_init(aw,ax,az)
    type(myaxis), intent(in) :: aw,ax,az

    pi = acos(-1.)

    nw=aw%n
    dw=aw%d * 2*pi
    ow=aw%o * 2*pi

    nx=ax%n
    dx=ax%d
    ox=ax%o

    nz=az%n
    dz=az%d
    oz=az%o

    nk = nx
    ok =   -pi/dx
    dk = 2.*pi/dx/nx

    allocate(av(ax%n),bv(ax%n),cv(ax%n))
    allocate(ko(az%n,ax%n),b2a(az%n,ax%n),s2a(az%n,ax%n),d2a(az%n,ax%n),c2a(az%n,ax%n))
    allocate(kda(az%n,ax%n),kca(az%n,ax%n))
    allocate(nu(az%n,ax%n),mu(az%n,ax%n))
  end subroutine fddtmrc_init
  !----------------------------------------------------------------
  real function getw(iw)
    integer, intent(in) :: iw
    real :: w
    w=ow + dw*(iw-1)
    getw=w
  end function getw

  real function getk(ik)
    integer, intent(in) :: ik
    real :: k
    k=ok + dk*(ik-1)
    getk=k
  end function getk
  !----------------------------------------------------------------
  subroutine dtmrc15(dat,wfl,rcp)
    complex, pointer     :: wfl(:,:)
    real,    pointer     :: rcp(:,:,:)
    complex, pointer     :: dat(:,:)
    real                 :: w
    integer              :: iz,ix,iw
    complex, pointer     :: frq(:)
    complex, dimension(:),allocatable :: km,kn
    complex              :: el,er
    complex, dimension(size(dat,1)) :: cd

    complex, dimension(:,:),allocatable :: a,b,c,d,s
    allocate( km(nx),kn(nx))
    allocate( a(nz,nx),b(nz,nx),c(nz,nx),d(nz,nx),s(nz,nx))

    call ctris_init(nx) 
    km = (0.,1.)*(-dz)/(4*dx)
    kn = (0.,1.)*(-dz)/(2*dx**2)

    a = cmplx(rcp(:,:,1),0.)
    b = cmplx(rcp(:,:,2),0.)
    c = cmplx(rcp(:,:,3),0.)
    d = cmplx(rcp(:,:,4),0.)
    s = cmplx(rcp(:,:,5),0.)

    b2a=b/(2*a)
    d2a=d/(2*a)
    !    c2a=c/(2*a)
    c2a=0.*c/(2*a)
    s2a=(2*s)**2/a

    do iw=1,nw

       w=getw(iw)
       frq => dat(:,iw)
       write(0,*) "migrc15",iw,nw,minval(cabs(wfl)),maxval(cabs(wfl))

       ko = sqrt( s2a*w**2 - b2a**2)

       do iz=1,nz
          mu(iz,:) = km/ko(iz,:) *   c2a(iz,:)
          nu(iz,:) =-kn/ko(iz,:) * ((c2a(iz,:)/ko(iz,:))**2 - 2.*d2a(iz,:))

          av =   -  (nu(iz,:) - mu(iz,:)) +    sixth 
          bv = 1.+2.*nu(iz,:)             - 2.*sixth
          cv =   -  (nu(iz,:) + mu(iz,:)) +    sixth

          cd( 1) = 0.
          cd(2:nx-1)=(      (nu(iz,2:nx-1) - mu(iz,2:nx-1))+sixth)*frq(1:nx-2) &
          +      (1.-2.*nu(iz,2:nx-1)             -2.*sixth)*frq(2:nx-1) &
          +      (      (nu(iz,2:nx-1) + mu(iz,2:nx-1))+sixth)*frq(3:nx)
          cd(nx) = 0.

          el = -(av( 1)+bv( 1))/cv( 1)
          er = -(cv(nx)+bv(nx))/av(nx)

          call tris(el, av, bv, cv, er, cd, frq)
          frq = frq*cexp( cmplx(-b2a(iz,:),ko(iz,:))*(-dz))

       end do
       wfl(:,iw) = frq
    end do


    call ctris_close()
    deallocate(km,kn,a,b,d,c,s)

  end subroutine dtmrc15
  !----------------------------------------------------------------
end module fddtmrc_
