module fdmigrc_NL_
  use fft
  use util_
  use tridiag

  implicit none 
  real, parameter, private :: sixth = 0.122996

  integer, private :: nz,nx,nw,nk
  real,    private :: dz,dx,dw,dk
  real,    private :: oz,ox,ow,ok
  real,    private :: pi

  complex, pointer, private :: av(:),bv(:),cv(:)
  real   , pointer, private :: b2a(:,:),s2a(:,:),d2a(:,:),c2a(:,:)
  complex, pointer, private :: nu(:,:),mu(:,:)
  real   , pointer, private :: ko(:,:)
contains
  !----------------------------------------------------------------
  subroutine fdmigrc_init(aw,ax,az)
    type(myaxis), intent(in) :: aw,ax,az
    pi = acos(-1.)
    nw=aw%n;  dw=aw%d*2*pi;  ow=aw%o*2*pi;
    nx=ax%n;  dx=ax%d    ;   ox=ax%o
    nz=az%n;  dz=az%d    ;   oz=az%o
    nk = nx;  ok = -pi/dx;   dk = 2.*pi/dx/nx
    allocate(av(ax%n),bv(ax%n),cv(ax%n))
    allocate(ko(az%n,ax%n),b2a(az%n,ax%n),d2a(az%n,ax%n))
    allocate(s2a(az%n,ax%n),c2a(az%n,ax%n))
    allocate(nu(az%n,ax%n),mu(az%n,ax%n))
  end subroutine fdmigrc_init
  !----------------------------------------------------------------
  real function getw(iw)
    integer, intent(in) :: iw
    real :: w
    w=ow + dw*(iw-1)
    getw=w
  end function getw
  !----------------------------------------------------------------
  real function getk(ik)
    integer, intent(in) :: ik
    real :: k
    k=ok + dk*(ik-1)
    getk=k
  end function getk
  !----------------------------------------------------------------
  subroutine migrc15_NL(dat,img,rcp,dx_init)
    integer                             :: iz,ix,iw
    real                                :: w
    real,    dimension(:)               :: dx_init
    real,    dimension(:  ),allocatable :: k1,k2,k3
    real,    dimension(:,:),allocatable :: a,b,c,d,s
    real,    pointer                    :: rcp(:,:,:)
    complex                             :: el,er
    complex, dimension(:),allocatable   :: km,kn,cd
    complex, pointer                    :: dat(:,:),frq(:),img(:,:)
    allocate( km(nx),kn(nx),k1(nx),k2(nx),k3(nx),cd(nx) )
    allocate( a(nz,nx),b(nz,nx),c(nz,nx),d(nz,nx),s(nz,nx))
    call ctris_init(nx) 
    km(1)    = (0.,1.)*(-dz)/(8.*dx_init(1))
    km(2:nx) = (0.,1.)*(-dz)/(4.*(dx_init(1:nx-1)+dx_init(2:nx)))
    kn(1:nx) = (0.,1.)*(-dz)/2.
    k1(1 )     = 1./(2.*dx_init(1 )**2)
    k1(2:nx)   = 1./( (dx_init(2:nx)+dx_init(1:nx-1))*dx_init(1:nx-1) )
    k2(1 )     =-1./(dx_init(1 )**2)
    k2(2:nx)   =-1./(dx_init(2:nx)*dx_init(1:nx-1) ) 
    k3(1 )     = 1./(2.*dx_init(1 )**2)
    k3(2:nx)   = 1./( (dx_init(2:nx)+dx_init(1:nx-1))*dx_init(2:nx) )

    a = rcp(:,:,1)
    b = rcp(:,:,2)
    c = rcp(:,:,3)
    d = rcp(:,:,4)
    s = rcp(:,:,5)

    b2a=0.*b/(2*a)
!!    b2a=b/(2*a)
    c2a=0.*c/(2*a)
!!    c2a=c/(2*a)
    d2a=d/(2*a)
    s2a=(2*s)**2/a
    do iw=1,nw
       w=getw(iw)
       frq => dat(:,iw)
       write(0,*) "migrc15",iw,nw,minval(cabs(img)),maxval(cabs(img))
       ko = sqrt( s2a*w**2 - b2a**2)
       do iz=1,nz

          mu(iz,:) = km/ko(iz,:) *   c2a(iz,:)
          nu(iz,:) =-kn/ko(iz,:) * ((c2a(iz,:)/ko(iz,:))**2 - 2.*d2a(iz,:))

          av =   -  k3*nu(iz,:) - mu(iz,:) 
          bv = 1.-  k2*nu(iz,:)           
          cv =   -  k1*nu(iz,:) + mu(iz,:) 
          el = -(av( 1)+bv( 1))/cv( 1)
          er = -(cv(nx)+bv(nx))/av(nx)

          cd( 1) = 0.
          cd(2:nx-1) = (  k3(2:nx-1)*nu(iz,2:nx-1) + mu(iz,2:nx-1) )*frq(3:nx  ) &
          +        ( 1.+  k2(2:nx-1)*nu(iz,2:nx-1)                 )*frq(2:nx-1) &
          +        (      k1(2:nx-1)*nu(iz,2:nx-1) - mu(iz,2:nx-1) )*frq(1:nx-2)
          cd(nx) = 0.

          call tris(el, av, bv, cv, er, cd, frq)
          frq = frq*cexp( -dz*cmplx(-b2a(iz,:),ko(iz,:)))
          img(iz,:) = img(iz,:)+frq

       end do
    end do

    deallocate(k1,k2,k3,km,kn,a,b,d,c,s,cd)
    call ctris_close()
  end subroutine migrc15_NL
  !----------------------------------------------------------------
end module fdmigrc_NL_
