module gazdag		!! Performs the phase shift migration
!  use sep
  implicit none
  integer, private                  :: m1,m2,qmode,n1
  real , private                    :: d1,d2,dz,slow,b1
  real , private                    :: acrust,bcrust,amantle,bmantle

contains

  !! --------------------------------------------------------------------
  !! Initialize matricies
  subroutine gazdag1_init(m1_in,m2_in,d1_in,d2_in,dz_in,slow_in,qmode_in,ac_in,am_in,bc_in,bm_in,b1_in,n1_in)
    integer                            :: m1_in,m2_in,qmode_in,n1_in
    real                               :: d1_in,d2_in,dz_in,slow_in
    real                               :: ac_in,am_in,bc_in,bm_in,b1_in
    m1=m1_in; m2=m2_in; qmode=qmode_in; d1=d1_in; d2=d2_in; dz=dz_in; slow=slow_in
    acrust=ac_in; bcrust=bc_in; amantle=am_in; bmantle=bm_in; b1=b1_in;n1=n1_in
write(0,*) 'd1_in',d1_in,d1
  end subroutine gazdag1_init

  !! --------------------------------------------------------------------
  !!  Do the actual phase shift migration
  subroutine gazdag1(adj,add,cdata,cmodl)
    real                                :: pi,sig1,sig2,w0,dw,kx0,dkx,qi
    complex		                :: ccu,ccd,cup,cu,cd,cux,cdx
    integer		                :: ikx,itau,iw,i1,ixp,ix,it
    real		                :: w,kx,vkx2u,vkx2d,dtau,xdir
    real, dimension(:),allocatable      :: uvel,dvel,xshift,avel,bvel
    complex, dimension(:,:)             :: cdata,cmodl
    complex, dimension(:,:), allocatable:: tmp
    logical                             :: adj,add
    allocate( uvel(m1),dvel(m1),xshift(m1),avel(m1),bvel(m1) )
    allocate( tmp(m1,m2) )
    xshift=0.;avel=0.;bvel=0.;uvel=0.;dvel=0.; tmp=0.

    !! . . Definitions
    pi = 3.14159265; sig1 = +1.; sig2 = -1.;
    kx0=-pi/d2; dkx= 2.*pi/(m2*d2);
    w0=-pi/d1;  dw=2.*pi/(m1*d1); qi=.5/(m1*d1)

    !! . . Create Velocity Profile
    do i1=1,m1
       if ( i1*dz.lt.b1) then
          avel(i1)=acrust;  bvel(i1)=bcrust
       else
          avel(i1)=amantle;  bvel(i1)=bmantle
       end if
    end do

    !! Choice of velocity. Scattering mode dependent
    !! uvel, dvel are up- and down-going wave velocities;
    if(qmode.eq.1) then
       write(0,*) 'Using Forward scattered P-P (q=1)'
       uvel=avel;  dvel=0.*avel; xdir=1.
    else if(qmode.eq.2) then
       write(0,*) 'Using Forward scattered P-S (q=2)'
       uvel=bvel;  dvel=-avel; xdir=slow/abs(slow)
    else if(qmode.eq.3) then
       write(0,*) 'Using Back-scattered P-P (q=3)'
       uvel=avel;  dvel=avel; xdir=1
    else if(qmode.eq.4) then
       write(0,*) 'Using Back-scattered P-S (q=4)'
       uvel=bvel;  dvel=avel;  xdir=slow/abs(slow)
    else if (qmode.eq.5) then
       write(0,*) 'Using Back-scattered S-P (q=5)'
       uvel=avel;  dvel=bvel;  xdir=slow/abs(slow)
    else if (qmode.eq.6) then
       write(0,*) 'Using Back-scattered S-S (q=6)'
       uvel=bvel;  dvel=bvel;  xdir=-slow/abs(slow)
    end if

    do itau=2,m1           !! . . Calculate X-shift
       xshift(itau)=xshift(itau-1)+0.5*dz*slow*uvel(itau)/sqrt(1-(0.5*slow*uvel(itau))**2) 
    end do

    if(adj) then           !! . . Begin Migration
       call ft1axis(0,sig1,m1,m2,cdata)
       call ft2axis(0,sig2,m1,m2,cdata)
    else
       call ft2axis(0,sig2,m1,m2,cmodl)
    end if
    kx: do ikx=2,m2         !! . . Loop over kx
       kx=kx0+(ikx-1)*dkx
       write(0,*) 'Doing ',ikx,' of ',m2
       if (adj==.false.) then
          do itau=2,m1     !! . . Horizontal Shift Correction
             cux=cexp(cmplx(0.,kx*xshift(itau))) 
             cmodl(itau,ikx)=-xdir*cmodl(itau,ikx)*conjg(cux)
          end do
       end if
       iw: do iw=m1/2,m1       !! . . Loop over frequency
          w=w0+(iw-1)*dw
          adjcall: if (adj) then
             cup=cdata(iw,ikx)
             adjtau: do itau=1,m1  !! . . Loop over depth
                cmodl(itau,ikx)=cmodl(itau,ikx)+cup
                if(qmode .le. 2) then 
                   vkx2u= uvel(itau)**2*kx**2
                   cd   = cexp(cmplx(0.,-0.25*w*dz/dvel(itau)/sqrt(1.-slow**2*dvel(itau)**2)))
                   ccu  = cexp((-0.25*dz/uvel(itau)*csqrt(cmplx(qi,-w)**2+vkx2u)))
                   cu   = cexp((-0.25*dz*kx**2*uvel(itau)/csqrt(cmplx(qi,-w)**2+vkx2u) ))
                   cup  = cup*conjg(ccu*cu*cd)
!                   vkx2u= 0.25*uvel(itau)**2*kx**2
!                   ccu  = cexp(-0.5*dz/uvel(itau)*csqrt(cmplx(qi,-w)**2+vkx2u))
!                   cu   = cexp(cmplx(0.,-dz*w*slow**2*uvel(itau)/sqrt(1-slow**2*uvel(itau)**2)))
!                   cup  = cup*conjg(2*ccu*cu)
                else
                   vkx2u= 0.25*uvel(itau)**2*kx**2
                   cd   = cexp(cmplx(0.,-w*dz/dvel(itau)/sqrt(1.-slow**2*dvel(itau)**2)))
                   ccu  = cexp((-dz/uvel(itau)*csqrt(cmplx(qi,-w)**2+vkx2u)))
                   cu   = cexp((-0.5*dz*kx**2*uvel(itau)/csqrt(cmplx(qi,-w)**2+vkx2u) ))
                   cup  = cup*conjg(cu*ccu*cd)
    !               vkx2u=.25*uvel(itau)**2*kx**2
   !                vkx2d=.25*dvel(itau)**2*kx**2 
  !                 ccu  = cexp(-dz/uvel(itau)*csqrt(cmplx(qi,-w)**2+vkx2u))
 !                  ccd  = cexp(-dz/dvel(itau)*csqrt(cmplx(qi,-w)**2+vkx2d))
!                   cd   = cexp(cmplx(0.,-0.5*dz*w*slow**2*dvel(itau)/sqrt(1-slow**2*dvel(itau)**2)))
!                   cu   = cexp(cmplx(0.,-0.5*dz*w*kx**2*uvel(itau)/csqrt(cmplx(qi,-w)**2+vkx2d)))
                end if
             end do adjtau
          else
             cdata(iw,ikx)=cmodl(m1,ikx)
             modtau: do itau=m1-1,1,-1
                if(qmode .le. 2) then 
                   vkx2u = .25*uvel(itau)**2*kx**2 
                   ccu   = cexp(-.5*dz/uvel(itau)*csqrt(cmplx(qi,-w)**2+vkx2u))
                   cu    = cexp(cmplx(0.,-dz*w*slow**2*uvel(itau)/sqrt(1-slow**2*uvel(itau)**2)))
                   cdata(iw,ikx)=cdata(iw,ikx)*ccu*cu+cmodl(itau,ikx)
                else
                   vkx2u= .25*uvel(itau)**2*kx**2
                   vkx2d= .25*dvel(itau)**2*kx**2 
                   ccu  = cexp(-dz/uvel(itau)*csqrt(cmplx(qi,-w)**2+vkx2u))
                   ccd  = cexp(-dz/dvel(itau)*csqrt(cmplx(qi,-w)**2+vkx2d))
                   cu   = cexp(cmplx(0.,-0.5*dz*w*slow**2*uvel(itau)/sqrt(1-slow**2*uvel(itau)**2)))
                   cd   = cexp(cmplx(0.,-0.5*dz*w*slow**2*dvel(itau)/sqrt(1-slow**2*dvel(itau)**2)))
                   cdata(iw,ikx)=cdata(iw,ikx)*ccu*ccd*cu*cd+cmodl(itau,ikx)
                end if
             end do modtau
          end if adjcall
       end do iw
       if (adj)  then  
          do itau=2,m1     !! . . (Single plane wave) Source Wavefield Horizontal Shift
             cux=cexp(cmplx(0.,kx*xshift(itau))) 
             cmodl(itau,ikx)=-xdir*cmodl(itau,ikx)*(cux)
          end do
       end if
    end do kx
    if (adj) then
       call ft2axis(1,sig2,m1,m2,cmodl)!! . . Final fft of spatial axis
    else 
       call ft1axis(1,sig1,m1,m2,cdata)!! . . Final fft of Time axis
       call ft2axis(1,sig2,m1,m2,cdata)!! . . Final fft of Spatial axis
    end if

  end subroutine gazdag1
 
end module gazdag
