module newgazdag		!! Performs the phase shift migration
  implicit none
  integer, private                  :: m1,m2,n1,qmode
  real , private                    :: d1,d2,dz,p1,p2,xdir,dx
contains

  !! --------------------------------------------------------------------
  !! Initialize matricies
  subroutine newgazdag1_init(n1_in,m1_in,m2_in,d1_in,d2_in,dz_in,p1_in,p2_in,xdir_in,qmode_in,dx_in) 
    integer                            :: n1_in,m1_in,m2_in,qmode_in
    real                               :: d1_in,d2_in,dz_in,p1_in,p2_in,xdir_in,dx_in
    n1=n1_in; m1=m1_in; m2=m2_in; d1=d1_in; d2=d2_in; dz=dz_in; p1=p1_in; p2=p2_in
    xdir=xdir_in;qmode=qmode_in; dx=dx_in
  end subroutine newgazdag1_init

  !! --------------------------------------------------------------------
  !!  Do the actual phase shift migration
  subroutine newgazdag1(adj,add,cdata,cmodl,uvel,dvel)
    real                                :: pi,sig1,sig2,w0,dw,kx0,dkx,qi
    complex		                :: ccu,ccd,cup,cu,cd,cux,cdx,cp2
    integer		                :: ikx,itau,iw,i1,ixp,ix,it
    real		                :: w,kx,vkx2u,vkx2d,dtau,slow,vqu,vqd
    real, dimension(:),allocatable      :: xshift,p2shift
    real, dimension(:)                  :: uvel,dvel
    complex, dimension(:,:)             :: cdata,cmodl
    complex, dimension(:,:), allocatable:: tmp
    logical                             :: adj,add
    allocate( xshift(m1) )
    allocate( tmp(m1,m2) )
    allocate( p2shift(m1))
    xshift=0.; tmp=0.; p2shift=0.

    !! . . Definitions
    pi = 3.14159265; sig1 = +1.; sig2 = -1.;
    kx0=-pi/d2; dkx= 2.*pi/(m2*d2);
    w0 =-pi/d1; dw = 2.*pi/(m1*d1); qi=.5/(m1*d1)
    slow=atan2(p2,p1)/abs(atan2(p2,p1))*sqrt(p1**2+p2**2)
    do itau=2,m1           !! . . Calculate X-shift
       xshift(itau) =xshift(itau-1) +0.5*dz*slow*uvel(itau)/sqrt(1-(0.5*slow*uvel(itau))**2) 
       p2shift(itau)=p2*dz 
    end do
    if(adj) then           !! . . Begin Migration
       call ft1axis(0,sig1,m1,m2,cdata)
       call ft2axis(0,sig2,m1,m2,cdata)
    else
       call ft2axis(0,sig2,m1,m2,cmodl)
    end if
    kx: do ikx=2,m2         !! . . Loop over kx
       kx=kx0+(ikx-1)*dkx 
       if (abs(kx).gt.pi/(dx)) then
          cycle
       end if
       write(0,*) 'Doing ',ikx,' of ',m2,' for mode',qmode,' of 7'
       if (adj==.false.) then
          do itau=2,m1     !! . . Horizontal Shift Correction
             cux=cexp(cmplx(0.,kx*xshift(itau))) 
             cmodl(itau,ikx)=-xdir*cmodl(itau,ikx)*conjg(cux)
          end do
       end if
       iw: do iw=m1/2,m1       !! . . Loop over frequency
          w=w0+(iw-1)*dw
          adjcall: if (adj) then
             cup=cdata(iw,ikx)
             adjtau: do itau=1,m1  !! . . Loop over depth
                cmodl(itau,ikx)=cmodl(itau,ikx)+cup
                if(qmode .le. 2) then 
                   vkx2u= .25*uvel(itau)**2*kx**2
                   ccu  = cexp(-.5*dz/uvel(itau)*csqrt(cmplx(qi,-w)**2+vkx2u))
                   cu   = cexp(cmplx(0.,-dz*w*slow**2*uvel(itau)/sqrt(1-slow**2*uvel(itau)**2)))
                   cp2  = cexp(cmplx(0.,-w*dz*abs(p2)))
                   !vqu   = 1./sqrt(1./uvel(itau)**2+p2**2)
                   !vkx2u= .25*vqu**2*kx**2
                   !ccu  = cexp(-.5*dz/vqu*csqrt(cmplx(qi,-w)**2+vkx2u))
                   !cu   = cexp(cmplx(0.,-dz*w*slow**2*vqu/sqrt(1-slow**2*vqu**2)))
                   cup  = cup*conjg(ccu*cu*cp2)
                else
                   vkx2u=.25*uvel(itau)**2*kx**2
                   vkx2d=.25*dvel(itau)**2*kx**2 
                   ccu  = cexp(-dz/uvel(itau)*csqrt(cmplx(qi,-w)**2+vkx2u))
                   ccd  = cexp(-dz/dvel(itau)*csqrt(cmplx(qi,-w)**2+vkx2d))
                   cd   = cexp(cmplx(0.,-0.5*dz*w*slow**2*uvel(itau)/sqrt(1-slow**2*uvel(itau)**2)))
                   cu   = cexp(cmplx(0.,-0.5*dz*w*slow**2*dvel(itau)/sqrt(1-slow**2*dvel(itau)**2)))
                   cp2  = cexp(cmplx(0.,-2*w*dz*abs(p2)))
                   !vqu   = 1./sqrt(1./uvel(itau)**2+p2**2)
                   !vqd   = 1./sqrt(1./dvel(itau)**2+p2**2)
                   !vkx2u=.25*vqu**2*kx**2
                   !vkx2d=.25*vqd**2*kx**2 
                   !ccu  = cexp(-dz/vqu*csqrt(cmplx(qi,-w)**2+vkx2u))
                   !ccd  = cexp(-dz/vqd*csqrt(cmplx(qi,-w)**2+vkx2d))
                   !cd   = cexp(cmplx(0.,-0.5*dz*w*slow**2/vqd/sqrt(1-slow**2*vqd**2)))
                   !cu   = cexp(cmplx(0.,-0.5*dz*w*slow**2*vqu/sqrt(1-slow**2*vqu**2)))
                   cup  = cup*conjg(ccu*ccd*cu*cd*cp2)
                end if
             end do adjtau
          else
             cdata(iw,ikx)=cmodl(m1,ikx)
             modtau: do itau=m1-1,1,-1
                if(qmode .le. 2) then 
                   vkx2u = .25*uvel(itau)**2*kx**2 
                   ccu   = cexp(-.5*dz/uvel(itau)*csqrt(cmplx(qi,-w)**2+vkx2u))
                   cu    = cexp(cmplx(0.,-dz*w*slow**2*uvel(itau)/sqrt(1-slow**2*uvel(itau)**2)))
                   cdata(iw,ikx)=cdata(iw,ikx)*ccu*cu+cmodl(itau,ikx)
                else
                   vkx2u= .25*uvel(itau)**2*kx**2
                   vkx2d= .25*dvel(itau)**2*kx**2 
                   ccu  = cexp(-dz/uvel(itau)*csqrt(cmplx(qi,-w)**2+vkx2u))
                   ccd  = cexp(-dz/dvel(itau)*csqrt(cmplx(qi,-w)**2+vkx2d))
                   cu   = cexp(cmplx(0.,-0.5*dz*w*slow**2*uvel(itau)/sqrt(1-slow**2*uvel(itau)**2)))
                   cd   = cexp(cmplx(0.,-0.5*dz*w*slow**2*dvel(itau)/sqrt(1-slow**2*dvel(itau)**2)))
                   cp2   = cexp(cmplx(0.,2.*p2shift(itau)*itau*dz))
                   cdata(iw,ikx)=cdata(iw,ikx)*ccu*ccd*cu*cd+cmodl(itau,ikx)
                end if
             end do modtau
          end if adjcall
       end do iw
       if (adj)  then  
          do itau=2,m1     !! . . (Single plane wave) Source Wavefield Horizontal Shift
             cux=cexp(cmplx(0.,kx*xshift(itau))) 
             cmodl(itau,ikx)=-xdir*cmodl(itau,ikx)*(cux)
          end do
       end if
    end do kx
    if (adj) then
       call ft2axis(1,sig2,m1,m2,cmodl)!! . . Final fft of spatial axis
    else 
       call ft1axis(1,sig1,m1,m2,cdata)!! . . Final fft of Time axis
       call ft2axis(1,sig2,m1,m2,cdata)!! . . Final fft of Spatial axis
    end if

  end subroutine newgazdag1

end module newgazdag
