module R2PS_baz 
use SEP
  implicit none
  integer, private              :: n1,n2,n3
  real   , private              :: alpha,beta,p1,p2

contains
  subroutine R2PS_baz_init(n1_in,n2_in,n3_in,a_in,b_in,p1_in,p2_in)
    integer                      :: n1_in,n2_in,n3_in
    real                         :: a_in,b_in,p1_in,p2_in
    n1=n1_in; n2=n2_in; n3=n3_in; alpha=a_in; beta=b_in; p1=p1_in; p2=p2_in
  end subroutine R2PS_baz_init

  subroutine R2PS_baz(adj,add,data,modl)
    !! . . Declarations
    integer                               :: ii,jj
    real                                  :: Frp,Fzp,Frsv,Fzsv,qa,qb,delta,pi
    real                                  :: pmag,sina,cosa,slow,invfact
    real, dimension(n1,n2,n3)             :: data,modl
    real, dimension(:,:,:),allocatable    :: tmp  
    logical                               :: adj,add
    allocate( tmp(n1,n2,n3) )
    pi=acos(-1.)
    if (p1 .lt. 0.) then
       pmag=-sqrt(p1**2+p2**2)   !! . . Magnitudes and Angles
    else
!       pmag=-sqrt(p1**2+p2**2)  !! . . Magnitudes and Angles
       pmag=-sqrt(p1**2+p2**2)  !! . . Magnitudes and Angles
    end if
    sina=sin(atan2(p2,p1))
!    cosa=sqrt(1.-sina**2)
    sina=sin(atan2(p2,p1))
    cosa=cos(atan2(p2,p1))
    slow=pmag                !! . . Constant Definitions
    qa=sqrt(1./(alpha*alpha) - slow**2)
    qb=sqrt(1./(beta * beta) - slow**2)
    delta = 1.-4.*slow**2*beta**2 + 4.*slow**4*beta**4 + 4*beta**4*slow**2*qa*qb
    Frp = 4*alpha*beta**2*slow*qa*qb       /delta
    Fzp =-2*alpha*qa*(1-2.*slow**2*beta**2)/delta
    Frsv= 2*beta *qb*(1-2.*slow**2*beta**2)/delta
    Fzsv= 4.*beta**3*slow*qa*qb             /delta
    invfact=1./(-Frsv*Fzp+Frp*Fzsv)
    if (adj==.false.) then   !! . . Model to Data
       do ii=1,n1            !! . . Decompose section from [SV,SH,P] to [R,T,Z]
          do jj=1,n2 
!             tmp(ii,jj,1)= Frsv*modl(ii,jj,1) + Fzsv*modl(ii,jj,3)
!             tmp(ii,jj,2)= -2.*  modl(ii,jj,2)
!             tmp(ii,jj,3)= Frp *modl(ii,jj,1) + Fzp *modl(ii,jj,3)
             tmp(ii,jj,1)= -Fzp/invfact*modl(ii,jj,1) + Frp /invfact*modl(ii,jj,3)
             tmp(ii,jj,2)=        - 0.5*modl(ii,jj,2)
             tmp(ii,jj,3)= Fzsv/invfact*modl(ii,jj,1) - Frsv/invfact*modl(ii,jj,3)
          end do
       end do

       do ii=1,n1            !! . . Rotate Sections from [R,T,Z] to [u1,u2,u3]
          do jj=1,n2
             data(ii,jj,1)= cosa*tmp(ii,jj,1) - sina*tmp(ii,jj,2)  !! Ur
             data(ii,jj,2)= sina*tmp(ii,jj,1) + cosa*tmp(ii,jj,2)  !! Ut
             data(ii,jj,3)=      tmp(ii,jj,3)                      !! Uz
          end do
       end do
    else                     !! . . Data to Model
       do ii=1,n1            !! . . Rotation to go from [u1,u2,u3] to [Ur,Ut,Uz]
          do jj=1,n2
             tmp(ii,jj,1)=  cosa*data(ii,jj,1) + sina*data(ii,jj,2)
             tmp(ii,jj,2)= -sina*data(ii,jj,1) + cosa*data(ii,jj,2)
             tmp(ii,jj,3)=       data(ii,jj,3)
          end do
       end do
       do ii=1,n1            !! . . From [Ur,Ut,Uz] to [SV,SH,P]
          do jj=1,n2       
             modl(ii,jj,1)= Frsv*tmp(ii,jj,1) + Fzsv*tmp(ii,jj,3)
             modl(ii,jj,2)= -2.* tmp(ii,jj,2)
             modl(ii,jj,3)= Frp *tmp(ii,jj,1) + Fzp *tmp(ii,jj,3)
!             modl(ii,jj,1)= -Fzp/invfact*tmp(ii,jj,1) + Frp/invfact*tmp(ii,jj,3)
!             modl(ii,jj,2)=        - 0.5*tmp(ii,jj,2)
!             modl(ii,jj,3)= Fzsv/invfact*tmp(ii,jj,1) - Frsv/invfact*tmp(ii,jj,3)
          end do
       end do
    end if
  end subroutine R2PS_baz
end module R2PS_baz
