module steppingr1_mod
  use sep 
  use igrad1
  use fft
  use triangle
  implicit none
  real(KIND=8), parameter, private :: sixth = 0.122996
  integer     ,private   :: nvx,nvz,nx,nw,ang_win,ang_num,abcd_win,abcd_num
  real(KIND=8),private   :: eps,dtau,gridang,pi
  real(KIND=8),private   :: dvx,dvz,dx,dw,ovx,ovz,ox,ow
contains
  subroutine stepping_init1(nvx_in,nvz_in,nx_in,nw_in,dvx_in,dvz_in,dx_in,dw_in,&
  ovx_in,ovz_in,ox_in,ow_in,dtau_in,eps_in,gridang_in,ang_win_in,ang_num_in,&
  abcd_win_in,abcd_num_in)
    integer        :: nvx_in,nvz_in,nx_in,nw_in
    integer        :: ang_win_in,ang_num_in,abcd_win_in,abcd_num_in
    real(KIND=8)   :: dvx_in,dvz_in,dx_in,dw_in
    real(KIND=8)   :: ovx_in,ovz_in,ox_in,ow_in
    real(KIND=8)   :: eps_in,dtau_in,gridang_in
    nvx=nvx_in; nvz=nvz_in; nx=nx_in; nw=nw_in 
    dvx=dvx_in; dvz=dvz_in; dx=dx_in; dw=dw_in
    ovx=ovx_in; ovz=ovz_in; ox=ox_in; ow=ow_in
    dtau=dtau_in;eps=eps_in;gridang=gridang_in
    ang_win=ang_win_in  ; ang_num=ang_num_in
    abcd_win=abcd_win_in; abcd_num=abcd_num_in
    pi=dacos(-1.)
  end subroutine stepping_init1
  !####################################################################
  subroutine model_points(x,z)
    integer                                    :: i2
    integer,         dimension(:), allocatable :: ixx,izz
    real(KIND=8),    dimension(:)              :: x,z
    allocate( ixx(nvx),izz(nvz) )
    ixx(1:nvx) = (/ (i2,i2=1,nvx) /)  
    izz(1:nvz) = (/ (i2,i2=1,nvz) /) 
    x=dble((ixx-1)*dvx+ovx)               
    z=dble((izz-1)*dvz+ovz)
    deallocate(ixx,izz)
  end subroutine model_points
  !####################################################################
  subroutine get_start_slow(x,z,xsta,zsta,vel,ssta)
    integer                                :: iiz,iix,i2,ix
    real(KIND=8),dimension(:,:)            :: xsta,zsta,ssta
    real(KIND=8),dimension(:)              :: x,z
    real(KIND=8),dimension(:,:)            :: vel
    integer,     dimension(:), allocatable :: ztmp,xtmp
    allocate( ztmp(1), xtmp(1) )
    do i2=1,4
       do ix=1,nx
          xtmp=minloc(dabs(x-xsta(i2,ix))); iix=xtmp(1)
          ztmp=minloc(dabs(z-zsta(i2,ix))); iiz=ztmp(1)
          ssta(i2,ix)=1./vel(iiz,iix)
       end do
    end do
    deallocate(ztmp,xtmp)
  end subroutine get_start_slow
  !!####################################################################
  subroutine Step_calculation(xsta,zsta,vel,data,x,z,s)
    integer                                    :: ip,iix,iiz,na,ih,xsgn,zsgn
    integer        ,dimension(:)  ,allocatable :: xloc,zloc
    real(KIND=8)                               :: dx1,dx2,dz1,dz2
    real(KIND=8)                               :: da,db,dc,dd,wa,wb,wc,wd
    real(KIND=8)   ,dimension(:)               :: x,z,s
    real(KIND=8)   ,dimension(:)  ,allocatable :: v
    real(KIND=8)   ,dimension(:)  ,allocatable :: dxadd,dzadd,pxstep,pzstep
    real(KIND=8)   ,dimension(:,:)             :: xsta,zsta,vel
    real(KIND=8)   ,dimension(:,:),allocatable :: dWdx_W,dWdz_W,K,tau,gamma 
    real(KIND=8)   ,dimension(:,:),allocatable :: dtdx,dtdz,dgdx,dgdz,ds
    real(KIND=8)   ,dimension(:,:),allocatable :: dxsta,dzsta,dxstep,dzstep
    complex(KIND=8),dimension(:,:)             :: data
    complex(KIND=8),dimension(:,:),allocatable :: dWdt,dWdg
    allocate( xloc(1), zloc(1) )
    allocate( dxstep(3,nx),dzstep(3,nx),ds(3,nx) )
    allocate( dWdx_W(3,nx),dWdz_W(3,nx),dtdx(3,nx),dtdz(3,nx) )
    allocate( dgdx(3,nx),dgdz(3,nx),pxstep(nx),pzstep(nx) )
    allocate( dWdt(3,nx), dWdg(3,nx), K(3,nx) )
    allocate( tau(3,nx),gamma(3,nx) )
    allocate( dxsta(3,nx),dzsta(3,nx))
    allocate( dxadd(nx),dzadd(nx),v(nx) )

    !! . . Find location on velocity model
    do ip=1,nx 
       do ih=1,3
          xloc=minloc(dabs(x-xsta(ih+1,ip))); iix=xloc(1)
          zloc=minloc(dabs(z-zsta(ih+1,ip))); iiz=zloc(1)

          !! Distances for weighting
          dx1=dble(x(iix  )-xsta(ih+1,ip))
          dz1=dble(z(iiz  )-zsta(ih+1,ip))
          if (iix .gt. 1 .and. iix .lt. nvx) then
             if (dx1 .gt. 0.) then
                dx2=dble(x(iix-1)-xsta(ih+1,ip))
                xsgn=-1
             else
                dx2=dble(x(iix+1)-xsta(ih+1,ip))
                xsgn=1
             end if
          else
             dx2=dble(x(iix))
             xsgn=0
          end if

          if (iiz .gt. 1 .and. iiz .lt. nvz) then
             if (dz1 .gt. 0.) then
                dz2=dble(z(iiz-1)-zsta(ih+1,ip))
                zsgn=-1
             else
                dz2=dble(z(iiz+1)-zsta(ih+1,ip))
                zsgn=1
             end if
          else
             dz2=dble(z(iiz))
             zsgn=0
          end if
          da=dsqrt(dx1**2+dz1**2); db=dsqrt(dx1**2+dz2**2)
          dc=dsqrt(dx2**2+dz1**2); dd=dsqrt(dx2**2+dz2**2)
          wa=dble(1/(1+da/(db+eps)+da/(dc+eps)+da/(dd+eps)))
          wb = wa*da/(db+eps); wc = wa*da/(dc+eps)
          wd = wa*da/(dd+eps)
          if (dx1 .eq. 0. .and. dz1 .eq. 0.) then
             wa=1.; wb=0.;wc=0.;wd=0.
          end if
          ds(ih,ip)=-dble(dtau)*&
          (wa*vel(iiz,iix     ) + wb*vel(iiz+zsgn,iix      )+&
          wc*vel(iiz,iix+xsgn) + wd*vel(iiz+zsgn,iix+xsgn))
          ds(ih,ip)=-dble(dtau)*vel(iiz,iix)
       end do
       if (dx1 .eq. 0. .and. dz1 .eq. 0.) then 
          s(ip)=1./vel(iiz,iix)
       else 
          s(ip)=1./(wa*vel(iiz,iix)+wb*vel(iiz+zsgn,iix)+wc*vel(iiz,iix+xsgn)&
          +wd*vel(iiz+zsgn,iix+xsgn))
       end if

       !write(0,*) s(ip) 
    end do

    !! . . dW/dtau
    dWdt = (data(2:4,:)-data(1:3,:))/dtau

    !! . . dW/dgamma
    dWdg(:,1)     =(data(2:4,2   )-data(2:4,1     ))/dx
    dWdg(:,2:nx-1)=(data(2:4,3:nx)-data(2:4,1:nx-2))/(2.*dx)
    dWdg(:,nx)    =(data(2:4,nx  )-data(2:4,nx-1  ))/dx

    !! . . Compute angles
    dxsta(3,2:nx-1) = xsta(4,3:nx)-xsta(4,1:nx-2)
    dzsta(3,2:nx-1) = zsta(4,3:nx)-zsta(4,1:nx-2)

    do ip=2,nx-1
       if ( dzsta(3,ip) .gt. 0.) then

          if ( dxsta(3,ip) .gt. 0.) then
             dtdx(3,ip)=-dabs(dzsta(3,ip)/dsqrt(dxsta(3,ip)**2+dzsta(3,ip)**2)*s(ip))
             dtdz(3,ip)=+dabs(dsqrt(s(ip)**2-dtdx(3,ip)**2))
          else
             dtdx(3,ip)=-dabs(dzsta(3,ip)/dsqrt(dxsta(3,ip)**2+dzsta(3,ip)**2)*s(ip))
             dtdz(3,ip)=-dabs(dsqrt(s(ip)**2-dtdx(3,ip)**2))
          end if
       else

          if ( dxsta(3,ip) .gt. 0.) then
             dtdx(3,ip)= +dabs(dzsta(3,ip)/dsqrt(dxsta(3,ip)**2+dzsta(3,ip)**2)*s(ip))
             dtdz(3,ip)= +dabs(dsqrt(s(ip)**2-dtdx(3,ip)**2))
          else
             dtdx(3,ip)= +dabs(dzsta(3,ip)/dsqrt(dxsta(3,ip)**2+dzsta(3,ip)**2)*s(ip))
             dtdz(3,ip)= -dabs(dsqrt(s(ip)**2-dtdx(3,ip)**2))
          end if

       end if
    end do

    dtdx(3,1 )=dtdx(3,2   )
    dtdx(3,nx)=dtdx(3,nx-1)
    dtdz(3,1 )=dtdz(3,2   )
    dtdz(3,nx)=dtdz(3,nx-1)

    dgdz(3,:) = s*dtdx(3,:)
    dgdx(3,:) =-s*dtdz(3,:)
    write(0,*) 'Min/max dtdx: ',minval(dtdx(3,:)*180./pi),maxval(dtdx(3,:)*180./pi)
    write(0,*) 'Min/max dtdz: ',minval(dtdz(3,:)*180./pi),maxval(dtdz(3,:)*180./pi)

    !! . . Compute Step
    dWdx_W = dble(aimag( (dWdt*dtdx+dWdg*dgdx)/( data(2:4,:)+eps )))
    dWdz_W = dble(aimag( (dWdt*dtdz+dWdg*dgdz)/( data(2:4,:)+eps )))
    K      = dsqrt (  dWdx_W**2+dWdz_W**2)
    dxstep = dWdx_W/K
    dzstep = dWdz_W/K
    dxadd  = ds(3,:)*dxstep(3,:)
    dzadd  = ds(3,:)*dzstep(3,:)

    !! . . Update station locations
    xsta(1:3,:)=xsta(2:4,:)
    zsta(1:3,:)=zsta(2:4,:)
    xsta(4,:)=xsta(3,:) + dxadd
    zsta(4,:)=zsta(3,:) + dzadd

    deallocate(xloc,zloc,dxstep,dzstep,dWdt,v,&
    dWdg,dtdx,dtdz,dgdx,dgdz,dWdx_W,dWdz_W,K,ds,tau,gamma,dzadd,dxadd)
  end subroutine Step_calculation
  !!####################################################################
  subroutine Ray_spreading(xsta,zsta,a,b,c,d,s)
    real(KIND=8),dimension(:,:)               :: xsta,zsta
    real(KIND=8),dimension(:,:), allocatable  :: h1,h2
    real(KIND=8),dimension(:  ), allocatable  :: m1,m2
    real(KIND=8),dimension(:,:), allocatable  :: azv,axv,jxv,jzv
    real(KIND=8),dimension(:  )               :: a,b,c,d,s
    allocate(  h1(2,nx),h2(2,nx) )
    allocate(  m1(nx),m2(nx) )
    allocate(  azv(2,nx),axv(2,nx),jxv(2,nx),jzv(2,nx) )

    !! h1 Calculate derivative along ray direction (tau)
    axv(1:2,:) =(xsta(3:4,:)-xsta(2:3,:))/dtau
    azv(1:2,:) =(zsta(3:4,:)-zsta(2:3,:))/dtau
    h1=sqrt(axv*axv+azv*azv)

    !! h2 Calculate derivative across ray direction
    jxv(1:2,1     ) =( xsta(3:4,2      ) -  xsta(3:4,1        ))/dx
    jxv(1:2,2:nx-1) =( xsta(3:4,3:nx   ) -  xsta(3:4,1:nx-2   ))/(2.*dx)
    jxv(1:2,nx    ) =( xsta(3:4,nx     ) -  xsta(3:4,nx-1     ))/dx
    jzv(1:2,1     ) =( zsta(3:4,2      ) -  zsta(3:4,1        ))/dx
    jzv(1:2,2:nx-1) =( zsta(3:4,3:nx   ) -  zsta(3:4,1:nx-2   ))/(2.*dx)
    jzv(1:2,nx    ) =( zsta(3:4,nx     ) -  zsta(3:4,nx-1     ))/dx
    h2=sqrt(jxv*jxv+jzv*jzv)
    h2(:,1 )=h2(:,2   ); 
    h2(:,nx)=h2(:,nx-1); 

    !! . . First cross derivative
    m1=( h1(2,:)*(h2(2,:)-h2(1,:))-&
         h2(2,:)*(h1(2,:)-h1(1,:)))/dtau

    !! . . Second cross-derivative
    m2(1)=       (h2(2,2)     *(h1(2,2)    -h1(2,1     ))&
                 +h1(2,2)     *(h2(2,2)    -h2(2,1     )))/dx
    m2(2:nx-1)=  (h2(2,2:nx-1)*(h1(2,3:nx) -h1(2,1:nx-2))+&
                  h1(2,2:nx-1)*(h2(2,3:nx) -h2(2,1:nx-2)))/(2.*dx)
    m2(nx)=      (h2(2,nx)    *(h1(2,nx)   -h1(2,nx-1  ))&
                 +h1(2,nx)    *(h2(2,nx)   -h2(2,nx-1  )))/dx

    !! . . Calculate Factors for finite difference
    a=1./(h1(2,:)*h1(2,:))
    b=m1/(h1(2,:)*h2(2,:)*h1(2,:)*h1(2,:))
    c=m2/(h1(2,:)*h2(2,:)*h2(2,:)*h2(2,:))
    d=1./(h2(2,:)*h2(2,:))
    deallocate(h1,h2,m1,m2,axv,azv,jxv,jzv)
  end subroutine Ray_spreading
  !######################################################################
  subroutine Jmigrc15(data,img,a,b,c,d,s)
    integer                                   :: iz,ix,iw,na,ip
    real                                      :: dtmp
    real(KIND=8)                              :: w
    real(KIND=8),    dimension(:)             :: a,b,c,d,s
    real(KIND=8),    dimension(:),allocatable :: b2a,c2a,d2a,s2a,ko
    complex(KIND=8)                           :: el,er,km,kn
    complex(KIND=8), dimension(:)             :: img,data
    complex(KIND=8), dimension(:),allocatable :: av,bv,cv,cd,nu,mu
    allocate( av(nx),bv(nx),cv(nx),cd(nx),ko(nx))
    allocate( b2a(nx),c2a(nx),d2a(nx),s2a(nx) )
    allocate( nu(nx),mu(nx),ko(nx) )

    w=2.*pi*ow 
    na=abcd_win 
    if (na .ne. 0) then
      b(1)=sum(b(1:2))/2
      c(1)=sum(c(1:2))/2
      d(1)=sum(d(1:2))/2
      do ip=2,na
        b(ip)=sum(b(ip-ceiling((ip-1)/2.):ip+ceiling((ip-1)/2.)))/&
                dble(2*ceiling((ip-1)/2.)+1)
        c(ip)=sum(c(ip-ceiling((ip-1)/2.):ip+ceiling((ip-1)/2.)))/&
                dble(2*ceiling((ip-1)/2.)+1)
        d(ip)=sum(d(ip-ceiling((ip-1)/2.):ip+ceiling((ip-1)/2.)))/&
                dble(2*ceiling((ip-1)/2.)+1)
      end do
      do ip=na+1,nx-na
        b(ip)=sum(b(ip-na:ip+na))/(2.*na+1.)
        c(ip)=sum(c(ip-na:ip+na))/(2.*na+1.)
        d(ip)=sum(d(ip-na:ip+na))/(2.*na+1.)
      end do
      do ip=nx-na+1,nx-1
        b(ip)=sum(b(ip-ceiling((nx-ip)/2.):ip+ceiling((nx-ip)/2.)))/&
                dble(2*ceiling((nx-ip)/2.)+1)
        c(ip)=sum(c(ip-ceiling((nx-ip)/2.):ip+ceiling((nx-ip)/2.)))/&
                dble(2*ceiling((nx-ip)/2.)+1)
        d(ip)=sum(d(ip-ceiling((nx-ip)/2.):ip+ceiling((nx-ip)/2.)))/&
                dble(2*ceiling((nx-ip)/2.)+1)
      end do
      b(nx)=sum(b(nx-1:nx))/2
      c(nx)=sum(c(nx-1:nx))/2
      d(nx)=sum(d(nx-1:nx))/2
    end if

    !! . . Divisions for solver
    b2a= dble( 0.*b     /(2.*a))
    c2a= dble( c        /(2.*a))
    d2a= dble( d        /(2.*a))
    s2a= dble( (2.*s)**2/a     )

    !! . . Variables for solver
    km = (0.,1.)*(-dtau)/(4.*dx)
    kn = (0.,1.)*(-dtau)/(2.*dx*dx)
    ko = sqrt(s2a*w**2 - b2a**2 )
    mu = km/ko *   c2a
    nu =-kn/ko * ((c2a/ko)**2 - d2a)
!    write(0,*) 'b2a', minval(abs(b2a)),maxval(abs(b2a))
!    write(0,*) 'a', minval(abs(a)),maxval(abs(a))
!    write(0,*) 'b', minval(abs(b)),maxval(abs(b))
!    write(0,*) 'ko', minval(abs(ko))
!    write(0,*) 'mu', minval(abs(mu))
!    write(0,*) 'nu', minval(abs(nu))

    !! . . Setup tridiagonal system
    av =    -  (nu - mu) +    sixth 
    bv = 1. +  2.*nu     - 2.*sixth          
    cv =    -  (nu + mu) +    sixth 
    cd(1) = 0.
    cd(2:nx-1)=(      (nu(2:nx-1) - mu(2:nx-1)) + sixth)*data(3:nx  )&
             + (1.- 2.*nu(2:nx-1) -           2.* sixth)*data(2:nx-1)&
             + (      (nu(2:nx-1) + mu(2:nx-1)) + sixth)*data(1:nx-2)
    cd(nx) = 0.
    el = -(av( 1)+bv( 1))/cv( 1)
    er = -(cv(nx)+bv(nx))/av(nx)

    call ctris_vector(el, av, bv, cv, er, cd, data)
    dtmp=dtau
    img = data*cexp(-dtmp*cmplx(-b2a,ko) )

    deallocate(av,bv,cv,ko,b2a,s2a,d2a,c2a,nu,mu)
  end subroutine Jmigrc15
  !----------------------------------------------------------------
  real function getw(iw)
    integer, intent(in) :: iw
    real :: w
    w=2*pi*(ow + dw*(iw-1))
    getw=w
  end function getw
  !----------------------------------------------------------------
  subroutine ctris_vector (el, a, b, c, er, d, q)
    complex(KIND=8),                intent (in)  :: el, er
    complex(KIND=8), dimension (:), intent (in)  :: a, b, c
    complex(KIND=8), dimension (:), intent (in)  :: d
    complex(KIND=8), dimension (:), intent (out) :: q
    complex(KIND=8), dimension (:), allocatable  :: cf, ce
    real(KIND=8),    dimension (:), allocatable  :: f, e
    integer                              :: i
    complex(KIND=8)                             :: den, endl, endr
    allocate (cf(nx), ce(nx),f(nx),e(nx))
    endl = c(1)*el + b(1)
    endr = a(nx)*er + b(nx)
    ce(1) = -a(1)/endl
    cf(1) =  d(1)/endl
    do i= 2, nx-1
       den   =  b(i)+c(i)*ce(i-1)  
       cf(i) = (d(i)-c(i)*cf(i-1))/den
       ce(i) =      -a(i)         /den
    end do
    q(nx) = (d(nx)-c(nx)*cf(nx-1))/(endr+c(nx)*ce(nx-1))
    do i= nx-1,1,-1
       q(i) = ce(i) * q(i+1) + cf(i)
    end do
    deallocate(cf,ce,f,e)
  end subroutine ctris_vector
  !######################################################################
  subroutine write_out(steps)
    integer                    :: steps
    real                       :: dtau1,dx1,ox1
    dx1=dx;ox1=ox;dtau1=dtau
    call to_history("n2",steps           ) 
    call to_history("n1",nx              )  
    call to_history("d2",dtau1           ) 
    call to_history("d1",dx1             )  
    call to_history("o2",0.              ) 
    call to_history("o1",ox1             )  
    call auxputch  ("n2","i",steps,"rays.H" )
    call auxputch  ("n1","i",nx   ,"rays.H" )
    call auxputch  ("d2","r",dtau1,"rays.H" )
    call auxputch  ("d1","r",dx1  ,"rays.H" )
    call auxputch  ("o2","r",0.   ,"rays.H" )
    call auxputch  ("o1","r",ox1  ,"rays.H" )
    call auxputch  ("esize","i",8 ,"rays.H" )
    call hclose()
  end subroutine write_out
  !######################################################################
end module steppingr1_mod
